from abc import ABC, ABCMeta, abstractmethod

from kivy.event import EventDispatcher
from kivy.properties import BooleanProperty
from setup_logger import setup_logger_devel


class AbstractReceiver(EventDispatcher, metaclass=ABCMeta):
    """受信抽象クラス
    """
    # ペアリングモード - 受信機
    pairing = BooleanProperty(False)
    # 切断状態 - 表示機
    disconnected = BooleanProperty(False)
    # 時刻ずれ - 表示機
    timeDiff = BooleanProperty(False)

    def __init__(self, received, erased, synced, pairing, received_on_pairing, long_pressed_on_pairing, processing, shutdown, eraser_pushed, **kwargs):
        """コンストラクタ
        
        Args:
            received (function): 呼出受信コールバック
            erased (function): 取消受信コールバック
            synced (function): 同期コールバック
            pairing (function): ペアリングボタン押下コールバック
            received_on_pairing (function): 呼出受信コールバック(ペアリングモード時)
            long_pressed_on_pairing (function): クリア長押しコールバック
            processing (function): 処理中表示コールバック
            shutdown (function): シャットダウンコールバック
            eraser_pushed (function): 消去機押下コールバック
        """
        super(AbstractReceiver, self).__init__(**kwargs)
        self.logger = setup_logger_devel(self.__class__.__name__)


    @abstractmethod
    def start(self):
        """開始
        """
        raise NotImplementedError()

    @abstractmethod
    def requestErase(self, data, eraseFinishedCb):
        """取消要求
        
        Args:
            data (dictionary): 取消データ
            eraseFinishedCb (function): 取消完了コールバック
        """
        raise NotImplementedError()

    @abstractmethod
    def requestSync(self, syncFinishedCb):
        """同期要求
        
        Args:
            syncFinishedCb (function): 同期完了コールバック
        """
        raise NotImplementedError()


    @abstractmethod
    def stop(self):
        """終了
        """
        raise NotImplementedError()

    @abstractmethod
    def pause(self):
        """一時停止
        """
        raise NotImplementedError()

    @abstractmethod
    def resume(self):
        """再開
        """
        raise NotImplementedError()
