import json
from datetime import datetime

from jnius import autoclass

import common_util as util
from absreceiver import AbstractReceiver
from base_demo_receiver import BaseDemoReceiver
from disp_dev.config_setting import ConfigSetting

if util.is_android_device():
    from android.broadcast import BroadcastReceiver
    from android_dev.sendbroadcast import send_broadcast_request, Action, Type

class DemoReceiver(BaseDemoReceiver):
    """デモ受信クラス
    """
    pass

class Receiver(AbstractReceiver):
    """受信クラス
    """
    RECV_SERVICE_NAME = util.PACKAGE_DOMAIN_NAME + '.' +'ServiceRecvservice'

    def __init__(self, received, erased, synced, pairing, received_on_pairing, long_pressed_on_pairing, processing, shutdown, eraser_pushed, **kwargs):
        """コンストラクタ

        Args:
            received (function): 呼出受信コールバック
            erased (function): 取消受信コールバック
            synced (function): 同期コールバック
            pairing (function): ペアリングボタン押下コールバック
            received_on_pairing (function): 呼出受信コールバック(ペアリングモード時)
            long_pressed_on_pairing (function): クリア長押しコールバック
            processing (function): 処理中表示コールバック
            shutdown (function): シャットダウンコールバック
            eraser_pushed (function): 消去機押下コールバック
        """
        super(Receiver, self).__init__(received, erased, synced,  pairing, received_on_pairing, long_pressed_on_pairing, processing, shutdown, eraser_pushed, **kwargs)
        self.service = None
        self.config = ConfigSetting()
        self.received = received
        self.erased = erased
        self.synced = synced
        self.br = None
        self.processing = processing

    def start(self):
        """開始
        """
        self.logger.info("receiver start.")
        try:
            self.logger.debug("is_android_device={0}".format(util.is_android_device()))
            if util.is_android_device():
                self.br = BroadcastReceiver(self.on_broadcast, actions=[Action.RESPONSE.value])
                self.br.start()

                # サービス起動
                self.service = autoclass(self.RECV_SERVICE_NAME)
                mActivity = autoclass('org.kivy.android.PythonActivity').mActivity
                argument = "{}"
                self.service.start(mActivity, argument)


        except:
            self.logger.error("start failed.", exc_info=True)

        self.logger.debug("service started.")

    def requestErase(self, data, eraseFinishedCb):
        """取消要求
        
        Args:
            data (dictionary): 取消データ
            eraseFinishedCb (function): 取消完了コールバック
        """
        self.eraseFinishedCb = eraseFinishedCb
        send_broadcast_request(Type.ERASE_REQ, data)

 
    def requestSync(self, syncFinishedCb):
        """同期要求
        
        Args:
            syncFinishedCb (function): 同期完了コールバック
        """
        self.synced = syncFinishedCb
        send_broadcast_request(Type.SYNC_REQ, {})

    def stop(self):
        """終了
        """
        if util.is_android_device():
            # サービス停止
            mActivity = autoclass('org.kivy.android.PythonActivity').mActivity
            self.service.stop(mActivity)

            self.br.stop()

        self.logger.info("receiver stop.")


    def pause(self):
        """一時停止
        """
        self.logger.info("pause.")
        if util.is_android_device():

            self.br.stop()
            self.br = None

    def resume(self):
        """再開
        """
        self.logger.info("resume.")
        if util.is_android_device():
            self.br = BroadcastReceiver(self.on_broadcast, actions=[Action.RESPONSE.value])
            self.br.start()
            send_broadcast_request(Type.REQUEST_CONN_STATUS, {})


    def on_broadcast(self, context, intent):
        """ブロードキャストインテント受信
        
        Args:
            context (object): コンテクスト
            intent (object): インテント
        """
        extras = intent.getExtras()
        type = "".join(extras.get('type'))
        try:
            data = json.loads("".join(extras.get('data')))
            self.logger.debug("on_broadcast type={0}, data={1}".format(type, data))
            if type == Type.CALL.value:
                if self.received is not None:
                    self.received([data])
            elif type == Type.DEL.value:
                if self.erased is not None: self.erased([data])
            elif type == Type.ERASE_RES.value:
                if self.eraseFinishedCb is not None: self.eraseFinishedCb(True, data)
            elif type == Type.SYNC_RES.value:
                if self.synced is not None: self.synced(data["callList"], data["delList"])
            elif type == Type.SHOW_PROCESSING.value:
                if self.processing is not None: self.processing(True)
            elif type == Type.HIDE_PROCESSING.value:
                if self.processing is not None: self.processing(False)
            elif type == Type.NOTIFY_CONN_STATUS.value:
                if data["status"] > 1:
                    self.disconnected = False
                else:
                    self.disconnected = True
            elif type == Type.NOTIFY_TIME_DIFF.value:
                self.timeDiff = data["timeDiff"]

        except:
            self.logger.error("on_broadcast failed.", exc_info=True)
		

if __name__ == "__main__":
    # テスト
    import kivy
    from kivy.app import App
    from kivy.uix.label import Label
    from kivy.clock import Clock
    class TestApp(App):
        def build(self):
            self.receiver = Receiver(self.received, self.erased, self.synced, None, None, None)
            return Label(text='Debugging now...')

        def on_start(self):
            self.receiver.start()
            Clock.schedule_interval(lambda dt: self.receiver.requestSync(self.synced), 30)

        def stop(self):
            self.receiver.stop()

        def received(self, datalist):
            print('received:{}'.format( datalist))

        def erased(self, datalist):
            print('erased:{}'.format( datalist))

        def synced(self, callList, delList):
            print('synced: callList={0}, \n delList={1}'.format( callList, delList))

    try:
        app = TestApp()
        app.run()
    except KeyboardInterrupt:
        app.stop()
