import json
import os
import pathlib
import threading
import time
from datetime import datetime
from enum import IntEnum
from socket import AF_INET, SOCK_STREAM, socket, timeout

from jnius import autoclass, cast

import common_util as util
from disp_dev.config_setting import ConfigSetting
from resources import Resources
from setup_logger import setup_logger_devel
from telegram import Telegram

if util.is_android_device():
    from android.broadcast import BroadcastReceiver
    from android_dev.sendbroadcast import send_broadcast_response, Action, Type

class Status(IntEnum):
    S1 = 1
    S2 = 2
    S3 = 3
    S4 = 4
    S5 = 5

class WatchServer():

    NOTIFICATION_CHANNEL_ID = "BeyondEC"

    def __init__(self, config):
        """コンストラクタ
        
        Args:
            config (object): 設定オブジェクト
        """
        try:
            self.logger = setup_logger_devel(self.__class__.__name__)
            self.resources = Resources()
            self.status = Status.S1
            # send_broadcast_response(Type.NOTIFY_CONN_STATUS, {"status": int(Status.S1)})
            self.logger.info("status = {}".format(self.status))
            self.config = config
            self.socket = None
            self.healthDt = None
            self.seqNo = 0
            self.br = None
            self.callList = []
            self.recv_lock = threading.Lock()
            self.vibrator = None
            if util.is_android_device():
                self.initSound()
                self.initVibrator()

                # Backgroundサービスだと、一定時間経過すると強制終了させられるので
                # 通知を表示して、Foregroundサービスにする

                service = autoclass('org.kivy.android.PythonService').mService

                new_notification = self.create_notification(self.resources.getResource("receiver_service:notification_connecting"))
                #Below sends the notification to the notification bar; nice but not a foreground service.
                service.startForeground(1, new_notification)
                self.logger.debug("__init__ end")
        except:
            self.logger.error("__init__ failed.", exc_info=True)

    def create_notification(self, message, num=0):
        """通知生成

        Args:
            message (string): 通知メッセージ
            num (int, optional): 呼び出し数. Defaults to 0.

        Returns:
            object: 通知
        """
        if util.is_android_device():
            Context = autoclass('android.content.Context')
            Intent = autoclass('android.content.Intent')
            PendingIntent = autoclass('android.app.PendingIntent')
            AndroidString = autoclass('java.lang.String')
            NotificationBuilder = autoclass('android.app.Notification$Builder')
            Notification = autoclass('android.app.Notification')
            BuildVersion  = autoclass('android.os.Build$VERSION')
            # BuildVersionCodes  = autoclass('android.os.Build$VERSION_CODES')

            service = autoclass('org.kivy.android.PythonService').mService

            PythonActivity = autoclass('org.kivy.android' + '.PythonActivity')
            notification_service = service.getSystemService(
                Context.NOTIFICATION_SERVICE)
            app_context = service.getApplication().getApplicationContext()
            notification_builder = NotificationBuilder(app_context)
            title = AndroidString(self.resources.getResource("receiver_service:notificationTitle").encode('utf-8'))
            message = AndroidString(message.encode('utf-8'))
            # app_class = service.getApplication().getClass()
            notification_intent = Intent(app_context, PythonActivity)
            notification_intent.setFlags(Intent.FLAG_ACTIVITY_CLEAR_TOP |
                Intent.FLAG_ACTIVITY_SINGLE_TOP | Intent.FLAG_ACTIVITY_NEW_TASK)
            notification_intent.setAction(Intent.ACTION_MAIN)
            notification_intent.addCategory(Intent.CATEGORY_LAUNCHER)
            pendingIntentFlag = 0
            if BuildVersion.SDK_INT >= 31:  # BuildVersionCodes.S Android12
                pendingIntentFlag = pendingIntentFlag | PendingIntent.FLAG_IMMUTABLE

            intent = PendingIntent.getActivity(service, 0, notification_intent, pendingIntentFlag)
            notification_builder.setContentTitle(title)
            notification_builder.setContentText(message)
            notification_builder.setContentIntent(intent)
            # Drawable = autoclass("{}.R$drawable".format(service.getPackageName()))
            # icon = getattr(Drawable, 'icon')
            Drawable = autoclass("{}.R$mipmap".format(service.getPackageName()))
            icon = getattr(Drawable, 'ic_launcher')
            notification_builder.setSmallIcon(icon)
            notification_builder.setAutoCancel(True)
            notification_builder.setDefaults(0)
            notification_builder.setNumber(num)

            # if BuildVersion.SDK_INT >= BuildVersionCodes.O:
            if BuildVersion.SDK_INT >= 26:  # BuildVersionCodes.O: Android 8
                NotificationChannel = autoclass('android.app.NotificationChannel')
                NotificationManager = autoclass('android.app.NotificationManager')
                channelTitle = self.resources.getResource("receiver_service:channelTitle")
                channelDescription = ""
                channel = NotificationChannel(self.NOTIFICATION_CHANNEL_ID, channelTitle, NotificationManager.IMPORTANCE_LOW)
                channel.setDescription(channelDescription)
                channel.setSound(None, None)
                channel.setShowBadge(False)
                notification_service.createNotificationChannel(channel)

                notification_builder.setChannel(self.NOTIFICATION_CHANNEL_ID)

            new_notification = notification_builder.getNotification()
            return new_notification

        return None

    def update_notification(self, message, num=0):
        """通知更新

        Args:
            message (string): 通知メッセージ
        """
        if util.is_android_device():
            new_notification = self.create_notification(message, num)
            Context = autoclass('android.content.Context')
            service = autoclass('org.kivy.android.PythonService').mService
            notification_service = service.getSystemService(
                Context.NOTIFICATION_SERVICE)
            notification_service.notify(1, new_notification)

    def update_notification_calling(self, num):
        """通知更新(呼び出し数)

        Args:
            num (int): 通知メッセージ
        """
        if num > 0:
            self.update_notification(self.resources.getResource("receiver_service:notification_calling").format(num), num)
        else:
            self.update_notification(self.resources.getResource("receiver_service:notification_receiving"), num)

    def release(self):
        """release
        """
        if util.is_android_device():
            for no in self.mediaPlayers:
                self.mediaPlayers[no].release()
            self.br.stop()


    def accept_request(self):
        """ブロードキャストレシーバーの待ち受け開始
        """
        if util.is_android_device():
            self.br = BroadcastReceiver(self.on_broadcast, actions=[Action.REQUEST.value])
            self.br.start()


    def on_broadcast(self, context, intent):
        """ブロードキャストインテント受信
        
        Args:
            context (object): コンテキスト
            intent (object): インテント
        """
        extras = intent.getExtras()
        type = "".join(extras.get('type'))
        try:
            data = json.loads("".join(extras.get('data')))
            if type == Type.ERASE_REQ.value:
                self.eraseData(data)
            elif type == Type.SYNC_REQ.value:
                self.syncData(data)
            elif type == Type.RECONNECT_REQ.value:
                self.reconnect(data)
            elif type == Type.STOP_SOUND.value:
                self.stopSoundVibration(data)
            elif type == Type.SOUND_CONFIG_UPDATED.value:
                self.soundConfigUpdated(data)
            elif type == Type.REQUEST_CONN_STATUS.value:
                self.notifyConnStatus()

        except:
            self.logger.error("on_broadcast failed.", exc_info=True)


    def initVibrator(self):
        """バイブ初期化
        """
        if util.is_android_device():
            Context = autoclass('android.content.Context')
            service = autoclass('org.kivy.android.PythonService').mService
            self.vibrator = cast("android.os.Vibrator", service.getSystemService(Context.VIBRATOR_SERVICE))

    def vibrate(self, mill, loop):
        """バイブ
        
        Args:
            mill (int): 振動時間（ミリ秒）
            loop (bool): ループ
        """
        if util.is_android_device():

            self.stopVibrate()
            BuildVersion  = autoclass('android.os.Build$VERSION')
            BuildVersionCodes  = autoclass('android.os.Build$VERSION_CODES')
            # if BuildVersion.SDK_INT >= BuildVersionCodes.O:
            if BuildVersion.SDK_INT >= 26:  # BuildVersionCodes.O Android 8
                VibrationEffect = autoclass('android.os.VibrationEffect')
                if loop:
                    vibrationEffect = VibrationEffect.createWaveform([mill, mill], [VibrationEffect.DEFAULT_AMPLITUDE, 0], 0)
                else:
                    vibrationEffect = VibrationEffect.createOneShot(mill, VibrationEffect.DEFAULT_AMPLITUDE)

                self.vibrator.vibrate(vibrationEffect)
            else:
                if loop:
                    self.vibrator.vibrate([0, mill, mill, mill], 2)
                else:
                    self.vibrator.vibrate(mill)
    def stopVibrate(self):
        if self.vibrator is not None:
            self.vibrator.cancel()

    def initSound(self):
        self.mediaPlayers = {}
        self.mediaPlayersForLoop = {}
        self.nowPlayingMediaPlayer = None
        self.nowPlayingSoundNo = None
        if util.is_android_device():
            MediaPlayer = autoclass('android.media.MediaPlayer')
            AudioManager = autoclass('android.media.AudioManager')
            with open(pathlib.Path("./sound.json").resolve()) as f:
                soundList = json.loads(f.read())
                for s in soundList:
                    mPlayer = MediaPlayer()
                    mPlayer.setDataSource(s["path"])
                    mPlayer.setAudioStreamType(AudioManager.STREAM_NOTIFICATION)
                    mPlayer.prepare()
                    self.mediaPlayers[s["no"]] = mPlayer

            with open(pathlib.Path("./sound_loop.json").resolve()) as f:
                soundList = json.loads(f.read())
                for s in soundList:
                    mPlayer = MediaPlayer()
                    mPlayer.setDataSource(s["path"])
                    mPlayer.setAudioStreamType(AudioManager.STREAM_NOTIFICATION)
                    mPlayer.prepare()
                    self.mediaPlayersForLoop[s["no"]] = mPlayer

    def playSound(self, soundNo, loop):
        """音声鳴動
        
        Args:
            soundNo (string): 音番号
            loop (boolean): ループ再生
        """
        volume = float(self.config.volume) / 10.0
        if loop == False and soundNo in self.mediaPlayers:
            self.mediaPlayers[soundNo].setVolume(volume, volume)
            self.mediaPlayers[soundNo].start()

        elif loop and soundNo in self.mediaPlayersForLoop:
            if soundNo != self.nowPlayingSoundNo and self.nowPlayingMediaPlayer is not None:
                self.nowPlayingMediaPlayer.stop()
                self.nowPlayingMediaPlayer.prepare()

            if self.nowPlayingMediaPlayer == None or self.nowPlayingMediaPlayer.isPlaying() == False:
                self.mediaPlayersForLoop[soundNo].setVolume(volume, volume)
                self.mediaPlayersForLoop[soundNo].setLooping(True)
                self.mediaPlayersForLoop[soundNo].start()
                self.nowPlayingMediaPlayer = self.mediaPlayersForLoop[soundNo]
                self.nowPlayingSoundNo = soundNo

    def stopAllSoundVibration(self):
        """呼出音とバイブを止める
        """
        if self.nowPlayingMediaPlayer is not None:
            self.nowPlayingMediaPlayer.stop()
            self.nowPlayingMediaPlayer.prepare()
            self.nowPlayingMediaPlayer = None

        self.stopVibrate()

    def stopSoundVibration(self, data):
        for c in self.callList:
            if c["address"] == data["address"] and c["modelNo"] == data["modelNo"] and c["buttonNo"] == data["buttonNo"]:
                c["checked"] = True

        self.playLoopSound()

    def playLoopSound(self):
        if self.config.loop:
            uncheckedCallList = [c for c in self.callList if c["checked"] == False]
            if len(uncheckedCallList) > 0:
                self.playSound(uncheckedCallList[0]["sound"], True)
                self.stopVibrate()
                if self.config.vibration:
                    self.vibrate(1000, True)
            else:
                self.stopAllSoundVibration()
        else:
            if self.nowPlayingMediaPlayer is not None:
                self.nowPlayingMediaPlayer.stop()
                self.nowPlayingMediaPlayer.prepare()
                self.nowPlayingMediaPlayer = None

        if self.config.vibration == False:
            self.stopVibrate()

    def loop_proc(self):
        """ループ処理
        """
        self.accept_request()
        while True:
            if self.socket is None:
                self.connect_server(self.config.server_ip, self.config.server_port)

            if self.status == Status.S3:
                self.register()

            if self.status == Status.S4:
                self.syncData({})

            if self.status >= Status.S5:
                self.health_check()

            self.recv_watcher()

            # if self.status == Status.S1:
            #     time.sleep(5)
            # else:
            time.sleep(0.5)

        self.disconnect()

    def recv_watcher(self):
        """受信データ監視
        """
        # self.logger.debug("status = {}".format(self.status))
        if self.socket is None:
            self.connect_server(self.config.server_ip, self.config.server_port)

        if self.socket is None:
            return True

        try:
            raw_data = bytearray(b'')
            end = False
            self.recv_lock.acquire()
            while end == False:
                self.socket.settimeout(0.1)
                _raw_data = self.socket.recv(1024)

                if not _raw_data:
                    self.logger.info("Closed connection to {}".format(self.socket.getpeername()))
                    self.disconnect()
                    return True

                self.logger.debug("recv data = {}".format(_raw_data.hex()))
                raw_data.extend(_raw_data)
                if raw_data[-1:] == b'\n':
                    end = True
            
            self.logger.info("recv data = {}".format(raw_data.hex()))
            self.handle_telegram(raw_data)

        except timeout:
            pass
            # self.logger.debug("timeout.", exc_info=True)
        except:
            self.logger.error("recv_watcher failed.", exc_info=True)
            self.disconnect()
        finally:
            self.recv_lock.release()
            return True

        return True

    def connect_server(self, ip_addr, port):
        """サーバ接続
        
        Args:
            ip_addr (string): IPアドレス
            port (int): ポート番号
        """        
        try:
            self.update_notification(self.resources.getResource("receiver_service:notification_connecting"))
            self.socket = socket(AF_INET, SOCK_STREAM)
            self.socket.settimeout(3)
            self.socket.connect((ip_addr, port))
            self.status = Status.S2
            send_broadcast_response(Type.NOTIFY_CONN_STATUS, {"status": int(Status.S2)})
            self.logger.info("status = {}".format(self.status))

        except timeout:
            self.logger.error("connect_server timeout.")
            self.disconnect()
        except:
            self.logger.error("connect_server failed.", exc_info=True)
            self.disconnect()

    def disconnect(self):
        """サーバ切断
        """
        if self.socket is not None:
            self.socket.close()
            self.socket = None

        self.status = Status.S1
        send_broadcast_response(Type.NOTIFY_CONN_STATUS, {"status": int(Status.S1)})
        self.logger.info("status = {}".format(self.status))

    def handle_telegram(self, raw_data):
        """電文ハンドリング
        
        Args:
            raw_data (bytes): 電文データ
        
        Returns:
            dict: 解析後電文データ
        """
        recv_data = []
        pos = 0
        while True:
            idx = raw_data.find(b'\r\n', pos)
            if idx >= 0:
                _raw_data = raw_data[pos:idx+2]
                t = Telegram()
                if t.pre_check_telegram(_raw_data) == False:
                    continue

                _recv_data = t.parse_recv_data(_raw_data)

                recv_data.append(_recv_data)

                if _recv_data["telgNo"] == Telegram.TELEG_SYSINFO:
                    self.handle_sysinfo(_recv_data)
                elif _recv_data["telgNo"] == Telegram.TELEG_REGISTER:
                    self.handle_register(_recv_data)
                elif _recv_data["telgNo"] == Telegram.TELEG_HEALTH_CHECK:
                    self.handle_health_check(_recv_data)
                elif _recv_data["telgNo"] == Telegram.TELEG_CALL_DATA:
                    self.handle_call_data(_recv_data)
                self.handle_time_diff(_recv_data)
                pos = idx + 2
            else:
                break

        return recv_data

    def handle_time_diff(self, recv_data):
        """時刻ずれ検知処理

        Args:
            recv_data (dict): 受信データ
        """
        if "timeDiff" in recv_data:
            send_broadcast_response(Type.NOTIFY_TIME_DIFF, {"timeDiff": recv_data["timeDiff"]})

    def handle_sysinfo(self, recv_data):
        """受信機情報通知受信後処理
        
        Args:
            recv_data (dict): 受信データ
        Returns:
            dict: 受信データ
        """
        if self.status == Status.S2:
            self.config.logLevel = recv_data["logLevel"]
            self.config.groups = recv_data["groups"]
            self.status = Status.S3
            send_broadcast_response(Type.NOTIFY_CONN_STATUS, {"status": int(Status.S3)})
            self.logger.info("status = {}".format(self.status))
        else:
            self.logger.error("status error. status = {}".format(self.status))
        return recv_data

    def handle_register(self, recv_data):
        """登録応答電文(R)受信後処理
        
        Args:
            recv_data (dict): 受信データ
        Returns:
            dict: 受信データ
        """
        if self.status == Status.S3:
            self.status = Status.S4
            send_broadcast_response(Type.NOTIFY_CONN_STATUS, {"status": int(Status.S4)})
            self.logger.info("status = {}".format(self.status))
            self.update_notification(self.resources.getResource("receiver_service:notification_receiving"))
        else:
            self.logger.error("status error. status = {}".format(self.status))
        return recv_data

    def handle_health_check(self, recv_data):
        """ヘルスチェック受信後処理
        
        Args:
            recv_data (dict): 受信データ
        
        Returns:
            dict: 受信データ
        """
        if self.status >= Status.S5:
            self.healthDt = datetime.now()
        else:
            self.logger.error("status error. status = {}".format(self.status))
        return recv_data

    def handle_call_data(self, data):
        """呼び出しデータ受信後処理
        
        Args:
            data (dict): 受信データ
        
        Returns:
            dict: 受信データ
        """
        if self.status >= Status.S5:
            if data["command"] == Telegram.COMMAND_S:
                # data["checked"] = self.isCheckedCallData(data)
                # send_broadcast_response(Type.CALL, data)
                self.playSound(data["sound"], False)
                if self.config.vibration:
                    self.vibrate(1000, self.config.loop)

                self.addCallList(data)
                data["checked"] = self.isCheckedCallData(data)
                send_broadcast_response(Type.CALL, data)

            elif data["command"] == Telegram.COMMAND_C:
                send_broadcast_response(Type.DEL, data)

                self.delCallList(data)

            self.playLoopSound()
            self.update_notification_calling(len(self.callList))


        else:
            self.logger.error("status error. status = {}".format(self.status))

        return data

    def handle_del_res(self):
        pass

    def register(self):
        """登録
        """
        def __proc_async():
            # send_broadcast_response(Type.SHOW_PROCESSING, {})
            groups = [g["groupId"] for g in self.config.groupList]
            name = self.config.name
            t = Telegram()
            send_data = t.create_telegram_register_D(groups, name,  self.get_seqNo())
            recv_data = self.send_sync(send_data)
            if recv_data is not None:
                self.handle_telegram(recv_data)
            else:
                self.disconnect()
            # send_broadcast_response(Type.HIDE_PROCESSING, {})

        if self.status == Status.S3:
            th = threading.Thread(target=__proc_async)
            th.start()



    def health_check(self):
        """ヘルスチェック
        """
        if self.healthDt is None or (datetime.now() - self.healthDt).seconds > 60:
            def __proc_async():
                # send_broadcast_response(Type.SHOW_PROCESSING, {})
                t = Telegram()
                groups = [g["groupId"] for g in self.config.groupList]
                send_data = t.create_telegram_healthcheck_D(groups, self.config.name, self.get_seqNo())
                recv_data = self.send_sync(send_data)
                if recv_data is not None:
                    self.handle_telegram(recv_data)
                else:
                    self.disconnect()
                # send_broadcast_response(Type.HIDE_PROCESSING, {})

            if self.status >= Status.S5:
                th = threading.Thread(target=__proc_async)
                th.start()

    def send_async(self, socket, data, callback=None):
        """非同期送信
        
        Args:
            socket (object): ソケット
            data (bytes): 送信データ
            callback (function, optional): コールバック関数. Defaults to None.
        """
        if socket is None:
            if callback is not None:
                callback(False)
            return

        def __proc_async():
            result = True
            try:
                self.logger.info("send data = {}".format(data.hex()))
                socket.sendall(data)
            except:
                self.logger.error("send_async failed.", exc_info=True)
                result = False
            finally:
                if callback is not None:
                    callback(result)

        th = threading.Thread(target=__proc_async)
        th.start()


    def send_sync(self, send_data):
        """同期送信
        
        Args:
            send_data (bytes): 送信データ
        
        Returns:
            bytes: 受信データ
        """
        if socket is None:
            return None

        flg = False
        cnt = 0
        recv_data = None
        end = False
        self.recv_lock.acquire()
        while flg == False:
            try:
                recvarr = bytearray(b'')
                self.logger.info("send data = {}".format(send_data.hex()))
                self.socket.settimeout(3)
                self.socket.sendall(send_data)
                
                while end == False:
                    _recv_data = self.socket.recv(1024)
                    if len(_recv_data) > 0:
                        self.logger.debug("recv data = {}".format(_recv_data.hex()))
                        # recvarr += _recv_data
                        recvarr.extend(_recv_data)
                        if recvarr[-1:] == b'\n':
                            end = True

                recv_data = recvarr
                self.logger.info("recv data = {}".format(recv_data.hex()))
                flg = True
            except timeout:
                self.logger.info("send_sync timeout.", exc_info=True)
            except:
                self.logger.error("send_sync failed.", exc_info=True)
                self.disconnect()
                flg = True
            finally:
                cnt += 1

            if cnt >= 3:
                flg = True
        self.recv_lock.release()

        return recv_data

    def get_seqNo(self):
        """シーケンス番号取得
        
        Returns:
            int: シーケンス番号
        """
        # TODO: 排他処理？
        self.seqNo += 1
        if self.seqNo > 255:
            self.seqNo = 1
        return self.seqNo

    def eraseData(self,  reqData):
        """削除（非同期）
        
        Args:
            reqData (dict): 削除データ
        """
        def __proc_async():
            # send_broadcast_response(Type.SHOW_PROCESSING, {})
            t = Telegram()
            send_data = t.create_telegram_del_data_D(Telegram.COMMAND_C, reqData, self.config.name, self.get_seqNo())
            recv_data = self.send_sync(send_data)
            ret = False if recv_data is None else True 
            if ret:
                self.handle_telegram(recv_data)
            else:
                self.disconnect()

            resData = {"result": ret, "data": reqData}
            send_broadcast_response(Type.ERASE_RES, resData)
            # send_broadcast_response(Type.HIDE_PROCESSING, {})

        if self.status >= Status.S5:
            th = threading.Thread(target=__proc_async)
            th.start()

    def syncData(self, reqData):
        """データ同期(非同期実行)
        
        Args:
            reqData (dict): 要求データ
        """

        def __proc_async():
            send_broadcast_response(Type.SHOW_PROCESSING, {})
            t = Telegram()
            groups = [g["groupId"] for g in self.config.groupList]
            send_data = t.create_telegram_sync_req_D(groups, self.config.name, self.get_seqNo())
            recv_data = self.send_sync(send_data)
            if recv_data is not None:
                parsed_recv_datalist = self.handle_telegram(recv_data)
                for resData in parsed_recv_datalist:
                    if "callList" in resData and "delList" in resData:
                        for data in resData["callList"]:
                            data["checked"] = self.isCheckedCallData(data)

                        self.logger.debug("resData = {}".format(resData))

                        send_broadcast_response( Type.SYNC_RES, resData)
                        if self.status == Status.S4:
                            self.status = Status.S5 
                            send_broadcast_response(Type.NOTIFY_CONN_STATUS, {"status": int(Status.S5)})
                            self.logger.info("status = {}".format(self.status))
                        
                        self.callList.clear()
                        self.callList.extend(resData["callList"])

                        self.playLoopSound()
                        self.update_notification_calling(len(self.callList))
            else: 
                self.disconnect()
            send_broadcast_response(Type.HIDE_PROCESSING, {})

        if self.status >= Status.S4:
            th = threading.Thread(target=__proc_async)
            th.start()

    def reconnect(self, reqData):
        """再接続
        
        Args:
            reqData (dict): 要求データ
        """
        if self.status >= Status.S2:
            self.disconnect()
            send_broadcast_response(Type.HIDE_PROCESSING, {})

    def addCallList(self, data):
        if len([c for c in self.callList if c["address"] == data["address"] and c["modelNo"] == data["modelNo"] and c["buttonNo"] == data["buttonNo"]]) == 0:
            data["checked"] = False
            self.callList.append(data)
        else:
            for c in self.callList:
                if c["address"] == data["address"] and c["modelNo"] == data["modelNo"] and c["buttonNo"] == data["buttonNo"]:
                    c["checked"] = False

    def delCallList(self, data):
        self.callList = [c for c in self.callList if not (c["address"] == data["address"] and c["modelNo"] == data["modelNo"] and c["buttonNo"] == data["buttonNo"])]

    def isCheckedCallData(self, data):
        return len([c for c in self.callList if c["address"] == data["address"] and c["modelNo"] == data["modelNo"] and c["buttonNo"] == data["buttonNo"] and c["checked"]]) > 0
    
    def soundConfigUpdated(self,data):
        self.stopAllSoundVibration()
        self.playLoopSound()

    def notifyConnStatus(self):
        send_broadcast_response(Type.NOTIFY_CONN_STATUS, {"status": self.status})


if __name__ == '__main__':
    logger = setup_logger_devel(__name__)
    try:
        # param = os.environ.get('PYTHON_SERVICE_ARGUMENT', '{}')
        # logger.debug(param)


        logger.info("receiver_service main start.", exc_info=True)
        config = ConfigSetting()
        watchServer = WatchServer(config)

        watchServer.loop_proc()

    except:
        logger.error("receiver_service main failed.", exc_info=True)

    finally:
        watchServer.release()

    logger.info("receiver_service main end.", exc_info=True)

