import os

import common_util as util
from setup_logger import setup_logger_devel

if util.is_android_device():
    from android_dev.receiver import Receiver
elif util.is_receiver_device():
    from recv_dev.receiver import Receiver
    from recv_dev.storage import Caller, CallerButton, Eraser
    from kivy.app import App
    import subprocess

from kivy.clock import Clock

class Controller():
    """
    Controller
    """
    def __init__(self, screen):
        """
        コンストラクタ
        @param screen スクリーン
        """ 
        self.logger = setup_logger_devel(self.__class__.__name__)
        self.screen = screen
        self.screen.bind(on_request_erase=lambda __, data, *args: self.requestErase(data))
        self.receiver = Receiver(self.received, self.eraserReceived, self.syncFinished, self.pairing, self.received_on_pairing, self.long_pressed_on_pairing, self.processing, self.shutdown, self.eraser_pushed)
        # self.receiver.bind(pairing=self.on_pairing_change)
        self.receiver.bind(disconnected=self.on_disconnected_change)
        self.receiver.bind(timeDiff=self.on_timediff_change)

    def start(self):
        """
        開始
        """
        #受信機開始
        self.receiver.start()

    def pause(self):
        """Pause
        """
        self.receiver.pause()

    def resume(self):
        """Resume
        """
        self.receiver.resume()      
        self.screen.addProcQueue(self.screen.clearCards, [])
        self.screen.addProcQueue(self.screen.clearHistCards, [])
        self.receiver.requestSync(self.syncFinished)

    def received(self, datalist):
        """受信機から呼び出し時に呼ばれるコールバック関数
        
        Args:
            datalist (list)): 受信データ（リスト） 
        """
        self.logger.debug("received datalist={}".format(datalist))
        # スクリーンにデータの表示を要求する
        # self.screen.addCards(datalist)
        self.screen.addProcQueue(self.screen.addCards, [datalist])

    def eraserReceived(self, datalist):
        """受信機から削除時に呼ばれるコールバック関数
        
        Args:
            datalist (list)): 削除データ（リスト） 
        """
        # スクリーンにデータの削除を要求する
        self.logger.debug("eraserReceived datalist={}".format(datalist))
        # self.screen.delCards(datalist)
        # self.screen.addHistCards(datalist)
        self.screen.addProcQueue(self.screen.delCards, [datalist])
        self.screen.addProcQueue(self.screen.addHistCards, [datalist])
    
    def requestErase(self, data):
        """削除要求
        
        Args:
            data (dictionary)): 削除データ
        """
        self.receiver.requestErase( data, self.eraseRequestFinished)

    def eraseRequestFinished(self, success, data):
        """削除完了
        
        Args:
            success (bool): 削除結果
            data (dictionary): 削除データ
        """
        pass

    def syncFinished(self, calllist, dellist):
        """同期完了
        
        Args:
            calllist (list): 呼び出しデータのリスト
            dellist (list): 削除データのリスト
        """
        # self.screen.clearCards()
        # self.screen.clearHistCards()
        # self.screen.addCards(calllist)
        # self.screen.addHistCards(dellist)
        self.screen.addProcQueue(self.screen.clearCards, [])
        self.screen.addProcQueue(self.screen.clearHistCards, [])
        self.screen.addProcQueue(self.screen.addCards, [calllist])
        self.screen.addProcQueue(self.screen.addHistCards, [dellist])

    def stop(self):
        """終了
        """
        self.receiver.stop()

    def pairing(self):
        """ペアリングボタン押下時コールバック
        """
        if util.is_receiver_device():
            if self.screen.manager.current_screen.name == 'mainscreen':
                # ペアリングモードに遷移する
                self.receiver.pairing = True
                self.dispPairingScreen()

            elif self.screen.manager.current_screen.name == 'pairingscreen':
                pairingScreen = self.screen.manager.get_screen('pairingscreen')
                if pairingScreen.isShowingDupMsg():
                    pairingScreen.popupCloseDuplication()
                else:
                    self.receiver.pairing = False
                    self.dispMainScreen()


    def dispPairingScreen(self):
        """ペアリングモード画面表示
        """
        if util.is_receiver_device():
            self.screen.manager.current = 'pairingscreen'

    def dispMainScreen(self):
        """メイン画面表示
        """
        if util.is_receiver_device():
            self.screen.manager.current = 'mainscreen'

    def received_on_pairing(self, registered, strength, data, isCaller, registerdOtherType=False):
        """ペアリング時、既存の呼出機から呼び出しされたときに呼ばれるコールバック関数
        
        Args:
            registered (bool):登録済み（重複）
            strength (int):電波強度
            data (dict): 呼出機(消去機)データ 
            isCaller(bool):呼出機か？(True:呼出機, False:消去機)
            registerdOtherType:他のタイプで登録済み
        """
        self.logger.debug(self.screen.manager.current_screen)
        if self.screen.manager.current_screen.name == 'pairingscreen':
            pairingScreen = self.screen.manager.get_screen('pairingscreen')
            print(data)
            if registerdOtherType:
                # 消去機(呼出機)として登録済み
                pairingScreen.popupOpenOtherType(data, strength)
            else:
                if isCaller == pairingScreen.callerRegisterMode:
                    if registered:
                        # ペアリング画面の重複ダイアログを表示する。
                        pairingScreen.popupOpenDuplication(data, strength)
                    else:
                        # ペアリング画面の登録ダイアログを表示する。
                        pairingScreen.popupOpenRegister(data, strength)

    def long_pressed_on_pairing(self):
        """ペアリング時にクリアボタンが長押しされたときに呼ばれるコールバック関数
        """
        if self.screen.manager.current_screen.name == 'pairingscreen':
            pairingScreen = self.screen.manager.get_screen('pairingscreen')
            address = pairingScreen.getDeleteAddress()
            if len(address) > 0:
                if pairingScreen.callerRegisterMode:
                    # 呼出機を削除する
                    c = Caller()
                    cb = CallerButton()

                    key = {"address": address}
                    c.delete(key)
                    cb.delete(key)
                else:
                    # 消去機を削除する
                    e = Eraser()

                    key = {"address": address}
                    e.delete(key)

            pairingScreen.popupCloseDuplication()

    def processing(self, show):
        """処理中表示コールバック
        
        Args:
            show (bool): True=表示、False=非表示
        """
        if show:
            self.screen.showProcessing()
        else:
            self.screen.hideProcessing()

    def shutdown(self):
        """シャットダウンコールバック
        """
        if util.is_receiver_device():
            self.screen.manager.current = 'shutdownscreen'
            Clock.schedule_once(self._shutdown, 3)

    def _shutdown(self, dt):
        """シャットダウン実行
        """
        App.get_running_app().stop()
        subprocess.run(["poweroff"])

    def on_disconnected_change(self, instance, disconnected):
        """接続状態変更イベントハンドラー
        
        Args:
            instance (object): イベント発生オブジェクト
            disconnected (bool): 接続状態(True:切断)
        """
        self.screen.disconnected = disconnected

    def on_timediff_change(self, instance, timeDiff):
        """時刻ずれイベントハンドラー

        Args:
            instance (object): イベント発生オブジェクト
            timeDiff (bool): 表示機と受信機の時刻ずれ（True:あり）
        """
        self.screen.timeDiff = timeDiff

    def eraser_pushed(self, eraserName, buttonNo):
        """消去機押下コールバック

        Args:
            eraserName (string): 消去機名
            buttonNo (string): ボタン番号
        """
        self.screen.eraserPushed(eraserName, buttonNo)
