import json

import common_util as util
from kivy.config import ConfigParser


class ConfigSetting():
    """設定
    """

    def __init__(self):
        """コンストラクタ
        """
        self.config = None

    def get_config_path(self):
        """INIファイルのパス取得
        
        Returns:
            string: INIファイルのパス
        """
        if util.is_android_device():
            return "../.ecomonitor.ini"
        else:
            return "ecomonitor.ini"

    def load(self):
        """読み込み
        """
        if self.config is None:
            self.config = ConfigParser()
            self.config.read(self.get_config_path())

    def reload(self):
        """再読み込み
        """
        if self.config is None:
            self.config = ConfigParser()

        self.config.read(self.get_config_path())

    def getDefaultConfig(self):
        """デフォルト値取得
        
        Returns:
            tuple: デフォルト値
        """

        default_config = ('basic', {
            'name': 'My Name',
            'server_ip': '',
            'server_port': '5001',
            'udp_port': '5000',
            'http_port': '80',
            'volume': '10',
            'vibration': '0',
            'loop': '0',
            'melody': '0',
            'group': '',
            'language': util.get_system_language(),
            'groups': '[]',
            'logLevel': '10',
        })

        return default_config

    def getValue(self, section, option, default):
        """値取得
        
        Args:
            section (string)): セクション
            option (string): オプション
            default (string): デフォルト値
        
        Returns:
            string: 値
        """
        self.reload()
        return self.config.getdefault(section, option, default)

    def getValueInt(self, section, option, default):
        """値取得
        
        Args:
            section (string)): セクション
            option (string): オプション
            default (int): デフォルト値
        
        Returns:
            int: 値
        """
        self.reload()
        return self.config.getdefaultint(section, option, default)

    def getBasicValue(self, option, default):
        """値取得(basicセクション)
        
        Args:
            option (string): オプション
            default (string): デフォルト値
        
        Returns:
            string: 値
        """
        return self.getValue("basic", option, default)

    def getBasicValueInt(self, option, default):
        """値取得(basicセクション)
        
        Args:
            option (string): オプション
            default (int): デフォルト値
        
        Returns:
            int: 値
        """
        return self.getValueInt("basic", option, default)

    def setValue(self, section, option, value):
        """値設定
        
        Args:
            section (string)): セクション
            option (string): オプション
            value (string): 値
        """
        self.load()
        self.config.set(section, option, value)
        self.config.write()

    def setBasicValue(self, option, value):
        """値設定(basicセクション)
        
        Args:
            option (string): オプション
            value (string): 値
        """
        self.setValue("basic", option, value)

    @property
    def name(self):
        """名称
        
        Returns:
            string: 名称
        """
        return self.getBasicValue("name", None)

    @name.setter
    def name(self, v):
        """名称
        
        Args:
            v (string): 設定値
        """
        self.setBasicValue("name", v)

    @property
    def server_ip(self):
        """サーバIPアドレス
        
        Returns:
            string: サーバIPアドレス
        """
        return self.getBasicValue("server_ip", None)

    @server_ip.setter
    def server_ip(self, v):
        """サーバIPアドレス
        
        Args:
            v (string): 設定値
        """
        self.setBasicValue("server_ip", v)

    @property
    def server_port(self):
        """サーバポート番号
        
        Returns:
            int: サーバポート番号
        """
        return self.getBasicValueInt("server_port", 5001)

    @server_port.setter
    def server_port(self, v):
        """サーバポート番号
        
        Args:
            v (int): 設定値
        """
        self.setBasicValue("server_port", v)

    @property
    def udp_port(self):
        """UDPポート番号
        
        Returns:
            int: UDPポート番号
        """
        return self.getBasicValueInt("udp_port", 5000)

    @udp_port.setter
    def udp_port(self, v):
        """UDPポート番号
        
        Args:
            v (int): 設定値
        """
        self.setBasicValue("udp_port", v)


    @property
    def http_port(self):
        """HTTPポート番号
        
        Returns:
            int: HTTPポート番号
        """  
        return self.getBasicValueInt("http_port", 8080)

    @http_port.setter
    def http_port(self, v):
        """HTTPポート番号
        
        Args:
            v (int): 設定値
        """
        self.setBasicValue("http_port", v)

    @property
    def volume(self):
        """音量
        
        Returns:
            int: 音量
        """
        return self.getBasicValueInt("volume", 10)

    @volume.setter
    def volume(self, v):
        """音量
        
        Args:
            v (int): 設定値
        """
        self.setBasicValue("volume", v)

    @property
    def vibration(self):
        """バイブレーション
        
        Returns:
            bool: バイブレーション有無
        """
        return True if self.getBasicValueInt("vibration", 0) == 1 else False

    @vibration.setter
    def vibration(self, v):
        """バイブレーション
        
        Args:
            v (bool): 設定値
        """
        if v == True:
            self.setBasicValue("vibration", "1")
        else:
            self.setBasicValue("vibration", "0")

    @property
    def loop(self):
        """呼出音・バイブレーションループ
        
        Returns:
            bool: ループ有無
        """
        return True if self.getBasicValueInt("loop", 0) == 1 else False

    @loop.setter
    def loop(self, v):
        """呼出音・バイブレーションループ
        
        Args:
            v (bool): 設定値
        """
        if v == True:
            self.setBasicValue("loop", "1")
        else:
            self.setBasicValue("loop", "0")

    @property
    def melody(self):
        """メロディー
        
        Returns:
            int: メロディーファイルのindex
        """
        return self.getBasicValueInt("melody", 0)

    @melody.setter
    def melody(self, v):
        """メロディー
        
        Args:
            v (int): 設定値
        """
        self.setBasicValue("melody", v)


    @property
    def group(self): 
        """自身が所属しているグループ
        
        Returns:
            string: グループIDをカンマ区切りで列挙したもの(ex: 1,2,3)
        """
        return self.getBasicValue("group", "")

    @group.setter
    def group(self, v):
        """自身が所属しているグループ
        
        Args:
            v (string): グループIDをカンマ区切りで列挙したもの(ex: 1,2,3)
        """
        self.setBasicValue("group", v)

    @property
    def groupList(self):
        """自身が所属しているグループのリスト
        
        Returns:
            list: グループリスト
                [
                    {"groupId":1, "name":"Aグループ"},
                    {"groupId":2, "name":"Bグループ"},
                ]
        """
        groupList = []
        myGroups = self.group.split(",")
        for g in self.groups:
            if len([mg for mg in myGroups if mg==str(g["groupId"])]) > 0:
                groupList.append(g)

        allGroup = '全て' if self.language =="ja" else 'All Group'
        if len(groupList) == 0:
            groupList.append({
                "groupId": 0,
                "name": allGroup
            })

        return groupList


    @property
    def language(self):
        """言語
        
        Returns:
            string: 言語
        """
        return self.getBasicValue("language", util.get_system_language())

    @language.setter
    def language(self, v):
        """言語
        
        Args:
            v (string): 言語
        """
        self.setBasicValue("language", v)

    @property
    def groups(self):
        """受信機に登録してあるグループ
        
        Returns:
            list: グループリスト
                [
                    {"groupId":1, "name":"Aグループ"},
                    {"groupId":2, "name":"Bグループ"},
                ]
        """
        return json.loads(self.getBasicValue("groups", "[]"))

    @groups.setter
    def groups(self, v):
        """受信機に登録してあるグループ
        
        Args:
            v (list): グループリスト
                [
                    {"groupId":1, "name":"Aグループ"},
                    {"groupId":2, "name":"Bグループ"},
                ]
        """
        self.setBasicValue("groups", json.dumps(v))

    @property
    def logLevel(self):
        """ログレベル
        
        Returns:
            int: ログレベル
                RITICAL 50
                ERROR   40
                WARNING 30
                INFO    20
                DEBUG   10
                NOTSET  0
        """
        return self.getBasicValueInt("logLevel", 10)

    @logLevel.setter
    def logLevel(self, v):
        """ログレベル
        
        Args:
            v (int): ログレベル
                RITICAL 50
                ERROR   40
                WARNING 30
                INFO    20
                DEBUG   10
                NOTSET  0
        """
        self.setBasicValue("logLevel", v)
