import json
import pathlib
import re
import threading
from socket import AF_INET, SOCK_STREAM, socket, timeout

import netifaces

import common_util as util
import disp_dev.send_udp as udp
from disp_dev.config_setting import ConfigSetting
from kivy.base import EventLoop
from kivy.graphics import Color
from kivy.metrics import dp, sp
from kivy.properties import (BooleanProperty, ListProperty, NumericProperty,
                             StringProperty)
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.recycleview import RecycleView
from kivy.uix.screenmanager import Screen
from kivy.uix.textinput import TextInput
from resources import Resources
from setup_logger import setup_logger_devel
from telegram import Telegram

if util.is_android_device():
    from android_dev.sendbroadcast import Type, send_broadcast_request
    from jnius import autoclass, cast

class BaseConfigScreen(Screen):
    """設定画面基底クラス
    """
    background_color =ListProperty([0,0,0,0])
    headerHeight = NumericProperty(1)

    def __init__(self, **kwargs):
        """コンストラクタ
        """
        super(BaseConfigScreen, self).__init__(**kwargs)
        self.logger = setup_logger_devel(self.__class__.__name__)
        self.setAppHeight(self.isLandscape(), self.size)
        self.bind(size=self.on_size)

    def on_size(self, instance, new_size):
        """on_size
        
        Args:
            instance (object): イベント発生オブジェクト
            new_size (tuple): サイズ
        """
        self.setAppHeight(self.isLandscape(), new_size)

    def isLandscape(self):
        """横表示判定
        
        Returns:
            bool: 判定結果
        """
        if self.size[0] >= self.size[1]:
            return True
        else:
            return False

    def setAppHeight(self, landscape, size):
        """スクリーンのサイズに合わせてヘッダー等のサイズを設定する

        Args:
            landscape (bool): 横長か？
            size (tuple): Screenのサイズ
        """
        if landscape:
            self.headerHeight = size[1] * 0.133
        else:
            self.headerHeight = size[1] * 0.075

    def getRgba(self, color, alpha=None):
        """RGBA取得
        
        Args:
            color (string): #ffffff形式の色コード
            alpha (float, optional): 不透明度. Defaults to None.
        
        Returns:
            list: RGBA
        """
        rr = int(color[1:3], 16) / 255.0
        gg = int(color[3:5], 16) / 255.0
        bb = int(color[5:7], 16) / 255.0
        a = 1
        if alpha is not None:
            a = alpha
        return [rr,gg,bb,a]

class ConfigScreen(BaseConfigScreen):
    """設定画面
    """

    termName = StringProperty("")
    ipAddress = StringProperty("")
    serverIpAddress = StringProperty("")
    volume = NumericProperty(0)
    group = StringProperty("")

    lang = StringProperty("")
    res = Resources()
    langDict = {
        "ja": {
            "英語(English)":"en",
            "日本語(Japanese)":"ja"
        },
        "en": {
            "English":"en",
            "Japanese":"ja"
        }
    }
    

    def __init__(self, **kwargs):
        """コンストラクタ
        """
        super(ConfigScreen, self).__init__(**kwargs)
        self.configSetting = ConfigSetting()
        self.__org_group = ""
        self.updateProp()

    def on_enter(self):
        """画面表示時に実行される
        """
        self.__org_group = self.configSetting.group
        self.updateProp()
        self.updateGroups()
        self.getReceiverSysInfo()

    def updateProp(self):
        """プロパティ更新（設定値を画面に反映）
        """
        self.termName = self.configSetting.name
        self.serverIpAddress = self.configSetting.server_ip

        self.volume = self.configSetting.volume

        self.lang = self.configSetting.language

        # addresses = netifaces.ifaddresses('wlan0')
        # if netifaces.AF_INET in addresses:
        #     self.ipAddress = addresses[netifaces.AF_INET][0]['addr']
        try:
            # ipAddress, subnetMask, router = self.getNetworkInfo()
            ipAddress, _, _ = getNetworkInfo()
            self.ipAddress = ipAddress
        except:
            self.logger.error("exception", exc_info=True)


    def updateConfig(self):
        """設定値を更新
        """
        group = []
        for d in filter(lambda d: d["checked"] == True, self.ids.groupRv.data):
            if d["groupId"] == 0:
                self.configSetting.group=""
                break
            else:
                group.append(str(d["groupId"]))
        if len(group) > 0:
            self.configSetting.group = ",".join(group)

        if self.__org_group != self.configSetting.group:
            if util.is_android_device():
                # 同期要求
                send_broadcast_request(Type.SYNC_REQ, {})


    def commtest(self):
        """通信テスト
        """

        def __async_proc():
            s = socket(AF_INET, SOCK_STREAM)
            s.settimeout(3)

            msg = ""
            try:
                s.connect((self.configSetting.server_ip, self.configSetting.server_port))
            except timeout:
                msg = self.getCommTestErrorMessage("configscreen:error_connect_timeout")
                self.logger.info(msg, exc_info=True)
            except:
                msg = self.getCommTestErrorMessage("configscreen:error_connect")
                self.logger.error(msg, exc_info=True)

            try:
                if len(msg) == 0:

                    # 受信機情報通知を受信する
                    raw_data = b''
                    end = False
                    while end == False:
                        _raw_data = s.recv(1024)

                        if not _raw_data:
                            self.logger.info("Closed connection to {}".format(s.getpeername()))
                            break

                        self.logger.debug("recv data = {}".format(_raw_data.hex()))
                        raw_data += _raw_data
                        if raw_data[-1:] == b'\n':
                            end = True
                    
                    s.close()

                    self.logger.info("recv data = {}".format(raw_data.hex()))
 
                    t = Telegram()
                    if t.pre_check_telegram(raw_data):
                        data = t.parse_recv_data(raw_data)
                        if data["telgNo"] == t.TELEG_SYSINFO:
                            # self.configSetting.language = data["language"]
                            # self.configSetting.groups = data["groups"]
                            # self.updateGroups()
                            msg = self.res.getResource("configscreen:commtest_success")
                        else:
                            msg = self.getCommTestErrorMessage("configscreen:error_telgNo")
                    else:
                        msg = self.getCommTestErrorMessage("configscreen:error_checkNG")

            except timeout:
                msg = self.getCommTestErrorMessage("configscreen:error_recv_timeout")
                self.logger.info(msg, exc_info=True)
            except:
                msg = self.getCommTestErrorMessage("configscreen:error_recv")
                self.logger.error(msg, exc_info=True)

            self.ids.commtestResultLbl.text = msg
            self.ids.processing.opacity = 0
            self.ids.processing.height = 0
            # self.ids.commtestResultLbl.height = dp(40)
            self.ids.commtestResultLbl.size = self.ids.commtestResultLbl.texture_size
            self.ids.commtestResultLbl.opacity = 1

        self.ids.processing.opacity = 1
        self.ids.processing.height = dp(20)

        self.ids.commtestResultLbl.text = ""
        self.ids.commtestResultLbl.height = 0
        self.ids.commtestResultLbl.opacity = 0
        th = threading.Thread(target=__async_proc)
        th.start()

    def getCommTestErrorMessage(self, resourceId):
        """通信テスト失敗メッセージ取得
        
        Args:
            resourceId (string): リソースID
        
        Returns:
            string: メッセージ
        """
        return self.res.getResource("configscreen:commtest_failed").format(self.res.getResource(resourceId))

    def getReceiverSysInfo(self):
        """受信機情報を取得する
        """

        def __async_proc():
            s = socket(AF_INET, SOCK_STREAM)
            s.settimeout(3)

            msg = ""
            try:
                s.connect((self.configSetting.server_ip, self.configSetting.server_port))
            except timeout:
                self.logger.info(msg, exc_info=True)
            except:
                self.logger.error(msg, exc_info=True)

            try:
                if len(msg) == 0:

                    # 受信機情報通知を受信する
                    raw_data = b''
                    end = False
                    while end == False:
                        _raw_data = s.recv(1024)

                        if not _raw_data:
                            self.logger.info("Closed connection to {}".format(s.getpeername()))
                            break

                        self.logger.debug("recv data = {}".format(_raw_data.hex()))
                        raw_data += _raw_data
                        if raw_data[-1:] == b'\n':
                            end = True
                    
                    s.close()

                    self.logger.info("recv data = {}".format(raw_data.hex()))
 
                    t = Telegram()
                    if t.pre_check_telegram(raw_data):
                        data = t.parse_recv_data(raw_data)
                        if data["telgNo"] == t.TELEG_SYSINFO:
                            self.configSetting.groups = data["groups"]
                            self.updateGroups()
                            msg = self.res.getResource("configscreen:commtest_success")
                        else:
                            msg = self.getCommTestErrorMessage("configscreen:error_telgNo")
                    else:
                        msg = self.getCommTestErrorMessage("configscreen:error_checkNG")

            except timeout:
                msg = self.getCommTestErrorMessage("configscreen:error_recv_timeout")
                self.logger.info(msg, exc_info=True)
            except:
                msg = self.getCommTestErrorMessage("configscreen:error_recv")
                self.logger.error(msg, exc_info=True)


        th = threading.Thread(target=__async_proc)
        th.start()

    def updateGroups(self):
        """グループ表示更新
        """
        data = [{"text": self.res.getResource("configscreen:allGroup"), "checked": False, "groupId": 0}]

        for g in self.configSetting.groups:
            data.append({"text": g["name"], "checked": False, "groupId": g["groupId"]})

        confGroup = self.configSetting.groupList

        if len(list(filter(lambda g: g["groupId"] == 0, confGroup))) > 0:
            #"全て"が選択
            for d in data:
                # 全てTrue
                d["checked"] = True

        else:
            for d in data:
                # 設定値に含まれる場合は✓
                cg = list(filter(lambda g: g["groupId"] ==  d["groupId"], confGroup))
                if len(cg) > 0:
                    d["checked"] = True
                

        self.ids.groupRv.data = data

    def open_websetting(self):
        """WEB設定を開く
        """
        if util.is_display_device():

            if util.is_android_device():
                # ブラウザを起動する
                PythonActivity = autoclass('org.renpy.android.PythonActivity')
                Intent = autoclass('android.content.Intent')
                Uri = autoclass('android.net.Uri')

                http_port = str(self.configSetting.http_port)
                if http_port == "80":
                    http_port = ""
                else:
                    http_port = ":" + http_port
                intent = Intent()
                intent.setAction(Intent.ACTION_VIEW)
                intent.setData(Uri.parse("http://"+ self.configSetting.server_ip + http_port + "/login.html"))
                currentActivity = cast('android.app.Activity', PythonActivity.mActivity)
                currentActivity.startActivity(intent)

    def getLanguageText(self, lang):
        """言語選択Spinnerに初期表示する値を取得する。

        Args:
            lang ([type]): [description]

        Returns:
            [type]: [description]
        """
        text = ""
        if not lang in self.langDict:
            lang = "ja"
        
        textList = [k for k, v in self.langDict[lang].items() if v == lang]
        if len(textList) > 0:
            text = textList[0]

        return text


    def getLanguageValues(self, lang):
        """言語選択Spinnerに表示する値のリストを取得する。

        Args:
            lang (string): 言語コード

        Returns:
            list: 言語選択Spinnerに表示する値のリスト
        """
        if not lang in self.langDict:
            lang = "ja"

        return sorted(list(self.langDict[lang].keys()))


    def selectLanguage(self, lang, text):
        """Spinnerで設定された言語を反映する。

        Args:
            lang (string): 言語コード
            text (string): Spinnerで選択された値
        """
        if not lang in self.langDict:
            lang = "ja"

        langList = [v for k, v in self.langDict[lang].items() if k == text]
        if len(langList) > 0:
            self.configSetting.language = langList[0]
            self.lang = langList[0]


class GroupSelectableLabel(BoxLayout):
    """グループ選択ラベル
    """
    index = None
    checked = BooleanProperty(False)
    text = StringProperty("")
    groupId = NumericProperty(0)

    def __init__(self, **kwargs):
        """コンストラクタ
        """
        super(GroupSelectableLabel, self).__init__(**kwargs)
        self.logger = setup_logger_devel(self.__class__.__name__)
        
    def on_touch_down(self, touch):
        """on_touch_down
        
        Args:
            touch (object): touch情報
        
        Returns:
            bool: Trueを返すとイベント処理終了
        """
        ''' Add selection on touch down '''
        if super(GroupSelectableLabel, self).on_touch_down(touch):
            return True
        if self.collide_point(*touch.pos):
            self.parent.parent.updateSelection({"text":self.text, "checked": self.checked, "groupId": self.groupId})
            return True

    def calcFontSize(self, strlen, size):
        """フォントサイズを計算する
        
        Args:
            strlen (int): 文字列長（半角）
            size (tuple): 表示領域のサイズ
        
        Returns:
            float: フォントサイズ
        """
        if strlen == 0:
            strlen = 1
        font_size = size[0] / strlen * 2
        font_size = int(font_size * 0.9)
        
        if font_size > sp(20):
            font_size = sp(20)
        return font_size


class GroupRV(RecycleView):
    """グループリサイクルビュー
    """
 
    # カスタムイベント
    __events__ = ('on_update_selection',  )

    def __init__(self, **kwargs):
        """コンストラクタ
        """
        super(GroupRV, self).__init__(**kwargs)
        self.logger = setup_logger_devel(self.__class__.__name__)

    def on_update_selection(self, *args):
        pass

    def updateSelection(self, selData):
        """選択
        
        Args:
            selData (dict): 選択データ
        """
        self.logger.debug('updateSelection({0})'.format(selData))
        if selData["groupId"] == 0:
            #全てを選択
            if selData["checked"] == False:
                # 全グループTrueにする。
                for d in self.data:
                    d["checked"] = True
            else:
                # False
                for d in self.data:
                    d["checked"] = False
        else:
            self.data[0]["checked"] = False
            for d in self.data:
                if d["groupId"] == selData["groupId"]:
                    d["checked"] = not d["checked"]
                    break

        self.refresh_from_data()

        self.dispatch('on_update_selection', selData)

    def getHeight(self, root_size, _):
        height = len(self.data) * dp(40)
        if height > root_size[1] / 2:
            height = root_size[1] / 2
        return height

class NameInputScreen(BaseConfigScreen):
    """名前入力画面
    """
    termName = StringProperty("")

    lang = StringProperty("")
    res = Resources()

    def __init__(self, **kwargs):
        """コンストラクタ
        """
        super(NameInputScreen, self).__init__(**kwargs)
        self.configSetting = ConfigSetting()
        self.updateProp()

    def on_enter(self):
        """画面表示時に実行される
        """
        self.updateProp()
        self.ids.termNameInput.bind_textedit()
        # self.ids.termNameInput.imetext = ""

    def on_leave(self):
        """画面削除時に実行される
        """
        self.ids.termNameInput.unbind_textedit()
        # self.ids.termNameInput.imetext = ""

    def updateProp(self):
        """プロパティ更新（設定値を画面に反映）
        """
        self.termName = self.configSetting.name
        self.lang = self.configSetting.language

    def updateConfig(self):
        """設定値を更新
        """
        self.configSetting.name = self.ids.termNameInput.text


class ServerIpConfigScreen(BaseConfigScreen):
    """受信機設定画面
    """
    serverIpAddress = StringProperty("")

    lang = StringProperty("")
    res = Resources()

    def __init__(self, **kwargs):
        """コンストラクタ
        """
        super(ServerIpConfigScreen, self).__init__(**kwargs)
        self.configSetting = ConfigSetting()
        self.updateProp()
        self.__org_server_ip = ""

    def on_enter(self):
        """画面表示時に実行される
        """
        self.__org_server_ip = self.configSetting.server_ip
        self.updateProp()

    def updateProp(self):
        """プロパティ更新（設定値を画面に反映）
        """
        self.serverIpAddress = self.configSetting.server_ip
        self.lang = self.configSetting.language

    def updateConfig(self):
        """設定値を更新
        """
        if self.__org_server_ip != self.configSetting.server_ip:
            if util.is_android_device():
                # 再接続要求
                send_broadcast_request(Type.RECONNECT_REQ, {})

    def findReceiver(self):
        """受信機検索
        """

        def __update_selection(instance, selData):
            m = re.match(r'(.*)\(.*\)', selData["text"])
            self.configSetting.server_ip = m.group(1)
            self.updateProp()

        self.ids.receiverRv.data = []
        self.ids.receiverRv.bind(on_update_selection=__update_selection)

        self.ids.findReceiverLbl1.opacity = 0
        self.ids.findReceiverLblErr.opacity = 0
        self.ids.findReceiverLblErr.height = 0
        self.ids.findReceiverLbl2.opacity = 0
        self.ids.processing.opacity = 1
        self.ids.processing.height = dp(40)

        def __async_proc():
            ipList = {}
            cnt = 0
            while cnt < 10 and len(ipList) == 0:
                ret = udp.request_server_ipaddress(port=self.configSetting.udp_port)
                if ret is not None:
                    ipList[ret["ipAddress"]] = ret["macAddress"]
                cnt += 1

            data = []
            for ip, mac in ipList.items():
                data.append({"text": ip + "(" + mac + ")", "checked": False})
            
            self.ids.receiverRv.data = data

            if len(data) > 0:
                self.ids.findReceiverLbl1.opacity = 1
                self.ids.findReceiverLbl2.opacity = 1
                self.ids.findReceiverLblErr.opacity = 0
                self.ids.findReceiverLblErr.height = 0
            else:
                self.ids.findReceiverLbl1.opacity = 1
                self.ids.findReceiverLbl2.opacity = 0
                self.ids.findReceiverLblErr.opacity = 1
                self.ids.findReceiverLblErr.height = dp(40)

            self.ids.processing.opacity = 0
            self.ids.processing.height = 0

        th = threading.Thread(target=__async_proc)
        th.start()


class ReceiverRV(RecycleView):
    """受信機リサイクルビュー
    """

    # カスタムイベント
    __events__ = ('on_update_selection',  )

    def __init__(self, **kwargs):
        """コンストラクタ
        """
        super(ReceiverRV, self).__init__(**kwargs)
        self.logger = setup_logger_devel(self.__class__.__name__)

    def on_update_selection(self, *args):
        pass

    def updateSelection(self, selData):
        """選択
        
        Args:
            selData (dict): 選択データ
        """
        self.logger.debug('updateSelection({0})'.format(selData))
        for d in self.data:
            if d["text"] == selData["text"]:
                d["checked"] = True
            else:        
                d["checked"] = False

        self.refresh_from_data()

        self.dispatch('on_update_selection', selData)

    def getHeight(self, root_size, _):
        height = len(self.data) * dp(40)
        if height > root_size[1] / 2:
            height = root_size[1] / 2
        return height


class IpInputScreen(BaseConfigScreen):
    """サーバーIPアドレス入力画面
    """
    serverIpAddress = StringProperty("")

    lang = StringProperty("")
    res = Resources()

    def __init__(self, **kwargs):
        """コンストラクタ
        """
        super(IpInputScreen, self).__init__(**kwargs)
        self.configSetting = ConfigSetting()
        self.__org_server_ip = ""
        self.updateProp()

    def on_enter(self):
        """画面表示時に実行される
        """
        self.__org_server_ip = self.configSetting.server_ip
        self.updateProp()
        self.ids.textinput.bind_textedit()
        # self.ids.textinput.imetext = ""

    def on_leave(self):
        """画面削除時に実行される
        """
        self.ids.textinput.unbind_textedit()
        # self.ids.textinput.imetext = ""

    def updateProp(self):
        """プロパティ更新（設定値を画面に反映）
        """
        self.serverIpAddress = self.configSetting.server_ip
        self.lang = self.configSetting.language

    def updateConfig(self):
        """設定値を更新
        """
        self.configSetting.server_ip = self.ids.textinput.text
        if self.__org_server_ip != self.configSetting.server_ip:
            if util.is_android_device():
                # 再接続要求
                send_broadcast_request(Type.RECONNECT_REQ, {})


class SoundConfigScreen(BaseConfigScreen):
    """呼出音設定画面
    """
    volume = NumericProperty(0)
    vibration = BooleanProperty(False)
    loop = BooleanProperty(False)

    lang = StringProperty("")
    res = Resources()

    def __init__(self, **kwargs):
        """コンストラクタ
        """
        super(SoundConfigScreen, self).__init__(**kwargs)
        self.configSetting = ConfigSetting()
        self.updateProp()
        if util.is_android_device():
            MediaPlayer = autoclass('android.media.MediaPlayer')
            self.mPlayer = MediaPlayer()


    def on_enter(self):
        """画面表示時に実行される
        """
        # self.configSetting.reload()
        self.updateProp()

    def updateProp(self):
        """プロパティ更新（設定値を画面に反映）
        """
        self.volume = self.configSetting.volume
        self.vibration = self.configSetting.vibration 
        self.loop = self.configSetting.loop 
        self.lang = self.configSetting.language

    def updateConfig(self):
        """設定値を更新
        """
        self.configSetting.volume = int(self.ids.volume.value)
        self.configSetting.vibration = self.ids.vibration.active
        self.configSetting.loop = self.ids.loop.active
        if util.is_android_device():
            # 温泉設定変更を通知する
            send_broadcast_request(Type.SOUND_CONFIG_UPDATED, {})

    def on_touch_up(self, touch):
        """on_touch_up
        
        Args:
            touch (object): touch情報
        
        Returns:
            bool: Trueを返すとイベント処理終了
        """
        self.logger.debug("on_touch_up touch_pos={}".format(touch.pos))
        if self.ids.volume.collide_point(*self.ids.volume.to_widget(*touch.pos)):
            super(SoundConfigScreen, self).on_touch_up(touch)
            self.playSound()
            return True
        else:
            return super(SoundConfigScreen, self).on_touch_up(touch)

    def playSound(self):
        """音を鳴らす
        """
        if util.is_android_device():
            soundList = []
            with open(pathlib.Path("./sound.json").resolve()) as f:
                soundList = json.loads(f.read())

            self.logger.debug("soundList={}".format(soundList))
            if len(soundList) > 0: 
                MediaPlayer = autoclass('android.media.MediaPlayer')
                AudioManager = autoclass('android.media.AudioManager')

                self.mPlayer.stop()
                self.mPlayer.release()

                self.mPlayer = MediaPlayer()
                self.mPlayer.setDataSource(soundList[0]["path"])
                self.mPlayer.setAudioStreamType(AudioManager.STREAM_NOTIFICATION)
                self.mPlayer.prepare()

                volume = float(self.ids.volume.value) / 10.0
                self.mPlayer.setVolume(volume, volume)
                self.mPlayer.start()

        
class SelectableLabel(BoxLayout):
    """選択ラベル
    """

    index = None
    checked = BooleanProperty(False)
    text = StringProperty("")

    def __init__(self, **kwargs):
        """コンストラクタ
        """
        super(SelectableLabel, self).__init__(**kwargs)
        self.logger = setup_logger_devel(self.__class__.__name__)

    def on_touch_down(self, touch):
        """on_touch_down
        
        Args:
            touch (object): touch情報
        
        Returns:
            bool: Trueを返すとイベント処理終了
        """

        ''' Add selection on touch down '''
        if super(SelectableLabel, self).on_touch_down(touch):
            return True
        if self.collide_point(*touch.pos):
            self.parent.parent.updateSelection({"text":self.text, "checked": self.checked})
            return True

    def calcFontSize(self, strlen, size):
        """フォントサイズを計算する
        
        Args:
            strlen (int): 文字列長（半角）
            size (tuple): 表示領域のサイズ
        
        Returns:
            float: フォントサイズ
        """
        if strlen == 0:
            strlen = 1
        font_size = size[0] / strlen * 2
        font_size = int(font_size * 0.9)
        
        if font_size > sp(20):
            font_size = sp(20)
        return font_size



class TermInfoScreen(BaseConfigScreen):
    """端末情報表示画面
    """
    ipAddress = StringProperty("")
    subnetMask = StringProperty("")
    router = StringProperty("")
    aplVer = StringProperty("")

    lang = StringProperty("")
    res = Resources()

    def __init__(self, **kwargs):
        """コンストラクタ
        """
        super(TermInfoScreen, self).__init__(**kwargs)
        self.configSetting = ConfigSetting()
        self.updateProp()

    def on_enter(self):
        """画面表示時に実行される
        """
        self.updateProp()

    def updateProp(self):
        """プロパティ更新（設定値を画面に反映）
        """
        # addresses = netifaces.ifaddresses('wlan0')
        # if netifaces.AF_INET in addresses:
        #     self.ipAddress = addresses[netifaces.AF_INET][0]['addr']
        #     self.subnetMask = addresses[netifaces.AF_INET][0]['netmask']

        # #default gateway
        # if netifaces.AF_INET in netifaces.gateways():
        #     for (addr, ifn, is_default ) in netifaces.gateways()[netifaces.AF_INET]:
        #         if ifn == "wlan0":
        #             self.router = addr

        try:
            # ipAddress, subnetMask, router = self.getNetworkInfo()
            ipAddress, subnetMask, router = getNetworkInfo()
            self.ipAddress = ipAddress
            self.subnetMask = subnetMask
            self.router = router
            
            self.aplVer = self.getAplVersion()
            self.lang = self.configSetting.language
        except:
            self.logger.error("exception", exc_info=True)

    def getAplVersion(self):
        """APLバージョン取得
        
        Returns:
            string: APLバージョン
        """
        version = ""
        if util.is_android_device():
            # Androidアプリのバージョン名を取得する。
            mActivity = autoclass('org.kivy.android.PythonActivity').mActivity
            app_context = mActivity.getApplication().getApplicationContext()
            pm = app_context.getPackageManager()
            packageInfo = pm.getPackageInfo(app_context.getPackageName(), 0)
            version = packageInfo.versionName

        return version

    # def getNetworkInfo(self):
    #     """ IPアドレス、サブネットマスク、デフォルトゲートウェイを取得する
    #     """
        
    #     ipAddress = ""
    #     subnetMask = ""
    #     router = ""
    #     if util.is_android_device():
    #         try:
    #             mActivity = autoclass('org.kivy.android.PythonActivity').mActivity
    #             Context = autoclass('android.content.Context')
    #             BuildVersion  = autoclass('android.os.Build$VERSION')
    #             BuildVersionCodes  = autoclass('android.os.Build$VERSION_CODES')
    #             # Android 11以降はAndroidのAPIでIPアドレス等を取得する。
    #             if BuildVersion.SDK_INT >= 30:
    #                 connectivityManager = mActivity.getSystemService(
    #                 Context.CONNECTIVITY_SERVICE)
    #                 currentNetwork = connectivityManager.getActiveNetwork()
    #                 linkProperties = connectivityManager.getLinkProperties(currentNetwork)
    #                 addrList = linkProperties.getLinkAddresses()
    #                 addrList = addrList.toArray([])
    #                 for addr in addrList:
    #                     m = re.match(r'([0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3})/([0-9]+)', addr.toString())
    #                     if m:
    #                         ipAddress = m.group(1)
    #                         prefx = int(m.group(2))
    #                         mask = "1" * prefx + "0" * (32 - prefx) 
    #                         subnetMask = "{0}.{1}.{2}.{3}".format(int(mask[0:8], 2), int(mask[8:16],2), int(mask[16:24],2), int(mask[24:32],2))
    #                         break


    #                 routeList = linkProperties.getRoutes()
    #                 routeList = routeList.toArray([])
    #                 for route in routeList:
    #                     if route.isDefaultRoute():
    #                         m = re.match(r'.*/([0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3})', route.getGateway().toString())
    #                         if m:
    #                             router = m.group(1)
    #                             break

    #             else:
    #                 addresses = netifaces.ifaddresses('wlan0')
    #                 if netifaces.AF_INET in addresses:
    #                     ipAddress = addresses[netifaces.AF_INET][0]['addr']
    #                     subnetMask = addresses[netifaces.AF_INET][0]['netmask']

    #                 #default gateway
    #                 if netifaces.AF_INET in netifaces.gateways():
    #                     for (addr, ifn, is_default ) in netifaces.gateways()[netifaces.AF_INET]:
    #                         if ifn == "wlan0":
    #                             router = addr
    #         except:
    #             self.logger.error("exception", exc_info=True)

    #     return ipAddress, subnetMask, router

class MenuBoxLayout(BoxLayout):
    """MenuBoxLayout
    """

    title = StringProperty("")

    # カスタムイベント
    __events__ = ('on_back',  )

    def __init__(self, **kwargs):
        """コンストラクタ
        """
        super(MenuBoxLayout, self).__init__(**kwargs)
        self.logger = setup_logger_devel(self.__class__.__name__)

    def on_back(self, *args):
        pass


class TitleValueButton(BoxLayout):
    """TitleValueButton
    """

    title = StringProperty("")
    value = StringProperty("")
    clickable = BooleanProperty(True)

    # カスタムイベント
    __events__ = ('on_press', 'on_release', )

    def __init__(self, **kwargs):
        """コンストラクタ
        """
        super(TitleValueButton, self).__init__(**kwargs)
        self.logger = setup_logger_devel(self.__class__.__name__)

    def on_press(self, *args):
        pass

    def on_release(self, *args):
        pass

    def on_touch_down(self, touch):
        """on_touch_down
        
        Args:
            touch (object): touch情報
        
        Returns:
            bool: Trueを返すとイベント処理終了
        """
        if self.collide_point(*touch.pos) :
            if self.clickable == True:
                self.dispatch('on_press', touch)
            return True
        else:
            return super(TitleValueButton, self).on_touch_down(touch)

    def on_touch_up(self, touch):
        """on_touch_up
        
        Args:
            touch (object): touch情報
        
        Returns:
            bool: Trueを返すとイベント処理終了
        """
        if self.collide_point(*touch.pos) :
            if self.clickable == True:
                self.dispatch('on_release', touch)
            return True
        else:
            return super(TitleValueButton, self).on_touch_up(touch)



class TextInputIME(TextInput):
    """日本語入力中の文字が表示されない問題に対応するための改造版TextInput
    （入力中文字をイベントで検知し、プロパティに保持する。）
    """
    imetext = StringProperty("")

    def __init__(self, **kwargs):
        """コンストラクタ
        """
        super(TextInputIME, self).__init__(**kwargs)
        self.logger = setup_logger_devel(self.__class__.__name__)
        # EventLoop.window.bind(on_textedit = self._on_textedit)
        # # self.bind(cursor_col=self.on_cursor_change)
        # self.bind(text=self.on_text)

    def bind_textedit(self):
        """bind
        """
        EventLoop.window.bind(on_textedit = self._on_textedit)
        self.bind(text=self.on_text)

    def unbind_textedit(self):
        """unbind
        """
        EventLoop.window.unbind(on_textedit = self._on_textedit)
        self.unbind(text=self.on_text)
        self.imetext = ""

    def _on_textedit(self, window, text):
        """on_texteditイベントハンドラ
        
        Args:
            window (object): [description]
            text (string): [description]
        """
        self.imetext = text

    # def on_cursor_change(self, instance, col):
    #     self.imetext = ""

    def on_text(self, instance, text):
        self.imetext = ""


class TextInputJPN(BoxLayout):
    """日本語入力中の文字が表示されない問題に対応するための改造版TextInput
    （Labelで入力中文字を表示する）
    """
    text = StringProperty("")
    imetext = StringProperty("")

    def __init__(self, **kwargs):
        """コンストラクタ
        """
        super(TextInputJPN, self).__init__(**kwargs)
        self.logger = setup_logger_devel(self.__class__.__name__)

    def on_kv_post(self, __):
        """on_kv_post
        """
        self.ids.textinput.bind(text=self.on_text)
        self.ids.textinput.bind(imetext=self.on_imetext)
        self.ids.textinput.bind(focus=self.on_focus)
        # self.ids.textinput.bind(cursor=self.on_cursor)

    def on_text(self, instance, text):
        """on_text
        
        Args:
            instance (object): イベント発生オブジェクト
            text (string): text
        """
        self.text = text

    def on_imetext(self, instance, imetext):
        """on_imetext
        
        Args:
            instance (object): イベント発生オブジェクト
            imetext (string): imetext
        """
        self.imetext = imetext
        if len(imetext) > 0:
            self.ids.imelabel.opacity = 1
        else:
            self.ids.imelabel.opacity = 0

    def on_focus(self, instance, value):
        """on_focus
        
        Args:
            instance (object): イベント発生オブジェクト
            value (bool): focus
        """
        if value:
            self.ids.imelabel.opacity = 1
        else:
            # TextInputのフォーカスが外れたら入力中文字表示を消したいので。
            self.ids.imelabel.opacity = 0

    def bind_textedit(self):
        """bind
        """
        self.ids.textinput.bind_textedit()

    def unbind_textedit(self):
        """unbind
        """
        self.ids.textinput.unbind_textedit()


def getNetworkInfo():
    """ IPアドレス、サブネットマスク、デフォルトゲートウェイを取得する
    """
    
    ipAddress = ""
    subnetMask = ""
    router = ""
    if util.is_android_device():
        mActivity = autoclass('org.kivy.android.PythonActivity').mActivity
        Context = autoclass('android.content.Context')
        BuildVersion  = autoclass('android.os.Build$VERSION')
        BuildVersionCodes  = autoclass('android.os.Build$VERSION_CODES')
        # Android 11以降はAndroidのAPIでIPアドレス等を取得する。
        if BuildVersion.SDK_INT >= 30:
            connectivityManager = mActivity.getSystemService(
            Context.CONNECTIVITY_SERVICE)
            currentNetwork = connectivityManager.getActiveNetwork()
            linkProperties = connectivityManager.getLinkProperties(currentNetwork)
            addrList = linkProperties.getLinkAddresses()
            addrList = addrList.toArray([])
            for addr in addrList:
                m = re.match(r'([0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3})/([0-9]+)', addr.toString())
                if m:
                    ipAddress = m.group(1)
                    prefx = int(m.group(2))
                    mask = "1" * prefx + "0" * (32 - prefx) 
                    subnetMask = "{0}.{1}.{2}.{3}".format(int(mask[0:8], 2), int(mask[8:16],2), int(mask[16:24],2), int(mask[24:32],2))
                    break


            routeList = linkProperties.getRoutes()
            routeList = routeList.toArray([])
            for route in routeList:
                if route.isDefaultRoute():
                    m = re.match(r'.*/([0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3})', route.getGateway().toString())
                    if m:
                        router = m.group(1)
                        break

        else:
            addresses = netifaces.ifaddresses('wlan0')
            if netifaces.AF_INET in addresses:
                ipAddress = addresses[netifaces.AF_INET][0]['addr']
                subnetMask = addresses[netifaces.AF_INET][0]['netmask']

            #default gateway
            if netifaces.AF_INET in netifaces.gateways():
                for (addr, ifn, is_default ) in netifaces.gateways()[netifaces.AF_INET]:
                    if ifn == "wlan0":
                        router = addr

    return ipAddress, subnetMask, router
