"""
Beyond EC main.py
"""

__author__ = "森下"
__version__ = "0.0.1"
__date__    = "2019.12.10"


import common_util as util
import control
import kivy
from kivy.app import App
from kivy.core.text import DEFAULT_FONT, LabelBase

from kivy.core.window import Window
from kivy.lang import Builder
from kivy.resources import resource_add_path
from kivy.uix.screenmanager import NoTransition, Screen, ScreenManager
from kivy.uix.widget import Widget
from mainscreen import MainScreen
from setup_logger import setup_logger_devel

kivy.require('1.11.1')

if util.is_display_device():
    from disp_dev.config_setting import ConfigSetting
else:
    from recv_dev.config import TerminalInfo

# kvファイルを画面ごとに分離してバラで読み込む
Builder.load_file('mainscreen_common.kv')
if util.is_display_device():
    from disp_dev.configscreen import ConfigScreen, NameInputScreen, ServerIpConfigScreen, IpInputScreen, SoundConfigScreen, TermInfoScreen
    Builder.load_file('disp_dev/mainscreen.kv')
    Builder.load_file('disp_dev/configscreen.kv')

else:
    from pairingscreen import PairingScreen
    from shutdownscreen import ShutdownScreen
    Builder.load_file('mainscreen.kv')
    Builder.load_file('pairingscreen.kv')
    Builder.load_file('shutdownscreen.kv')

    from recv_dev.storage import Database
    from recv_dev.config import NetworkInfo
    from recv_dev.netconfig import NetworkStatus


class EcoMonitorApp(App):
    """ECOモニタアプリケーション
    """
    def __init__(self):
        super(EcoMonitorApp, self).__init__()
        self.logger = setup_logger_devel(self.__class__.__name__)

    def build(self):
        """build
        """

        # 画面表示
        self.sm = ScreenManager()
        self.screen = MainScreen(name="mainscreen")
        self.sm.add_widget(self.screen)
        if util.is_display_device():
            self.sm.add_widget(ConfigScreen(name="configscreen"))
            self.sm.add_widget(NameInputScreen(name="nameinputscreen"))
            self.sm.add_widget(ServerIpConfigScreen(name="serveripconfigscreen"))
            self.sm.add_widget(IpInputScreen(name="ipinputscreen"))
            self.sm.add_widget(SoundConfigScreen(name="soundconfigscreen"))
            self.sm.add_widget(TermInfoScreen(name="terminfoscreen"))

        else:
            self.sm.transition = NoTransition()
            self.sm.add_widget(PairingScreen(name="pairingscreen"))
            self.sm.add_widget(ShutdownScreen(name="shutdownscreen"))


        self.sm.current = 'mainscreen'

        # 制御部
        self.controller = control.Controller(self.screen)

        return self.sm

    def on_start(self):
        """on_start
        """
        # データの初期化
        self.init_data()
        # ネットワークの初期化
        self.init_network()

        if util.is_android_device():
            Window.bind(on_keyboard=self.on_keyboard_handler)
        elif util.is_receiver_device():
            if TerminalInfo().dummyMode:
                Window.bind(on_key_down=self.on_key_down_for_dummy)


        # 一拍おいて開始する
        # Clock.schedule_once( lambda dt: self.startControl(), 0)
        self.startControl()
        
    def startControl(self):
        """開始
        """
        # 画面表示開始
        self.screen.start()
        # 制御部開始
        self.controller.start()

    def on_pause(self):
        """on_pause
        
        Returns:
            bool: True
        """
        self.controller.pause()
        return True

    def on_resume(self):
        """on_resume
        """
        self.controller.resume()

    def on_stop(self):
        """on_stop
        
        Returns:
            bool: True
        """
        self.controller.stop()
        self.screen.stop()
        return True

    def build_config(self, config):
        if util.is_display_device():
            conf = ConfigSetting()
            config.setdefaults(*conf.getDefaultConfig())

    def init_data(self):
        """データの初期化
        """
        if util.is_receiver_device():
            # DBの初期処理（テーブル作成）
            d = Database()
            d.install()
            d.upgrade()
            d.setup()

    def init_network(self):
        """ネットワーク初期化
        """
        if util.is_receiver_device():
            netconf = NetworkInfo()
            netstatus = NetworkStatus()
            # if netconf.eth["enabled"] == False:
            #     netstatus.if_up_down("eth0", False)

            if netconf.wlan["enabled"] == False:
                netstatus.if_up_down("wlan0", False)

    def on_keyboard_handler(self, window, key,*largs):
        """キー入力イベントハンドラー

        Args:
            window (object): window
            key (int): キー

        Returns:
            bool: Trueを返すとイベント処理終了
        """
        # Backキー制御
        if key == 27:
            if hasattr(self.sm.current_screen.ids, "menubox"):
                self.sm.current_screen.ids.menubox.dispatch("on_back")
                return True

        
        return False

    def on_key_down_for_dummy(self, window, keycode, scancode, codepoint, modifiers, *largs):
        """キー入力イベントハンドラー

        Args:
            window (object): window
            keycode (int): キーコード
            codepoint (string): 文字
            modifiers(list): 拡張

        Returns:
            bool: Trueを返すとイベント処理終了
        """
        print("{0}, {1}, {2}, {3}".format(keycode, scancode, codepoint, modifiers))
        keyEventInfoList = [
            {
                "key": {
                        "codepoint": 'f1'
                    },
                "command": {
                    "file": "/tmp/dummy_serial",
                    "text": "F01BBBBB,A1,01,100"
                }
            },
            {
                "key": {
                        "codepoint": 'f2'
                    },
                "command": {
                    "file": "/tmp/dummy_serial",
                    "text": "F02BBBBB,A1,01,100"
                }
            },
            {
                "key": {
                        "codepoint": 'f3'
                    },
                "command": {
                    "file": "/tmp/dummy_serial",
                    "text": "F03BBBBB,A1,01,100"
                }
            },
            {
                "key": {
                        "codepoint": 'f4'
                    },
                "command": {
                    "file": "/tmp/dummy_serial",
                    "text": "F04BBBBB,A1,01,100"
                }
            },
            {
                "key": {
                        "codepoint": 'f5'
                    },
                "command": {
                    "file": "/tmp/dummy_serial",
                    "text": "F05BBBBB,A1,01,100"
                }
            },
            {
                "key": {
                        "codepoint": 'f6'
                    },
                "command": {
                    "file": "/tmp/dummy_serial",
                    "text": "F06BBBBB,A1,01,100"
                }
            },
            {
                "key": {
                        "codepoint": 'f7'
                    },
                "command": {
                    "file": "/tmp/dummy_serial",
                    "text": "F07BBBBB,A1,01,100"
                }
            },
            {
                "key": {
                        "codepoint": 'f8'
                    },
                "command": {
                    "file": "/tmp/dummy_serial",
                    "text": "F08BBBBB,A1,01,100"
                }
            },
            {
                "key": {
                        "codepoint": 'f9'
                    },
                "command": {
                    "file": "/tmp/dummy_serial",
                    "text": "F09BBBBB,A1,01,100"
                }
            },
            {
                "key": {
                        "codepoint": 'f10'
                    },
                "command": {
                    "file": "/tmp/dummy_serial",
                    "text": "F10BBBBB,A1,01,100"
                }
            },
            {
                "key": {
                        "codepoint": 'f11'
                    },
                "command": {
                    "file": "/tmp/dummy_serial",
                    "text": "F11BBBBB,A1,01,100"
                }
            },
            {
                "key": {
                        "codepoint": 'f12'
                    },
                "command": {
                    "file": "/tmp/dummy_serial",
                    "text": "F12BBBBB,A1,01,100"
                }
            },
            {
                "key": {
                        "scancode": 0x02,
                        "codepoint": '1'
                    },
                "command": {
                    "file": "/tmp/dummy_serial",
                    "text": "15AAAAAA,B5,01,100"
                }
            },
            {
                "key": {
                        "scancode": 0x03,
                        "codepoint": '2'
                    },
                "command": {
                    "file": "/tmp/dummy_serial",
                    "text":"15AAAAAA,B5,02,100"
                }
            },
            {
                "key": {
                        "scancode": 0x04,
                        "codepoint": '3'
                    },
                "command": {
                    "file": "/tmp/dummy_serial",
                    "text":"15AAAAAA,B5,03,100"
                }
            },
            {
                "key": {
                        "scancode": 0x05,
                        "codepoint": '4'
                    },
                "command": {
                    "file": "/tmp/dummy_serial",
                    "text":"15AAAAAA,B5,04,100"
                }
            },
            {
                "key": {
                        "scancode": 0x06,
                        "codepoint": '5'
                    },
                "command": {
                    "file": "/tmp/dummy_serial",
                    "text":"15AAAAAA,B5,05,100"
                }
            },
            {
                "key": {
                        "scancode": 0x07,
                        "codepoint": '6'
                    },
                "command": {
                    "file": "/tmp/dummy_serial",
                    "text": "60AAAAAA,B5,01,100"
                }
            },
            {
                "key": {
                        "scancode": 0x08,
                        "codepoint": '7'
                    },
                "command": {
                    "file": "/tmp/dummy_serial",
                    "text": "60AAAAAA,B5,02,100"
                }
            },
            {
                "key": {
                        "scancode": 0x09,
                        "codepoint": '8'
                    },
                "command": {
                    "file": "/tmp/dummy_serial",
                    "text": "60AAAAAA,B5,03,100"
                }
            },
            {
                "key": {
                        "scancode": 0x0a,
                        "codepoint": '9'
                    },
                "command": {
                    "file": "/tmp/dummy_serial",
                    "text": "60AAAAAA,B5,04,100"
                }
            },
            {
                "key": {
                        "scancode": 0x0b,
                        "codepoint": '0'
                    },
                "command": {
                    "file": "/tmp/dummy_serial",
                    "text": "60AAAAAA,B5,05,100"
                }
            },
            {
                "key": {
                        "codepoint": 'insert'
                    },
                "command": {
                    "file": "/tmp/gpio_pairing",
                    "text": ""
                }
            },
            {
                "key": {
                        "codepoint": 'delete'
                    },
                "command": {
                    "file": "/tmp/gpio_clear",
                    "text": "long"
                }
            },
            {
                "key": {
                        "codepoint": 'enter',
                        "modifiers": ['shift', 'ctrl']
                    },
                "command": {
                    "file": "/tmp/gpio_shutdown",
                    "text": ""
                }
            },
        ]

        def _find_key_event(keycode, scancode, codepoint, modifiers, keyEventInfoList):
            for keyEventInfo in keyEventInfoList:
                check = True
                if "keycode" in keyEventInfo["key"] and keycode != keyEventInfo["key"]["keycode"]:
                    check = False
                if "scancode" in keyEventInfo["key"] and scancode != keyEventInfo["key"]["scancode"]:
                    check = False
                if "codepoint" in keyEventInfo["key"] and codepoint != keyEventInfo["key"]["codepoint"]:
                    check = False
                if "modifiers" in keyEventInfo["key"] and set(modifiers) != set(keyEventInfo["key"]["modifiers"]):
                    check = False
            
                if check:
                    return keyEventInfo

            return None

        try:
            keyEventInfo = _find_key_event(keycode, scancode, codepoint, modifiers, keyEventInfoList)
            if keyEventInfo is not None:
                with open(keyEventInfo["command"]["file"], "w") as f:
                    f.write(keyEventInfo["command"]["text"])
                return True
            elif keycode == 282:
                # F1のデフォルト動作を潰す
                return True

        except:
            pass



if __name__ == '__main__':
    logger = setup_logger_devel(__name__)
    logger.info("start app")
    
    from kivy.config import Config
    Config.set('kivy', 'log_level', "critical")
    Config.set('kivy', 'log_enable', 0)
    if util.is_receiver_device():
        Config.set('graphics', 'fullscreen', '1')
        Config.set('graphics', 'show_cursor', '0')
        Config.set('graphics', 'borderless', '1')
        Config.set('graphics', 'resizable', '0')
    Config.write()

    # Full Screen
    Window.fullscreen = True

    # カーソルを消す
    Window.show_cursor = False
    
    # 日本語フォントをセット
    # resource_add_path('./fonts')    
    # resource_add_path('./fonts/TakaoFonts_00303.01')    
    # LabelBase.register(DEFAULT_FONT, 'TakaoGothic.ttf')
    resource_add_path('./fonts/mplus-fonts')    
    # LabelBase.register(DEFAULT_FONT, 'mplus-1p-regular.otf', fn_bold='mplus-1p-bold.otf')
    LabelBase.register(DEFAULT_FONT, 'mplus-1m-regular.otf', fn_bold='mplus-1m-bold.otf')

    try:
        app = EcoMonitorApp()
        app.run()
    except KeyboardInterrupt:
        app.stop()
    except:
        logger.error("app failed", exc_info=True)

    logger.info("end app")
