"""
Beyond EC mainscreen.py
"""

import pathlib
import queue
import threading
import time
import unicodedata
from datetime import datetime, timedelta


import common_util as util
from kivy.animation import Animation
from kivy.clock import Clock
from kivy.graphics import Color, Line, Rectangle, RoundedRectangle, Triangle
from kivy.properties import (BooleanProperty, ListProperty, NumericProperty,
                             ObjectProperty, StringProperty)
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.button import ButtonBehavior
from kivy.uix.image import Image
from kivy.uix.label import Label, CoreLabel
from kivy.uix.screenmanager import Screen
from kivy.metrics import dp
from recv_dev.storage import Eraser
from resources import Resources
from setup_logger import setup_logger_devel
# import tracemalloc
# tracemalloc.start()

if util.is_android_device():
    from android_dev.sendbroadcast import Type, send_broadcast_request
    from jnius import autoclass, cast

if util.is_display_device():
    from disp_dev.config_setting import ConfigSetting
    import  disp_dev.send_udp as udp
else:
    from recv_dev.config import TerminalInfo

person_img_texture = Image(source=str(pathlib.Path('./images/history-man.png').resolve())).texture
clock_img_texture = Image(source=str(pathlib.Path('./images/history-time.png').resolve())).texture
bell_img_texture = Image(source=str(pathlib.Path('./images/bell.png').resolve())).texture
checked_img_texture = Image(source=str(pathlib.Path('./images/checked.png').resolve())).texture

class BaseCardWidget(BoxLayout):
    """BaseCardWidget
    """
    r = NumericProperty(0)
    histMode = BooleanProperty(False)
    data = ObjectProperty(None)
    color = StringProperty('R')
    recvDatetime = StringProperty("")
    groupNameProp = StringProperty("")
    line1Prop = StringProperty("")
    line2Prop = StringProperty("")
    datetimeProp = StringProperty("")
    delName = StringProperty("")
    groupNameSize = ListProperty([1, 1])
    groupNamePos = ListProperty([0, 0])
    spaceSize = ListProperty([1, 1])
    spacePos = ListProperty([0, 0])
    lineWidth = NumericProperty(1)

    res = Resources()

    def __init__(self, histMode = False, **kwargs):
        """コンストラクタ
        
        Args:
            histMode (bool, optional): 履歴モード. Defaults to False.
        """

        self.logger = setup_logger_devel(self.__class__.__name__)        
        self.histMode = histMode
        self.timedelta = None
        self.bind(data=self.on_data)

        super(BaseCardWidget, self).__init__(**kwargs)

        self.lineWidth = self.size[0] * 0.015
        self.bind(size=self.on_size)
        self.bind(pos=self.on_pos)

    def on_kv_post(self, __):
        """on_kv_post
        """
        # self.setPosSize()
        self.ids.groupName.bind(pos=lambda _, __: self.setPosSize())
        self.ids.groupName.bind(size=lambda _, __: self.setPosSize())
        self.ids.space.bind(pos=lambda _, __: self.setPosSize())
        self.ids.space.bind(size=lambda _, __: self.setPosSize())

    def setPosSize(self):
        """各種pos,sizeをプロパティに設定する。
        """
        # self.groupNamePos = self.ids.groupName.pos
        # self.groupNameSize = self.ids.groupName.size
        if hasattr(self.ids, "groupName"):
            self.groupNamePos = self.ids.groupName.pos[0] + self.lineWidth, self.ids.groupName.pos[1] 
            self.groupNameSize = self.ids.groupName.size[0] - self.lineWidth * 2, self.ids.groupName.size[1] - self.lineWidth

        if hasattr(self.ids, "space"):
            self.spacePos = self.ids.space.pos
            self.spaceSize = self.ids.space.size



    def on_data(self, instance, new_data):
        """on_data　受信データ変更イベントハンドラー
        
        Args:
            instance (object): イベント発生オブジェクト
            new_data ([type]): 受信データ
        """
        data = self.data
        if not "address" in data:
            return
        self.address = data['address']        
        self.modelNo = data['modelNo']
        self.buttonNo = data['buttonNo']
        self.color = data['color']

        self.line1Prop = data['line1']
        self.line2Prop = data['line2']

        self.priority = data['priority']

        if self.histMode == False:
            if len(self.recvDatetime) == 0:
                self.recvDatetime = data['ins_dt']
            if len(self.recvDatetime) > 0:
                self.datetimeProp = self.createTimeText(self.recvDatetime)
            self.blink = True if data['recall'] == 1 else False
        else:
            self.recvDatetime = data['upd_dt']
            dt = datetime.strptime(self.recvDatetime, "%Y-%m-%d %H:%M:%S.%f")
            self.datetimeProp = self.res.getResource("cardwidget:handled").format(dt)

        self.groupNameProp = data['groupName'] if 'groupName' in data else ''
        self.delName = data['displayName'] if 'displayName' in data else ''

        if util.is_android_device() and hasattr(self.ids, "callStopButton"):
            self.ids.callStopButton.data = data

    def on_size(self, instance, new_size):
        """on_size
        
        Args:
            instance (object): イベント発生オブジェクト
            new_size (tuple): サイズ
        """
        self.lineWidth = new_size[0] * 0.015
        self.correct_groundName_pos_size()
        self.setPosSize()

    def on_pos(self, instance, new_pos):
        """on_pos
        
        Args:
            instance (object): イベント発生オブジェクト
            new_pos (tuple): サイズ
        """
        self.correct_groundName_pos_size()
        self.setPosSize()

    def ask_update(self):
        """表示更新を要求する
        """
        if self.canvas is not None:
            self.canvas.ask_update()

    def getBaseColor(self, color):
        """基本色取得

        Args:
            color (string): 色（'R':赤、'G':緑、'B':青、'Y':黄、'W':白...）

        Returns:
            tuple: 色（[r, g, b, a]）
        """
        return self.getRgba(self.get_base_color_code(color))

    def getOutColor(self, color):
        """枠色取得

        Args:
            color (string): 色（'R':赤、'G':緑、'B':青、'Y':黄、'W':白...）

        Returns:
            tuple: 色（[r, g, b, a]）
        """
        return self.getRgba(self.get_out_color_code(color))

    def get_base_color_code(self, color):
        """基本色コード取得

        Args:
            color (string): 色（'R':赤、'G':緑、'B':青、'Y':黄、'W':白...）

        Returns:
            string: 色コード（"#e16cb5"等）
        """
        code = "#e16cb5"
        if color=="R":
            code = "#e16cb5"
        elif color=="G":
            code = "#5ad294"
        elif color=="B":
            code = "#4c9eeb"
        elif color=="Y":
            code = "#ebc64c"
        elif color=="W":
            code = "#ffffff"
        return code

    def get_out_color_code(self, color):
        """枠色コード取得

        Args:
            color (string): 色（'R':赤、'G':緑、'B':青、'Y':黄、'W':白...）

        Returns:
            string: 色コード（"#e16cb5"等）
        """
        code = "#e5c2d8"
        if color=="R":
            code = "#e5c2d8"
        elif color=="G":
            code = "#bde1cf"
        elif color=="B":
            code = "#b8d2e9"
        elif color=="Y":
            code = "#e9deb9"
        elif color=="W":
            code = "#efefef"
        return code

    def getRgba(self, color, alpha=None):
        """RGBA取得
        
        Args:
            color (string): #ffffff形式の色コード
            alpha (float, optional): 不透明度. Defaults to None.
        
        Returns:
            list: RGBA
        """
        rr = int(color[1:3], 16) / 255.0
        gg = int(color[3:5], 16) / 255.0
        bb = int(color[5:7], 16) / 255.0
        a = 1
        if alpha is not None:
            a = alpha
        return [rr,gg,bb,a]



    def calcFontSize(self, strlen, size):
        """フォントサイズを計算する
        
        Args:
            strlen (int): 文字列長（半角）
            size (tuple): 表示領域のサイズ
        
        Returns:
            float: フォントサイズ
        """
        if strlen == 0:
            strlen = 1
        font_size = size[0] / strlen * 2
        font_size = int(font_size * 0.9)
        
        return font_size
        
    def cut_hankaku_len(self, s, l):
        """指定された半角での文字列の長さで文字列をカットする

        Args:
            s (string): 文字列
            l (int): 半角での文字列の長さ

        Returns:
            string: カットした文字列
        """
        ret = ""
        cnt = 0
        for c in s:
            j = unicodedata.east_asian_width(c)
            if 'F' == j:
                # F	全角。全角英数など。
                cnt += 2
            elif 'H' == j:
                # H	半角。半角カナなど。
                cnt += 1
            elif 'W' == j:
                # W	全角。漢字やかななど。
                cnt += 2
            elif 'Na' == j:
                # Na	半角。半角英数など。
                cnt += 1
            elif 'A' == j:
                # A	曖昧。ギリシア文字など。
                cnt += 2
            else:
                # N	中立。アラビア文字など。
                cnt += 1

            if cnt > l:
                break
            ret += c

        return ret

    def createTimeText(self, dt):
        """経過時間表示文字列を作成する
        
        Args:
            dt (string): datetime
        
        Returns:
            string: 経過時間表示文字列
        """
        text = ''
        if dt is None or len(dt) == 0:
            return text
        now = datetime.now()

        if self.timedelta is not None:
            # 受信機との表示ずれを修正するために、表示機の時刻遅れ分の秒数を加算する
            now = now + timedelta(seconds=self.timedelta.delta)

        # time delta
        td = now - datetime.strptime(dt, "%Y-%m-%d %H:%M:%S.%f")

        if td.days >= 0:
            if td.seconds > 60:
                text = self.res.getResource("cardwidget:minutes").format(int(td.seconds/60))
            elif td.seconds >= 10:
                text = self.res.getResource("cardwidget:seconds").format(int(td.seconds/10))
            else:
                text = self.res.getResource("cardwidget:justnow")
        else:
            text = self.res.getResource("cardwidget:justnow")

        return text

    def updateTimeLabel(self):
        """経過時間ラベルを更新する
        """
        if self.histMode == False:
            self.datetimeProp = self.createTimeText(self.recvDatetime)

            self.correct_groundName_pos_size()
            self.setPosSize()
            # self.__debug_pos_check()


    def showStart(self):
        """表示開始アニメーション実行
        """
        anim = Animation(opacity=0.0, duration=0)+ Animation(opacity=1, duration=0.3)
        anim.repeat = False
        anim.start(self)

    def doRemoveAction(self, dist_pos, callback = None):
        """削除アニメーション実行
        
        Args:
            dist_pos (tuple): 移動先
            callback (function), optional):  アニメーション実行後に呼び出すコールバック関数. Defaults to None.
        """
        anim = Animation(opacity=0.3, pos = dist_pos, duration=0.1)
        if callback != None:
            anim.bind(on_complete=callback)
        anim.start(self)

    def correct_groundName_pos_size(self):
        """groundNameの位置とサイズを修正する
            （kvファイルの定義通りにならないときがあるため）
        """
        if isinstance(self, CardWidget):
            if hasattr(self.ids, "datetime"):
                self.ids.datetime.pos = self.pos
                self.ids.datetime.size = (self.size[0],  self.size[1] * self.ids.datetime.size_hint_y)

            if hasattr(self.ids, "space"):
                self.ids.space.pos = (self.pos[0],  self.pos[1] + self.ids.datetime.size[1])
                self.ids.space.size = (self.size[0],  self.size[1] * self.ids.space.size_hint_y)

            if hasattr(self.ids, "line2"):
                self.ids.line2.pos = (self.pos[0],  self.pos[1] + self.ids.datetime.size[1] + self.ids.space.size[1])
                self.ids.line2.size = (self.size[0],  self.size[1] * self.ids.line2.size_hint_y)

            if hasattr(self.ids, "line1"):
                self.ids.line1.pos = (self.pos[0],  self.pos[1] + self.ids.datetime.size[1] + self.ids.space.size[1]
                                                    + self.ids.line2.size[1])
                self.ids.line1.size = (self.size[0],  self.size[1] * self.ids.line1.size_hint_y)

            if hasattr(self.ids, "groupName"):
                self.ids.groupName.pos = (self.pos[0],  self.pos[1] + self.ids.datetime.size[1] + self.ids.space.size[1]
                                                    + self.ids.line2.size[1] + self.ids.line1.size[1])
                self.ids.groupName.size = (self.size[0],  self.size[1] * self.ids.groupName.size_hint_y)


    def __debug_pos_check(self):
        # debug
        if isinstance(self, CardWidget):
            top1 = round(self.groupNamePos[1] + self.groupNameSize[1] + self.lineWidth ,1 )
            top2 = round(self.pos[1] + self.size[1], 1)
            if top1!= round(top2,1):
                self.logger.warn("!!! top1={0}, top2={1}".format(top1, top2 ))
            else:
                self.logger.info("... top1={0}, top2={1}".format(top1, top2 ))

    def clean(self):
        if self.data is not None:
            self.data.clear()
        del self.logger


class CardWidget(BaseCardWidget):
    """CardWidget
    """
    blink = BooleanProperty(False)

    def __init__(self, **kwargs):
        """コンストラクタ
        """
        super(CardWidget, self).__init__(histMode=False, **kwargs)
        # self.drawBackground()

        self.bind(color=lambda _,__: self.drawBackground())
        self.bind(r=lambda _,__: self.updateBackground())
        self.bind(pos=lambda _,__: self.updateBackground())
        self.bind(size=lambda _,__: self.updateBackground())
        self.bind(lineWidth=lambda _,__: self.updateBackground())
        self.bind(groupNamePos=lambda _,__: self.updateBackground())
        self.bind(groupNameSize=lambda _,__: self.updateBackground())
        self.bind(blink=self.on_blink)
        self.bind(datetimeProp=lambda _,__: self.updateBackground())
        self.blinkAnim = Animation(opacity=0.2, duration=0.2) + Animation(opacity=1.0, duration=0.2) + Animation(opacity=1.0, duration=1) 
        self.blinkAnim.repeat = True

    def on_kv_post(self, __):
        """on_kv_post
        """
        super(CardWidget, self).on_kv_post(__)
        self.drawBackground()

    def drawBackground(self):
        """背景図形の描画
        """
        self.clearCanvas()
        if hasattr(self.ids, "background1"):
            with self.ids.background1.canvas.before:
                Color(*self.getOutColor(self.color))
                # 外枠の線のための四角
                self.rect1 = RoundedRectangle(pos=self.pos, size=self.size, radius=[self.r])

        if hasattr(self.ids, "background2"):
            (groupName_label, line1_label, line2_label, datetime_label) = self.__getLabelTexture()
            with self.ids.background2.canvas.before:

                Color(*self.getBaseColor(self.color))
                # ベースの四角
                self.rect2 = RoundedRectangle(pos=self.__rect2pos(),
                                size=self.__rect2size(),
                                radius=[self.r])

                Color(*self.getOutColor(self.color))
                # グループ名表示領域
                self.rect3 = RoundedRectangle(pos=self.groupNamePos,
                                size=self.groupNameSize,
                                radius=[self.r, self.r, 0, 0])

                Color(0, 0, 0, 1)
                self.groupName_rect = Rectangle(pos=self.ids.groupName.pos,
                                            size=self.ids.groupName.size,
                                            texture=groupName_label.texture)

                Color(0, 0, 0, 1)
                self.line1_rect = Rectangle(pos=self.ids.line1.pos,
                                            size=self.ids.line1.size,
                                            texture=line1_label.texture)

                Color(0, 0, 0, 1)
                self.line2_rect = Rectangle(pos=self.ids.line2.pos,
                                            size=self.ids.line2.size,
                                            texture=line2_label.texture)

                Color(0, 0, 0, 1)
                self.datetime_rect = Rectangle(pos=self.ids.datetime.pos,
                                            size=self.ids.datetime.size,
                                            texture=datetime_label.texture)

    def updateBackground(self):
        """背景図形の更新
        """
        # 外枠の線のための四角
        if hasattr(self, "rect1"):
            self.rect1.pos = self.pos
            self.rect1.size = self.size
            self.rect1.radius=[self.r]

        # ベースの四角
        if hasattr(self, "rect2"):
            self.rect2.pos = self.__rect2pos()
            self.rect2.size = self.__rect2size()
            self.rect2.radius=[self.r]

        # グループ名表示領域
        if hasattr(self, "rect3"):
            self.rect3.pos = self.groupNamePos
            self.rect3.size = self.groupNameSize
            self.rect3.radius=[self.r, self.r, 0, 0]

        (groupName_label, line1_label, line2_label, datetime_label) = self.__getLabelTexture()
        if len(self.groupNameProp) > 0:
            if hasattr(self, "groupName_rect"):
                self.groupName_rect.texture = groupName_label.texture
                self.groupName_rect.pos = self.ids.groupName.pos
                self.groupName_rect.size = self.ids.groupName.size

        if len(self.line1Prop) > 0:
            if hasattr(self, "line1_rect"):
                self.line1_rect.texture = line1_label.texture
                self.line1_rect.pos = self.ids.line1.pos
                self.line1_rect.size = self.ids.line1.size

        if len(self.line2Prop) > 0:
            if hasattr(self, "line2_rect"):
                self.line2_rect.texture = line2_label.texture
                self.line2_rect.pos = self.ids.line2.pos
                self.line2_rect.size = self.ids.line2.size

        if len(self.datetimeProp) > 0:
            if hasattr(self, "datetime_rect"):
                self.datetime_rect.texture = datetime_label.texture
                self.datetime_rect.pos = self.ids.datetime.pos
                self.datetime_rect.size = self.ids.datetime.size

        if hasattr(self.ids, "callStopButton"):
            self.ids.callStopButton.updateBackground()

    def __rect2pos(self):
        """ベースの四角のpos

        Returns:
            tuple: 位置
        """
        return [self.pos[0] + self.lineWidth, self.pos[1]  + self.lineWidth]
        
    def __rect2size(self):
        """ベースの四角のsize

        Returns:
            tuple: サイズ
        """
        return [self.size[0] - self.lineWidth * 2, self.size[1]  - self.lineWidth * 2]

    # def __refreshTexture(self):
    #     """表示文字列のTextureを更新する
    #     """
    #     self.groupName_label = CoreLabel(text=self.cut_hankaku_len(self.groupNameProp, 34),
    #                                      font_size=self.calcFontSize(34, self.ids.groupName.size),
    #                                      text_size=self.ids.groupName.size,
    #                                      halign='center',
    #                                      valign='middle')
    #     self.groupName_label.refresh()

    #     self.line1_label = None
    #     self.line1_label = CoreLabel(text=self.cut_hankaku_len(self.line1Prop, 20),
    #                                      font_size=self.calcFontSize(20, self.ids.line1.size),
    #                                      text_size=self.ids.line1.size,
    #                                      halign='center',
    #                                      valign='middle')
    #     self.line1_label.refresh()

    #     self.line2_label = None
    #     self.line2_label = CoreLabel(text=self.cut_hankaku_len(self.line2Prop, 12),
    #                                      font_size=self.calcFontSize(12, self.ids.line2.size),
    #                                      text_size=self.ids.line2.size,
    #                                      halign='center',
    #                                      valign='middle')
    #     self.line2_label.refresh()

    #     self.datetime_label = CoreLabel(text=self.datetimeProp,
    #                                      font_size=self.calcFontSize(20, self.ids.datetime.size),
    #                                      text_size=self.ids.datetime.size,
    #                                      halign='center',
    #                                      valign='middle')
    #     self.datetime_label.refresh()

    def __getLabelTexture(self):
        """表示文字列のTextureを取得する
        """
        groupName_label = CoreLabel(text=self._label_text(self.cut_hankaku_len(self.groupNameProp, 34)),
                                         font_size=self.calcFontSize(34, self.ids.groupName.size),
                                         text_size=self.ids.groupName.size,
                                         halign='center',
                                         valign='middle')
        groupName_label.refresh()

        line1_label = CoreLabel(text=self._label_text(self.cut_hankaku_len(self.line1Prop, 20)),
                                         font_size=self.calcFontSize(20, self.ids.line1.size),
                                         text_size=self.ids.line1.size,
                                         halign='center',
                                         valign='middle')
        line1_label.refresh()

        line2_label = CoreLabel(text=self._label_text(self.cut_hankaku_len(self.line2Prop, 12)),
                                         font_size=self.calcFontSize(12, self.ids.line2.size),
                                         text_size=self.ids.line2.size,
                                         halign='center',
                                         valign='middle')
        line2_label.refresh()

        datetime_label = CoreLabel(self._label_text(text=self.datetimeProp),
                                         font_size=self.calcFontSize(20, self.ids.datetime.size),
                                         text_size=self.ids.datetime.size,
                                         halign='center',
                                         valign='middle')
        datetime_label.refresh()


        return (groupName_label, line1_label, line2_label, datetime_label)

    def _label_text(self, text):
        """ テキストラベル表示文字取得
            空文字が設定されていた場合は空白を返す
        """

        if text is None or text == "":
            return " "
        else:
            return text

    def on_blink(self, instance, new_blink):
        if new_blink:
            self.blinkStart()
        else:
            self.blinkStop()

    def blinkStart(self):
        """点滅アニメーション実行
        """
        self.blinkAnim.start(self.ids.background1)

    def blinkStop(self):
        """点滅アニメーション実行
        """
        self.blinkAnim.cancel(self.ids.background1)

    def clean(self):
        """ 描画の後処理を行う。（メモリーリーク対策）
        """
        super(CardWidget, self).clean()

        self.blinkAnim.stop(self.ids.background1)
        del self.blinkAnim

        if hasattr(self.ids, "callStopButton"):
            self.ids.callStopButton.clean()


        self.clearCanvas()
        self.canvas.clear()
        self.canvas.before.clear()
        self.canvas.after.clear()

    def clearCanvas(self):
        if hasattr(self, "rect1"):
            del self.rect1
        if hasattr(self, "rect2"):
            del self.rect2
        if hasattr(self, "rect3"):
            del self.rect3
        if hasattr(self, "groupName_rect"):
            del self.groupName_rect
        if hasattr(self, "line1_rect"):
            del self.line1_rect
        if hasattr(self, "line2_rect"):
            del self.line2_rect
        if hasattr(self, "datetime_rect"):
            del self.datetime_rect

        if hasattr(self.ids, "background1"):
            self.ids.background1.canvas.clear()
            self.ids.background1.canvas.before.clear()
            self.ids.background1.canvas.after.clear()

        if hasattr(self.ids, "background2"):
            self.ids.background2.canvas.clear()
            self.ids.background2.canvas.before.clear()
            self.ids.background2.canvas.after.clear()


class HistCardWidget(BaseCardWidget):
    """HistCardWidget
    """
    handledSize = ListProperty([1, 1])
    handledPos = ListProperty([0, 0])

    # person_img_path = StringProperty(str(pathlib.Path('./images/history-man.png').resolve())) 
    # clock_img_path = StringProperty(str(pathlib.Path('./images/history-time.png').resolve())) 
    # person_img_texture = Image(source=str(pathlib.Path('./images/history-man.png').resolve())).texture
    # clock_img_texture = Image(source=str(pathlib.Path('./images/history-time.png').resolve())).texture

    def __init__(self, **kwargs):
        """コンストラクタ
        """
        super(HistCardWidget, self).__init__(histMode=True, **kwargs)
        self.drawBackground()
        self.bind(color=lambda _,__: self.drawBackground())
        self.bind(r=lambda _,__: self.updateBackground())
        self.bind(pos=lambda _,__: self.updateBackground())
        self.bind(size=lambda _,__: self.updateBackground())
        self.bind(lineWidth=lambda _,__: self.updateBackground())
        self.bind(spacePos=lambda _,__: self.updateBackground())
        self.bind(spaceSize=lambda _,__: self.updateBackground())
        self.bind(handledPos=lambda _,__: self.updateBackground())
        self.bind(handledSize=lambda _,__: self.updateBackground())
        self.bind(groupNamePos=lambda _,__: self.updateBackground())
        self.bind(groupNameSize=lambda _,__: self.updateBackground())

    def on_kv_post(self, __):
        """on_kv_post
        """
        super(HistCardWidget, self).on_kv_post(__)
        self.setPosSize()
        self.ids.handled.bind(pos=lambda _, __: self.setPosSize())
        self.ids.handled.bind(size=lambda _, __: self.setPosSize())

    def setPosSize(self):
        """各種pos,sizeをプロパティに設定する。
        """
        super(HistCardWidget, self).setPosSize()
        if hasattr(self.ids, "handled"):
            self.handledPos = self.ids.handled.pos
            self.handledSize = self.ids.handled.size


    def drawBackground(self):
        """背景図形の描画
        """
        # self.canvas.before.clear()
        self.clearCanvas()

        with self.canvas.before:
            Color(*self.getOutColor(self.color))
            # 枠線のための四角
            self.rect1 = RoundedRectangle(pos=self.__rect1pos(),
                                        size=self.__rect1size(),
                                        radius=[self.r])

            Color(*self.getBaseColor(self.color))
            # ベースの四角
            self.rect2 = RoundedRectangle(pos=self.__rect2pos(),
                                        size=self.__rect2size(),
                                        radius=[self.r])

            Color(*self.getOutColor(self.color))
            # グループ名表示領域
            self.rect3 = RoundedRectangle(pos=self.groupNamePos,
                            size=self.groupNameSize,
                            radius=[self.r, self.r, 0, 0])

            Color(*self.getBaseColor(self.color))
            # 吹き出しの境目の枠線を消すための四角
            self.rect4 = Rectangle(pos=self.__rect4pos(),
                            size=self.__rect4size())
            # 吹き出しの三角
            self.triangle= Triangle(points=self.__triangle_points())

            Color(*self.getOutColor(self.color))
            # 吹き出しの外枠の線
            self.line = Line(width=self.lineWidth/2, cap='square', joint='miter',
                points=self.__line_points())

            Color(*self.getBaseColor(self.color))
            # 対処済みの四角
            self.rect5 = RoundedRectangle(pos=self.handledPos,
                            size=self.handledSize,
                            radius=[self.r])

        # with self.canvas:
            # self.__refreshTexture()
            (datetime_label, delName_label) = self.__getLabelTexture()

            self.clock_img_rect = RoundedRectangle(pos=self.__clock_img_rect_pos(datetime_label),
                      size=self.__clock_img_rect_size(datetime_label),
                      radius=[self.r])

            Color(0, 0, 0, 1)
            self.datetime_rect = Rectangle(pos=self.__datetime_rect_pos(datetime_label), 
                                        size=self.__datetime_rect_size(datetime_label),
                                         texture=datetime_label.texture)

            Color(*self.getBaseColor(self.color))
            self.person_img_rect = Rectangle(pos=self.__person_img_rect_pos(delName_label),
                      size=self.__person_img_rect_size(delName_label))

            Color(0, 0, 0, 1)
            self.delName_rect = Rectangle(pos=self.__delName_rect_pos(delName_label),
                                    size=self.__delName_rect_size(datetime_label),
                                    texture=delName_label.texture)

        # if hasattr(self.ids, "clock_img"):
        #     # self.ids.clock_img.texture = self.clock_img_texture
        #     self.ids.clock_img.texture = clock_img_texture
        # if hasattr(self.ids, "person_img"):
        #     # self.ids.person_img.texture = self.person_img_texture
        #     self.ids.person_img.texture = person_img_texture

    def updateBackground(self):
        """背景図形の更新
        """
        # 枠線のための四角
        if hasattr(self, "rect1"):
            self.rect1.pos = self.__rect1pos()
            self.rect1.size = self.__rect1size()
            self.rect1.radius=[self.r]

        # ベースの四角
        if hasattr(self, "rect2"):
            self.rect2.pos = self.__rect2pos()
            self.rect2.size = self.__rect2size()
            self.rect2.radius = [self.r]

        # グループ名表示領域
        if hasattr(self, "rect3"):
            self.rect3.pos = self.groupNamePos
            self.rect3.size = self.groupNameSize
            self.rect3.radius=[self.r, self.r, 0, 0]

        # 吹き出しの境目の枠線を消すための四角
        if hasattr(self, "rect4"):
            self.rect4.pos = self.__rect4pos()
            self.rect4.size = self.__rect4size()

        # 吹き出しの三角
        if hasattr(self, "triangle"):
            self.triangle.points = self.__triangle_points()

        # 吹き出しの外枠の線
        if hasattr(self, "line"):
            self.line.width=self.lineWidth/2
            self.line.points= self.__line_points()

        # 対処済みの四角
        if hasattr(self, "rect5"):
            self.rect5.pos = self.handledPos
            self.rect5.size = self.handledSize
            self.rect5.radius=[self.r]


        # if hasattr(self.ids, "clock_img"):
        #     # self.ids.clock_img.texture = self.clock_img_texture
        #     self.ids.clock_img.texture = clock_img_texture

        #下段表示文字列のTextureを更新する
        # self.__refreshTexture()
        (datetime_label, delName_label) = self.__getLabelTexture()

        if hasattr(self, "clock_img_rect"):
            self.clock_img_rect.texture = clock_img_texture
            self.clock_img_rect.pos = self.__clock_img_rect_pos(datetime_label)
            self.clock_img_rect.size = self.__clock_img_rect_size(datetime_label)

        if hasattr(self, "datetime_rect"):
            self.datetime_rect.texture = datetime_label.texture
            self.datetime_rect.pos = self.__datetime_rect_pos(datetime_label)
            self.datetime_rect.size = self.__datetime_rect_size(datetime_label)

        # if hasattr(self.ids, "person_img"):
        #     # self.ids.person_img.texture = self.person_img_texture
        #     self.ids.person_img.texture = person_img_texture

        if hasattr(self, "person_img_rect"):
            self.person_img_rect.texture = person_img_texture
            self.person_img_rect.pos = self.__person_img_rect_pos(delName_label)
            self.person_img_rect.size = self.__person_img_rect_size(delName_label)

        if hasattr(self, "delName_rect"):
            self.delName_rect.texture = delName_label.texture
            self.delName_rect.pos = self.__delName_rect_pos(delName_label)
            self.delName_rect.size = self.__delName_rect_size(datetime_label)

        # self.canvas.ask_update()

    def __rect1pos(self):
        """枠線のための四角のpos

        Returns:
            tuple: 位置
        """
        return [self.spacePos[0], self.spacePos[1] + self.spaceSize[1]]
    def __rect1size(self):
        """枠線のための四角のsize

        Returns:
            tuple: サイズ
        """
        return [self.size[0], self.size[1] - self.handledSize[1] - self.spaceSize[1]]

    def __rect2pos(self):
        """ベースの四角のpos

        Returns:
            tuple: 位置
        """
        return [self.spacePos[0] + self.lineWidth, self.spacePos[1] + self.spaceSize[1] + self.lineWidth]
    def __rect2size(self):
        """ベースの四角のsize

        Returns:
            tuple: サイズ
        """
        return [self.size[0] - self.lineWidth * 2, self.size[1] - self.handledSize[1] - self.spaceSize[1] - self.lineWidth * 2]

    def __rect4pos(self):
        """吹き出しの境目の枠線を消すための四角のpos

        Returns:
            tuple: 位置
        """
        return [self.spacePos[0] + self.size[0]/2 - self.size[0]*0.02, self.spacePos[1] + self.spaceSize[1]]
    def __rect4size(self):
        """吹き出しの境目の枠線を消すための四角のsize

        Returns:
            tuple: サイズ
        """
        return [self.size[0]*0.02 *2, self.lineWidth]

    def __triangle_points(self):
        """吹き出しの三角のpoints

        Returns:
            tuple: points
        """
        return (self.spacePos[0] + self.size[0]/2 - self.size[0]*0.02, self.spacePos[1] + self.spaceSize[1], self.spacePos[0] + self.size[0]/2, self.spacePos[1],self.spacePos[0] + self.size[0]/2 + self.size[0]*0.02, self.spacePos[1] + self.spaceSize[1])

    def __line_points(self):
        """吹き出しの外枠の線のpoints

        Returns:
            tuple: points
        """
        return ( self.spacePos[0] + self.size[0]/2 - self.size[0]*0.02, self.spacePos[1] + self.spaceSize[1], self.spacePos[0] + self.size[0]/2, self.spacePos[1] + self.lineWidth, self.spacePos[0] + self.size[0]/2 + self.size[0]*0.02, self.spacePos[1] + self.spaceSize[1])

    # def __refreshTexture(self):
    #     """下段表示文字列のTextureを更新する
    #     """
    #     self.datetime_label = CoreLabel(text=self.datetimeProp,
    #                                      font_size=self.calcFontSize(22, (self.handledSize[0] * 0.7, self.handledSize[1])),
    #                                      padding_x=dp(5))
    #     self.datetime_label.refresh()

    #     self.delName_label = CoreLabel(text=self.cut_hankaku_len(self.delName, 20),
    #                                     font_size=self.calcFontSize(22, (self.handledSize[0] * 0.7, self.handledSize[1])),
    #                                     padding_x=dp(5))
    #     self.delName_label.refresh()

    def __getLabelTexture(self):
        """下段表示文字列のTextureを取得する
        """
        datetime_label = CoreLabel(text=self.datetimeProp,
                                         font_size=self.calcFontSize(22, (self.handledSize[0] * 0.7, self.handledSize[1])),
                                         padding_x=dp(5))
        datetime_label.refresh()

        delName_label = CoreLabel(text=self.cut_hankaku_len(self.delName, 20),
                                        font_size=self.calcFontSize(22, (self.handledSize[0] * 0.7, self.handledSize[1])),
                                        padding_x=dp(5))
        delName_label.refresh()
        
        return (datetime_label, delName_label)

    def __clock_img_rect_pos(self, datetime_label):
        """時計アイコンのための四角のpos

        Returns:
            tuple: 位置
        """

        icon_size = self.__clock_img_rect_size(datetime_label)
        area_width = self.size[0] / 3
        area_height = self.handledSize[1]
        texture_size = datetime_label.texture.size
        x = self.pos[0] + (area_width - (icon_size[0] + texture_size[0])) / 2
        y = self.pos[1] + (area_height - icon_size[1]) / 2
        return [x, y]

    def __clock_img_rect_size(self, datetime_label):
        """時計アイコンのための四角のsize

        Returns:
            tuple: サイズ
        """
        size = datetime_label.texture.size
        return [size[1], size[1]]

    def __datetime_rect_pos(self, datetime_label):
        """時刻表示のための四角のpos

        Returns:
            tuple: サイズ
        """
        icon_size = self.__clock_img_rect_size(datetime_label)
        area_width = self.size[0] / 3
        area_height = self.handledSize[1]
        texture_size = datetime_label.texture.size
        x = self.pos[0] + icon_size[0] + (area_width - (icon_size[0] + texture_size[0])) / 2
        y = self.pos[1] + (area_height - icon_size[1]) / 2
        return [x, y]

    def __datetime_rect_size(self, datetime_label):
        """時刻表示のための四角のsize

        Returns:
            tuple: サイズ
        """
        return datetime_label.texture.size

    def __person_img_rect_pos(self, delName_label):
        """人アイコンのための四角のpos

        Returns:
            tuple: 位置
        """
        icon_size = self.__person_img_rect_size(delName_label)
        area_width = self.size[0] * 2 / 3
        area_height = self.handledSize[1]
        texture_size = delName_label.texture.size
        x = self.pos[0]  + (self.size[0] / 3)+ (area_width - (icon_size[0] + texture_size[0])) / 2
        y = self.pos[1] + (area_height - icon_size[1]) / 2
        return [x, y]

    def __person_img_rect_size(self, delName_label):
        """人アイコンのための四角のsize

        Returns:
            tuple: サイズ
        """
        size = delName_label.texture.size
        return [size[1], size[1]]

    def __delName_rect_pos(self, delName_label):
        """削除者名表示のための四角のpos

        Returns:
            tuple: サイズ
        """
        icon_size = self.__person_img_rect_size(delName_label)
        area_width = self.size[0] * 2 / 3
        area_height = self.handledSize[1]
        texture_size = delName_label.texture.size
        x = self.pos[0]  + (self.size[0] / 3) + icon_size[0] + (area_width - (icon_size[0] + texture_size[0])) / 2
        y = self.pos[1] + (area_height - icon_size[1]) / 2
        return [x, y]

    def __delName_rect_size(self, datetime_label):
        """削除者名表示のための四角のsize

        Returns:
            tuple: サイズ
        """
        return list(datetime_label.texture.size)

    def clean(self):
        """ 描画の後処理を行う。（メモリーリーク対策）
        """
        super(HistCardWidget, self).clean()
        self.clearCanvas()

        self.canvas.clear()
        self.canvas.before.clear()
        self.canvas.after.clear()

    def clearCanvas(self):
        if hasattr(self, "rect1"):
            del self.rect1
        if hasattr(self, "rect2"):
            del self.rect2
        if hasattr(self, "rect3"):
            del self.rect3
        if hasattr(self, "rect4"):
            del self.rect4
        if hasattr(self, "triangle"):
            del self.triangle
        if hasattr(self, "line"):
            del self.line
        if hasattr(self, "rect5"):
            del self.rect5
        if hasattr(self, "clock_img_rect"):
            del self.clock_img_rect
        if hasattr(self, "datetime_rect"):
            del self.datetime_rect
        if hasattr(self, "person_img_rect"):
            del self.person_img_rect
        if hasattr(self, "delName_rect"):
            del self.delName_rect

        self.canvas.before.clear()


class MainScreen(Screen):
    """スクリーンクラス
    """
    # 時刻プロパティ
    nowTime = StringProperty('')
    # 呼び出し中件数
    queueNum = NumericProperty(0)
    # selectedTabName = StringProperty('')

    landscape = BooleanProperty(False)
    histMode = BooleanProperty(False)

    headerHeight = NumericProperty(1)
    statusHeight = NumericProperty(1)
    tabHeight = NumericProperty(1)
    card_space = NumericProperty(1)

    version = StringProperty('')
    disconnected = BooleanProperty(False)
    timeDiff = BooleanProperty(False)

    selectedIndex = NumericProperty(0)


    __widget_lock = threading.Lock()
    __wait_events = []
    __proc_queue = queue.Queue()



    lang = StringProperty('')
    res = Resources()

    MAX_HIST_NUM = 100

    # カスタムイベント
    # 制御部に消去を要求する
    __events__ = ('on_request_erase',  )

    def on_request_erase(self, *args):
        pass

    def __init__(self, **kwargs):
        """コンストラクタ
        """
        super(MainScreen, self).__init__(**kwargs)
        self.logger = setup_logger_devel(self.__class__.__name__)
        self.cards = []
        self.hist_cards = []
        self.histMode = False
        self.dragging_widget = None
        self.drag_prev_pos = None
        self.drag_start_pos = None
        self.drag_start_time = None
        self.updateTimeLabels()
        # self.selectedTabName = self.res.getResource("mainscreen:calling")
        self.landscape = self.isLandscape()
        self.setAppHeight(self.landscape, self.size)
        self.bind(size=self.on_size)
        if util.is_receiver_device():
            self.lang = TerminalInfo().language
            self.version = TerminalInfo().firmwareVersion
        else:
            self.lang = ConfigSetting().language

        self.timedelta = None
        if util.is_display_device:
            self.bind(timeDiff=self.on_timediff_change)
            self.timedelta = TimeDelta()

        if util.is_receiver_device():
            # self.selectFrameEvent = None
            self.selectFrame = SelectFrame()
            self.selectFrame.eraserName = ""
            self.bind(selectedIndex=self.on_selectedIndex_change)

        self.prev_snap = None

        # Controlerからの呼出はQueueに格納されているので
        # それを定期的に実行する
        Clock.schedule_interval(self._run_proc_queue, 0.05)

    def on_kv_post(self, __):
        """on_kv_post
        """
        self.ids.sc1_scroll.bind(size=self.on_sc1_scroll_size)
        self.ids.sc2_scroll.bind(size=self.on_sc2_scroll_size)
        if hasattr(self.ids, "ctltab"):
            self.ids.ctltab.landscape = self.landscape

    def on_enter(self):
        """画面表示時に実行される
        """
        if util.is_receiver_device():
            self.lang = TerminalInfo().language
            self.updateSelectFrame()
        else:
            self.lang = ConfigSetting().language
            for card in self.cards:
                card.updateBackground()


    def on_size(self, instance, new_size):
        """on_size
        
        Args:
            instance (object): イベント発生オブジェクト
            new_size (tuple): サイズ
        """
        self.landscape = self.isLandscape()
        self.logger.debug("landscape={}".format(self.landscape))
        self.logger.debug("size={}".format(new_size))
        self.setAppHeight(self.landscape, new_size)

        if hasattr(self.ids, "ctltab"):
            self.ids.ctltab.landscape = self.landscape

    def isLandscape(self):
        """横表示判定
        
        Returns:
            bool: 判定結果
        """
        if self.size[0] >= self.size[1]:
            return True
        else:
            return False

    def setAppHeight(self, landscape, size):
        """スクリーンのサイズに合わせてヘッダー等のサイズを設定する

        Args:
            landscape (bool): 横長か？
            size (tuple): Screenのサイズ
        """
        if util.is_display_device():
            # 表示機
            if landscape:
                self.headerHeight = size[1] * 0.133
                self.statusHeight = size[1] * 0.11
                self.tabHeight = size[1] * 0.18
                self.card_space = size[1] * 0.04
            else:
                self.headerHeight = size[1] * 0.075
                self.statusHeight = size[1] * 0.062
                self.tabHeight = size[1] * 0.1
                self.card_space = size[1] * 0.03
        else:
            # 受信機
            self.headerHeight = size[1] * 0.088
            self.statusHeight = size[1] * 0.088

    def on_sc1_scroll_size(self, instance, new_size):
        """on_sc1_scroll_size
        
        Args:
            instance (object): イベント発生オブジェクト
            new_size (tuple): size
        """
        if new_size[1] > 0:
            self.changeTabDisp()

    def on_sc2_scroll_size(self, instance, new_size):
        """on_sc2_scroll_size
        
        Args:
            instance (object): イベント発生オブジェクト
            new_size (tuple): size
        """
        if new_size[1] > 0:
            self.changeTabDisp()

    def on_timediff_change(self, instance, timeDiff):
        """時刻ずれイベントハンドラー

        Args:
            instance (object): イベント発生オブジェクト
            timeDiff (bool): 表示機と受信機の時刻ずれ（True:あり）
        """
        self.timedelta.request_timedelta(force=True)


    def on_selectedIndex_change(self, instance, selectedIndex):
        self.updateSelectFrame()

    def start(self):
        """開始
        """

        # 経過時間更新スケジュール開始
        Clock.schedule_interval(lambda dt: self.updateTimeLabels(), 2)

        if util.is_receiver_device():
            Clock.schedule_interval(lambda dt: self.updateSelectFrame(), 10)


    def stop(self):
        """
        終了
        """
        pass

    def addProcQueue(self, method, args):
        """ Controlerからのメソッド呼出をQueueに格納する。
        Args:
            method (func): 呼び出すメソッド
            args (array): 引数の配列.
        """
        self.__proc_queue.put({"method": method, "args": args})                    

    def _run_proc_queue(self, dt):
        """ Queueに格納されたメソッド呼出しを実行する。
        """

        if not self.__proc_queue.empty():
            locked = self.__widget_lock_acquire()
            if locked:
                clearEvents = True
                procInfo = self.__proc_queue.get_nowait()
                if procInfo is not None:
                    procInfo["method"](*procInfo["args"])
                    if procInfo["method"].__name__ == self.delCards.__name__:
                        clearEvents = False

                self.__widget_lock_release(clearEvents)

    def __widget_lock_acquire(self):
        """ ロック取得
        Returns:
            result: True:ロック取得、False：ロックが取得できなかった（処理中）
        """
        ret = self.__widget_lock.acquire(blocking=False)
        if ret:
            for e in self.__wait_events: #削除アニメーションの実行を待つ
                ret = e.wait(0.1)
                if ret == False:
                    break
            if ret == False:
                self.__widget_lock.release()

        return ret
    
    def __widget_lock_release(self, clearEvents=True):
        """ロック解放
        Args:
            clearEvents (bool): 削除アニメーション完了イベント配列のクリア
        """
        if clearEvents:
            self.__wait_events.clear()
        self.__widget_lock.release()

    def addCards(self, datalist):
        """カード追加
        
        Args:
            datalist (list): 追加データのリスト
            sort (bool, optional): ソートフラグ. Defaults to False.
        """
        sort = len([d for d in datalist if d["priority"] == 1]) > 0 
        self.add_cards(datalist, False, sort=sort)


    def add_cards(self, datalist, hist, sort=False):
        """カード追加
        
        Args:
            datalist (list): 追加データのリスト
            hist (bool): 履歴フラグ
            sort (bool, optional): ソートフラグ. Defaults to False.
        """
        try:
            cards = None
            if hist:
                cards = self.hist_cards
            else:
                cards = self.cards

            if hist:
                sc = self.ids.sc2
            else:
                sc = self.ids.sc1

            for d in datalist:
                oldcard = None
                if hist == False:
                    for card in cards:
                        if  (d['address'] == card.address
                            and d['modelNo'] == card.modelNo
                            and d['buttonNo'] == card.buttonNo):
                            oldcard = card
                            break

                if oldcard != None:
                    # 呼出中の場合、既存のカードは追加しない
                    oldcard.data['recall'] = d['recall']
                    oldcard.blink = True if d['recall'] == 1 else False
                    if 'checked' in d:
                        oldcard.data['checked'] = d['checked']
                        oldcard.updateBackground()


                else:
                    num = len(cards)
                    # サイズ、位置を計算
                    size = self.calc_card_size(num, num + 1)
                    pos = self.calc_card_pos(num, num + 1)

                    if self.landscape:
                        pos = (sc.size[0], pos[1])
                    else:
                        pos = (pos[0], -size[1])

                    # 追加
                    # self.logger.debug('add card_pos={}'.format(pos))
                    # self.logger.debug('add card_size={}'.format(size))

                    if hist:
                        newcard = HistCardWidget(pos=pos, size=size)
                    else:
                        newcard = CardWidget(pos=pos, size=size)

                    newcard.timedelta = self.timedelta
                    newcard.data = d.copy()

                    if util.is_display_device():
                        newcard.opacity = 0

                    sc.add_widget(newcard, num)

                    if util.is_display_device():
                        # 表示アニメーション開始
                        newcard.showStart()


                    cards.append(newcard)
                    num = num + 1

                    # snapshot = tracemalloc.take_snapshot()
                    # if self.prev_snap is not None:
                    #     stats = snapshot.compare_to(self.prev_snap, 'lineno')
                    #     for stat in stats[:10]:
                    #         print(stat)

                    # self.prev_snap = snapshot

                    # snapshot = tracemalloc.take_snapshot()
                    # top_stats = snapshot.statistics('lineno')

                    # print("[ Top 10 ]")
                    # for stat in top_stats[:10]:
                    #     print(stat)

            if hist:
                if len(cards) > self.MAX_HIST_NUM:
                    sc.clear_widgets()
                    __cards = cards[-self.MAX_HIST_NUM:]
                    for card in __cards:
                        sc.add_widget(card)
                    self.hist_cards.clear()
                    self.hist_cards += __cards

            # 件数表示の更新
            self.showStatus() 

            # カードの表示を更新
            doAnim = False
            if util.is_display_device():
                doAnim = True

            self.move_cards(hist=hist, sort=sort, doAnim=doAnim)

            if util.is_display_device():
                if hist and len(self.hist_cards) > 0:
                    # 最新の履歴へスクロール
                    self.ids.sc2_scroll.scroll_to(self.hist_cards[-1])
    
            self.updateSelectFrame()

        except:
            self.logger.error("add_cards failed.", exc_info=True)

        

    def calc_card_size(self, idx, num):
        """カードのサイズを計算する
        
        Args:
            idx (int): インデックス
            num (int): カード数
        
        Returns:
            tuple: サイズ
        """
        if self.histMode:
            sc_scroll = self.ids.sc2_scroll
        else:
            sc_scroll = self.ids.sc1_scroll

        if util.is_display_device():
            # 表示機
            screen_size = sc_scroll.size

            width = 0
            height = 0
            if screen_size[0] / screen_size[1] > 1.56:
                # 横長
                margin = self.card_space
                height = screen_size[1] - margin * 2
                width = height * 1.56
            else:
                # 縦長
                margin_ratio = 0.1
                margin = screen_size[0] * margin_ratio
                width = screen_size[0] - margin * 2
                height = width/1.56
            return width, height

        else:
            # 受信機
            screen_size = sc_scroll.size
            if num < 1:
                return None
            elif num == 1:
                # 1画面レイアウト
                margin = self.size[1] * 0.055
                height = screen_size[1] - margin * 2
                return height * 1.56, height
            elif num == 2:
                # 2画面レイアウト
                margin = screen_size[0] * 0.031
                width = (screen_size[0] - margin * 3)/2
                return width, width/1.56
            elif num >= 3 and num <= 4:
                # 3～4画面レイアウト
                margin = self.size[1] * 0.055
                height = (screen_size[1] - margin * 3)/2
                return height * 1.56, height

            elif num >= 5 and num <= 7:
                # 5～7画面レイアウト
                margin3 = screen_size[0] * 0.031
                margin4 = screen_size[0] * 0.025
                margin5 = screen_size[0] * 0.025

                if idx < 3:
                    width = (screen_size[0] - margin3 * 2 - margin4 * 2 )/3
                    return width, width/1.56
                else:
                    width = (screen_size[0] - margin3 * 2 - margin5 * 3 )/4
                    return width, width/1.56

            else:
                # 12画面レイアウト
                margin3 = screen_size[0] * 0.031
                margin4 = screen_size[0] * 0.025
                width = (screen_size[0] - margin3 * 2 - margin4 * 3 )/4
                return width, width/1.83

    
    def calc_card_pos(self, idx, num):
        """カードの位置を計算する
        
        Args:
            idx (int): インデックス
            num (int): カード数
        
        Returns:
            tuple: 位置
        """
        card_size = self.calc_card_size(idx, num)

        if self.histMode:
            sc = self.ids.sc2
            sc_scroll = self.ids.sc2_scroll
        else:
            sc = self.ids.sc1
            sc_scroll = self.ids.sc1_scroll

        if util.is_display_device():
            # 表示機
            if self.landscape:
                pos_start = (self.card_space, sc.pos[1] + (sc_scroll.height - card_size[1])/2)
                pos = (pos_start[0] + (card_size[0] + self.card_space) * idx , pos_start[1]  )
            else:
                area_height = (card_size[1] + self.card_space) * num + self.card_space
                if area_height < sc_scroll.height:
                    area_height = sc_scroll.height
                pos_start = ( sc.pos[0] + (sc_scroll.width - card_size[0])/2, sc.pos[1] + area_height - card_size[1] - self.card_space)
                pos = ( pos_start[0] , pos_start[1]  - (card_size[1] + self.card_space) * idx )
        else:
            # 受信機
            pos = (0, 0)
            base_pos = sc_scroll.pos
            screen_size = sc_scroll.size
            if num == 1:
                margin = self.size[1] * 0.055
                pos = (base_pos[0] + (screen_size[0] - card_size[0])/2, base_pos[1] + margin)

            elif num == 2:
                # 2画面レイアウト
                margin = screen_size[0] * 0.031
                pos = (base_pos[0] + card_size[0] * idx + margin * (idx + 1), base_pos[1] + (screen_size[1] - card_size[1])/2)

            elif num >= 3 and num <= 4:
                # 3～4画面レイアウト
                margin = self.size[1] * 0.055
                start_x = base_pos[0] + (screen_size[0] - (card_size[0] * 2 +  margin))/2
                if idx < 2:
                    n = 1
                else:
                    n = 0
                pos = (start_x + card_size[0] * (idx % 2) + margin * (idx % 2), base_pos[1] + card_size[1] * n + margin * (n + 1))

            elif num >= 5 and num <= 7:
                # 5～7画面レイアウト
                card_size1 = self.calc_card_size(0, 5)
                card_size2 = self.calc_card_size(3, 5)
                margin3 = screen_size[0] * 0.031
                margin4 = screen_size[0] * 0.025
                margin5 = screen_size[0] * 0.025
                margin7 = self.size[1] * 0.055
                start_y = base_pos[1] + (screen_size[1] - (card_size1[1] + card_size2[1] +  margin7))/2

                if idx < 3:
                    pos = (base_pos[0] + margin3 + card_size[0] * idx + margin4 * idx, start_y + card_size2[1] + margin7)
                else:
                    pos = (base_pos[0] + margin3 + card_size[0] * (idx - 3) + margin5 * (idx - 3), start_y)

            elif num >= 8:
                # 12画面レイアウト
                margin3 = screen_size[0] * 0.031
                margin4 = screen_size[0] * 0.025
                margin5 = self.size[1] * 0.044
                start_y = base_pos[1] + (screen_size[1] - (card_size[1] * 3 +  margin5 * 2))/2

                if idx < 12:
                    if idx < 4:
                        n = 2
                    elif idx < 8:
                        n = 1
                    else:
                        n = 0
                    pos = (base_pos[0] + margin3 + card_size[0] * (idx % 4) + margin4 * (idx % 4), start_y + card_size[1] * n + margin5 * n)
                else:
                    #13画面目以降は隠す
                    pos = (base_pos[0] - card_size[0], base_pos[1] - card_size[1])

        return pos
                
    def delCards(self, datalist, waitAnim=True):
        """カード削除
        
        Args:
            datalist (list): 削除対象データのリスト
        """

        for d in datalist:
            # del
            event = None
            if waitAnim == True:
                event = threading.Event()
                self.__wait_events.append(event)

            self.delCard(d, event)

    def delCard(self, data, event):
        """カード削除
        
        Args:
            data (dict): 削除対象データ
        """
        oldcard = None
        for card in self.cards:
            if  (data['address'] == card.address
                and data['modelNo'] == card.modelNo
                and data['buttonNo'] == card.buttonNo):
                oldcard = card
                break

        if oldcard != None:
            # if self.histMode == False:
            if self.histMode == False and util.is_display_device():
                # 削除アニメーション実行
                if self.landscape:
                    dist_pos = (oldcard.pos[0], - oldcard.size[1])
                else:
                    dist_pos = (self.ids.sc1_scroll.pos[0] + self.ids.sc1_scroll.width, oldcard.pos[1])

                oldcard.blinkStop()
                oldcard.doRemoveAction(dist_pos, lambda anim, *args: self._del_anim_callback(oldcard, event))
            else:
                #すぐ消す
                # widgetを削除
                oldcard.blinkStop()
                self.ids.sc1.remove_widget(oldcard)
                self.cards.remove(oldcard)
                oldcard.clean()
                del oldcard
                # 件数表示更新
                self.showStatus()

                # カードの表示を更新
                doAnim = False
                if util.is_display_device():
                    doAnim = True

                self.move_cards(hist=False, sort=True, doAnim=doAnim)

                self.updateSelectFrame()
                if event is not None:
                    event.set()
        else:
            if event is not None:
                event.set()


                # import gc
                # gc.collect()
                # snapshot = tracemalloc.take_snapshot()
                # top_stats = snapshot.statistics('lineno')

                # print("[ Top 10 ]")
                # for stat in top_stats[:10]:
                #     print(stat)

    def _del_anim_callback(self, oldcard, event):
        """削除アニメーション実行後に呼ばれるコールバック関数
        
        Args:
            oldcard (object): 削除Widget
        """
        try:
            self.logger.debug('_del_anim_callback enter.')

            # widgetを削除
            self.ids.sc1.remove_widget(oldcard)
            if oldcard in self.cards:
                self.cards.remove(oldcard)
                oldcard.clean()
                del oldcard

            # 件数表示更新
            self.showStatus()

            # カードの表示を更新
            self.move_cards(hist=False, sort=True)
        except:
            self.logger.error("_del_anim_callback failed.", exc_info=True)
        finally:
            if event is not None:
                event.set()
            self.logger.debug('_del_anim_callback exit.')


    def addHistCards(self, datalist):
        """履歴領域への表示
        
        Args:
            datalist (list): 対象データのリスト
        """
        if util.is_display_device():
            self.add_cards(datalist, hist=True)


    def move_cards(self, hist, sort = False, doAnim = True):
        """カードの移動
        
        Args:
            hist (bool): 移動対象 True:履歴、False:呼出中
            sort (bool, optional): ソートフラグ. Defaults to False.
            doAnim (bool, optional): アニメーション表示. Defaults to True.
        """

        cards = None

        if hist:
            cards = self.hist_cards
            sc = self.ids.sc2
        else:
            cards = self.cards
            sc = self.ids.sc1

        if sort and len(cards) > 0:
            __cards = sorted(cards, key=lambda x: x.recvDatetime)
            __cards = sorted(__cards, key=lambda x: x.priority, reverse=True)
            sc.clear_widgets()
            idx = 0
            for c in __cards:
                sc.add_widget(c, idx)
                idx += 1

            if hist:
                self.hist_cards.clear()
                self.hist_cards += __cards
            else:
                self.cards.clear()
                self.cards += __cards

        card_num = len(cards)
        last_idx  = card_num - 1
        first_pos = self.calc_card_pos(0, card_num)
        first_size = self.calc_card_size(0, card_num)
        last_pos = self.calc_card_pos(last_idx, card_num)
        last_size = self.calc_card_size(last_idx, card_num)

        # 画面スクロールのためにサイズの設定変更を行う。
        self.setSizeForScroll(hist, first_pos, first_size, last_pos, last_size)

        idx = 0
        for card in cards:
            # カードの位置とサイズを算出し、Animationで変更する。
            size = self.calc_card_size(idx, card_num)
            pos = self.calc_card_pos(idx, card_num)
            # self.logger.debug('move card_pos:{0}={1}'.format(idx, pos))
            if doAnim:
                anim = Animation(size = size, pos = pos, duration=0.1)
                anim.start(card)
            else:
                card.size = size
                card.pos = pos
            

            # if idx == 0:
            #     first_pos = pos
            #     first_size = size

            # last_pos = pos
            # last_size = size
            idx += 1


        # # 画面スクロールのためにサイズの設定変更を行う。
        # self.setSizeForScroll(hist, first_pos, first_size, last_pos, last_size)

        self.showStatus()


    def setSizeForScroll(self, hist, first_pos, first_size, last_pos, last_size):
        """画面スクロールのためにサイズの設定変更を行う。
        
        Args:
            hist (bool): 対象 True:履歴、False:呼出中
            first_pos (tuple): 最初のカードの位置
            first_size (tuple): 最初のカードのサイズ
            last_pos (tuple): 最後のカードの位置
            last_size (tuple): 最後のカードのサイズ
        """
        if util.is_display_device():
            # 表示機
            if hist:
                sc = self.ids.sc2
                sc_scroll = self.ids.sc2_scroll
            else:
                sc = self.ids.sc1
                sc_scroll = self.ids.sc1_scroll

            if self.landscape:
                if ( last_pos is not None  
                    and last_pos[0] + last_size[0] > sc_scroll.size[0]) :
                    sc.width = last_pos[0] + last_size[0] + self.card_space
                else:
                    sc.width = sc_scroll.width
            else:
                if ( first_pos is not None  
                    and first_pos[1] + first_size[1] > sc_scroll.size[1]) :
                    sc.height = first_pos[1] + first_size[1] + self.card_space
                else:
                    sc.height = sc_scroll.height
        else:
            # 受信機
            # スクロールさせなくてもよい
            self.ids.sc1.width = self.ids.sc1_scroll.width
            self.ids.sc1.height = self.ids.sc1_scroll.height


    def clearCards(self):
        """カードクリア
        """
        for card in self.cards:
            self.ids.sc1.remove_widget(card)
            card.clean()

        self.cards.clear()
        self.showStatus()

    def clearHistCards(self):
        """履歴領域クリア
        """
        for card in self.hist_cards:            
            self.ids.sc2.remove_widget(card)

        self.hist_cards.clear()
        self.showStatus()

    def updateTimeLabels(self):
        """経過時間表示の更新
        """
        if util.is_receiver_device():
            self.lang = TerminalInfo().language
        else:
            self.lang = ConfigSetting().language

        #時刻プロパティの更新
        now = datetime.now()
        self.nowTime = "{0:%H:%M}".format(now)

        for c in self.cards:
            c.updateTimeLabel()
        

    
    def showStatus(self):
        """件数表示
        """
        # self.logger.debug("histMode={0}, calling={1},history={2}".format(self.histMode, len(self.cards),len(self.hist_cards)))
        if self.histMode:
            self.queueNum = len(self.hist_cards)
        else:
            self.queueNum = len(self.cards)

    def on_touch_down(self, touch):
        """touch_downイベントハンドラ
        
        Args:
            touch (object): touch情報
        
        Returns:
            bool: Trueを返すとイベント処理終了
        """

        if self.histMode:
            cards = self.hist_cards
            sc = self.ids.sc2
        else:
            cards = self.cards
            sc = self.ids.sc1


        self.logger.debug('on_touch_down={}'.format(touch))

        self.dragging_widget = None
        self.drag_prev_pos = None
        self.drag_start_pos = None
        self.drag_start_time = None

        if self.histMode == False:
            nowtime = time.time()

            for c in cards:
                if c.collide_point(*c.to_widget(*touch.pos)) :
                    pos  = sc.to_widget(*touch.pos)
                    self.logger.debug('{0} is touched. {1}'.format( c, pos))
                    self.dragging_widget = c
                    self.drag_start_pos = pos
                    self.drag_start_time = nowtime
                    self.drag_prev_pos = pos
                    break

        if hasattr(self.ids, "configBtn") and self.ids.configBtn.collide_point(*self.ids.configBtn.to_widget(*touch.pos)) :
            return super(MainScreen, self).on_touch_down(touch)

        # if hasattr(self.ids, "menubox") and self.ids.menubox.collide_point(*self.ids.menubox.to_widget(*touch.pos)) :
        #     return super(MainScreen, self).on_touch_down(touch)

        if hasattr(self.ids, "ctltab") and self.ids.ctltab.collide_point(*self.ids.ctltab.to_widget(*touch.pos)) :
            return super(MainScreen, self).on_touch_down(touch)

        if sc.collide_point(*sc.to_widget(*touch.pos)) :
            return super(MainScreen, self).on_touch_down(touch)

        return True


    def on_touch_move(self, touch):
        """touch_moveイベントハンドラ
        
        Args:
            touch (object): touch情報
        
        Returns:
            bool: Trueを返すとイベント処理終了
        """
        self.logger.debug('on_touch_move={}'.format(touch))

        if self.histMode:
            sc = self.ids.sc2
        else:
            sc = self.ids.sc1

        if self.dragging_widget is not None:
            if self.drag_prev_pos is not None:
                pos  = sc.to_widget(*touch.pos)
                if self.landscape:
                    # 横画面の場合、縦方向の動きのみを反映
                    self.dragging_widget.pos = (self.dragging_widget.pos[0],  self.dragging_widget.pos[1] + pos[1] - self.drag_prev_pos[1])
                else:
                    # 縦画面の場合、横方向の動きのみを反映
                    self.dragging_widget.pos = (self.dragging_widget.pos[0] + pos[0] - self.drag_prev_pos[0],  self.dragging_widget.pos[1])
                self.drag_prev_pos = pos

        if sc.collide_point(*sc.to_widget(*touch.pos)) :
            return super(MainScreen, self).on_touch_move(touch)

        return True

    def on_touch_up(self, touch):
        """touch_upイベントハンドラ
        
        Args:
            touch (object): touch情報
        
        Returns:
            bool: Trueを返すとイベント処理終了
        """

        self.logger.debug('on_touch_up={}'.format(touch))
        nowtime = time.time()

        if self.histMode:
            sc = self.ids.sc2
        else:
            sc = self.ids.sc1

        if self.dragging_widget is not None:
            if self.drag_prev_pos is not None:
                pos  = sc.to_widget(*touch.pos)
                v = 0
                if self.landscape:
                    # 横画面の場合、縦方向の動きのみを反映
                    self.dragging_widget.pos = (self.dragging_widget.pos[0],  self.dragging_widget.pos[1] + pos[1] - self.drag_prev_pos[1])
                    if self.dragging_widget.top < sc.y + sc.height/2:
                        v = abs(pos[1] - self.drag_start_pos[1]) / (nowtime - self.drag_start_time)
                else:
                    # 縦画面の場合、横方向の動きのみを反映
                    self.dragging_widget.pos = (self.dragging_widget.pos[0] + pos[0] - self.drag_prev_pos[0],  self.dragging_widget.pos[1])
                    if self.dragging_widget.x > sc.x + sc.width/2:
                        v = abs(pos[0] - self.drag_start_pos[0]) / (nowtime - self.drag_start_time)
                if v > 1000: # 秒速1000pixel
                    self.logger.debug('fast')
                    delData = self.dragging_widget.data.copy()
                    # self.delCards([self.dragging_widget.data], lock=False)
                    # self.dispatch('on_request_erase', self.dragging_widget.data)
                    self.delCards([delData], waitAnim=False)
                    Clock.schedule_once(lambda dt: self.dispatch('on_request_erase', delData), 0.1)
                else:
                    # self.moveCards()
                    self.move_cards(hist=self.histMode)

        self.dragging_widget = None
        self.drag_start_time = None
        self.drag_prev_pos = None
        self.drag_start_pos = None

        if hasattr(self.ids, "configBtn") and self.ids.configBtn.collide_point(*self.ids.configBtn.to_widget(*touch.pos)) :
            return super(MainScreen, self).on_touch_up(touch)

        # if hasattr(self.ids, "menubox") and self.ids.menubox.collide_point(*self.ids.menubox.to_widget(*touch.pos)) :
        #     return super(MainScreen, self).on_touch_up(touch)

        # if hasattr(self.ids, "menubox"):
        #     self.ids.menubox.opacity = 0

        if hasattr(self.ids, "ctltab") and self.ids.ctltab.collide_point(*self.ids.ctltab.to_widget(*touch.pos)) :
            return super(MainScreen, self).on_touch_up(touch)

        if sc.collide_point(*sc.to_widget(*touch.pos)) :
            return super(MainScreen, self).on_touch_up(touch)

        return True


    def tabChanged(self, tabName):
        """タブが切り替わったときに呼ばれる
        
        Args:
            tabName (string): タブ名
        """
        self.logger.debug("tabChanged {}".format(tabName))

        if tabName == "calling":
            self.histMode = False
            # self.selectedTabName = self.res.getResource("mainscreen:calling")

        elif tabName == "history":
            self.histMode = True
            # self.selectedTabName = self.res.getResource("mainscreen:history")

        self.changeTabDisp()

    def getSelectedTabName(self, tabName, lang):
        """選択されたタブ名取得（多言語）

        Args:
            tabName (string): タブ名(eng)
            lang (string): 言語

        Returns:
            string: 選択されたタブ名（多言語）
        """
        if hasattr(self.ids, "ctltab"):
            if self.ids.ctltab.selectedTab == 'calling':
                return self.ids.ctltab.tabNames[0]
            else:
                return self.ids.ctltab.tabNames[1]

    def changeTabDisp(self):
        """タブの状態により、履歴、呼出中どちらを見せるかと、スクロール領域のサイズを切り替える
        """
        if util.is_display_device():
            if self.histMode:
                self.ids.sc1_scroll.size_hint_y = None
                self.ids.sc1_scroll.height = 0
                self.ids.sc1_scroll.opacity = 0

                self.ids.sc2_scroll.size_hint_y = None
                self.ids.sc2_scroll.height = self.size[1] - self.headerHeight - self.statusHeight - self.tabHeight
                self.ids.sc2_scroll.opacity = 1

                self.move_cards(hist=True, doAnim=False)
                if len(self.hist_cards):
                    # 最新の履歴へスクロール
                    self.ids.sc2_scroll.scroll_to(self.hist_cards[-1])
            else:            
                self.ids.sc2_scroll.size_hint_y = None
                self.ids.sc2_scroll.height = 0
                self.ids.sc2_scroll.opacity = 0

                self.ids.sc1_scroll.size_hint_y = None
                self.ids.sc1_scroll.height = self.size[1] - self.headerHeight - self.statusHeight - self.tabHeight
                self.ids.sc1_scroll.opacity = 1

                self.move_cards(hist=False, doAnim=False)
        else:
            self.ids.sc2_scroll.height = 0
            self.ids.sc2_scroll.opacity = 0

            self.ids.sc1_scroll.size_hint_y = None
            self.ids.sc1_scroll.height = self.size[1] - self.headerHeight - self.statusHeight
            self.ids.sc1_scroll.opacity = 1

            self.move_cards(hist=False, doAnim=False)


    def showProcessing(self):
        """処理中表示
        """
        if hasattr(self.ids, "processing"):
            self.ids.processing.opacity = 1

        if util.is_display_device():
            if self.histMode:
                self.ids.sc2_scroll.opacity = 0

    def hideProcessing(self):
        """処理中非表示
        """
        if hasattr(self.ids, "processing"):
            self.ids.processing.opacity = 0

        if util.is_display_device():
            if self.histMode:
                self.ids.sc2_scroll.opacity = 1

    def getRgba(self, color, alpha=None):
        """RGBA取得
        
        Args:
            color (string): #ffffff形式の色コード
            alpha (float, optional): 不透明度. Defaults to None.
        
        Returns:
            list: RGBA
        """
        rr = int(color[1:3], 16) / 255.0
        gg = int(color[3:5], 16) / 255.0
        bb = int(color[5:7], 16) / 255.0
        a = 1
        if alpha is not None:
            a = alpha
        return [rr,gg,bb,a]

    def getAppNameColor(self, disconnected):
        """AppNameの色を取得する。（表示機用）

        Args:
            disconnected (bool): 接続状態(True:切断)

        Returns:
            list: 色[R,G,B,A]
        """
        if disconnected:
            return [1, 0, 0, 1]
        else:
            return [1, 1, 1, 1]

    def eraserPushed(self, eraserName, buttonNo):
        """消去機押下

        Args:
            eraserName (string): 消去機名
            buttonNo (string): ボタン番号
        """
        self.selectFrame.eraserName = eraserName
        if buttonNo == "05":
            #消去要求イベントを発行
            if self.selectedIndex >= 0 and self.selectedIndex < len(self.cards):
                data = self.cards[self.selectedIndex].data
                data["displayName"] = self.selectFrame.eraserName
                self.dispatch('on_request_erase', data)
            #     if self.selectedIndex >= len(self.cards) - 1 and self.selectedIndex != 0:
            #         self.selectedIndex = len(self.cards) - 2
            # elif len(self.cards) > 0 and self.selectedIndex < 0:
            #     self.selectedIndex = 0
            # elif len(self.cards) > 0 and self.selectedIndex >= len(self.cards):
            #     self.selectedIndex = len(self.cards) - 1

        else:
            #選択枠の移動
            upDown = 0
            leftRight = 0
            if buttonNo == "01":
                #上
                upDown = -1
            elif buttonNo == "02":
                #右
                leftRight = 1
            elif buttonNo == "03":
                #下
                upDown = 1
            elif buttonNo == "04":
                #左
                leftRight = -1

            self.selectedIndex = self.calc_next_sel_idx(len(self.cards), self.selectedIndex, upDown, leftRight)


    def updateSelectFrame(self):
        """選択枠の表示更新
        """
        if util.is_receiver_device():
            #消去機が登録されているか？
            er = Eraser()
            els = er.select({})
            if len(els) == 0:
                self.selectFrame.hide()
                return

            if hasattr(self.ids, "sc1"):
                sc = self.ids.sc1
                if len([f for f in sc.children if isinstance(f, SelectFrame)]) == 0:
                    sc.add_widget(self.selectFrame, 9999) 

            if self.selectedIndex < 0:
                self.selectedIndex = 0
            elif self.selectedIndex >= len(self.cards) and self.selectedIndex != 0:
                if len(self.cards) > 12:
                    self.selectedIndex = 11 
                else:                    
                    self.selectedIndex = len(self.cards) - 1 

            # if self.selectedIndex >= 0 and self.selectedIndex < len(self.cards) and self.selectedIndex < 12:
            if len(self.cards) > 0:
                card_num = len(self.cards)
                pos = self.calc_card_pos(self.selectedIndex, card_num)
                size = self.calc_card_size(self.selectedIndex, card_num)
                self.selectFrame.size = size
                self.selectFrame.pos = pos
                self.selectFrame.show()

            else:
                self.selectFrame.hide()


    def calc_next_sel_idx(self, card_num, selIdx, upDown, leftRight):
        """選択枠の移動先Indexを計算する。

        Args:
            card_num (int): 呼出機の数
            selIdx (int): 選択中Index
            upDown (int): 上下（上：-1, 下：1）への移動
            leftRight (int): 左右（左：-1, 右：1）への移動

        Returns:
            int: 移動先Index
        """

        #選択中の呼出機の段数(0～)を算出
        now_dan_idx = self.__calc_dan_idx(card_num, selIdx)
        if now_dan_idx >= 0:
            #上下への移動
            if upDown != 0:
                #上下移動の場合の移動先を計算する
                nextIdx = self.__calc_idx_updown(card_num, selIdx, upDown)
                return nextIdx

            #左右への移動
            elif leftRight !=0:
                nextIdx = selIdx + leftRight
                next_dan_idx = self.__calc_dan_idx(card_num, nextIdx)
                if now_dan_idx == next_dan_idx:
                    return nextIdx
                else:
                    #段が違う場合は、選択枠は移動できない
                    return selIdx
        else:
            return selIdx

    def __calc_dan_num(self, card_num):
        if card_num <= 0:
            return 0
        elif card_num <=2:
            return 1 
        elif card_num <=8:
            return 2
        else:
            return 3

    def __calc_dan_idx(self, card_num, idx):
        if idx < 0:
            return -1
        if card_num < 0:
            return -1
        if idx >= card_num:
            return -1

        if card_num <=2:
            return 0 
        elif card_num <=4:
            return idx // 2
        elif card_num <=7:
            idx2 = idx // 3
            return idx2 if idx2 < 2 else 1
        elif card_num <=12:
            return idx // 4
        else:
            if idx < 12:
                return idx // 4
            else:
                return -1

    def __calc_idx_updown(self, card_num, idx, upDown):
        if idx < 0:
            return idx
        if card_num < 0:
            return idx
        if idx >= card_num:
            return idx
        if upDown != 1 and upDown != -1:
            return idx

        now_dan_idx = self.__calc_dan_idx(card_num, idx)
        dan_num = self.__calc_dan_num(card_num)

        next_dan_idx = now_dan_idx + upDown
        if next_dan_idx >= 0 and next_dan_idx < dan_num:
            nextIdx = idx
            if card_num <=4:
                nextIdx = idx + 2*upDown
            elif card_num <=7:
                if card_num == 7 and idx == 6:
                    nextIdx = idx + 4*upDown
                else:
                    nextIdx = idx + 3*upDown
            else:
                nextIdx = idx + 4*upDown

            if nextIdx < 0:
                nextIdx = 0
            elif nextIdx >= card_num:
                nextIdx = card_num - 1
            return nextIdx
        else:
            return idx

class ControlTab(BoxLayout):
    """コントロールタブ
    """
    selectedTab = StringProperty('')
    tabNames = ListProperty(["",""])
    landscape = BooleanProperty(False)

    # カスタムイベント
    __events__ = ('on_tab_changed',  )

    def on_tab_changed(self, *args):
        pass

    def __init__(self, **kwargs):
        """コンストラクタ
        """
        super(ControlTab, self).__init__(**kwargs)
        self.logger = setup_logger_devel(self.__class__.__name__)
        self.bind(selectedTab=self.on_selectedTab)

    def on_selectedTab(self, instance, new_val):
        """selectedTabのイベントハンドラ
        
        Args:
            instance (object): イベント発生オブジェクト
            new_val (string): タブ名
        """
        self.dispatch('on_tab_changed', new_val)

    def callingButtonPressed(self):
        """呼出中ボタンが押された
        """
        self.logger.debug('callingButtonPressed.')
        self.selectedTab = 'calling'

    def historyButtonPressed(self):
        """履歴ボタンが押された
        """
        self.logger.debug('historyButtonPressed.')
        self.selectedTab = 'history'

class TabButton(BoxLayout):
    """タブボタン
    """
    selected = BooleanProperty(False)
    background_color = ListProperty() 
    selected_background_color = ListProperty() 
    backgroundColor = ListProperty()
    image_source = StringProperty('')
    label_text = StringProperty('')
    label_color = ListProperty([])
    selected_label_color = ListProperty([])
    labelColor = ListProperty([])
    labelFontSize = NumericProperty(0)
    landscape = BooleanProperty(False)
    selectSize = ListProperty([1, 1])
    notSelectSize = ListProperty([1, 1])

    # カスタムイベント
    __events__ = ('on_press', 'on_release', )

    def __init__(self, **kwargs):
        """コンストラクタ
        """
        super(TabButton, self).__init__(**kwargs)
        self.logger = setup_logger_devel(self.__class__.__name__)
        self.bind(background_color=lambda _, __: self.setColor())
        self.bind(selected_background_color=lambda _, __: self.setColor())
        self.bind(label_color=lambda _, __: self.setColor())
        self.bind(selected_label_color=lambda _, __: self.setColor())
        self.bind(selected=lambda _, __: self.setColor())
        self.selectSize = self.size
        self.notSelectSize = self.size[0], self.size[1] * 9.4 / 10
        self.bind(size=lambda _, __: self.setSelectSize())
        self.bind(landscape=lambda _, __: self.setSelectSize())

    def setColor(self):
        """背景色、文字色の設定
        """
        if self.selected:
            self.backgroundColor = self.selected_background_color
            self.labelColor = self.selected_label_color
        else:
            self.backgroundColor = self.background_color
            self.labelColor = self.label_color

    def setSelectSize(self):
        """選択時のバックグラウンドのサイズを設定する
        """
        if self.landscape:
            # self.selectSize = self.size[0], self.size[1] * 18/16.6
            self.selectSize = self.size
            self.notSelectSize = self.size[0], self.size[1] * 16.6/18
        else:
            # self.selectSize = self.size[0], self.size[1] * 10/9.4
            self.selectSize = self.size
            self.notSelectSize = self.size[0], self.size[1] * 9.4/10

    def on_press(self, *args):
        """on_pressのデフォルトイベントハンドラ
        """
        self.logger.debug('on_press')
        self.selected = True

    def on_release(self, *args):
        pass

    def on_touch_down(self, touch):
        """on_touch_down
        
        Args:
            touch (object): touch情報
        
        Returns:
            bool: Trueを返すとイベント処理終了
        """
        self.logger.debug('on_touch_down pos={}'.format(touch.pos))
        if self.collide_point(*touch.pos) :
            self.logger.debug('dispatch on_press pos={}'.format(touch.pos))
            # on_press発火
            self.dispatch('on_press', touch)
            return True
        else:
            return super(TabButton, self).on_touch_down(touch)

    def on_touch_up(self, touch):
        """on_touch_up
        
        Args:
            touch (object): touch情報
        
        Returns:
            bool: Trueを返すとイベント処理終了
        """
        self.logger.debug('on_touch_up pos={}'.format(touch.pos))
        if self.collide_point(*touch.pos) :
            # on_release発火
            self.dispatch('on_release', touch)
            return True
        else:
            return super(TabButton, self).on_touch_up(touch)


class TimeDelta():
    """サーバとの時刻の差分
    """
    def __init__(self):
        """コンストラクタ
        """
        self.logger = setup_logger_devel(self.__class__.__name__)
        self.delta = 0
        self.prevtime = 0

        if util.is_display_device():
            self.request_timedelta()     
            Clock.schedule_interval(lambda dt: self.request_timedelta(), 10)

    def request_timedelta(self, force=False):
        """差分をサーバから取得する
        """
        if util.is_display_device():
            now = time.time()
            if force or self.delta == 0 or now - self.prevtime > 60:
                self.delta = udp.request_timedelta(ConfigSetting().server_ip)                
                self.logger.debug("delta={}".format(self.delta))
                self.prevtime = now

class SelectFrame(BoxLayout):
    """SelectFrame
    """
    background_color =ListProperty([0,0,0,0])
    frame_color =ListProperty([0,0,0,0])
    lineWidth = NumericProperty(1)
    eraserName = StringProperty("")
    hide_interval = NumericProperty(10.0)


    def __init__(self, **kwargs):
        """コンストラクタ
        
        """
        self.logger = setup_logger_devel(self.__class__.__name__)        
        super(SelectFrame, self).__init__(**kwargs)
        self.opacity = 0
        self.lineWidth = self.size[0] * 0.015
        self.eraserName = ""
        self.bind(size=self.on_size)

    def on_size(self, instance, new_size):
        """on_size
        
        Args:
            instance (object): イベント発生オブジェクト
            new_size (tuple): サイズ
        """
        self.lineWidth = new_size[0] * 0.015

    def hide(self):
        self.opacity = 0

    def show(self):
        self.opacity = 1

    def isShown(self):
        return self.opacity == 1

class CallStopButton(ButtonBehavior, BoxLayout):
    """呼出停止ボタン
    """

    background_color = ListProperty([0,0,0,1]) 
    checked = BooleanProperty(False)
    data = ObjectProperty(None)

    def __init__(self, **kwargs):
        """コンストラクタ
        
        """
        self.logger = setup_logger_devel(self.__class__.__name__)        
        super(CallStopButton, self).__init__(**kwargs)
        self.drawBackground()
        # self.bind(size=lambda _, __: self.updateBackground())
        # self.bind(pos=lambda _, __: self.updateBackground())
        self.bind(checked=lambda _, __: self.updateBackground())
        self.bind(data=lambda _, __: self.updateBackground())

    def on_kv_post(self, __):
        """on_kv_post
        """
        super(CallStopButton, self).on_kv_post(__)
        self.drawBackground()

    def drawBackground(self):
        """背景図形の描画
        """
        self.canvas.before.clear()

        with self.canvas.before:
            Color(*self.background_color)
            self.button_img_rect = Rectangle(pos=self.pos,
                    size=self.size)


    def updateBackground(self):
        """背景図形の更新
        """
        if self.data is not None and "checked" in self.data:
            self.checked = self.data["checked"]

        if hasattr(self, "button_img_rect"):
            if self.checked:
                self.button_img_rect.texture = checked_img_texture
            else:
                self.button_img_rect.texture = bell_img_texture
            self.button_img_rect.pos = self.pos
            self.button_img_rect.size = self.size

        self.updateCheckButton()


    def updateCheckButton(self):
        if util.is_display_device() and ConfigSetting().loop and self.data is not None:
            self.opacity = 1
        else:
            self.opacity = 0

    def buttonPressed(self):
        """チェックボタン押下時動作
        """

        if self.opacity > 0 and self.data is not None and self.checked == False:
            self.data["checked"] = True
            self.checked = self.data["checked"]
            if util.is_android_device():
                send_broadcast_request(Type.STOP_SOUND ,self.data)
            self.logger.debug("data={}".format(self.data))

    def clean(self):
        del self.logger
        del self.button_img_rect
        self.canvas.clear()
        self.canvas.before.clear()
        self.canvas.after.clear()

