import base64
import hashlib
import hmac
import time

from setup_logger import setup_logger_devel


class OneTimePasswd:
    """ワンタイムパスワード

    RFC6238 Time-based One-time Password Algorithm (TOTP)　に従う
    """
    ONEDAY = 24 * 60 * 60   #1日
    # JSTSHIFT =  9 * 60 * 60 #JST：0:00に切り替えるためにUnixTimeをシフトする。
    TIMEZONE_SHIFT = - time.timezone #0:00に切り替えるためにUnixTimeをシフトする。
    def __init__(self, step=ONEDAY, shift=TIMEZONE_SHIFT, add_secret = None):
        """コンストラクタ
        
        Args:
            step (int), optional): パスワード切り替え間隔(秒). Defaults to ONEDAY.
            shift (int, optional): UnixTimeシフト（秒）.（1日単位で切り替える場合に、０時に切り替えるために指定する） Defaults to TIMEZONE_SHIFT.
            add_secret (bytes, optional): 追加のsecret bytes. Defaults to None.
        """
        self.logger = setup_logger_devel(self.__class__.__name__)
        self.secret = self.get_base_secret()
        if add_secret is not None:
            self.secret = self.secret + add_secret
        self.step = step
        self.shift = shift

    @property
    def secret_b32(self):
        """secretをBase32で返却する
        
        Returns:
            string: secret(Base32)
        """
        return base64.b32encode(self.secret).decode("utf-8")


    def calc(self,dt=None, keta=6): 
        """[summary]
        
        Args:
            dt (datetime, optional): 日時. Defaults to None.
            keta (int, optional): 桁数. Defaults to 6.
        
        Returns:
            string: ワンタイムパスワード
        """
        unixtime = time.time()
        if dt is not None:
            unixtime = dt.timestamp()

        messagetime = int(unixtime + self.shift)//self.step
        hexmessagetime = "{0:016x}".format(messagetime)
        d=hmac.new(self.secret,bytes.fromhex(hexmessagetime),hashlib.sha256).digest()

        offset = d[19] & 0xf
        sn = (d[offset]  & 0x7f) << 24 | \
        (d[offset+1] & 0xff) << 16 | \
        (d[offset+2] & 0xff) <<  8 | \
        (d[offset+3] & 0xff)        

        fmt = "{:0" + str(keta) + "d}"
        return fmt.format(sn % (10 ** keta))

    def verify(self, otpw, dt=None, keta=6):
        """ワンタイムパスワードの検証
        
        Args:
            otpw (string): ワンタイムパスワード
            dt (datetime, optional): 日時. Defaults to None.
            keta (int, optional): 桁数. Defaults to 6.
        
        Returns:
            bool: 検証結果
        """
        if otpw == self.calc(dt, keta):
            return True
        else:
            return False

    def get_base_secret(self):
        """Base secret取得
        
        Returns:
            bytes: Base Secret
        """
        return "ecomonitor".encode("utf-8")

class LoginDailyPasswd(OneTimePasswd):
    """ログイン用Dailyパスワード
    """
    def __init__(self, user):
        """コンストラクタ
        
        Args:
            user (string): ログインユーザー
        """
        super(LoginDailyPasswd, self).__init__(add_secret = user.encode())
        self.logger = setup_logger_devel(self.__class__.__name__)

    def get_base_secret(self):
        """Base secret取得
        
        Returns:
            bytes: Base Secret
        """
        return "ecomoni_login".encode("utf-8")


class UDPOneTimePasswd(OneTimePasswd):
    """UDPブロードキャスト用ワンタイムパスワード(60秒)
    """
    def __init__(self):
        super(UDPOneTimePasswd, self).__init__(step=60, shift=0, add_secret = None)
        self.logger = setup_logger_devel(self.__class__.__name__)

    def get_base_secret(self):
        """Base secret取得
        
        Returns:
            bytes: Base Secret
        """
        return "ecomonitor".encode("utf-8")
