from datetime import datetime

import netifaces

from kivy.clock import Clock
from kivy.properties import NumericProperty, StringProperty, BooleanProperty
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.popup import Popup
from kivy.uix.screenmanager import Screen
from recv_dev.config import TerminalInfo
from recv_dev.httpserver import WebSession
from resources import Resources
from setup_logger import setup_logger_devel


class PairingScreen(Screen):
    """ペアリングモード画面
    """

    headerHeight = NumericProperty(1)
    termNo = StringProperty('')
    ethIpAddress = StringProperty('')
    wlanIpAddress = StringProperty('')
    # signalStrength = StringProperty('')
    # 時刻プロパティ
    nowTime = StringProperty('')
    version = StringProperty('')
    lang = StringProperty('')
    res = Resources()
    callerRegisterMode = BooleanProperty(True)
    displayingWebPage = StringProperty('')


    def __init__(self, **kwargs):
        """コンストラクタ
        """
        super(PairingScreen, self).__init__(**kwargs)
        self.logger = setup_logger_devel(self.__class__.__name__)
        self.popup_dup = None
        self.popup_reg = None
        self.popup_otherType = None
        self.reg_close_event = None
        self.otherType_close_event = None
        self.headerHeight = self.height * 0.088
        self.clock_event = None
        self.bind(size=self.on_size)

    def on_enter(self):
        """画面表示時に実行される
        """
        self.updateProp()

        def __interval_proc(dt):
            self.updateTime(dt)
            self.updateIPAddress(dt)
            self.updateCallerRegisterMode()

        self.clock_event = Clock.schedule_interval( __interval_proc, 2)

    def on_leave(self):
        """画面削除時に実行される
        """
        if self.clock_event is not None:
            self.clock_event.cancel()
            self.clock_event = None

        self.popupCloseDuplication()
        self.popupCloseRegister()
        self.popupCloseOtherType()

    def on_size(self, instance, new_size):
        """on_size
        
        Args:
            instance (object): イベント発生オブジェクト
            new_size (tuple): サイズ
        """
        self.headerHeight = self.height * 0.088

    def updateProp(self):
        """画面表示用プロパティ更新
        """
        terminfo = TerminalInfo()
        self.lang = terminfo.language

        self.termNo = terminfo.termNo
        self.version = terminfo.firmwareVersion

        # addresses = netifaces.ifaddresses("eth0")
        # if netifaces.AF_INET in addresses:
        #     self.ethIpAddress = addresses[netifaces.AF_INET][0]["addr"]

        # addresses = netifaces.ifaddresses("wlan0")
        # if netifaces.AF_INET in addresses:
        #     self.wlanIpAddress = addresses[netifaces.AF_INET][0]["addr"]
        self.updateIPAddress(None)

        self.updateTime(None)
        self.updateCallerRegisterMode()

    def updateTime(self, dt):
        """時刻の更新
        
        Args:
            dt (timedelta): delta-time
        """
        now = datetime.now()
        #時刻プロパティの更新
        self.nowTime = "{0:%H:%M}".format(now)

    def updateIPAddress(self, dt):
        """IPアドレスの更新

        Args:
            dt (timedelta): delta-time
        """
        addresses = netifaces.ifaddresses("eth0")
        if netifaces.AF_INET in addresses:
            self.ethIpAddress = addresses[netifaces.AF_INET][0]["addr"]

        addresses = netifaces.ifaddresses("wlan0")
        if netifaces.AF_INET in addresses:
            self.wlanIpAddress = addresses[netifaces.AF_INET][0]["addr"]

    def updateCallerRegisterMode(self):
        self.displayingWebPage = WebSession().get_displaying_page()
        self.callerRegisterMode = self.displayingWebPage != "eraser.html"

    def getScreenMessage(self, lang, displayingWebPage):
        return self.res.getResource("pairing:screenMessage", lang) 
        # if displayingWebPage == "caller.html":
        #     return self.res.getResource("pairing:screenMessageCaller", lang)
        # elif displayingWebPage == "eraser.html":
        #     return self.res.getResource("pairing:screenMessageEraser", lang)
        # else:
        #     return self.res.getResource("pairing:screenMessage", lang) 

    def popupOpenDuplication(self, data, strength):
        """重複メッセージダイアログを表示する

        Args:
            callerData (dict): 呼出機データ
            strength (int): 電波強度
        """
        if self.callerRegisterMode:
            content = MessagePopupDup(callerData=data, strength=strength)
        else:
            content = MessagePopupDupEraser(eraserData=data, strength=strength)

        if self.popup_dup is None:
            # self.popup_dup = Popup(title='', content=content, size_hint=(0.5, 0.5), auto_dismiss=False, separator_color=[0,0,0,0], background='images/popup_background.png')
            self.popup_dup = BgPopup(title='', content=content, size_hint=(0.5, 0.5), auto_dismiss=False, separator_color=[0,0,0,0])
            self.popup_dup.open()
        else:
            self.popup_dup.content = content


    def popupCloseDuplication(self):
        """重複メッセージダイアログを消す
        """
        if self.popup_dup is not None:
            self.popup_dup.dismiss()
            self.popup_dup = None

    def getDeleteAddress(self):
        """重複メッセージダイアログに表示されている呼出機アドレスを返却する

        Returns:
            string: 呼出機アドレス
        """
        if self.popup_dup is not None:
            return self.popup_dup.content.address
        else:
            return ""


    def isShowingDupMsg(self):
        """重複メッセージダイアログが表示中か？

        Returns:
            bool: 表示中＝True
        """
        return self.popup_dup is not None

    def popupOpenRegister(self, data, strength):
        """登録メッセージダイアログを表示する

        Args:
            data (dict): 呼出機データ
            strength (int): 電波強度
        """
        if self.callerRegisterMode:
            content = MessagePopupReg(callerData=data, strength=strength)
        else:
            content = MessagePopupRegEraser(eraserData=data, strength=strength)
        if self.popup_reg is None:
            self.popup_reg = BgPopup(title='', content=content, size_hint=(0.5, 0.5), auto_dismiss=False, separator_color=[0,0,0,0])
            self.popup_reg.open()
            self.reg_close_event = Clock.schedule_once(lambda dt: self.popupCloseRegister(), 5)

        else:
            if self.reg_close_event is not None:
                self.reg_close_event.cancel()
                self.reg_close_event = None

            self.popup_reg.content = content
            self.reg_close_event = Clock.schedule_once(lambda dt: self.popupCloseRegister(), 5)

    def popupCloseRegister(self):
        """登録メッセージダイアログを消す
        """
        if self.reg_close_event is not None:
            self.reg_close_event.cancel()
            self.reg_close_event = None

        if self.popup_reg is not None:
            self.popup_reg.dismiss()
            self.popup_reg = None

    def popupOpenOtherType(self, data, strength):
        """呼出機(消去機)として登録済みメッセージダイアログを表示する

        Args:
            data (dict): 呼出機データ
            strength (int): 電波強度
        """
        if self.callerRegisterMode:
            content = MessagePopupRegisteredAsEraser(strength=strength, eraserData=data)
        else:
            content = MessagePopupRegisteredAsCaller(strength=strength, callerData=data)
        if self.popup_otherType is None:
            self.popup_otherType = BgPopup(title='', content=content, size_hint=(0.5, 0.5), auto_dismiss=False, separator_color=[0,0,0,0])
            self.popup_otherType.open()
            self.otherType_close_event = Clock.schedule_once(lambda dt: self.popupCloseOtherType(), 5)

        else:
            if self.otherType_close_event is not None:
                self.otherType_close_event.cancel()
                self.otherType_close_event = None

            self.popup_otherType.content = content
            self.otherType_close_event = Clock.schedule_once(lambda dt: self.popupCloseOtherType(), 5)

    def popupCloseOtherType(self):
        """呼出機(消去機)として登録済みメッセージダイアログを消す
        """
        if self.otherType_close_event is not None:
            self.otherType_close_event.cancel()
            self.otherType_close_event = None

        if self.popup_otherType is not None:
            self.popup_otherType.dismiss()
            self.popup_otherType = None


    def getRgba(self, color, alpha=None):
        """RGBA取得
        
        Args:
            color (string): #ffffff形式の色コード
            alpha (float, optional): 不透明度. Defaults to None.
        
        Returns:
            list: RGBA
        """
        rr = int(color[1:3], 16) / 255.0
        gg = int(color[3:5], 16) / 255.0
        bb = int(color[5:7], 16) / 255.0
        a = 1
        if alpha is not None:
            a = alpha
        return [rr,gg,bb,a]

class BgPopup(Popup):
    """backgroundを変更したPopup
    """
    pass
 
class MessagePopup(BoxLayout):
    """Popupのコンテンツ
    """
    callerName1 = StringProperty('')
    callerName2 = StringProperty('')
    eraserName = StringProperty('')
    address = StringProperty('')
    modelNo = StringProperty('')
    strength = NumericProperty(0)
    res = Resources()

    def __init__(self, strength=None, callerData=None, eraserData=None, **kwargs):
        """コンストラクタ
        
        Args:
            callerData (dict): 呼出機データ
        """
        super(MessagePopup, self).__init__(**kwargs)
        self.logger = setup_logger_devel(self.__class__.__name__)
        if callerData is not None:
            self.callerName1 = callerData["line1"]
            self.callerName2 = callerData["line2"]
            self.address = callerData["address"]
            self.modelNo = callerData["modelNo"]
        elif eraserData is not None:
            self.eraserName = eraserData["name"]
            self.address = eraserData["address"]
            self.modelNo = eraserData["modelNo"]

        if strength is not None:
            self.strength = strength

class MessagePopupDup(MessagePopup):
    """Popupのコンテンツ - 重複画面
    """
    pass

class MessagePopupReg(MessagePopup):
    """Popupのコンテンツ - 登録画面
    """
    pass

class MessagePopupDupEraser(MessagePopup):
    """Popupのコンテンツ - 重複画面
    """
    pass

class MessagePopupRegEraser(MessagePopup):
    """Popupのコンテンツ - 登録画面
    """
    pass


class MessagePopupRegisteredAsEraser(MessagePopup):
    pass

class MessagePopupRegisteredAsCaller(MessagePopup):
    pass
