import json
import pathlib
import traceback
import common_util as util
import os

class Config():
    """Configクラス
    """
    CONFIG_FILE = ""
    INIT_DATA = "{}"

    def __init__(self):
        """コンストラクタ
        """
        self.dict = None
        self.load()

    def load(self):
        """読み込み
        """
        try:
            if os.path.exists(self.CONFIG_FILE) == False:
                try:
                    self.dict = json.loads(self.INIT_DATA)
                    self.save()
                except:
                    traceback.print_exc()

            with open(self.CONFIG_FILE) as f:
                self.dict = json.loads(f.read())
        except:
            traceback.print_exc()

    def save(self):
        """保存
        """
        try:
            with open(self.CONFIG_FILE, "w") as f:
                f.write(json.dumps(self.dict, indent=4))
        except:
            traceback.print_exc()

    def initialize(self):
        """初期化
        """
        try:
            self.dict = json.loads(self.INIT_DATA)
        except:
            traceback.print_exc()

class TerminalInfo(Config):
    """端末設定クラス
    """
    CONFIG_FILE = pathlib.Path("recv_dev/config/terminal.json").resolve()
    INIT_DATA = """
{
    "termNo": "",
    "modelName": "BEC-100",
    "firmwareVersion": "1.0.0",
    "timezone": "Asia/Tokyo",
    "country": "",
    "language": "",
    "logLevel": 10
}    """

    def initialize(self):
        """初期化
        """
        self.termNo = ""
        self.timezone = "Asia/Tokyo"
        self.country = ""
        self.language = ""
        self.logLevel = 10


    @property
    def terminalInfo(self):
        """端末設定
        
        Returns:
            dictionary: 端末設定
        """
        if self.dict is not None and "language" in self.dict and len(self.dict["language"]) == 0:
            self.dict["language"] = util.get_system_language()
        if self.dict is not None and "country" in self.dict and len(self.dict["country"]) == 0:
            self.dict["country"] = util.get_system_country()
        return self.dict

    @property
    def termNo(self):
        """端末番号
        
        Returns:
            string: 端末番号
        """
        if self.dict is not None and "termNo" in self.dict:
            return self.dict["termNo"]
        else:
            return None

    @termNo.setter
    def termNo(self, termNo):
        """端末番号
        
        Args:
            termNo (string): 端末番号
        """
        if self.dict is not None:
            self.dict["termNo"] = termNo

    @property
    def firmwareVersion(self):
        """ファームウェアバージョン

        Returns:
            string: ファームウェアバージョン
        """
        if self.dict is not None and "firmwareVersion" in self.dict:
            return self.dict["firmwareVersion"]
        else:
            return None

    @property
    def timezone(self):
        """タイムゾーン
        
        Returns:
            string: タイムゾーン
        """
        if self.dict is not None and "timezone" in self.dict:
            return self.dict["timezone"]
        else:
            return None

    @timezone.setter
    def timezone(self, timezone):
        """タイムゾーン
        
        Args:
            timezone (string): タイムゾーン
        """
        if self.dict is not None:
            self.dict["timezone"] = timezone

    @property
    def country(self):
        """国
        
        Returns:
            country: 国
        """
        if self.dict is not None and "country" in self.dict:
            if self.dict["country"] is not None and len(self.dict["country"]):
                return self.dict["country"]
            else:
                return util.get_system_country()
        else:
            return util.get_system_country()

    @country.setter
    def country(self, country):
        """国
        
        Args:
            country (string): 国
        """
        if self.dict is not None:
            self.dict["country"] = country

    @property
    def language(self):
        """言語
        
        Returns:
            language: 言語
        """
        if self.dict is not None and "language" in self.dict:
            if self.dict["language"] is not None and len(self.dict["language"]) > 0:
                return self.dict["language"]
            else:
                return util.get_system_language()
        else:
            return util.get_system_language()

    @language.setter
    def language(self, language):
        """言語
        
        Args:
            language (string): 言語
        """
        if self.dict is not None:
            self.dict["language"] = language

    @property
    def dummyMode(self):
        """ダミーモード
        
        Returns:
            bool: ダミーモード=True, 通常モード=False
        """
        if self.dict is not None and "dummyMode" in self.dict:
            return self.dict["dummyMode"]
        else:
            return False

    @dummyMode.setter
    def dummyMode(self, val):
        """ダミーモード
        
        Args:
            val (bool): 設定値
        """
        if self.dict is not None:
            self.dict["dummyMode"] = val

    @property
    def logLevel(self):
        """ログレベル
        
        Returns:
            int: ログレベル
                RITICAL 50
                ERROR   40
                WARNING 30
                INFO    20
                DEBUG   10
                NOTSET  0
        """
        if self.dict is not None and "logLevel" in self.dict:
            return self.dict["logLevel"]
        else:
            # logging.DEBUG
            return 10

    @logLevel.setter
    def logLevel(self, logLevel):
        """ログレベル
        
        Args:
            logLevel (int): ログレベル
                RITICAL 50
                ERROR   40
                WARNING 30
                INFO    20
                DEBUG   10
                NOTSET  0
        """
        if self.dict is not None:
            self.dict["logLevel"] = logLevel



class OtherInfo(Config):
    """その他設定クラス
    """
    CONFIG_FILE = pathlib.Path("recv_dev/config/other.json").resolve()
    INIT_DATA = """
{
    "sound":{
        "volume": 30
    },
    "timeSetting":{
        "ntp": true,
        "ntpServer": "ntp.nict.jp",
        "interval": "H01"
    },
    "logClear": {
        "interval": "W01",
        "days": 90
    }
}    """
#     INIT_DATA = """
# {
#     "sound":{
#         "volume": 30
#     },
#     "timeSetting":{
#         "ntp": true,
#         "ntpServer": "ntp.nict.jp",
#         "interval": "H01"
#     },
#     "logServer":{
#         "ipAddress": "",
#         "startTime": "",
#         "interval": "D01"
#     }
# }    """

    @property
    def otherInfo(self):
        """その他設定
        
        Returns:
            dictionary: その他設定
        """
        return self.dict

    @otherInfo.setter
    def otherInfo(self, otherInfo):
        """その他設定
        
        Args:
            otherInfo (dictionary): dictionary
        """
        self.dict = otherInfo

    @property
    def sound(self):
        """音量設定
        
        Returns:
            dictionary: 音量設定
        """
        if self.dict is not None and "sound" in self.dict:
            return self.dict["sound"]
        else:
            return None

    @sound.setter
    def sound(self, sound):
        """音量設定
        
        Args:
            sound (dictionary): 音量設定
        """
        if self.dict is not None:
            self.dict["sound"] = sound


    @property
    def timeSetting(self):
        """時計合わせ設定
        
        Returns:
            dictionary: 時計合わせ設定
        """
        if self.dict is not None and "timeSetting" in self.dict:
            return self.dict["timeSetting"]
        else:
            return None

    @timeSetting.setter
    def timeSetting(self, timeSetting):
        """時計合わせ設定
        
        Args:
            timeSetting (dictionary): 時計合わせ設定
        """
        if self.dict is not None:
            self.dict["timeSetting"] = timeSetting

    @property 
    def logClear(self):
        """ログクリア設定

        Returns:
            dict: ログクリア設定
        """
        if self.dict is not None and "logClear" in self.dict:
            return self.dict["logClear"]
        else:
            return None

    @logClear.setter
    def logClear(self, logClear):
        """ログクリア設定

        Args:
            logClear (dict): ログクリア設定
        """
        if self.dict is not None:
            self.dict["logClear"] = logClear

    # @property
    # def logServer(self):
    #     """ログサーバー設定
        
    #     Returns:
    #         dictionary: ログサーバー設定
    #     """
    #     if self.dict is not None and "logServer" in self.dict:
    #         return self.dict["logServer"]
    #     else:
    #         return None

    # @logServer.setter
    # def logServer(self, logServer):
    #     """ログサーバー設定
        
    #     Args:
    #         logServer (dictionary): ログサーバー設定
    #     """
    #     if self.dict is not None and "logServer" in self.dict:
    #         self.dict["logServer"] = logServer

class NetworkInfo(Config):
    """ネットワーク設定クラス
    """
    CONFIG_FILE = pathlib.Path("recv_dev/config/network.json").resolve()
    INIT_DATA = """
{
    "eth": {
        "enabled": true,
        "autoIp": true,
        "ipAddress": "",
        "subnetMask": "",
        "defaultGw": "",
        "autoDNS": true,
        "dns1": "",
        "dns2": ""
    },
    "wlan": {
        "enabled": true,
        "ssidSettings": {
        }
    }
}    """

    @property
    def networkInfo(self):
        """ネットワーク設定
        
        Returns:
            dictionary: ネットワーク設定
        """
        return self.dict

    @networkInfo.setter
    def networkInfo(self, otherInfo):
        """ネットワーク設定
        
        Args:
            networkInfo (dictionary): ネットワーク設定
        """
        self.dict = otherInfo

    @property
    def eth(self):
        """有線LAN設定
        
        Returns:
            dictionary: 有線LAN設定
        """
        if self.dict is not None and "eth" in self.dict:
            return self.dict["eth"]
        else:
            return None

    @eth.setter
    def eth(self, eth):
        """有線LAN設定
        
        Args:
            eth (dictionary): 有線LAN設定
        """
        if self.dict is not None and "eth" in self.dict:
            self.dict["eth"] = eth

    @property
    def wlan(self):
        """無線LAN設定
        
        Returns:
            dictionary: 無線LAN設定
        """
        if self.dict is not None and "wlan" in self.dict:
            return self.dict["wlan"]
        else:
            return None

    @wlan.setter
    def wlan(self, wlan):
        """無線LAN設定
        
        Args:
            wlan (dictionary): 無線LAN設定
        """
        if self.dict is not None and "wlan" in self.dict:
            self.dict["wlan"] = wlan

class CallerSetting(Config):
    """呼出機設定クラス
    """
    CONFIG_FILE = pathlib.Path("recv_dev/config/caller.json").resolve()
    INIT_DATA = """
{
    "A": {
        "longPressTime": 2,
        "timeout": 10
    },
    "common": {
        "dupCheckIntervalMs": 20
    }
}    """

    @property
    def callerSetting(self):
        """呼出機設定
        
        Returns:
            dictionary: 呼出機設定
        """
        return self.dict

    @callerSetting.setter
    def callerSetting(self, callerSetting):
        """ネットワーク設定
        
        Args:
            callerSetting (dictionary): 呼出機設定
        """
        self.dict = callerSetting

    @property
    def modelA(self):
        """モデルA設定
        
        Returns:
            dictionary: モデルA設定
        """
        if self.dict is not None and "A" in self.dict:
            return self.dict["A"]
        else:
            return None

    @modelA.setter
    def modelA(self, modelASetting):
        """モデルA設定
        
        Args:
            modelASetting (dictionary): モデルA設定
        """
        if self.dict is not None and "A" in self.dict:
            self.dict["A"] = modelASetting

    @property
    def common(self):
        """共通設定
        
        Returns:
            dictionary: 共通設定
        """
        if self.dict is not None and "common" in self.dict:
            return self.dict["common"]
        else:
            self.dict["common"] = {
                "dupCheckIntervalMs": 20
            }
            self.save()
            return self.dict["common"]

    @common.setter
    def common(self, commonSetting):
        """共通設定
        
        Args:
            commonSetting (dictionary): モデルA設定
        """
        if self.dict is not None and "common" in self.dict:
            self.dict["common"] = commonSetting
