import os
import select
import time

import RPi.GPIO as GPIO

from absthread import AbstractThread


class GPIOControl(AbstractThread):
    """スイッチ制御クラス
    """
    SW4_PIN=12
    SW5_PIN=11
    PAIRING_PIN=SW4_PIN
    CLEAR_PIN=SW5_PIN
    SHUTDOWN_REQ_PIN=16
    SWITCH_ON = GPIO.LOW
    SWITCH_OFF = GPIO.HIGH

    PRESS_SHORT = 0
    PRESS_LONG = 5
    PRESS_LONG2 = 10

    def __init__(self, sw_callbacks, shutdown_callback, dummyMode=False):
        """コンストラクタ

        Args:
            sw_callbacks (dictionary): スイッチ押下時コールバック関数
            shutdown_callback (function): シャットダウン要求時コールバック
            dummyMode (bool, optional): ダミーモード. Defaults to False.
        """

        super(GPIOControl, self).__init__()
        self.dummyMode = dummyMode
        self.sw_callbacks = sw_callbacks
        self.shutdown_callback = shutdown_callback
        self.time = {
            self.PAIRING_PIN: 0,
            self.CLEAR_PIN: 0,
        }
        self.status = {
            self.PAIRING_PIN: self.SWITCH_OFF,
            self.CLEAR_PIN: self.SWITCH_OFF,
        }

        self.callback_status = {
            self.PAIRING_PIN: {
               self.PRESS_SHORT: False, 
               self.PRESS_LONG: False, 
               self.PRESS_LONG2: False, 
            },
            self.CLEAR_PIN: {
               self.PRESS_SHORT: False, 
               self.PRESS_LONG: False, 
               self.PRESS_LONG2: False, 
            }
        }
        if self.dummyMode:
            if os.path.exists("/tmp/gpio_pairing") == False:
                os.mkfifo("/tmp/gpio_pairing")
            if os.path.exists("/tmp/gpio_clear") == False:
                os.mkfifo("/tmp/gpio_clear")
            if os.path.exists("/tmp/gpio_shutdown") == False:
                os.mkfifo("/tmp/gpio_shutdown")

    def pre_proc(self):
        """前処理
        """
        if self.dummyMode == False:
            GPIO.setmode(GPIO.BOARD)
            GPIO.setup(self.PAIRING_PIN, GPIO.IN)
            GPIO.setup(self.CLEAR_PIN, GPIO.IN)
            GPIO.setup(self.SHUTDOWN_REQ_PIN, GPIO.IN)

    def thread_proc(self):
        """スレッドループ処理
        
        Returns:
            [bool]: 処理結果：Falseを返すとスレッド終了
        """
        if self.dummyMode == False:
            t = time.time()

            if GPIO.input(self.SHUTDOWN_REQ_PIN) == self.SWITCH_ON:
                if self.shutdown_callback is not None:
                    self.shutdown_callback()
                    return False

            inputs = {
                self.PAIRING_PIN: GPIO.input(self.PAIRING_PIN),
                self.CLEAR_PIN: GPIO.input(self.CLEAR_PIN)
            }

            self.__handle_input(inputs, t)

            time.sleep(0.1)
        else:
            # ダミーモード
            selectlist = [
                os.open("/tmp/gpio_pairing",  os.O_RDONLY | os.O_NONBLOCK),
                os.open("/tmp/gpio_clear",  os.O_RDONLY | os.O_NONBLOCK),
                os.open("/tmp/gpio_shutdown",  os.O_RDONLY | os.O_NONBLOCK),
            ]
            (rready, _, _) = select.select(selectlist, [], [], 0.1)
            for fd in rready:
                c = os.read(fd, 1024)
                c = c.decode().strip('\n')
                if fd == selectlist[2]:
                    if self.shutdown_callback is not None:
                        self.shutdown_callback()

                else:
                    key = ""
                    if fd == selectlist[0]:
                        pin = self.PAIRING_PIN
                        key = "pairing"
                    elif fd == selectlist[1]:
                        pin = self.CLEAR_PIN
                        key = "clear"

                    if self.sw_callbacks[key] is not None:
                        if c == "long":
                            self.sw_callbacks[key](pin, self.PRESS_LONG)
                        elif c == "long2":
                            self.sw_callbacks[key](pin, self.PRESS_LONG2)
                        else:
                            self.sw_callbacks[key](pin, 0)

            for fd in selectlist:
                os.close(fd)

        return True

    def __handle_input(self, inputs, t):
        """入力値処理
        
        Args:
            inputs (dictionary): 入力値 {12:1, 11:0}
            t (float): 時間
        """
        for (pin, inData) in inputs.items():
            if self.status[pin] != inData:
                # 状態が変わった
                if inData == self.SWITCH_ON:
                    # ON
                    self.time[pin] = t
                else:
                    # OFF
                    for k in self.callback_status[pin]:
                        self.callback_status[pin][k] = False
            else:
                pass

            if inData == self.SWITCH_ON:
                if self.sw_callbacks is not None:
                    key = ""
                    if pin == self.PAIRING_PIN:
                        key = "pairing"
                    else:
                        key = "clear"
                    if key in self.sw_callbacks:
                        if self.sw_callbacks[key] is not None:
                            if t - self.time[pin] >= self.PRESS_SHORT and self.callback_status[pin][self.PRESS_SHORT] == False:
                                self.sw_callbacks[key](pin, t - self.time[pin])
                                self.callback_status[pin][self.PRESS_SHORT] = True
                            elif t - self.time[pin] >= self.PRESS_LONG and self.callback_status[pin][self.PRESS_LONG] == False:
                                # 3秒長押し
                                self.sw_callbacks[key](pin, t - self.time[pin])
                                self.callback_status[pin][self.PRESS_LONG] = True
                            elif t - self.time[pin] >= self.PRESS_LONG2 and self.callback_status[pin][self.PRESS_LONG2] == False:
                                # 10秒長押し
                                self.sw_callbacks[key](pin, t - self.time[pin])
                                self.callback_status[pin][self.PRESS_LONG2] = True

            self.status[pin] = inData

    def post_proc(self):
        """後処理
        """
        if self.dummyMode == False:
            GPIO.cleanup()



if __name__ == "__main__": 
    def callback1(pin, time):
        print("button pressed {0}. {1}s".format(pin, time))
    def callback2(pin, time):
        print("button pressed {0}. {1}s".format(pin, time))
    def callback3():
        print("shutdown")

    callbacks = {
        "pairing": callback1,
        "clear":  callback2
    }

    s = GPIOControl(callbacks, callback3, True)
    try:
        s.start()
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        s.stop()
