import csv
import io
import json
import os
import pathlib
import shutil
import subprocess
import time
import uuid
import zipfile
from datetime import datetime
from http import cookies

import netifaces
from ntplib import NTPClient

import common_util as util
from onetimepasswd import LoginDailyPasswd
from recv_dev.config import CallerSetting, NetworkInfo, OtherInfo, TerminalInfo
from recv_dev.iwlist import IwList
from recv_dev.netconfig import DhcpcdConf, NetworkStatus, WpaConf
from recv_dev.storage import (Caller, CallerButton, CallerButtonView, CallLog,
                              Database, Display, Eraser, Group, WebUser)
from setup_logger import setup_logger_devel


class BaseMyHttpHandler():
    """基底HTTPハンドラー
    """

    @property
    def directory(self):
        return self.handlerObj.directory

    def __init__(self, simpleHTTPRequestHandlerObject):
        """コンストラクタ
        
        Args:
            simpleHTTPRequestHandlerObject (object): SimpleHTTPRequestHandlerのオブジェクト
        """
        self.logger = setup_logger_devel(self.__class__.__name__)
        # self.logger = logger
        self.handlerObj = simpleHTTPRequestHandlerObject

    def get_json_param(self):
        """JSONパラメータの取得
        
        Returns:
            dictionary: JSONパラメータ
        """
        return self.handlerObj.get_json_param()
 
    def get_param(self):
        """パラメータ取得(POST)
        
        Returns:
            dictionary: パラメータ
        """
        return self.handlerObj.get_param()

    def get_param_of_get(self):
        """パラメータ取得(GET)
        
        Returns:
            dictionary: パラメータ
        """
        return self.handlerObj.get_param_of_get()


    def get_multipart_form_data(self):
        """multipart/form-data取得
        
        Returns:
            list: multipart/form-data
        """
        return self.handlerObj.get_multipart_form_data()

    def response_json(self, response, cookie = None):
        """JSON応答
        
        Args:
            response (dictionary): 応答データ
            cookie (object, optional): cookie. Defaults to None.
        """
        self.handlerObj.response_json(response, cookie)

    def response_json_error(self, opt = None):
        """JSON応答(エラー)
        
        Args:
            opt (dictionary, optional): オプション. Defaults to None.
        """
        response = {"status": "error"}
        if opt is not None:
            for (key, value) in opt.items():
                response[key] = value 
        self.response_json(response)

    def response_json_login_error(self):
        """JSON応答(ログインエラー)
        """
        self.response_json_error({"error":"login"})

    def write_html_file(self, path, cookie = None):
        """HTMLファイル出力
        
        Args:
            path (string): ファイルパス
            cookie (object, optional): cookie. Defaults to None.
        """
        self.handlerObj.write_html_file(path, cookie)

    def send_response(self, code):
        self.handlerObj.send_response(code)

    def send_header(self, header, value):
        self.handlerObj.send_header(header, value)

    def end_headers(self):
        self.handlerObj.end_headers()

    def get_sessionId_from_cookie(self):
        """セッションIDをCookieから取得
        
        Returns:
            string: セッションID
        """
        return self.handlerObj.get_sessionId_from_cookie()

    def add_session_info(self, sessionInfo):
        """セッション情報追加
        
        Args:
            sessionInfo (dictionary): セッション情報
        """
        self.handlerObj.add_session_info(sessionInfo)

    def get_session_info(self, sessionId):
        """セッション情報取得
        
        Args:
            sessionId (string): セッションID
        
        Returns:
            dictionary: セッション情報
        """
        return self.handlerObj.get_session_info(sessionId)

    def remove_sessio_info(self, sessionId):
        """セッション情報削除
        
        Args:
            sessionId (string): セッションID
        """
        self.handlerObj.remove_sessio_info(sessionId)

    def check_login(self, priv=None):
        """ログインチェック
        
        Args:
            priv (int, optional): 権限. Defaults to None.
        
        Returns:
            bool: 結果
        """
        return self.handlerObj.check_login(priv)


class LoginHandler(BaseMyHttpHandler):
    """ログインハンドラー
    """

    def get_path_dictionary(self):
        """ハンドリングパスディクショナリーの返却
        
        Returns:
            dictionary: ハンドリングパスディクショナリー
        """
        path_dict = {
            "post":{
                '/login': self.login,
                '/logoff': self.logoff,
                '/set/password': self.setPassword,
            },
            "get":{

            }
        }
        return path_dict

    def login(self):
        """ログイン
        
        Returns:
            bool: Trueは処理完了
        """
        sessionId = self.get_sessionId_from_cookie()

        if sessionId is not None:
            self.remove_sessio_info(sessionId)

        (user, passwd) = self.__get_login_param()
        self.logger.debug('user={0}, password={1}'.format(user, passwd))

        priv = self.__get_privilege(user, passwd)

        if priv > 0:
            cookie = cookies.SimpleCookie()
            sessionId = str(uuid.uuid4())
            cookie["SESSIONID"] = sessionId
            cookie["SESSIONID"]["Path"] = "/"
            cookie["PRIVILEGE"] = priv
            cookie["PRIVILEGE"]["Path"] = "/"
            cookie["USER"] = user
            cookie["USER"]["Path"] = "/"
            cookie["LANGUAGE"] = TerminalInfo().language
            cookie["LANGUAGE"]["Path"] = "/"
            self.add_session_info({
                "sessionId": sessionId,
                "user": user,
                "privilege": priv,
                "time": time.time(), 
                "page": "" 
            })

            response = {"status": "ok"}
            self.response_json(response, cookie)

        else:
            self.response_json_login_error()

        return True


    def logoff(self):
        """ログオフ
        
        Returns:
            bool: Trueは処理完了
        """
        sessionId = self.get_sessionId_from_cookie()

        if sessionId is not None:
            self.remove_sessio_info(sessionId)

        cookie = cookies.SimpleCookie()
        cookie["SESSIONID"] = ""
        cookie["SESSIONID"]["expires"] = "expires=Fri, 31-Dec-1999 23:59:59 GMT"
        cookie["PRIVILEGE"] = ""
        cookie["PRIVILEGE"]["expires"] = "expires=Fri, 31-Dec-1999 23:59:59 GMT"
        cookie["USER"] = ""
        cookie["USER"]["expires"] = "expires=Fri, 31-Dec-1999 23:59:59 GMT"
        cookie["LANGUAGE"] = ""
        cookie["LANGUAGE"]["expires"] = "expires=Fri, 31-Dec-1999 23:59:59 GMT"

        response = {"status": "ok"}
        self.response_json(response, cookie)

        return True

    def setPassword(self):
        """パスワード変更
        
        Returns:
            bool: Trueは処理完了
        """
        if self.check_login() == False:
            self.response_json_login_error()
            return True

        param = self.get_json_param()
        self.logger.debug("param = {}".format(param))
        user = None
        oldpasswd = None
        newpasswd = None
        if "user" in param and "oldpasswd" in param and "newpasswd" in param:
            user = param["user"]
            oldpasswd = param["oldpasswd"]
            newpasswd = param["newpasswd"]
        else:
            self.response_json_error()
            return True


        sessionId = self.get_sessionId_from_cookie()        
        sessionInfo = self.get_session_info(sessionId)

        if sessionInfo["user"] != user:
            self.response_json_error()
            return True


        try:
            priv = self.__get_privilege(user, oldpasswd)
            if priv > 0:
                w = WebUser()
                hash = w.create_hash_salt(newpasswd)
                key = {
                    "userId":user
                }

                value = {
                    "hash":hash[0],
                    "salt":hash[1]
                }

                w.update(key, value)

                response = {"status": "ok"}
                self.response_json(response)

            else:
                self.response_json_error()

        except:
            self.logger.error("setPassword failed.", exc_info=True)
            self.response_json_error()

        return True

    def __get_login_param(self):
        """ログインパラメータの取得
        
        Returns:
            (string, string): ユーザ, パスワード
        """
        param = self.get_json_param()

        self.logger.debug("param = {}".format(param))
        user = None
        passwd = None

        if "user" in param:
            user = param["user"]
        if "passwd" in param:
            passwd = param["passwd"]

        return (user, passwd)

    def __get_privilege(self, user, passwd):
        """権限取得
        
        Args:
            user (string): ユーザ名
            passwd (string): パスワード
        
        Returns:
            int: 権限
        """
        u = WebUser()
        users = u.select({"userId":user})
        if len(users) == 0:
            return 0
        
        hash1 = users[0]["hash"]
        salt = users[0]["salt"]
        privilege = users[0]["privilege"]
        hash2 = u.calc_hash(passwd, salt)

        if hash1 == hash2:
            return privilege
        else:
            # デイリーパスワードチェック
            dailyPw = LoginDailyPasswd(user=user)
            if dailyPw.verify(passwd):
                return privilege

        return 0


class TerminalInfoHandler(BaseMyHttpHandler):
    """端末情報ハンドラー
    """

    def get_path_dictionary(self):
        """ハンドリングパスディクショナリーの返却
        
        Returns:
            dictionary: ハンドリングパスディクショナリー
        """
        path_dict = {
            "post":{
                '/get/terminalInfo': self.getTerminalInfo,
                '/set/terminalInfo': self.setTerminalInfo,
            },
            "get":{
            }
        }
        return path_dict

    def getTerminalInfo(self):
        """端末情報取得
        
        Returns:
            bool: Trueは処理完了
        """

        if self.check_login() == False:
            self.response_json_login_error()
            return True

        try:
            terminal = TerminalInfo()
            termInfo = terminal.terminalInfo
            termInfo["macAddress1"] = netifaces.ifaddresses('eth0')[netifaces.AF_LINK][0]['addr']
            termInfo["macAddress2"] = netifaces.ifaddresses('wlan0')[netifaces.AF_LINK][0]['addr']
            response = {"status": "ok",
                        "result": termInfo}
            self.response_json(response)

        except:
            self.logger.error("getTerminalInfo failed.", exc_info=True)
            self.response_json_error()

        return True


    def setTerminalInfo(self):
        """端末情報設定
        
        Returns:
            bool: Trueは処理完了
        """
        if self.check_login() == False:
            self.response_json_login_error()
            return True

        param = self.get_json_param()
        self.logger.debug("param = {}".format(param))

        try:
            cookie = None            
            terminal = TerminalInfo()
            wpa_conf = WpaConf()

            if "termNo" in param:
                terminal.termNo = param["termNo"]
            if "timezone" in param:
                terminal.timezone = param["timezone"]

                # タイムゾーン システム設定
                self.__set_timezone(terminal.timezone)

            if "country" in param:
                terminal.country = param["country"]

                # wpa_supplicant.confの更新
                if wpa_conf.country != terminal.country:
                    wpa_conf.country = terminal.country
                    wpa_conf.save()
                    wpa_conf.apply()

            if "language" in param:
                terminal.language = param["language"]

                # cookieの更新
                cookie = cookies.SimpleCookie()                
                cookie["LANGUAGE"] = terminal.language
                cookie["LANGUAGE"]["Path"] = "/"

            terminal.save()
            response = {"status": "ok"}
            self.response_json(response, cookie)

        except:
            self.logger.error("setTerminalInfo failed.", exc_info=True)
            self.response_json_error()

        return True

    def __set_timezone(self, timezone):
        """タイムゾーン設定
        
        Args:
            timezone (string): タイムゾーン
        """
        try:
            subprocess.run(["timedatectl", "set-timezone", timezone], check = True)
        except:
            self.logger.error("__set_timezone failed.", exc_info=True)

class GroupHandler(BaseMyHttpHandler):
    """グループハンドラー
    """
    def get_path_dictionary(self):
        """ハンドリングパスディクショナリーの返却
        
        Returns:
            dictionary: ハンドリングパスディクショナリー
        """
        path_dict = {
            "post":{
                '/get/groups': self.getGroups,
                '/update/groups': self.updateGroups,
                '/delete/groups': self.deleteGroups,
                '/add/group': self.addGroup,
            },
            "get":{
            }
        }
        return path_dict

    def getGroups(self):
        """グループ取得
        
        Returns:
            bool: Trueは処理完了
        """
        if self.check_login() == False:
            self.response_json_login_error()
            return True

        try:
            g = Group()

            key = {}
            orderBy = {"groupId": ""}

            result = g.select(key, orderBy)

            response = {"status": "ok",
                        "result": result}
            self.response_json(response)

        except:
            self.logger.error("getGroups failed.", exc_info=True)
            self.response_json_error()

        return True

    def updateGroups(self):
        """グループ更新
        
        Returns:
            bool: Trueは処理完了
        """
        if self.check_login() == False:
            self.response_json_login_error()
            return True

        param = self.get_json_param()
        self.logger.debug("param = {}".format(param))
        deleteList = None
        updateList = None
        # addList = None

        if "delete" in param and "update" in param:
            deleteList = param["delete"]
            updateList = param["update"]
            # addList = param["add"]
        else:
            self.response_json_error()
            return True

        try:
            g = Group()

            # 削除
            delKeys = []
            for d in deleteList:
                delKey = { "groupId": d}
                delKeys.append(delKey)

            if len(delKeys) > 0:
                g.deletemany(delKeys)

            # 更新
            keyValueList = []
            for u in updateList:
                key = {"groupId": u["groupId"]}
                value = {"name": u["name"]}
                keyValueList.append( (key, value) )

            if len(keyValueList) > 0:
                g.updatemany(keyValueList)

            # # 追加
            # insList = []
            # for a in addList:
            #     insList.append(a)

            # if len(insList) > 0:
            #     g.insertmany(insList)
 
            response = {"status": "ok"}
            self.response_json(response)

        except:
            self.logger.error("updateGroups failed.", exc_info=True)
            self.response_json_error()

        return True

    def deleteGroups(self):
        """グループ削除
        
        Returns:
            bool: Trueは処理完了
        """
        if self.check_login() == False:
            self.response_json_login_error()
            return True

        param = self.get_json_param()
        self.logger.debug("param = {}".format(param))
        deleteList = None

        if "delete" in param:
            deleteList = param["delete"]
        else:
            self.response_json_error()
            return True

        try:
            g = Group()

            # 削除
            delKeys = []
            for d in deleteList:
                delKey = { "groupId": d}
                delKeys.append(delKey)

            if len(delKeys) > 0:
                g.deletemany(delKeys)

            response = {"status": "ok"}
            self.response_json(response)

        except:
            self.logger.error("deleteGroups failed.", exc_info=True)
            self.response_json_error()

        return True


    def addGroup(self):
        """グループ追加
        
        Returns:
            bool: Trueは処理完了
        """
        if self.check_login() == False:
            self.response_json_login_error()
            return True

        param = self.get_json_param()
        self.logger.debug("param = {}".format(param))
        name = None

        if "name" in param:
            name = param["name"]
        else:
            self.response_json_error()
            return True

        try:
            g = Group()

            groupId = g.nextGroupId()
            if groupId <= 99:
                groupInfo = {"groupId": groupId, "name": name}
                g.insert(groupInfo)

                response = {"status": "ok",
                            "result": groupInfo}
                self.response_json(response)

            else:
                self.response_json_error()

        except:
            self.logger.error("deleteGroups failed.", exc_info=True)
            self.response_json_error()

        return True

class CallerHandler(BaseMyHttpHandler):
    """呼出機ハンドラー
    """

    def get_path_dictionary(self):
        """ハンドリングパスディクショナリーの返却
        
        Returns:
            dictionary: ハンドリングパスディクショナリー
        """
        path_dict = {
            "post":{
                '/get/callers': self.getCallers,
                '/update/callers': self.updateCallers,
                '/delete/callers': self.deleteCallers,
            },
            "get":{
            }
        }        
        return path_dict

    def getCallers(self):
        """呼出機取得
        
        Returns:
            bool: Trueは処理完了
        """
        if self.check_login() == False:
            self.response_json_login_error()
            return True

        param = self.get_json_param()
        self.logger.debug("param = {}".format(param))
        sound = False
        if "sound" in param:
            sound = param["sound"]

        try:
            c = CallerButtonView()

            key = {}
            orderBy = {"ins_dt": "", "address": "", "buttonNo": ""}

            callers = c.select(key, orderBy)

            result = {
                "callers": callers
            }

            if sound:
                sounds = None
                with open(pathlib.Path("./sound.json").resolve()) as f:
                    sounds = json.loads(f.read())

                result["sounds"] = sounds

            response = {"status": "ok",
                        "result": result}
            self.response_json(response)

        except:
            self.logger.error("getCallers failed.", exc_info=True)
            self.response_json_error()

        return True

    def updateCallers(self):
        """呼出機更新
        
        Returns:
            bool: Trueは処理完了
        """
        if self.check_login() == False:
            self.response_json_login_error()
            return True

        param = self.get_json_param()
        self.logger.debug("param = {}".format(param))

        deleteList = None
        updateList = None

        if "delete" in param and "update" in param:
            deleteList = param["delete"]
            updateList = param["update"]
        else:
            self.response_json_error()
            return True

        try:
            c = Caller()
            cb = CallerButton()

            # 削除
            delKeys = []
            for d in deleteList:
                delKey = { "address": d}
                delKeys.append(delKey)

            if len(delKeys) > 0:
                c.deletemany(delKeys)
                cb.deletemany(delKeys)



            updCallerList = []
            updCallerButtonList = []
            for caller in updateList:
                # if "line1" in caller and "line2" in caller:
                #     key = { "address": caller["address"]}
                #     value = {"line1": caller["line1"],
                #             "line2": caller["line2"]}
                #     updCallerList.append((key, value))

                if "line1" in caller and "line2" in caller and "call" in caller and "color" in caller and "priority" in caller:
                    key = { "address": caller["address"],
                            "buttonNo": caller["buttonNo"]}
                    value = {"line1": caller["line1"],
                            "line2": caller["line2"],
                            "call": caller["call"],
                            "color": caller["color"],
                            "sound": caller["sound"],
                            "priority": int(caller["priority"])}
                    updCallerButtonList.append((key, value))

            if len(updCallerList) > 0:
                c.updatemany(updCallerList)

            if len(updCallerButtonList) > 0:
                cb.updatemany(updCallerButtonList)
 
            response = {"status": "ok"}
            self.response_json(response)

        except:
            self.logger.error("updateCallers failed.", exc_info=True)
            self.response_json_error()

        return True


    def deleteCallers(self):
        """呼出機削除
        
        Returns:
            bool: Trueは処理完了
        """
        if self.check_login() == False:
            self.response_json_login_error()
            return True

        param = self.get_json_param()
        self.logger.debug("param = {}".format(param))

        deleteList = None

        if "delete" in param:
            deleteList = param["delete"]
        else:
            self.response_json_error()
            return True

        try:
            c = Caller()
            cb = CallerButton()

            # 削除
            delKeys = []
            for d in deleteList:
                delKey = { "address": d}
                delKeys.append(delKey)

            if len(delKeys) > 0:
                c.deletemany(delKeys)
                cb.deletemany(delKeys)
 
            response = {"status": "ok"}
            self.response_json(response)

        except:
            self.logger.error("deleteCallers failed.", exc_info=True)
            self.response_json_error()

        return True

class EraserHandler(BaseMyHttpHandler):
    """消去機ハンドラー
    """

    def get_path_dictionary(self):
        """ハンドリングパスディクショナリーの返却
        
        Returns:
            dictionary: ハンドリングパスディクショナリー
        """
        path_dict = {
            "post":{
                '/get/erasers': self.getErasers,
                '/update/erasers': self.updateErasers,
                '/delete/erasers': self.deleteErasers,
            },
            "get":{
            }
        }        
        return path_dict

    def getErasers(self):
        """消去機取得
        
        Returns:
            bool: Trueは処理完了
        """
        if self.check_login() == False:
            self.response_json_login_error()
            return True

        param = self.get_json_param()
        self.logger.debug("param = {}".format(param))
        sound = False
        if "sound" in param:
            sound = param["sound"]

        try:
            e = Eraser()

            key = {}
            orderBy = {"ins_dt": "", "address": ""}

            erasers = e.select(key, orderBy)

            result = {
                "erasers": erasers
            }

            response = {"status": "ok",
                        "result": result}
            self.response_json(response)

        except:
            self.logger.error("getErasers failed.", exc_info=True)
            self.response_json_error()

        return True

    def updateErasers(self):
        """消去機更新
        
        Returns:
            bool: Trueは処理完了
        """
        if self.check_login() == False:
            self.response_json_login_error()
            return True

        param = self.get_json_param()
        self.logger.debug("param = {}".format(param))

        deleteList = None
        updateList = None

        if "delete" in param and "update" in param:
            deleteList = param["delete"]
            updateList = param["update"]
        else:
            self.response_json_error()
            return True

        try:
            e = Eraser()

            # 削除
            delKeys = []
            for d in deleteList:
                delKey = { "address": d}
                delKeys.append(delKey)

            if len(delKeys) > 0:
                e.deletemany(delKeys)


            updEraserList = []
            for eraser in updateList:
                if "name" in eraser:
                    key = { "address": eraser["address"]}
                    value = {"name": eraser["name"]}
                    updEraserList.append((key, value))

            if len(updEraserList) > 0:
                e.updatemany(updEraserList)

            response = {"status": "ok"}
            self.response_json(response)

        except:
            self.logger.error("updateErasers failed.", exc_info=True)
            self.response_json_error()

        return True


    def deleteErasers(self):
        """消去機削除
        
        Returns:
            bool: Trueは処理完了
        """
        if self.check_login() == False:
            self.response_json_login_error()
            return True

        param = self.get_json_param()
        self.logger.debug("param = {}".format(param))

        deleteList = None

        if "delete" in param:
            deleteList = param["delete"]
        else:
            self.response_json_error()
            return True

        try:
            e = Eraser()

            # 削除
            delKeys = []
            for d in deleteList:
                delKey = { "address": d}
                delKeys.append(delKey)

            if len(delKeys) > 0:
                e.deletemany(delKeys)
 
            response = {"status": "ok"}
            self.response_json(response)

        except:
            self.logger.error("delateErasers failed.", exc_info=True)
            self.response_json_error()

        return True

class DisplayHandler(BaseMyHttpHandler):
    """表示機一覧ハンドラー
    """
    def get_path_dictionary(self):
        """ハンドリングパスディクショナリーの返却
        
        Returns:
            dictionary: ハンドリングパスディクショナリー
        """
        path_dict = {
            "post":{
                '/get/displays': self.getDisplays,
            },
            "get":{
            }
        }
        return path_dict

    def getDisplays(self):
        """表示機一覧取得
        
        Returns:
            bool: Trueは処理完了
        """
        if self.check_login() == False:
            self.response_json_login_error()
            return True

        try:
            d = Display()

            key = {}
            orderBy = {"datetime": "DESC"}

            result = d.select(key, orderBy)

            response = {"status": "ok",
                        "result": result}
            self.response_json(response)

        except:
            self.logger.error("getDisplays failed.", exc_info=True)
            self.response_json_error()

        return True

class OtherInfoHandler(BaseMyHttpHandler):
    """その他設定ハンドラー
    """
    FIRMWARE_TEMP_DIR = "/home/pi/temp/firmware/"

    def get_path_dictionary(self):
        """ハンドリングパスディクショナリーの返却
        
        Returns:
            dictionary: ハンドリングパスディクショナリー
        """
        path_dict = {
            "post":{
                '/get/otherInfo': self.getOtherInfo,
                '/update/otherInfo': self.updateOtherInfo,
                '/upload/firmware': self.uploadFirmware,
                '/set/time': self.setTime,
                # '/test/logsend': self.testLogSend,
                '/download/log': self.downloadLog,
                '/restore/factorySetting': self.restoreFactorySetting,
            },
            "get":{
            }
        }
        return path_dict

    def getOtherInfo(self):
        """その他の設定取得
        
        Returns:
            bool: Trueは処理完了
        """
        if self.check_login() == False:
            self.response_json_login_error()
            return True

        try:
            otherInfo = OtherInfo().otherInfo

            callerSetting = CallerSetting().callerSetting
            #呼出機設定を付加する
            otherInfo["caller"] = callerSetting

            response = {"status": "ok",
                        "result": otherInfo}
            self.response_json(response)

        except:
            self.logger.error("getOtherInfo failed.", exc_info=True)
            self.response_json_error()

        return True

    def updateOtherInfo(self):
        """その他の設定更新
        
        Returns:
            bool: Trueは処理完了
        """

        if self.check_login() == False:
            self.response_json_login_error()
            return True

        param = self.get_json_param()
        self.logger.debug("param = {}".format(param))

        try:
            o = OtherInfo()
            c = CallerSetting()
            modelASetting = c.modelA
            if "sound" in param:
                o.sound = param["sound"]
            if "timeSetting" in param:
                o.timeSetting = param["timeSetting"]
            if "caller" in param and "A" in param["caller"] and "longPressTime" in param["caller"]["A"]:
                modelASetting["longPressTime"] = param["caller"]["A"]["longPressTime"]
                c.modelA = modelASetting

            o.save()
            c.save()

            response = {"status": "ok"}
            self.response_json(response)

        except:
            self.logger.error("updateOtherInfo failed.", exc_info=True)
            self.response_json_error()

        return True

    def uploadFirmware(self):
        """ファームウェアアップロード
        
        Returns:
            bool: Trueは処理完了
        """
        if self.check_login() == False:
            self.response_json_login_error()
            return True

        try:
            param = self.get_multipart_form_data()

            zipFile = self.__get_zip_file(param)
            if zipFile is None:
                self.response_json_error()
                return True

            fileName = zipFile[0]
            fileData = zipFile[1]

            os.makedirs(self.FIRMWARE_TEMP_DIR, exist_ok = True)

            with open(os.path.join(self.FIRMWARE_TEMP_DIR, fileName) , "wb") as f:
                f.write(fileData)

            response = {"status": "ok"}
            self.response_json(response)

            subprocess.Popen(["reboot"])

        except:
            self.logger.error("uploadFirmware failed.", exc_info=True)
            self.response_json_error()

        return True

    def __get_zip_file(self, param):
        """ZIPファイル取得
        
        Args:
            param (list): パラメータ(multipart/form-data)
        
        Returns:
            (string, bytes): ファイル名, ファイルデータ
        """
        if param is None or len(param) == 0:
            return None

        if param[0]["type"] != "application/x-zip-compressed" and  param[0]["type"] != "application/zip":
            return None
        
        return (param[0]["filename"], param[0]["value"])
        
    def setTime(self):
        """時計合わせ
        
        Returns:
            bool: Trueは処理完了
        """

        if self.check_login() == False:
            self.response_json_login_error()
            return True

        param = self.get_json_param()
        self.logger.debug("param = {}".format(param))

        setTimeType = None
        if "type" in param:
            setTimeType = param["type"]
        else:
            self.response_json_error()
            return True

        time_val = None
        if setTimeType == "ntp":
            # NTP
            server = None
            if "server" in param:
                server = param["server"]
            else:
                self.response_json_error()
                return True

            try:
                client = NTPClient()
                res = client.request(server)
                dt = datetime.fromtimestamp(res.tx_time)

                time_val = "{:%Y/%m/%d %H:%M:%S}".format(dt)
            except:
                self.logger.error("setTime(NTP) failed.", exc_info=True)
                self.response_json_error()
                return True


        else:
            # 手入力
            date = None
            time = None
            if "date" in param and "time" in param:
                date = param["date"]
                time = param["time"]
            else:
                self.response_json_error()
                return True

            time_val = date + " " + time

        try:
            subprocess.run(["date", "-s", time_val], check = True)

            result = {"time": "{:%Y/%m/%d %H:%M:%S}".format(datetime.now())}
            response = {"status": "ok",
                        "result": result}
            self.response_json(response)

        except:
            self.logger.error("setTime failed.", exc_info=True)
            self.response_json_error()

        return True


    # def testLogSend(self):
    #     """ログサーバー通信テスト
        
    #     Returns:
    #         bool: Trueは処理完了
    #     """

    #     if self.check_login() == False:
    #         self.response_json_error()
    #         return True

    #     param = self.get_json_param()
    #     self.logger.debug("param = {}".format(param))

    #     ipAddress = None
    #     if "ipAddress" in param:
    #         ipAddress = param["ipAddress"]
    #     else:
    #         self.response_json_error()
    #         return True

    #     try:
    #         # TODO: 通信テスト

    #         response = {"status": "ok"}
    #         self.response_json(response)
    #     except:
    #         self.logger.error("testLogSend failed.", exc_info=True)
    #         self.response_json_error()

    #     return True


    def downloadLog(self):
        """内部ログダウンロード
        
        Returns:
            bool: Trueは処理完了
        """

        if self.check_login() == False:
            self.response_json_login_error()
            return True

        try:
            DEV_LOG_DIR = "./log"
            SYS_LOG_DIR = "/var/log"
            TEMP_DIR = os.path.join(self.directory, "temp") 
            dirname = "log_" + str(uuid.uuid4())
            dirpath = os.path.join(TEMP_DIR, dirname)

            os.makedirs(dirpath, exist_ok = True)
            # app log
            shutil.copytree(DEV_LOG_DIR, os.path.join(dirpath, "applog"), symlinks=True)
            # syslog
            shutil.copytree(SYS_LOG_DIR, os.path.join(dirpath, "syslog"), symlinks=True)

            # 呼び出しログ
            calllog = CallLog()
            logs = calllog.select({}, {"datetime": "ASC"})
            with open(os.path.join(dirpath, "applog/calllog.csv"), "w") as f:
                writer = csv.writer(f)
                writer.writerow(["date","time","category","message","sent"])
                for log in logs:
                    writer.writerow([
                        "{:%Y/%m/%d}".format(datetime.strptime(log["datetime"], "%Y-%m-%d %H:%M:%S.%f")),
                        "{:%H:%M:%S}".format(datetime.strptime(log["datetime"], "%Y-%m-%d %H:%M:%S.%f")),
                        log["category"],
                        log["message"],
                        log["sent"]
                        ])

            # ZIPファイル作成
            shutil.make_archive(dirname, "zip", dirpath)
            zipFileName = dirname + ".zip"
            shutil.rmtree(dirpath)
            shutil.move(zipFileName, TEMP_DIR)

            result = {
                "path": "temp/" + zipFileName
            }
            response = {
                "status": "ok",
                "result": result
            }
            self.response_json(response)
        except:
            self.logger.error("downloadLog failed.", exc_info=True)
            self.response_json_error()

        return True

    def restoreFactorySetting(self):
        """工場出荷状態に戻す
        
        Returns:
            bool: Trueは処理完了
        """
        if self.check_login() == False:
            self.response_json_login_error()
            return True

        try:
            d = Database()
            d.uninstall()
            d.install()
            d.setup()

            t = TerminalInfo()
            o = OtherInfo()
            n = NetworkInfo()
            c = CallerSetting()
            t.initialize()
            o.initialize()
            n.initialize()
            c.initialize()
            t.save()
            o.save()
            n.save()
            c.save()

            wpa_conf = WpaConf()
            wpa_conf.clear_network()
            wpa_conf.save()

            dhcp_conf = DhcpcdConf()
            info = {
                    "autoIp": True,
                    "autoDNS": True
                    }
            dhcp_conf.set_info("eth0", info)
            dhcp_conf.save()

            response = {"status": "ok"}
            self.response_json(response)
    
            subprocess.Popen(["reboot"])

        except:
            self.logger.error("restoreFactorySetting failed.", exc_info=True)
            self.response_json_error()

        return True


class NetConfigHandler(BaseMyHttpHandler):
    """ネットワーク設定ハンドラー
    """
    def get_path_dictionary(self):
        """ハンドリングパスディクショナリーの返却
        
        Returns:
            dictionary: ハンドリングパスディクショナリー
        """
        path_dict = {
            "post":{
                '/get/networkInfo': self.getNetworkInfo,
                '/get/networkSetting': self.getNetworkSetting,
                '/scan/wifi': self.scanWifi,
                '/update/networkSetting': self.updateNetworkSetting,
                '/get/ssidSetting': self.getSsidSetting,
                '/save/ssidSetting': self.saveSsidSetting,
                '/delete/ssidSetting': self.deleteSsidSetting,
            },
            "get":{
            }
        }
        return path_dict

    def getNetworkInfo(self):
        """ネットワーク情報取得
        
        Returns:
            bool: Trueは処理完了
        """
        if self.check_login() == False:
            self.response_json_login_error()
            return True

        try:
            dhcpconf = DhcpcdConf()
            networkStatus = NetworkStatus()
            networkInfo = NetworkInfo()

            # 無線LAN
            wlan = {
                "enabled": networkInfo.wlan["enabled"],
                "ssid": "",
                "ipAddress": "",
                "subnetMask": "",
                "defaultGw": "",
                "dns1": "",
                "dns2": ""
            }

            # dhcpcd.confよりDHCPの設定を取得
            if wlan["enabled"]:
                wlan_dhcpcd = dhcpconf.get_info("wlan0")
                wlan["autoIp"] = wlan_dhcpcd["autoIp"]
                wlan["autoDNS"] = wlan_dhcpcd["autoDNS"]

                #現時点のIPアドレスを取得
                wlan_status = networkStatus.get_net_status("wlan0")
                if "ssid" in wlan_status: wlan["ssid"] = wlan_status["ssid"]
                if "ipAddress" in wlan_status: wlan["ipAddress"] = wlan_status["ipAddress"]
                if "subnetMask" in wlan_status: wlan["subnetMask"] = wlan_status["subnetMask"]
                if "defaultGw" in wlan_status: wlan["defaultGw"] = wlan_status["defaultGw"]

                # DNS
                dnslist = wlan_status["dns"]
                i = 1
                for dns in dnslist:
                    wlan["dns{}".format(i)] = dns
                    i += 1
                    

            # 有線LAN
            eth = {
                "enabled": networkInfo.eth["enabled"],
                "ipAddress": "",
                "subnetMask": "",
                "defaultGw": "",
                "dns1": "",
                "dns2": "",
            }

            if eth["enabled"]:
                # dhcpcd.confよりDHCPの設定を取得
                eth_dhcpcd = dhcpconf.get_info("eth0")
                eth["autoIp"] = eth_dhcpcd["autoIp"]
                eth["autoDNS"] = eth_dhcpcd["autoDNS"]

                #現時点のIPアドレスを取得
                eth_status = networkStatus.get_net_status("eth0")
                if "ipAddress" in eth_status: eth["ipAddress"] = eth_status["ipAddress"]
                if "subnetMask" in eth_status: eth["subnetMask"] = eth_status["subnetMask"]
                if "defaultGw" in eth_status: eth["defaultGw"] = eth_status["defaultGw"]



                # DNS
                dnslist = eth_status["dns"]
                i = 1
                for dns in dnslist:
                    eth["dns{}".format(i)] = dns
                    i += 1


            result = {
                "wlan": wlan,
                "eth": eth
            }
            response = {"status": "ok",
                        "result": result}
            self.response_json(response)

        except:
            self.logger.error("getNetworkInfo failed.", exc_info=True)
            self.response_json_error()

        return True

    def getNetworkSetting(self):
        """ネットワーク設定取得
        
        Returns:
            bool: Trueは処理完了
        """
        if self.check_login() == False:
            self.response_json_login_error()
            return True

        try:
            networkInfo = NetworkInfo()
            networkStatus = NetworkStatus()


            result = networkInfo.networkInfo

            wlan_status = networkStatus.get_net_status("wlan0")
            if "ssid" in wlan_status: result["wlan"]["ssid"] = wlan_status["ssid"]

            response = {"status": "ok",
                        "result": result}
            self.response_json(response)

        except:
            self.logger.error("getNetworkInfo failed.", exc_info=True)
            self.response_json_error()

        return True

    def scanWifi(self):
        """Wifiスキャン
        
        Returns:
            bool: Trueは処理完了
        """
        if self.check_login() == False:
            self.response_json_login_error()
            return True

        try:
            networkInfo = NetworkInfo()
            networkStatus = NetworkStatus()
            if networkInfo.wlan["enabled"] == False:
                #Wifiスキャンするために一時的にwlan0をアップする
                networkStatus.if_up_down("wlan0", True)

            iwlist = IwList()
            result = iwlist.scan()

            if networkInfo.wlan["enabled"] == False:
                networkStatus.if_up_down("wlan0", False)

            response = {"status": "ok",
                        "result": result}
            self.response_json(response)

        except:
            self.logger.error("scanWifi failed.", exc_info=True)
            self.response_json_error()

        return True

    def updateNetworkSetting(self):
        """ネットワーク設定更新
        
        Returns:
            bool: Trueは処理完了
        """
        if self.check_login() == False:
            self.response_json_login_error()
            return True

        param = self.get_json_param()

        self.logger.debug("param = {}".format(param))
        wlan = None
        eth = None
        if "wlan" in param and "eth" in param:
            wlan = param["wlan"]
            eth = param["eth"]
        else:
            self.response_json_error()
            return True

        try:
            wpa_conf = WpaConf()
            dhcp_conf = DhcpcdConf()
            netInfo = NetworkInfo()
            networkStatus = NetworkStatus()

            # 設定ファイル更新
            netInfo.wlan["enabled"] = wlan["enabled"]

            wpa_conf.clear_network()

            if wlan["enabled"]:
                if "ssid" in wlan:
                    if wlan["ssid"] in netInfo.wlan["ssidSettings"]:
                        wlanSetting = netInfo.wlan["ssidSettings"][wlan["ssid"]]
                        # DHCP/IPアドレスの設定
                        info = self.__create_dhcpcd_param(wlanSetting)
                        dhcp_conf.set_info("wlan0", info)

                        # wpa_supplicant.confの更新
                        wpa_conf.set_network(wlanSetting["wpa_network"])


            # 設定ファイル更新
            netInfo.eth = eth

            if eth["enabled"]:
                # DHCP/IPアドレスの設定
                info = self.__create_dhcpcd_param(eth)
                dhcp_conf.set_info("eth0", info)


            netInfo.save()
            dhcp_conf.save()
            wpa_conf.save()

            #IPアドレスを削除
            networkStatus.if_flush_dev("eth0")
            networkStatus.if_flush_dev("wlan0")
            # wpa_conf.service_restart()
            
            #dhcpcd.serviceを再起動
            dhcp_conf.service_dhcpcd_restart()

            networkStatus.if_up_down("wlan0", False)
            if wlan["enabled"]:
                networkStatus.if_up_down("wlan0", True)

            # networkStatus.if_up_down("eth0", False)
            # if eth["enabled"]:
            #     networkStatus.if_up_down("eth0", True)

            response = {"status": "ok"}
            self.response_json(response)

        except:
            self.logger.error("updateNetworkSetting failed.", exc_info=True)
            self.response_json_error()

        return True

    def __create_dhcpcd_param(self, ifinfo):
        """dhcpcd設定用パラメータ生成
        
        Args:
            ifinfo (dictionary): インターフェース情報
        
        Returns:
            dictionary: dhcpcd設定用パラメータ
        """
        param = {
                "autoIp": ifinfo["autoIp"],
                "autoDNS": ifinfo["autoDNS"],
                }
        
        if param["autoIp"] == False:
            if ifinfo["ipAddress"] != "": param["ipAddress"] = ifinfo["ipAddress"]
            if ifinfo["subnetMask"] != "": param["subnetMask"] = ifinfo["subnetMask"]
            if ifinfo["defaultGw"] != "": param["defaultGw"] = ifinfo["defaultGw"]

        if param["autoDNS"] == False:
            param["dns"] = []
            if ifinfo["dns1"] != "":
                param["dns"].append(ifinfo["dns1"])
            if ifinfo["dns2"] != "":
                param["dns"].append(ifinfo["dns2"])

        return param

    def getSsidSetting(self):
        """SSID設定情報取得
        
        Returns:
            bool: Trueは処理完了
        """
        if self.check_login() == False:
            self.response_json_login_error()
            return True

        param = self.get_json_param()

        self.logger.debug("param = {}".format(param))

        ssid = None
        if "ssid" in param:
            ssid = param["ssid"]
        else:
            self.response_json_error()
            return True

        try:
            networkInfo = NetworkInfo()


            result = None
            if ssid in networkInfo.wlan["ssidSettings"]:
                result = networkInfo.wlan["ssidSettings"][ssid]

            response = {"status": "ok",
                        "result": result}
            self.response_json(response)

        except:
            self.logger.error("getSsidSetting failed.", exc_info=True)
            self.response_json_error()

        return True

    def saveSsidSetting(self):
        """SSID設定情報保存
        
        Returns:
            bool: Trueは処理完了
        """
        if self.check_login() == False:
            self.response_json_login_error()
            return True

        param = self.get_json_param()

        self.logger.debug("param = {}".format(param))
        ssid = None
        ssidSetting = None
        if "ssid" in param and "ssidSetting" in param:
            ssid = param["ssid"]
            ssidSetting = param["ssidSetting"]
        else:
            self.response_json_error()
            return True

        try:
            networkInfo = NetworkInfo()
            wpa_conf = WpaConf()            

            wpa_network = None

            password = ssidSetting["password"]
            encryption = ssidSetting["encryption"]

            if ssid in networkInfo.wlan["ssidSettings"] == False or len(password) > 0:
                encType = encryption["encType"]
                if encType == "wpa2":
                    wpa_network = wpa_conf.create_network_WPA2(ssid, password, group=encryption["cipher"], pairwise=encryption["cipher"])
                elif encType == "wpa":
                    wpa_network = wpa_conf.create_network_WPA(ssid, password, group=encryption["cipher"], pairwise=encryption["cipher"])
                elif encType == "wep":
                    wpa_network = wpa_conf.create_network_WEP(ssid, password)
                else:
                    wpa_network = wpa_conf.create_network_NONE(ssid)

                ssidSetting["wpa_network"] = wpa_network
            else:
                ssidSetting["wpa_network"] = networkInfo.wlan["ssidSettings"][ssid]["wpa_network"]

            del ssidSetting["password"]
            del ssidSetting["encryption"]

            networkInfo.wlan["ssidSettings"][ssid] = ssidSetting
            networkInfo.save()

            response = {"status": "ok"}
            self.response_json(response)

        except:
            self.logger.error("saveSsidSetting failed.", exc_info=True)
            self.response_json_error()

        return True

    def deleteSsidSetting(self):
        """SSID設定情報削除
        
        Returns:
            bool: Trueは処理完了
        """
        if self.check_login() == False:
            self.response_json_login_error()
            return True

        param = self.get_json_param()

        self.logger.debug("param = {}".format(param))
        ssid = None
        if "ssid" in param:
            ssid = param["ssid"]
        else:
            self.response_json_error()
            return True

        try:
            networkInfo = NetworkInfo()

            if ssid in networkInfo.wlan["ssidSettings"]:
                del networkInfo.wlan["ssidSettings"][ssid]

            networkInfo.save()

            response = {"status": "ok"}
            self.response_json(response)

        except:
            self.logger.error("deleteSsidSetting failed.", exc_info=True)
            self.response_json_error()

        return True

class CallLogHandler(BaseMyHttpHandler):
    """呼び出しログハンドラー
    """
    def get_path_dictionary(self):
        """ハンドリングパスディクショナリーの返却
        
        Returns:
            dictionary: ハンドリングパスディクショナリー
        """
        path_dict = {
            "post":{
                '/get/calllog': self.getCallLog,
            },
            "get":{
            }
        }
        return path_dict

    def getCallLog(self):
        """呼び出しログ取得
        """
        resultCode = "0000"
        param = self.get_param()
        self.logger.debug("param={}".format(param) )

        self.send_response(200)
        self.send_header('Content-type', 'text/plain')
        self.end_headers()

        passwd = ""
        if "user" in param:
            user = param["user"][0]
            if "passwd" in param:
                passwd = param["passwd"][0]

            if self.__check_user(user, passwd) == False:
                resultCode = "0100"
                self.handlerObj.wfile.write(resultCode.encode('utf-8'))
                return True

        else:
            resultCode = "0300"
            self.handlerObj.wfile.write(resultCode.encode('utf-8'))
            return True

        try:
            calllog = CallLog()
            logs = calllog.select({"sent":0}, {"datetime": "ASC"})

            dt = "{:%Y%m%d%H%M%S}".format(datetime.now())

            self.handlerObj.wfile.write(resultCode.encode('utf-8'))
            self.handlerObj.wfile.write(dt.encode('utf-8'))

            f = io.StringIO(newline='')
            writer = csv.writer(f,lineterminator='\r\n')
            writer.writerow(["date","time","category","message"])
            for log in logs:
                writer.writerow([
                    "{:%Y/%m/%d}".format(datetime.strptime(log["datetime"], "%Y-%m-%d %H:%M:%S.%f")),
                    "{:%H:%M:%S}".format(datetime.strptime(log["datetime"], "%Y-%m-%d %H:%M:%S.%f")),
                    log["category"],
                    log["message"]
                    ])

                calllog.update(log, {"sent": 1})

            self.handlerObj.wfile.write(f.getvalue().encode('utf-8'))
            f.close()
        except:
            self.logger.error("getCallLog failed.", exc_info=True)
            resultCode = "9900"
            self.handlerObj.wfile.write(resultCode.encode('utf-8'))
        return True

    def __check_user(self, user, passwd):
        """ユーザーチェック
        
        Args:
            user (string): ユーザ名
            passwd (string): パスワード
        
        Returns:
            bool: チェック結果
        """
        u = WebUser()
        users = u.select({"userId":user})
        if len(users) == 0:
            return False
        
        hash1 = users[0]["hash"]
        salt = users[0]["salt"]
        hash2 = u.calc_hash(passwd, salt)

        if hash1 == hash2:
            return True

        return False

class SpecialHandler(BaseMyHttpHandler):
    """特殊ユーザ用ハンドラー
    """
    def get_path_dictionary(self):
        """ハンドリングパスディクショナリーの返却
        
        Returns:
            dictionary: ハンドリングパスディクショナリー
        """
        path_dict = {
            "post":{
                '/get/special': self.getSpecial,
                '/set/ssh': self.setSsh,
                '/set/logLevel': self.setLogLevel,
                '/init/data': self.initData,
                '/reboot': self.reboot,
            },
            "get":{
            }
        }
        return path_dict


    def getSpecial(self):
        """特殊設定情報取得
        """
        if self.check_login(9) == False:
            self.response_json_login_error()
            return True
        try:

            # SSH状態取得
            cmd = ["systemctl", "is-active", "ssh.service"]
            sts = subprocess.run(cmd, stdout = subprocess.PIPE, stderr = subprocess.PIPE)\
                .stdout.decode().rstrip()

            ssh = False
            if sts.startswith("active"):
                ssh = True
            
            response = {"status": "ok",
                        "result": {
                                "ssh": ssh,
                                "logLevel": TerminalInfo().logLevel
                            }
                        }

            self.response_json(response)

        except:
            self.logger.error("getSpecial failed.", exc_info=True)
            self.response_json_error()

        return True


    def setSsh(self):
        """SSH開始・終了
        """
        if self.check_login(9) == False:
            self.response_json_login_error()
            return True

        param = self.get_json_param()

        self.logger.debug("param = {}".format(param))
        ssh = False
        if "ssh" in param:
            ssh = param["ssh"]
        else:
            self.response_json_error()
            return True

        try:
            action = "stop"
            if ssh:
                action = "start"
            cmd = ["systemctl", action, "ssh.service"]
            subprocess.run(cmd, check = True)
            
            response = {"status": "ok"}

            self.response_json(response)

        except:
            self.logger.error("setSsh failed.", exc_info=True)
            self.response_json_error()

        return True

    def setLogLevel(self):
        """ログレベル設定
        """
        if self.check_login(9) == False:
            self.response_json_login_error()
            return True

        param = self.get_json_param()

        self.logger.debug("param = {}".format(param))
        logLevel = False
        if "logLevel" in param:
            logLevel = param["logLevel"]

        try:
            terminalInfo = TerminalInfo()

            terminalInfo.logLevel = logLevel
            terminalInfo.save()
            
            response = {"status": "ok"}

            self.response_json(response)

        except:
            self.logger.error("setLogLevel failed.", exc_info=True)
            self.response_json_error()

        return True

    def initData(self):
        """データ初期化
        
        Returns:
            bool: Trueは処理完了
        """
        if self.check_login(9) == False:
            self.response_json_login_error()
            return True

        try:
            d = Database()
            d.uninstall()
            d.install()
            d.setup()

            t = TerminalInfo()
            o = OtherInfo()
            n = NetworkInfo()
            c = CallerSetting()
            t.initialize()
            o.initialize()
            n.initialize()
            c.initialize()
            t.save()
            o.save()
            n.save()
            c.save()
            response = {"status": "ok"}
            self.response_json(response)
        except:
            self.logger.error("initData failed.", exc_info=True)


        return True

    def reboot(self):
        """再起動
        
        Returns:
            bool: Trueは処理完了
        """
        if self.check_login(9) == False:
            self.response_json_login_error()
            return True

        response = {"status": "ok"}
        self.response_json(response)

        subprocess.Popen(["reboot"])

        return True

class DummyModeHandler(BaseMyHttpHandler):
    """ダミーモード用ハンドラー
    """
    def get_path_dictionary(self):
        """ハンドリングパスディクショナリーの返却
        
        Returns:
            dictionary: ハンドリングパスディクショナリー
        """
        path_dict = {
            "post":{
                '/call/dummy': self.callDummy,
            },
            "get":{
            }
        }
        return path_dict

    def callDummy(self):
        param = self.get_json_param()

        self.logger.debug("param = {}".format(param))
        target = ""
        command = ""
        if "target" in param:
            target = param["target"]
        if "command" in param:
            command = param["command"]

        try:
            if target == "/tmp/gpio_pairing" or \
                target == "/tmp/gpio_clear" or \
                target == "/tmp/gpio_shutdown" or \
                target == "/tmp/dummy_serial":

                with open(target, "w") as f:
                    f.write(command)
            
                response = {"status": "ok"}

                self.response_json(response)
            else:
                self.response_json_error()

        except:
            self.logger.error("setLogLevel failed.", exc_info=True)
            self.response_json_error()

        return True
