import io
import json
import os
import pathlib
import shutil
import threading
from cgi import FieldStorage
from functools import partial
from http import cookies
from http.server import SimpleHTTPRequestHandler, ThreadingHTTPServer
from urllib.parse import parse_qs, urlparse

from absthread import AbstractThread
from recv_dev.config import TerminalInfo
from recv_dev.httphandler import (CallerHandler, CallLogHandler, EraserHandler,
                                  DisplayHandler, DummyModeHandler,
                                  GroupHandler, LoginHandler, NetConfigHandler,
                                  OtherInfoHandler, SpecialHandler,
                                  TerminalInfoHandler)
from recv_dev.storage import Display
from setup_logger import setup_logger_devel
import time

class MyHTTPRequestHandler(SimpleHTTPRequestHandler):
    """HTTPハンドラー
    """
    DIRECTORY='./recv_dev/web'

    def __init__(self, *args, **kwargs):
        """コンストラクタ
        """
        self.logger = setup_logger_devel(self.__class__.__name__)
        self.PATH_DICT = {
            "post":{
            },
            "get":{
            }
        }

        loginHandler = LoginHandler(self)
        terminalInfoHandler = TerminalInfoHandler(self)
        otherInfoHandler = OtherInfoHandler(self)
        groupHandler = GroupHandler(self)
        callerHandler = CallerHandler(self)
        eraserHandler = EraserHandler(self)
        displayHandler = DisplayHandler(self)
        netConfigHandler = NetConfigHandler(self)
        specialHandler = SpecialHandler(self)
        callLogHandler = CallLogHandler(self)
        self.add_path_dictionary(loginHandler.get_path_dictionary())
        self.add_path_dictionary(terminalInfoHandler.get_path_dictionary())
        self.add_path_dictionary(otherInfoHandler.get_path_dictionary())
        self.add_path_dictionary(groupHandler.get_path_dictionary())
        self.add_path_dictionary(callerHandler.get_path_dictionary())
        self.add_path_dictionary(eraserHandler.get_path_dictionary())
        self.add_path_dictionary(displayHandler.get_path_dictionary())
        self.add_path_dictionary(netConfigHandler.get_path_dictionary())
        self.add_path_dictionary(specialHandler.get_path_dictionary())
        self.add_path_dictionary(callLogHandler.get_path_dictionary())
        if TerminalInfo().dummyMode:
            dummyModeHandler = DummyModeHandler(self)
            self.add_path_dictionary(dummyModeHandler.get_path_dictionary())

        self.webSession = WebSession()
        super(MyHTTPRequestHandler, self).__init__(*args, **kwargs)

    def add_path_dictionary(self, path_dict):
        """ハンドリングパスの追加
        
        Args:
            path_dict (dictionary): ハンドリングパスディクショナリー
        """
        for (action, _dict) in path_dict.items():
            if action in self.PATH_DICT:
                for (path, method) in _dict.items():
                    self.PATH_DICT[action][path] = method

    def do_GET(self):
        """GET
        """
        
        self.logger.info('path = {}'.format(self.path))

        parsed_path = urlparse(self.path)
        self.logger.debug('parsed: path = {}, query = {}'.format(parsed_path.path, parse_qs(parsed_path.query)))

        # self.logger.debug('headers\r\n-----\r\n{}-----'.format(self.headers))

        # pathに応じて処理実行
        if self.handleMethod('get', parsed_path.path):
            pass
        else:
            lang = TerminalInfo().language
            print("lang={}".format(lang))
            if parsed_path.path == "/" or parsed_path.path.endswith("html"):
                if self.check_login() == False:
                    # ログイン
                    self.write_html_file(os.path.join(lang, "login.html"))
                else:
                    # デフォルトの動作
                    # （pathで指定されたファイルを表示する）
                    if parsed_path.path == "/":
                        self.write_html_file(os.path.join(lang, "login.html"))
                    else:
                        self.write_html_file(os.path.join(lang, self.path.lstrip("/")))
            else:
                # デフォルトの動作
                # （pathで指定されたファイルを表示する）
                super().do_GET()


    def do_POST(self):
        """POST
        """

        self.logger.info('path = {}'.format(self.path))

        parsed_path = urlparse(self.path)
        self.logger.debug('parsed: path = {}, query = {}'.format(parsed_path.path, parse_qs(parsed_path.query)))

        # self.logger.debug('headers\r\n-----\r\n{}-----'.format(self.headers))


        # pathに応じて処理実行
        if self.handleMethod('post', parsed_path.path):
            pass
        else:
            # デフォルトの動作
            super().do_POST()

    def end_headers(self):
        """end_headersのオーバーライド
        """

        parsed_path = urlparse(self.path)
        if parsed_path.path == "/" or parsed_path.path.endswith("html") or parsed_path.path.endswith("js"):
            self.send_header("Cache-Control", "no-cache")
            self.send_header("Pragma", "no-cache")
            self.send_header("Expires", "0")

        SimpleHTTPRequestHandler.end_headers(self)

    def handleMethod(self, method, path):
        """メソッドハンドリング
        
        Args:
            method (string): method(get, post)
            path (string): パス
        
        Returns:
            bool: Trueは処理済みを示す
        """
        # pathに応じて処理を実行
        if method in self.PATH_DICT:
            handles =  self.PATH_DICT[method]

            if path in handles:
                return handles[path]()
        else:
            return False


    def get_param(self):
        """パラメータ取得(POST)
        
        Returns:
            dictionary: パラメータ
        """
        content_len=int(self.headers.get('content-length'))
        param = parse_qs(self.rfile.read(content_len).decode('utf-8'))

        return param

    def get_param_of_get(self):
        """パラメータ取得(GET)
        
        Returns:
            dictionary: パラメータ
        """
        parsed_path = urlparse(self.path)
        param = parse_qs(parsed_path.query)

        return param

    def get_json_param(self):
        """JSONパラメータの取得
        
        Returns:
            dictionary: JSONパラメータ
        """
        content_len=int(self.headers.get('content-length'))
        param = json.loads(self.rfile.read(content_len).decode('utf-8'))
        return param

    def get_multipart_form_data(self):
        """multipart/form-data取得
        
        Returns:
            list: multipart/form-data
        """
        content_len=int(self.headers.get('content-length'))
        fp = io.BytesIO(self.rfile.read(content_len))
        environ = {'REQUEST_METHOD': 'POST'}
        fs = FieldStorage(fp=fp, environ=environ, headers=self.headers)
        param = []
        for f in fs.list:
            field = {
                "name":f.name,
                "filename": f.filename,
                "type": f.type,
                "value": f.value
            }
            param.append(field)

        return param

    def response_json(self, response, cookie = None):
        """JSON応答
        
        Args:
            response (dictionary): 応答データ
            cookie (object, optional): cookie. Defaults to None.
        """
        self.send_response(200)
        self.send_header('Content-type', 'application/json')

        if cookie is not None:
                self.logger.debug("cookie.output = {}".format(cookie.output()))
                self.send_header(*cookie.output().split(":",1))

        self.end_headers()

        responseBody = json.dumps(response)
        self.logger.debug("response data = {}".format(responseBody))
        self.wfile.write(responseBody.encode('utf-8'))

    def write_html_file(self, path, cookie = None):
        """HTMLファイル出力
        
        Args:
            path (string): ファイルパス
            cookie (object, optional): cookie. Defaults to None.
        """
        filepath = os.path.join(pathlib.Path(self.directory).resolve(), path.lstrip('/'))
        print(filepath)
        if os.path.isfile(filepath):
            with open(filepath) as f:
                lines = f.readlines()
                self.send_response(200)
                if cookie is not None:
                    self.logger.debug("cookie.output = {}".format(cookie.output()))
                    self.send_header(*cookie.output().split(":",1))

                # self.send_header("Cache-Control", "no-cache")
                self.send_header('Content-type', 'text/html')
                self.end_headers()

                self.wfile.write("\r\n".join(lines).encode('utf-8'))

            sessionId = self.get_sessionId_from_cookie()
            if sessionId is not None:
                session = self.get_session_info(sessionId)
                if session is not None:
                    session["time"] = time.time()
                    session["page"] = os.path.basename(path)
                    self.add_session_info(session)
        else:
            self.send_response(404)

    def check_login(self, priv=None):
        """ログインチェック
        
        Args:
            priv (int, optional): 権限. Defaults to None.
        
        Returns:
            bool: 結果
        """
        sessionId = self.get_sessionId_from_cookie()
        if sessionId is None:
            return False
        
        sessionInfo = self.get_session_info(sessionId)
        self.logger.debug('sessionInfo={}'.format(sessionInfo))
        if sessionInfo is None:
            return False
        if priv is None:
            return True
        else:
            return sessionInfo["privilege"] >= priv

    def get_sessionId_from_cookie(self):
        """セッションIDをCookieから取得
        
        Returns:
            string: セッションID
        """
        if self.headers.get('Cookie') is None:
            return None

        cookie = cookies.SimpleCookie()
        cookie.load(self.headers.get('Cookie')) 

        if "SESSIONID" in cookie.keys():
            return cookie["SESSIONID"].value
        else:
            return None

    def add_session_info(self, session):
        """セッション情報追加
        
        Args:
            sessionInfo (dictionary): セッション情報
        """
        self.webSession.add_session_info(session)

    def get_session_info(self, sessionId):
        """セッション情報取得
        
        Args:
            sessionId (string): セッションID
        
        Returns:
            dictionary: セッション情報
        """
        return self.webSession.get_session_info(sessionId)

    def remove_sessio_info(self, sessionId):
        """セッション情報削除
        
        Args:
            sessionId (string): セッションID
        """
        self.webSession.remove_sessio_info(sessionId)

class Singleton(object):
    """シングルトン
    """
    def __new__(cls, *args, **kargs):
        if not hasattr(cls, "_instance"):
            cls._instance = super(Singleton, cls).__new__(cls)
        return cls._instance

class WebSession(Singleton):
    """Webセッション情報
    """
    webSessions = {}

    def add_session_info(self, session):
        """セッション情報追加
        
        Args:
            sessionInfo (dictionary): セッション情報
        """
        self.webSessions[session["sessionId"]] = session

    def get_session_info(self, sessionId):
        """セッション情報取得
        
        Args:
            sessionId (string): セッションID
        
        Returns:
            dictionary: セッション情報
        """
        if sessionId in self.webSessions:
            return self.webSessions[sessionId]
        else:
            return None

    def remove_sessio_info(self, sessionId):
        """セッション情報削除
        
        Args:
            sessionId (string): セッションID
        """
        if sessionId in self.webSessions:
            del self.webSessions[sessionId]

    def get_displaying_page(self):
        """表示中ページの取得

        Returns:
            string: 表示中ページ
        """
        sortedSessionList = sorted(self.webSessions.values(), key = lambda x:x["time"], reverse=True)
        if len(sortedSessionList) > 0:
            return sortedSessionList[0]["page"]
        else:
            return ""

class HTTPServerThread(AbstractThread):
    """HTTPサーバースレッド
    """
    def __init__(self, address, handler = partial(MyHTTPRequestHandler, directory=MyHTTPRequestHandler.DIRECTORY )):
        """コンストラクタ
        
        Args:
            address (tuple): 待ち受けアドレス, ポート番号
            handler (class, optional): HTTPRequestHandlerクラス. Defaults to partial(MyHTTPRequestHandler, directory=MyHTTPRequestHandler.DIRECTORY ).
        """
        super(HTTPServerThread, self).__init__()
        self.server = None
        self.address = address
        self.handler = handler

    def pre_proc(self):
        """前処理
        """
        self.__delete_temp_files()

    def thread_proc(self):
        """スレッドループ処理
        
        Returns:
            [bool]: 処理結果：Falseを返すとスレッド終了
        """
        with ThreadingHTTPServer(self.address, self.handler) as server:
            self.server = server
            server.serve_forever()
        return False

    def post_proc(self):
        """後処理
        """
        Display().delete({})

    def stop(self):
        """停止
        """
        if self.server is not None:
            self.server.shutdown()
        super(HTTPServerThread, self).stop()

    def __delete_temp_files(self):
        """tempファイル削除
        """
        temp_dir = os.path.join(MyHTTPRequestHandler.DIRECTORY, "temp")
        try:
            for f in os.listdir(temp_dir):
                path = os.path.join(temp_dir, f)
                if os.path.isfile(path):
                    os.remove(path)
                elif os.path.isdir(path):
                    shutil.rmtree(path)
        except:
            self.logger.error("__delete_temp_files failed.", exc_info=True)

if __name__ == "__main__":
    import time
    address = ('', 8080)
    serverThread = HTTPServerThread(address)
    try:
        serverThread.start()
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        serverThread.stop()
