import subprocess
import time
from datetime import datetime, timedelta

from ntplib import NTPClient

from absthread import AbstractThread
from recv_dev.config import OtherInfo
from recv_dev.storage import CallLog, IntervalJob
from setup_logger import setup_logger_devel


class BaseIntervalJob:
    def __init__(self):
        """コンストラクタ
        """
        self.logger = setup_logger_devel(self.__class__.__name__)

    def executeJob(self):
        pass

    def _interval_time_check(self, intervalStr):
        """実行間隔チェック
        
        Args:
            intervalStr (string): 実行間隔
        
        Returns:
            bool: チェック結果（Trueなら実行）
        """
        key = {"jobId": self.__class__.__name__}
        jobs = IntervalJob().select(key)
        if len(jobs) > 0:
            job = jobs[0]
            if job["datetime"] is None or len(job["datetime"]) == 0:
                return True
            else:
                unit = intervalStr[0:1]
                interval = int(intervalStr[1:])
                delta = timedelta()
                if unit == "M":
                    delta = timedelta(minutes=interval)
                elif unit == "H":
                    delta = timedelta(hours=interval)
                elif unit == "D":
                    delta = timedelta(days=interval)
                elif unit == "W":
                    delta = timedelta(weeks=interval)

                prevDt = datetime.strptime(job["datetime"], "%Y-%m-%d %H:%M:%S.%f")
                now = datetime.now()
                if now - prevDt > delta:
                    return True
                else:
                    return False
        else:
            return False

class NtpJob(BaseIntervalJob):
    def isNtp(self):
        """NTP実行可否
        
        Returns:
            bool: NTP実行可否
        """
        otherInfo = OtherInfo()
        if otherInfo.timeSetting["ntp"]:
            interval = otherInfo.timeSetting["interval"]
            return self._interval_time_check(interval)
        else:
            return False

    def executeJob(self):
        """ジョブ実行

        Returns:
            bool: JOB実施有無
        """
        ret = False
        if self.isNtp():
            try:
                self.logger.info("NTP start.")
                otherInfo = OtherInfo()
                client = NTPClient()
                res = client.request(otherInfo.timeSetting["ntpServer"])
                dt = datetime.fromtimestamp(res.tx_time)

                subprocess.run(["date", "-s", "{:%Y/%m/%d %H:%M:%S}".format(dt)], check = True)

                ret = True
                self.logger.info("NTP end.")

            except:
                self.logger.error("doNtp failed.", exc_info=True)

        return ret

class LogClearJob(BaseIntervalJob):
    def isLogClear(self):
        """ログクリア実行可否
        
        Returns:
            bool: ログクリア実行可否
        """
        otherInfo = OtherInfo()

        if otherInfo.logClear is None:
            otherInfo.logClear = {
                "interval": "W01",
                "days": 90
            }
            otherInfo.save()

        interval = otherInfo.logClear["interval"]
        return self._interval_time_check(interval)

    def executeJob(self):
        """ジョブ実行

        Returns:
            bool: JOB実施有無
        """
        ret = False
        if self.isLogClear():
            ret = self.__logClear()

        return ret

    def __logClear(self):
        """ログクリア

        Returns:
            bool: 実行結果
        """
        ret = False
        try:
            self.logger.info("LogClear start.")
            now = datetime.now()

            otherInfo = OtherInfo()
            days = otherInfo.logClear["days"]
            del_datetime = now - timedelta(days=days)
            cl = CallLog()
            cl.deleteOld(del_datetime, {"sent": 1})

            ret = True
            self.logger.info("LogClear end.")
        except:
            self.logger.error("LogClear failed.", exc_info=True)

        return ret
        

class IntervalJobThread(AbstractThread):
    """定期実行ジョブスレッド
    """
    def __init__(self):
        """コンストラクタ
        """
        super(IntervalJobThread, self).__init__()
        self.db = IntervalJob()
        self.prevDt = None

    def pre_proc(self):
        """前処理
        """
        pass

    def thread_proc(self):
        """スレッドループ処理
        """
        if self.prevDt is None or datetime.now() - self.prevDt > timedelta(minutes=1):
            jobs = self.db.select({})
            try:
                for job in jobs:
                    className = job["jobId"]
                    if className in globals():
                        cls = globals()[className]
                        jobInstance = cls()
                        
                        if jobInstance.executeJob():
                            key = {"jobId": className}
                            value = {"datetime": "{:%Y-%m-%d %H:%M:%S.%f}".format(datetime.now())}
                            self.db.update(key, value)
            except:
                self.logger.error("Job failed.", exc_info=True)

            self.prevDt = datetime.now()

        time.sleep(1)
        return True

    def post_proc(self):
        """後処理
        """
        pass


if __name__ == "__main__":
    import time
    jobThread = IntervalJobThread()
    try:
        jobThread.start()
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        jobThread.stop()
