import copy
import re
import subprocess

import netifaces

from setup_logger import setup_logger_devel


class WpaConf():
    """wpa_supplicant.confの設定を行う
    """

    WPA_CONF_FILE = "/etc/wpa_supplicant/wpa_supplicant.conf"

    @property
    def country(self):
        """国コード
        
        Returns:
            string: 国コード
        """
        return self.get("country")

    @country.setter
    def country(self, val):
        """国コード
        
        Args:
            val (string): 設定値
        """
        if self.conf is not None and "country" in self.conf:
            self.conf["country"] = val

    def __init__(self, filepath=WPA_CONF_FILE):
        """コンストラクタ
        
        Args:
            filepath (string, optional): wpa_suppliant.confのファイルパス. Defaults to WPA_CONF_FILE.
        """
        self.logger = setup_logger_devel(self.__class__.__name__)
        self.conf = None
        self.load(filepath)

    def load(self, filepath=WPA_CONF_FILE):
        """wpa_supplicant.confのロード
        
        Args:
            filepath (string, optional): wpa_suppliant.confのファイルパス. Defaults to WPA_CONF_FILE.
        """

        self.conf = self.parse(filepath)

    def parse(self, filepath=WPA_CONF_FILE):
        """wpa_supplicant.confのパース
        
        Args:
            filepath (string, optional): wpa_suppliant.confのファイルパス. Defaults to WPA_CONF_FILE.
        
        Returns:
            dictionary: wpa_supplicant.confのパース結果
        """

        wpa_conf = {}
        lines = []
        with open(filepath) as f:
            lines = f.readlines()

        networkFlg = False
        network = []
        subValue = {}
        for l in lines:
            if l.strip().startswith("#"):
                continue

            sp = l.split('=', 1)
            if networkFlg == False:
                if len(sp) >= 2:
                    if sp[0].strip() == "network" and sp[1].strip() == "{":
                        networkFlg = True
                        subValue = {}
                    else:
                        wpa_conf[sp[0].strip()] = sp[1].strip()
            else:
                if len(sp) >= 2:
                    subValue[sp[0].strip()] = sp[1].strip()
                else:
                    if sp[0].strip() == "}":
                        network.append(subValue)
                        networkFlg = False
                        subValue = {}

        # if len(network) > 0:
        wpa_conf["network"] = network

        return wpa_conf

    def dump(self, conf=None):
        """wpa_supplicant.confをダンプ
        
        Args:
            conf (dictionary, optional): wpa_supplicant設定情報. Defaults to None.
        
        Returns:
            string: wpa_supplicant設定情報
        """

        if conf is None:
            conf = self.conf

        ret = ""

        if conf is not None:
            for (k, v) in conf.items():
                if k == "network":
                    networks = v
                    for network in v:
                        ret = ret + "network={\n"
                        for (nk, nv) in network.items():
                            ret = ret + "        {0}={1}\n".format(nk, nv)

                        ret = ret + "}\n"
                else:
                    ret = ret + "{0}={1}\n".format(k, v)

        return ret

    def save(self, filepath=WPA_CONF_FILE):
        """wpa_supplicant.confを保存
        
        Args:
            filepath (string, optional): wpa_suppliant.confのファイルパス. Defaults to WPA_CONF_FILE.
        """
        with open(filepath, mode='w') as f:
            f.write(self.dump())

    def get(self, key = None):
        """値取得
        
        Args:
            key (string, optional): key. Defaults to None.
        
        Returns:
            object: value
        """
        if self.conf is None:
            return None

        if key == None:
            return self.conf

        if key in self.conf:
            return self.conf[key]
        else:
            return None

    def get_network(self):
        """network取得
        
        Returns:
            list: network情報
        """
        return self.get("network")

    def get_network_ssid(self, ssid):
        """network取得(ssid指定)
        
        Args:
            ssid (string): ssid
        
        Returns:
            dictionary: network情報
        """
        network = self.get_network()
        if network is None:
            return None
        # 絞り込み
        network = [n for n in network if ssid == n["ssid"] or '"' + ssid + '"' == n["ssid"]]
        if len(network) > 0:
            return network[0]
        else:
            return None
            
    def set_network(self, nw):
        """network設定
        
        Args:
            nw (dictionary): network情報
        """
        if self.conf is None:
            return

        if self.conf["network"] is None:
            self.conf["network"] = []

        # 既存のものは更新
        i = 0
        for n in self.conf["network"]:
            if nw["ssid"] == n["ssid"] or '"' + nw["ssid"] + '"' == n["ssid"]:
                self.conf["network"][i] = nw
                return
            i = i + 1

        # 新規のものは追加
        self.conf["network"].append(nw)

    def delete_network(self, ssid):
        """network削除
        
        Args:
            ssid (string): SSID
        """
        if self.conf is None:
            return

        if self.conf["network"] is None:
            return

        # 絞り込み
        self.conf["network"] = [n for n in self.conf["network"] if ssid != n["ssid"] and '"' + ssid + '"' != n["ssid"]]

    def clear_network(self):
        """networkクリア
        """
        network = self.get_network()
        if network is None:
            return

        network.clear()

    def create_network_WPA(self, ssid, passphrase, proto="WPA", pairwise="TKIP", group="TKIP", stealth=False):
        """WPA用network定義生成
        
        Args:
            ssid (string): SSID
            passphrase (string): パスフレーズ
            proto (str, optional): プロトコル. Defaults to "WPA".
            pairwise (str, optional): Pairwise Cipher. Defaults to "TKIP".
            group (str, optional): Group Cipher. Defaults to "TKIP".
            stealth (bool, optional): ステルスモード. Defaults to False.
        
        Returns:
            dictionary: network情報
        """
        network = self.__wpa_passphrase(ssid, passphrase)

        if network is None:
            return None

        network["key_mgmt"]= "WPA-PSK"
        network["proto"] = proto
        network["pairwise"] = pairwise
        network["group"] = group

        if stealth:
            network["scan_ssid"]='1'

        return network

    def create_network_WPA2(self, ssid, passphrase, proto="RSN", pairwise="CCMP", group="CCMP", stealth=False):
        """WPA2用network定義生成
        
        Args:
            ssid (string): SSID
            passphrase (string): パスフレーズ
            proto (str, optional): プロトコル. Defaults to "RSN".
            pairwise (str, optional): Pairwise Cipher. Defaults to "CCMP".
            group (str, optional): Group Cipher. Defaults to "CCMP".
            stealth (bool, optional): ステルスモード. Defaults to False.
        
        Returns:
            dictionary: network情報
        """
        return self.create_network_WPA(ssid, passphrase, proto, pairwise, group, stealth)

    def create_network_WEP(self, ssid, passphrase, keyidx='0', stealth=False):
        """WEP用network定義生成
        
        Args:
            ssid (string): SSID
            passphrase (string): パスフレーズ
            keyidx (str, optional): キーインデックス. Defaults to '0'.
            stealth (bool, optional): ステルスモード. Defaults to False.
        
        Returns:
            dictionary: network情報
        """
        network = self.create_network_NONE(ssid, stealth)
        network["wep_key" + keyidx] = '"' + passphrase +'"'
        network["wep_tx_keyidx"] = keyidx

        return network

    def create_network_NONE(self, ssid, stealth=False):
        """暗号化なしnetwork定義生成
        
        Args:
            ssid (string): SSID
            stealth (bool, optional): ステルスモード. Defaults to False.
        
        Returns:
            dictionary: network情報
        """

        network = {
            "ssid" : '"' + ssid +'"',
            "key_mgmt" : "NONE",
        }

        if stealth:
            network["scan_ssid"]='1'

        return network

    def __wpa_passphrase(self, ssid, passphrase):
        """wpa_passphraseコマンド実行
        
        Args:
            ssid ([type]): [description]
            passphrase ([type]): [description]
        
        Returns:
            [type]: [description]
        """

        cmd = ["wpa_passphrase", ssid, passphrase]
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        network = {}
        try:
            for line in iter(proc.stdout.readline,b''):
                l = line.strip().decode('utf-8')
                if l.startswith("#"):
                    continue

                sp = l.split('=', 1)
                if len(sp) >= 2:
                    if sp[0].strip() == "network" and sp[1].strip() == "{":
                        pass
                    else:
                        network[sp[0].strip()] = sp[1].strip()
                else:
                    if sp[0].strip() == "}":
                        break
        except:
            network = None
            self.logger.error("__wpa_passphrase failed.", exc_info=True)

        return network
 
    def ip_link_set_wlan0_down(self):
        """wlan0インターフェースdown
        """
        networkStatus = NetworkStatus()
        networkStatus.if_up_down("wlan0", False)

    def ip_link_set_wlan0_up(self):
        """wlan0インターフェースup
        """
        networkStatus = NetworkStatus()
        networkStatus.if_up_down("wlan0", True)

    def apply(self):
        """wpa_supplicant.confをシステムに反映する
        (dhcpcd.serviceをrestartする)
        """
        cmd = ["systemctl",  "restart", "dhcpcd.service"]
        subprocess.run(cmd)



class DhcpcdConf():
    """dhcpcd.conf設定クラス
    """

    DHCPCD_CONF_FILE = "/etc/dhcpcd.conf"
    def __init__(self, filepath=DHCPCD_CONF_FILE):
        """コンストラクタ
        
        Args:
            filepath (string, optional): dhcpcd.confファイルパス. Defaults to DHCPCD_CONF_FILE.
        """
        self.lines = []
        self.interfaces = {}
        self.load(filepath)

    def load(self, filepath=DHCPCD_CONF_FILE):
        """dhcpcd.confのロード
        
        Args:
            filepath (string, optional): dhcpcd.confファイルパス. Defaults to DHCPCD_CONF_FILE.
        """
        self.lines = []
        ls = []
        with open(filepath) as f:
            ls = f.readlines()

        ifname = ""
        interface = {}
        for l in ls:
            if len(ifname) == 0:
                m = re.search(r'^\s*interface (.*)$', l)
                if m is not None:
                    ifname = m.group(1).strip()
                    interface = { "autoIp": True, "autoDNS": True}
                    self.interfaces[ifname] = interface
                    continue
            else:
                m = re.search(r'^static ip_address=([0-9.]+)/([0-9]+)$', l)
                if m is not None:
                    if ifname in self.interfaces:
                        interface = self.interfaces[ifname]
                        interface["ipAddress"] = m.group(1)
                        interface["subnetPrefix"] = m.group(2)
                        interface["autoIp"] = False
                        self.interfaces[ifname] = interface

                    continue

                m = re.search(r'^static ip6_address=.*$', l)
                if m is not None:
                    continue

                m = re.search(r'^static routers=(.*)$', l)
                if m is not None:
                    if ifname in self.interfaces:
                        interface = self.interfaces[ifname]
                        interface["defaultGw"] = m.group(1)
                        self.interfaces[ifname] = interface

                    continue

                m = re.search(r'^static domain_name_servers=(.*)$', l)
                if m is not None:
                    if ifname in self.interfaces:
                        interface = self.interfaces[ifname]
                        interface["dns"] = m.group(1).split(' ')
                        interface["autoDNS"] = False
                        self.interfaces[ifname] = interface

                    continue

                m = re.search(r'^\s*interface (.*)$', l)
                if m is not None:
                    ifname = m.group(1).strip()
                    interface = { "autoIp": True, "autoDNS": True}
                    self.interfaces[ifname] = interface
                    continue

            self.lines.append(l)

    def get_info(self, ifname):
        """IP設定情報取得
        
        Args:
            ifname (string): インターフェース名
        
        Returns:
            dictionary: 設定情報 
                                = {
                                    "dns": False,
                                    "ipAddress": "172.16.x.x",
                                    "subnetMask": "255.255.255.0",
                                    "defaultGw": "172.16.x.x",
                                    "dns": [ "8.8.8.8", "x.x.x.x"]],
                                }
        """
        interface = {"autoIp": True, "autoDNS": True}
        if ifname in self.interfaces:
            interface = copy.copy(self.interfaces[ifname])

        if "subnetPrefix" in interface:
            interface["subnetMask"] = self.prefix_to_mask(int(interface["subnetPrefix"]))

        return interface

    def prefix_to_mask(self, prefx):
        """subnetプレフィックス→サブネットマスク
        
        Args:
            prefx (int): サブネットプレフィックス(ex:24)
        
        Returns:
            string: サブネットマスク(ex:"255.255.255.0")
        """
        mask = "1" * prefx + "0" * (32 - prefx) 
        return "{0}.{1}.{2}.{3}".format(int(mask[0:8], 2), int(mask[8:16],2), int(mask[16:24],2), int(mask[24:32],2))

    def mask_to_prefix(self, subnet_mask):
        """サブネットマスク→subnetプレフィックス
        
        Args:
            subnet_mask (string): サブネットマスク(ex:"255.255.255.0")
        
        Returns:
            int: サブネットプレフィックス(ex:24)
        """
        masks = subnet_mask.split('.')
        maskbin = ""
        for m in masks:
            maskbin = maskbin + "{0:08b}".format(int(m))
        return maskbin.count("1")

    def set_info(self, ifname, info):
        """IP設定情報更新
        
        Args:
            ifname (string): インターフェース名
            info (dictionary): 設定情報
                                = {
                                    "autoIp": False,
                                    "ipAddress": "172.16.x.x",
                                    "subnetMask": "255.255.255.0",
                                    "defaultGw": "172.16.x.x",
                                    "autoDNS": False,
                                    "dns": [ "8.8.8.8", "x.x.x.x"]],
                                }
        """

        self.interfaces[ifname] = info


    def save(self, filepath=DHCPCD_CONF_FILE):
        """ファイル書き込み
        
        Args:
            filepath (string, optional): dhcpcd.confファイルパス. Defaults to DHCPCD_CONF_FILE.
        """
        # ファイル書き込み
        with open(filepath, mode='w') as f:
            f.writelines(self.lines)

            for (ifname, interface) in self.interfaces.items():
                f.write("interface {}\n".format(ifname))

                autoIp = True
                if "autoIp" in interface and  interface["autoIp"] == False:
                    autoIp = False
                
                autoDNS = True
                if "autoDNS" in interface and  interface["autoDNS"] == False:
                    autoDNS = True

                for  (k, v) in interface.items():
                    if k == "ipAddress":
                        if autoIp == False:
                            prefix = ""
                            if "subnetPrefix" in interface:
                                prefix = interface["subnetPrefix"]
                            elif "subnetMask" in interface:
                                prefix = self.mask_to_prefix(interface["subnetMask"])

                            f.write("static ip_address={0}/{1}\n".format(interface["ipAddress"], prefix))
                        
                    elif k == "defaultGw":
                        if autoIp == False:
                            f.write("static routers={0}\n".format(interface["defaultGw"]))
                    else:
                        if k == "dns":
                            if autoDNS == True:
                                f.write("static domain_name_servers={0}\n".format(" ".join(interface["dns"])))

    def service_dhcpcd_restart(self):
        """dhcpcd.confの設定を反映するためにサービスをrestartする
        """
        # cmd = ["service",  "dhcpcd", "reload"]
        cmd = ["systemctl",  "restart", "dhcpcd.service"]
        subprocess.call(cmd)


class NetworkStatus:
    def __init__(self):
        """コンストラクタ
        """
        self.logger = setup_logger_devel(self.__class__.__name__)

    def get_net_status(self, ifname):
        """ネットワークステータス取得
        
        Args:
            ifname (string): インターフェース名
        
        Returns:
            dictionary: ネットワークステータス
        """
        status = {}
        addresses = netifaces.ifaddresses(ifname)
        if netifaces.AF_INET in addresses:
            status["ipAddress"] = addresses[netifaces.AF_INET][0]["addr"]
            status["subnetMask"] = addresses[netifaces.AF_INET][0]["netmask"]

        # default Gateway
        if netifaces.AF_INET in netifaces.gateways():
            for (addr, ifn, is_default) in netifaces.gateways()[netifaces.AF_INET]:
                if ifn == ifname:
                    status["defaultGw"] = addr

        if ifname == "wlan0":
            status["ssid"] = self.get_ssid()

        dns = self.get_dns()
        status["dns"] = dns
        return status

    def get_ssid(self):
        """接続中SSID取得
        
        Returns:
            string: SSID
        """
        cmd = 'wpa_cli -i wlan0 status | grep ^ssid'
        r = subprocess.run(cmd, shell = True, stdout = subprocess.PIPE, stderr = subprocess.PIPE)\
            .stdout.decode().rstrip()
        idx = r.find('ssid=')

        if idx >= 0:
            return r[idx + 5:]

    def get_dns(self):
        """使用中DNSサーバアドレス取得
        
        Returns:
            list: DNSサーバアドレス
        """
        dns = []
        resolv_conf = "/etc/resolv.conf"
        cmd = 'cat ' + resolv_conf + ' |grep nameserver'
        lines = subprocess.run(cmd, shell = True, stdout = subprocess.PIPE, stderr = subprocess.PIPE)\
            .stdout.decode('utf-8').split('\n')
        for line in lines:
            line = line.strip() 
            idx = line.find("nameserver")
            if idx >= 0:
                dns.append(line[idx + 11:])

        return dns

    def if_up_down(self, ifname, up):
        """インターフェースup/down
        
        Args:
            ifname (string): インターフェース名
            up (bool): up=True, down=False
        
        Returns:
            bool: 実行結果
        """
        ret = True
        cmd = ["ip",  "link",  "set", ifname]
        if up:
            cmd.append("up")
        else:       
            cmd.append("down")
        try:
            subprocess.run(cmd, check=True)
        except:
            self.logger.error("if_up_down failed.", exc_info=True)
            ret = False
        return ret

    def if_enabled(self, ifname):
        """インターフェースが有効か？
        
        Args:
            ifname (string): インターフェース名
        
        Returns:
            bool: 有効=True,無効=False
        """
        ret = False
        cmd = 'ip link show ' + ifname + '| grep ' + ifname
        try:
            r = subprocess.run(cmd, shell = True, stdout = subprocess.PIPE, stderr = subprocess.PIPE, check=True)\
                .stdout.decode().rstrip()
            idx = r.find("state")
            if idx >= 0:
                if r[idx + 6:].startswith("U"):
                    # UP
                    ret =  True
        except:
            self.logger.error("if_enabled failed.", exc_info=True)
        return ret

    def if_flush_dev(self, ifname):
        """インターフェースのIPアドレスを削除
        
        Args:
            ifname (string): インターフェース名
        
        Returns:
            bool: 実行結果
        """
        ret = True
        cmd = ["ip",  "addr",  "flush", "dev", ifname]

        try:
            subprocess.run(cmd, check=True)
        except:
            self.logger.error("if_flush_dev failed.", exc_info=True)
            ret = False
        return ret

    # def ip_link_set_wlan0_down(self, ifname, up):
    #     """wlan0インターフェースdown
    #     """
    #     cmd = "ip link set wlan0 down"
    #     subprocess.call(cmd.split())

    # def ip_link_set_wlan0_up(self):
    #     """wlan0インターフェースup
    #     """
    #     cmd = "ip link set wlan0 up"
    #     subprocess.call(cmd.split())

if __name__ == "__main__":
    # wpaConf = WpaConf('/home/pi/wpa_conf.conf')
    # conf = wpaConf.parse() 
    # print(conf)
    # print(wpaConf.dump(conf))

    # print(wpaConf.dump())
    # print(wpaConf.get())
    # print(wpaConf.get("country"))
    # print(wpaConf.get_network())
    # print(wpaConf.get_network_ssid("MSCREATIVE-a"))
    # nw2 = copy.copy(wpaConf.get_network_ssid("MSCREATIVE-a"))
    # nw2["ssid"]='"MSCREATIVE-g"'
    # wpaConf.set_network(nw2)
    # print(wpaConf.create_network_WPA("MSCREATIVE-a", "masa8444"))
    # print(wpaConf.create_network_WPA2("MSCREATIVE-a", "masa8444"))
    # print(wpaConf.create_network_WEP("MSCREATIVE-a", "masa8444"))
    # print(wpaConf.create_network_NONE("MSCREATIVE-a", stealth = True))
    # wpaConf.set_network(wpaConf.create_network_WPA2("MSCREATIVE-a", "masa8444"))
    # print(wpaConf.dump())

    dhcp = DhcpcdConf("/home/pi/dhcpcd2.conf")
    print(dhcp.lines)
    print(dhcp.interfaces)
    print(dhcp.get_info("eth0"))
    print(dhcp.get_info("wlan0"))

    info = dhcp.get_info("eth0")
    info["autoIp"] = False
    info["ipAddress"] = "172.16.1.61"
    info["subnetMask"] = "255.255.255.0"
    info["defaultGw"] = "172.16.1.1"
    info["dns"] = ["172.16.1.100", "172.16.1.101"]

    dhcp.set_info("wlan0", info)

    print(dhcp.get_info("eth0"))
    print(dhcp.get_info("wlan0"))

    dhcp.save("/home/pi/dhcpcd2.conf")

    # print(dhcp.get_info("eth0", "/home/pi/dhcpcd.conf"))
    # print(dhcp.get_info("wlan0", "/home/pi/dhcpcd.conf"))

    # info = dhcp.get_info("eth0", "/home/pi/dhcpcd.conf")
    # info["autoIp"] = False
    # info["ipAddress"] = "172.16.1.61"
    # info["subnetMask"] = "255.255.255.0"
    # info["defaultGw"] = "172.16.1.1"
    # info["dns"] = ["192.168.0.1","8.8.8.8"]
    # dhcp.set_info("wlan0", info, "/home/pi/dhcpcd2.conf")
    # print(dhcp.get_info("wlan0", "/home/pi/dhcpcd2.conf"))
    # info["autoIp"] = True
    # dhcp.set_info("eth0", info, "/home/pi/dhcpcd2.conf")
    # print(dhcp.get_info("eth0", "/home/pi/dhcpcd2.conf"))
    # print(dhcp.mask_to_prefix("255.255.255.0"))

    # wpaConf.save('/home/pi/wpa_conf.conf')
