import json
import pathlib
import subprocess
import threading
import time
from datetime import datetime

from absreceiver import AbstractReceiver
from absthread import AbstractThread
from base_demo_receiver import BaseDemoReceiver
# from kivy.app import App
from kivy.clock import Clock
from kivy.core.audio import SoundLoader
from ntplib import NTPClient
from resources import Resources
from setup_logger import setup_logger_devel
from telegram import Telegram

from recv_dev.config import CallerSetting, OtherInfo, TerminalInfo
from recv_dev.gpio import GPIOControl
from recv_dev.httpserver import HTTPServerThread, WebSession
from recv_dev.intervaljob import IntervalJobThread
from recv_dev.recvdev import DummyRecvDevice, RecvDevice
from recv_dev.storage import (Caller, CallerButton, CallerButtonView, CallList,
                              CallListView, CallLog, Database, DelList,
                              DelListView, Eraser, Group, IntervalJob)
from recv_dev.tcpserver import Status, TCPServerThread
from recv_dev.udpserver import UDPServerThread


class DemoReceiver(BaseDemoReceiver):
    """デモ受信機
    """
    pass

class Receiver(AbstractReceiver):
    """受信機
    """
    def __init__(self, received, erased, synced, pairing, received_on_pairing, long_pressed_on_pairing, processing, shutdown, eraser_pushed, **kwargs):
        """コンストラクタ
        
        Args:
            received (function): 呼出受信コールバック
            erased (function): 取消受信コールバック
            synced (function): 同期コールバック
            pairing (function): ペアリングボタン押下コールバック
            received_on_pairing (function): 呼出受信コールバック(ペアリングモード時)
            long_pressed_on_pairing (function): クリア長押しコールバック
            processing (function): 処理中表示コールバック
            shutdown (function): シャットダウンコールバック
            eraser_pushed (function): 消去機押下コールバック
        """
        super(Receiver, self).__init__(received, erased, synced, pairing, received_on_pairing, long_pressed_on_pairing, processing, shutdown,eraser_pushed, **kwargs)

        self.pairing = False
        self.pairingCb = pairing
        self.long_pressed_on_pairing = long_pressed_on_pairing
        self.shutdown = shutdown
        udp_address = ('', 5000)
        tcp_address = ('', 5001)
        http_address = ('', 80)
        switch_callbacks = {
            "pairing": self.sw_pairing_callback,
            "clear": self.sw_clear_callback
        }
        self.register = RecvDataRegister(dummyMode = TerminalInfo().dummyMode, received_on_pairing=received_on_pairing, eraser_pushed=eraser_pushed)
        tcpserverThread = TCPServerThread(tcp_address)
        self.threads = [
            self.register,
            UDPServerThread(udp_address),
            tcpserverThread,
            HTTPServerThread(http_address),
            IntervalJobThread(),
            WatchDataThread(tcpserverThread, received, erased),
            GPIOControl(switch_callbacks, self.shutdown_callback, dummyMode = TerminalInfo().dummyMode)
        ]

        if TerminalInfo().dummyMode:
            # ダミーモードの場合、debug用にsshを有効にする
            cmd = ["systemctl", "start", "ssh.service"]
            subprocess.run(cmd)

        self.bind(pairing=self.on_pairing_change)

        # 60秒間隔でスレッドの状態を監視する。
        self.thread_check_event = Clock.schedule_interval(lambda dt: self.checkThreads(), 60)



    def start(self):
        """開始
        
        Args:
            received (function): 呼出受信コールバック
            erased (function): 取消受信コールバック
            synced (function): 同期コールバック
        """
        self.logger.info("receiver start.")

        cl = CallList()
        cl.delete({})

        dl = DelList()
        dl.delete({})

        #初回NTP実行
        self.__do_first_ntp()

        for th in self.threads:
            th.start()


    def requestErase(self, data, eraseFinishedCb):
        """削除要求

        Args:
            data (dict): 削除データ
            eraseFinishedCb (function): 削除完了コールバック
        """
        key = {
            "address" : data["address"],
            "modelNo" : data["modelNo"],
            "buttonNo" : data["buttonNo"]
        }

        cl = CallList()
        calllist = cl.select(key)
        # ログ用
        clv = CallListView()
        vcalllist = clv.select(key)
        now = datetime.now()
        cl.delete(key)

        dl = DelList()
        for called in calllist:
            called["displayName"] = data["displayName"]
            called["upd_dt"] = "{0:%Y-%m-%d %H:%M:%S.%f}".format(now)
            called["displayed"] = "0"
            dl.insert(called)

        # 呼出ログ出力
        self.__insert_call_log(now, data, vcalllist)

        if eraseFinishedCb is not None:
            eraseFinishedCb(True, data)

    def __insert_call_log(self, dt, data, vcalllist):
        """呼出ログ出力

        Args:
            dt (datetime): 呼出日時
            data (dict): 消去要求データ
            vcalllist (list): 消去対象データのリスト
        """
        try:
            calllog = CallLog()
            for called in vcalllist:
                groupName = ""
                if called["groupId"] == 0:
                    groupName = "all"
                else:
                    groupName = called["groupName"]
                ipAddress = ""

                message = "address={0},model={1},button={2},strength=-{3}dBm,caller name1={4}, caller name2={5},group={6},color={7},device ip={8},device name={9}".format(
                    data["address"], data["modelNo"], data["buttonNo"], called["strength"], called["line1"], called["line2"], groupName, called["color"], ipAddress, data["displayName"] )

                calllog.insert({
                    "datetime": "{0:%Y-%m-%d %H:%M:%S.%f}".format(dt),
                    "category": CallLog.CATEGORY_RESPONSE,
                    "message": message,
                    "sent": 0
                })
        except:
            self.logger.error("__insert_call_log failed.", exc_info=True)

    def requestSync(self, syncFinishedCb):
        pass


    def stop(self):
        """終了
        """
        if self.thread_check_event is not None:
            self.thread_check_event.cancel()

        for th in self.threads:
            th.stop()

        # cl = CallList()
        # cl.delete({})

        # dl = DelList()
        # dl.delete({})
        
        self.logger.info("receiver stop.")


    def pause(self):
        pass

    def resume(self):
        pass


    def sw_pairing_callback(self, pin, t):
        """ペアリングスイッチコールバック
        
        Args:
            pin (int): PIN番号
            t (float): スイッチ押下時間(秒)
        """
        self.logger.debug("switch pressed {0}. {1}s".format(pin, t))
        if t < GPIOControl.PRESS_LONG:
            # 長押しではない
            try:
                if self.pairingCb is not None:
                    self.pairingCb()
                # self.pairing = not self.pairing
                # self.logger.debug("pairing mode {0}.".format(self.pairing))
                # self.register.pairing = self.pairing

                # #ログ出力
                # now = datetime.now()
                # calllog = CallLog()
                # message = "on" if self.pairing == True else "off"
                # calllog.insert({
                #     "datetime": "{0:%Y-%m-%d %H:%M:%S.%f}".format(now),
                #     "category": CallLog.CATEGORY_PAIRING,
                #     "message": message,
                #     "sent": 0
                # })
            except:
                self.logger.error("sw_pairing_callback failed", exc_info=True)


    def sw_clear_callback(self, pin, t):
        """クリアスイッチコールバック
        
        Args:
            pin (int): PIN番号
            t (float): スイッチ押下時間(秒)
        """
        self.logger.debug("switch pressed {0}. {1}s".format(pin, t))
        if t >= GPIOControl.PRESS_LONG:
            # 長押し
            if self.long_pressed_on_pairing is not None:
               self.long_pressed_on_pairing()

    def shutdown_callback(self):
        """シャットダウン要求コールバック
        電源ケーブルが抜かれたときに呼ばれる。
        """
        self.logger.info("shutdown start.")
        if self.shutdown is not None:
            self.shutdown()
        # App.get_running_app().stop()
        # subprocess.run(["poweroff"])

    def checkThreads(self):
        """スレッドの状態監視
        """
        self.logger.debug("--threads status--")
        for th in self.threads:
            self.logger.debug("{0} isRunning={1}".format(th, th.isRunning))
            if th.isRunning == False:
                self.logger.info("{} thread restart!".format(th))
                th.restart()

    def on_pairing_change(self, instance, pairing):
        """ペアリングモード変更イベントハンドラー
        
        Args:
            instance (object): イベント発生オブジェクト
            pairing (bool): ペアリングモード
        """
        self.logger.debug("pairing mode {0}.".format(pairing))
        self.register.pairing = pairing

        #ログ出力
        now = datetime.now()
        calllog = CallLog()
        message = "on" if self.pairing == True else "off"
        calllog.insert({
            "datetime": "{0:%Y-%m-%d %H:%M:%S.%f}".format(now),
            "category": CallLog.CATEGORY_PAIRING,
            "message": message,
            "sent": 0
        })

    def __do_first_ntp(self):
        """初回のNTPの実行
        """
        try:
            otherInfo = OtherInfo()
            if otherInfo.timeSetting["ntp"]:
                self.logger.info("NTP start.")
                client = NTPClient()
                res = client.request(otherInfo.timeSetting["ntpServer"])
                dt = datetime.fromtimestamp(res.tx_time)

                subprocess.run(["date", "-s", "{:%Y/%m/%d %H:%M:%S}".format(dt)], check = True)

                key = {"jobId": "NtpJob"}
                value = {"datetime": "{:%Y-%m-%d %H:%M:%S.%f}".format(datetime.now())}
                db = IntervalJob()
                db.update(key, value)

                self.logger.info("NTP end.")

        except:
            self.logger.error("doNtp failed.", exc_info=True)

class RecvDataRegister():
    """受信データ登録クラス
    """

    @property
    def isRunning(self):
        return self.dev.isRunning

    def __init__(self, dummyMode = False, received_on_pairing=None, eraser_pushed=None):
        """コンストラクタ
        
        Args:
            dummyMode (bool, optional): ダミーモード. Defaults to False.
            received_on_pairing (function): 呼出受信コールバック(ペアリングモード時). Defaults to None.
        """
        self.logger = setup_logger_devel(self.__class__.__name__)
        # self.logger_call = setup_logger_for_call(__name__)
        self.calllog = CallLog()
        self.received_on_pairing = received_on_pairing
        self.eraser_pushed = eraser_pushed

        self.pairing = False

        # 受信装置
        if dummyMode == True:
            # Dummy
            self.dev = DummyRecvDevice()

        else:
            self.dev = RecvDevice()



    def start(self):
        """開始
        """
        # 受信装置開始
        self.logger.info("start.")
        self.dev.start(self.recv_cb)

    def stop(self):
        """終了
        """
        # 受信装置終了
        self.dev.stop()
        self.logger.info("stop.")

    def restart(self):
        self.stop()
        self.start()

    def recv_cb(self, address, modelNo, buttonNo, strength, longPress = False):
        """受信装置からデータを受信

        Args:
            address (string): アドレス
            modelNo (string): モデルNo
            buttonNo (string): ボタンNo
            strength (int): 電波強度
            longPress (boolean): 長押し
        """

        self.logger.info('received: address={0}, modelNo={1}, buttonNo={2}, strength=-{3}dBm, longPress={4}, pairing mode={5}'.format(address, modelNo, buttonNo, strength, longPress, self.pairing))
        if self.pairing == False:
            # self.logger_call.info('{0},{1},{2},{3}'.format(address, modelNo, buttonNo, strength))
            recvDt = datetime.now()

            cbv = CallerButtonView()
            er = Eraser()
            

            if modelNo[0] == "A":
                key = {"address": address, "buttonNo": buttonNo}
                # 呼出機マスタの検索
                cl = cbv.select(key)
                if len(cl) == 0:
                    return

                # モデルA
                if longPress:
                    #長押しは消去
                    cl[0]["call"] = 100
                    self.__erase_data(address, modelNo, buttonNo, strength)
                else:
                    if cl[0]["call"] == 101:
                        # 無効
                        pass
                    elif cl[0]["call"] == 100:
                        # 消去
                        self.__erase_data(address, modelNo, buttonNo, strength)
                    else:
                        # 登録
                        self.__regist_data(address, modelNo, buttonNo, strength, cl[0]["call"], recvDt)

                # ログ書き込み
                self.__insert_call_log(recvDt, cl[0], strength)

            elif modelNo[0] == "B":
                key = {"address": address}
                el = er.select(key)
                if len(el) > 0:
                    if self.eraser_pushed is not None:
                        self.eraser_pushed(el[0]["name"], buttonNo)
                        # print("eraserName={}, buttonNo={}".format(el[0]["name"], buttonNo))

                else:
                    key = {"address": address, "buttonNo": buttonNo}
                    # 呼出機マスタの検索
                    cl = cbv.select(key)
                    if len(cl) == 0:
                        return

                    # モデルB
                    if cl[0]["call"] == 101:
                        # 無効
                        pass
                    elif cl[0]["call"] == 100:
                        # 消去
                        self.__erase_data(address, modelNo, buttonNo, strength)
                    else:
                        # 登録
                        self.__regist_data(address, modelNo, buttonNo, strength, cl[0]["call"], recvDt)

                    # ログ書き込み
                    self.__insert_call_log(recvDt, cl[0], strength)

        else:
            # ペアリングモード
            displayingPage = WebSession().get_displaying_page()
            if displayingPage == "eraser.html":
                if modelNo == "B5":
                    #消去機登録
                    self.__regist_eraser(address, modelNo, strength)
            elif displayingPage == "caller.html":
            # else:
                #呼出機登録
                self.__regist_caller(address, modelNo, strength)


    def __regist_data(self, address, modelNo, buttonNo, strength, groupId, recvDt):
        """呼び出しデータ登録
        
        Args:
            address (string): アドレス
            modelNo (string): モデルNo
            buttonNo (string): ボタンNo
            strength (int): 電波強度
            groupId (int): グループID
            recvDt (datetime): 受信日時
        """

        key = {"address": address, "buttonNo": buttonNo}
        t_calllist = CallList()

        tl = t_calllist.select(key)
        if len(tl) == 0:

            insData = {
                "address": address,
                "modelNo": modelNo,
                "buttonNo": buttonNo,
                "strength": strength,
                "groupId": groupId,
                "ins_dt": "{0:%Y-%m-%d %H:%M:%S.%f}".format(recvDt),
                "upd_dt": "{0:%Y-%m-%d %H:%M:%S.%f}".format(datetime.now()),
                "displayed": "0"
            }
            t_calllist.insert(insData)

        else: 
            # 登録済み
            updVal = {
                "recall": 1,
                "upd_dt": "{0:%Y-%m-%d %H:%M:%S.%f}".format(datetime.now()),
                "displayed": "0"
            }

            t_calllist.update(key, updVal)

    def __erase_data(self, address, modelNo, buttonNo, strength):
        """取消データ登録
        
        Args:
            address (string): アドレス
            modelNo (string): モデルNo
            buttonNo (string): ボタンNo
            strength (int): 電波強度
        """
        key = {"address": address}
        t_calllist = CallList()

        tl = t_calllist.select(key)
        if len(tl) > 0:
            t_calllist.delete(key)

            t_dellist = DelList()

            for t in tl:
                insData = {
                    "address": t["address"],
                    "modelNo": t["modelNo"],
                    "buttonNo": t["buttonNo"],
                    "groupId": t["groupId"],
                    "displayName": Resources().getResource("recvdataregister:delname"),
                    "strength": strength,
                    "recall": t["recall"],
                    "ins_dt": t["ins_dt"],
                    "upd_dt": "{0:%Y-%m-%d %H:%M:%S.%f}".format(datetime.now()),
                    "displayed": "0"
                }
                t_dellist.insert(insData)

    def __regist_caller(self, address, modelNo, strength):
        """呼出機登録
        
        Args:
            address (string): アドレス
            modelNo (string): モデルNo
            strength (int): 電波強度
        """
        try:
            key = {"address": address}

            e = Eraser()
            erasers = e.select(key)
            if len(erasers) > 0:
                #消去機として登録済み
                self.received_on_pairing(None, strength, erasers[0], False, True)

            else:
                cv = CallerButtonView()

                registered = False
                callers = cv.select(key)
                if len(callers) > 0:
                    registered = True
                    if self.received_on_pairing is not None:
                        self.received_on_pairing(registered, strength, callers[0], True)
                else:
                    buttonNum = int(modelNo[1:2]) if modelNo[1:2].isdecimal() else ord(modelNo[1:2].upper()) - ord('A') + 10

                    if buttonNum > 0:
                        c = Caller()
                        if len(c.select({'address': address})) > 0:
                            return

                        c.insert({'address': address, 'modelNo': modelNo, 'ins_dt': "{:%Y-%m-%d %H:%M:%S.%f}".format(datetime.now())})

                        cb = CallerButton()
                        insData = []
                        colors=['R', 'R', 'Y', 'G', 'B', 'W']
                        for i in range(1, buttonNum + 1):
                            if modelNo[0] == 'B':
                                color = colors[i] if i < len(colors) else colors[0]   
                                insData.append({'address': address, 'buttonNo': str(i).zfill(2), 'color': color})
                            else:
                                insData.append({'address': address, 'buttonNo': str(i).zfill(2)})

                        if len(insData) > 0:
                            cb.insertmany(insData)

                    if self.received_on_pairing is not None:
                        callers = cv.select(key)
                        if len(callers) > 0:
                            self.received_on_pairing(registered, strength, callers[0], True)

        except:
            self.logger.error("__regist_caller failed.", exc_info=True)


    def __regist_eraser(self, address, modelNo, strength):
        """消去機登録
        
        Args:
            address (string): アドレス
            modelNo (string): モデルNo
            strength (int): 電波強度
        """
        try:
            key = {"address": address}

            c = Caller()
            callers = c.select(key)
            if len(callers) > 0:
                #呼出機として登録済み
                self.received_on_pairing(None, strength, callers[0], True, True)

            else:
                e = Eraser()

                registered = False
                erasers = e.select(key)
                if len(erasers) > 0:
                    registered = True
                    if self.received_on_pairing is not None:
                        self.received_on_pairing(registered, strength, erasers[0], False)
                else:
                    e.insert({'address': address, 'modelNo': modelNo,
                        'name': Resources().getResource("recvdataregister:default_eraser_name"),
                        'ins_dt': "{:%Y-%m-%d %H:%M:%S.%f}".format(datetime.now())})

                    if self.received_on_pairing is not None:
                        erasers = e.select(key)
                        if len(erasers) > 0:
                            self.received_on_pairing(registered, strength, erasers[0], False)

        except:
            self.logger.error("__regist_eraser failed.", exc_info=True)

    def __insert_call_log(self, dt, cbv, strength):
        """呼び出しログ出力

        Args:
            dt (datetime): 呼出日時
            cbv (dict): 呼出機情報
            strength (int): 電波強度
        """
        try:
            groupName = ""
            if cbv["call"] == 0:
                groupName = "all"
            elif cbv["call"] == 100:
                groupName = "cancel"
            elif cbv["call"] == 101:
                groupName = "invalid"
            else:
                g = Group()
                gl = g.select({"groupId": cbv["call"]})
                if len(gl) > 0:
                    groupName = gl[0]["name"]

            message = "address={0},model={1},button={2},strength=-{3}dBm,caller name1={4}, caller name2={5},group={6},color={7}".format(
                cbv["address"], cbv["modelNo"], cbv["buttonNo"], strength, cbv["line1"], cbv["line2"], groupName, cbv["color"] )

            self.calllog.insert({
                "datetime": "{0:%Y-%m-%d %H:%M:%S.%f}".format(dt),
                "category": CallLog.CATEGORY_CALL,
                "message": message,
                "sent": 0
            })
        except:
            self.logger.error("__insert_call_log failed.", exc_info=True)



class WatchDataThread(AbstractThread):
    def __init__(self, tcpserver, received, erased):
        """コンストラクタ
        
        Args:
            tcpserver (object): TCPサーバスレッドオブジェクト
            received (function): 受信コールバック
            erased (function): 取消コールバック
        """
        super(WatchDataThread, self).__init__()
        self.soundInit()
        self.tcpserver = tcpserver
        self.received = received
        self.erased = erased

    def pre_proc(self):
        """前処理
        """
        pass

    def thread_proc(self):
        """スレッドループ処理
        
        Returns:
            [bool]: 処理結果：Falseを返すとスレッド終了
        """
        self.recv_watcher()
        time.sleep(0.05)
        return True

    def post_proc(self):
        """後処理
        """
        self.releaseSound()

    def recv_watcher(self):
        """受信データ監視
        """
        # 呼び出しチェック
        clv = CallListView()
        key = {'displayed': '0'}
        orderBy = {'upd_dt': 'ASC'}


        calledlist = clv.select(key, orderBy)
        if len(calledlist) > 0:
            for c in calledlist:
                self.playSound(c["sound"])

            # 受信コールバック関数呼び出し
            self.received(calledlist)

            cl = CallList()
            for c in calledlist:
                updkey = {
                    "address": c["address"],
                    "modelNo": c["modelNo"],
                    "buttonNo": c["buttonNo"],
                    "ins_dt": c["ins_dt"],
                    "upd_dt": c["upd_dt"],
                }
                cl.update(updkey, {'displayed': '1'})

            # 表示機に送信
            self.sendCallList(calledlist)

        # 消去履歴チェック
        dlv = DelListView()
        dellist = dlv.select(key, orderBy)
        if len(dellist) > 0:
            # 取消コールバック関数呼び出し
            self.erased(dellist)

            dl = DelList()
            for d in dellist:
                updkey = {
                    "address": d["address"],
                    "modelNo": d["modelNo"],
                    "buttonNo": d["buttonNo"],
                    "ins_dt": d["ins_dt"],
                    "upd_dt": d["upd_dt"],
                }
                dl.update(updkey, {'displayed': '1'})

            # 表示機に送信
            self.sendDelList(dellist)


    def sendCallList(self, callList):
        """呼び出しデータを表示機に送信
        
        Args:
            callList (list): 呼び出しデータのリスト
        """
        self.__send_list(Telegram.COMMAND_S, callList)

    def sendDelList(self, delList):
        """取消データを表示機に送信
        
        Args:
            delList (list): 取消データのリスト
        """
        self.__send_list(Telegram.COMMAND_C, delList)

    def __send_list(self, command, datalist):
        """データリスト送信
        
        Args:
            command (string)): コマンド
            datalist (list): データリスト
        """
        for data in datalist:
            connInfoList = self.getConnInfoListByGroupId(data["groupId"])
            for connInfo in connInfoList:
                if self.tcpserver.getStatus(connInfo["socket"]) == Status.S3:
                    t = Telegram()
                    send_data = \
                        t.create_telegram_call_data_R(command, data, \
                            self.tcpserver.get_seqNo(connInfo["socket"]))
                    self.tcpserver.send_async(connInfo["socket"], send_data)

    def getConnInfoListByGroupId(self, groupId):
        """接続情報取得
        
        Args:
            groupId (int): グループID
        
        Returns:
            [dictionary]: 接続情報
        """
        # 接続管理情報
        # conn_manage = self.tcpserver.conn_manage
        # conn_info_arr = conn_manage.values()
        # if groupId == 0:
        #     # 全てのグループ
        #     return conn_info_arr
        # else:
        #     return [connInfo for connInfo in conn_info_arr if groupId in connInfo["groupId"] or  0 in connInfo["groupId"]]
        return self.tcpserver.getConnInfoListByGroupId(groupId)

    def soundInit(self):
        """音声初期化
        """
        self.sounds = {}
        with open(pathlib.Path("./sound.json").resolve()) as f:
            soundList = json.loads(f.read())
            for s in soundList:
                self.sounds[s["no"]] = []
                for _ in range(5):
                    sound = SoundLoader.load(s["path"])
                    self.sounds[s["no"]].append(sound)
                    # sound.bind(on_stop=lambda _: self.logger.debug("stop sound"))

    def playSound(self, soundNo):
        """音声鳴動
        
        Args:
            soundNo (string): 音番号
        """
        # self.logger.debug("playSound")

        def __proc_async():
            if soundNo in self.sounds:
                volume = float(OtherInfo().sound["volume"]) / 100

                notPlayingSounds = [s for s in self.sounds[soundNo]  if s.state != 'play']
                if len(notPlayingSounds) > 0:
                    sound = notPlayingSounds[0]

                else:
                    sound =self.sounds[soundNo][0]
                    sound.stop()

                sound.volume = volume
                # self.logger.debug("play sound")
                sound.play()

        th = threading.Thread(target=__proc_async)
        th.start()

    def releaseSound(self):
        """音声解放
        """
        for no in self.sounds:
            for sound in self.sounds[no]:
                if sound.state == 'play':
                    sound.stop()
                sound.unload()


if __name__ == "__main__":
    # テスト
    import kivy
    from kivy.app import App
    from kivy.uix.label import Label
    class TestApp(App):
        def build(self):
            self.receiver = Receiver(self.received, self.erased, self.synced)
            return Label(text='Debugging now...')

        def on_start(self):
            self.receiver.start()

        def stop(self):
            self.receiver.stop()

        def received(self, datalist):
            print('received:{}'.format( datalist))
        def erased(self, datalist):
            print('erased:{}'.format( datalist))

        def synced(self, callList, delList):
            print('erased:{}, {}'.format( callList, delList))

    app = TestApp()
    try:
        app.run()
    except KeyboardInterrupt:
        app.stop()
