import os
import re
import select
import struct
import time
from abc import ABC, ABCMeta, abstractmethod
from kivy.clock import Clock

import serial
from absthread import AbstractThread

from recv_dev.config import CallerSetting


class AbstractRecvDevice(metaclass=ABCMeta):
    """受信装置抽象クラス
    """
    @abstractmethod
    def start(self, recvCb):
        """開始
        
        Args:
            recvCb (function): 受信コールバック関数
        
        Raises:
            NotImplementedError: [description]
        """
        raise NotImplementedError()

    @abstractmethod
    def stop(self):
        """停止
        
        Raises:
            NotImplementedError: [description]
        """
        raise NotImplementedError()

class DummyRecvDevice(AbstractThread, AbstractRecvDevice):
    """ダミー受信装置クラス
    """
    def __init__(self):
        """コンストラクタ
        """
        super(DummyRecvDevice, self).__init__()
        self.recvCb = None
        self.data = []
        self.idx = 0

        if os.path.exists("/tmp/dummy_serial") == False:
            os.mkfifo("/tmp/dummy_serial")


    def pre_proc(self):
        """前処理
        """
        pass

    def start(self, recvCb):
        """開始
        
        Args:
            recvCb (function): 受信コールバック関数
        """        
        self.recvCb = recvCb
        super(DummyRecvDevice, self).start()

    def thread_proc(self):
        """スレッドループ処理
        
        Returns:
            [bool]: 処理結果：Falseを返すとスレッド終了
        """
        # if self.recvCb is not None:
        #     data = self.getData()
        selectlist = [
            os.open("/tmp/dummy_serial",  os.O_RDONLY | os.O_NONBLOCK),
        ]
        (rready, _, _) = select.select(selectlist, [], [], 0.1)
        for fd in rready:
            data = os.read(fd, 1024)
            data = data.decode().strip('\n')
            datalist = data.split(",")
            try:
                if len(datalist) >= 4:
                    self.recvCb(datalist[0], datalist[1], datalist[2], int(datalist[3]))
            except:
                pass

        for fd in selectlist:
            os.close(fd)

        return True

    def post_proc(self):
        """
        後処理
        """
        pass


class RecvDevice(AbstractThread, AbstractRecvDevice):
    """受信装置クラス
    """

    TTY="/dev/ttyS0"
    BAUDRATE=57600
    BITESIZE=serial.EIGHTBITS
    PARITY=serial.PARITY_NONE
    STOPBITS=serial.STOPBITS_ONE

    def __init__(self):
        """コンストラクタ
        """
        super(RecvDevice, self).__init__()
        self.recvCb = None
        self.serial = serial.Serial(port=self.TTY, baudrate=self.BAUDRATE, bytesize=self.BITESIZE, parity=self.PARITY,stopbits=self.STOPBITS, timeout=0.2)
        self.__modelAManage = {}
        self.__dupCheckManage = {}
        self.__recv_data = []

    def start(self, recvCb):
        """開始
        
        Args:
            recvCb (function): 受信コールバック関数
        """        
        self.recvCb = recvCb
        super(RecvDevice, self).start()

    def pre_proc(self):
        """前処理
        """
        pass


    def thread_proc(self):
        """スレッドループ処理
        
        Returns:
            [bool]: 処理結果：Falseを返すとスレッド終了
        """
        if self.recvCb is not None:
            line = self.serial.readline()
            # line = self.serial.read(21)
            # #test用 仮
            # if len(line) >= 21:
            #     lineba = bytearray(line)
            #     lineba[8] = 0x41 #'A'
            #     lineba[9] = 0x31 #'1'

            #     cs = self.__calc_checksum(lineba)
            #     lineba[15] = ord(cs[0])
            #     lineba[16] = ord(cs[1])
            #     lineba[17] = ord(cs[2])
            #     lineba[18] = ord(cs[3])
            #     line = lineba
            if len(line) > 0:
                self.logger.debug("raw data = {0}, len={1}".format(line.hex(), len(line)))
                nowTime = time.time()

                self.__push_recv_data(line, nowTime)

            while True:
                line, nowTime = self.__pop_recv_data()
                if line is None:
                    break

                self.logger.debug("{0}, {1}".format(nowTime, line))

                if len(line) > 0 and self.check_telegram(line):
                    (address, modelNo, buttonNo, strength, _,) = struct.unpack_from("@" "8s2s2s3s4s", line, 0)
                    address = address.decode("utf-8")
                    modelNo = modelNo.decode("utf-8")
                    buttonNo = buttonNo.decode("utf-8")
                    strength = int(strength.decode("utf-8"))

                    # 重複チェック
                    if self.check_duplication(nowTime, address, modelNo, buttonNo, strength):

                        # 受信データ処理
                        self.handle_recvdata(nowTime, address, modelNo, buttonNo, strength)

        return True

    def __push_recv_data(self, data, recvTime):
        """受信データをキューに格納
        """
        if len(data) > 0:
            self.__recv_data.append({
                "recvtime": recvTime,
                "data": data
            })

    def __pop_recv_data(self):
        """受信データをキューから取り出す
        """
        recvTime = None
        recvData = None
        index = -1
        for i, data in  enumerate(self.__recv_data):
            # self.logger.debug("{0}, {1}".format(i, data))
            if recvData is None:
                recvData = data["data"]
            else:
                recvData = recvData + data["data"]

            self.logger.debug(recvData)
            if len(recvData) >= 2 and recvData[-2:] == b'\r\n':
                index = i
                recvTime = data["recvtime"]
                break

        if index >= 0:
            del self.__recv_data[0: index + 1]
            return recvData, recvTime
        else:
            return None, None

    def post_proc(self):
        """後処理
        """
        self.serial.close()

    def check_telegram(self, line):
        """電文チェック

        Args:
            line (bytes): 受信データ

        Returns:
            bool: チェック結果
        """
        ret = False
        if len(line) == 21:
            try:
                (_, modelNo, buttonNo, strength, checksum,) = struct.unpack_from("@" "8s2s2s3s4s", line, 0)
                if self.verify_checksum(line, checksum):

                    modelNo = modelNo.decode("utf-8")
                    buttonNo = buttonNo.decode("utf-8")
                    strength = strength.decode("utf-8")
                    # 電文チェック（共通）
                    if self.__check_telegram_common(line, modelNo, buttonNo, strength):
                        if modelNo[0] == 'A':
                            # 電文チェック（モデルA）
                            ret = self.__check_telegram_modelA(line, modelNo, buttonNo, strength)
                        else:
                            # 電文チェック（モデルB）
                            ret = self.__check_telegram_modelB(line, modelNo, buttonNo, strength)
                    else:
                        ret = False

                    if ret == True:                        
                        self.logger.info("recv data = {}".format(line.hex()))
                else:
                    self.logger.error("recv error (checksum error). data = {}".format(line.hex()))
            except:
                self.logger.error("recv error. data = {}".format(line.hex()), exc_info=True)
        elif len(line) > 0:
            self.logger.error("recv error (length error). data = {}".format(line.hex()))

        return ret

    def __check_telegram_common(self, line, modelNo, buttonNo, strength):
        """電文チェック（ボタン共通）
        
        Args:
            line (bytes): 受信データ
            modelNo (string): モデル番号
            buttonNo (string): ボタン番号
            strength (string): 電波強度

        Returns:
            bool: チェック結果
        """
        ret = False
        if modelNo[0] != 'A' and modelNo[0] != 'B':
            self.logger.error("recv error (Illegal modelNo). data = {0}, modelNo={1}".format(line.hex(), modelNo))
        elif not buttonNo.isdecimal():
            self.logger.error("recv error (Illegal buttonNo). data = {0}, buttonNo={1}".format(line.hex(), buttonNo))
        elif not strength.isdecimal():
            self.logger.error("recv error (Illegal strength). data = {0}, strength={1}".format(line.hex(), strength))
        else:
            ret = True
        return ret

    def __check_telegram_modelA(self, line, modelNo, buttonNo, strength):
        """電文チェック（モデルA）
        
        Args:
            line (bytes): 受信データ
            modelNo (string): モデル番号
            buttonNo (string): ボタン番号
            strength (string): 電波強度

        Returns:
            bool: チェック結果
        """
        ret = False
        if modelNo[1] != '1':
            self.logger.error("recv error (Illegal modelNo). data = {0}, modelNo={1}".format(line.hex(), modelNo))
        elif buttonNo != '00' and buttonNo != '20':
            self.logger.error("recv error (Illegal buttonNo). data = {0}, modelNo={1}, buttonNo={2}".format(line.hex(), modelNo, buttonNo))
        else:
            ret = True
        return ret

    def __check_telegram_modelB(self, line, modelNo, buttonNo, strength):
        """電文チェック（モデルB）
        
        Args:
            line (bytes): 受信データ
            modelNo (string): モデル番号
            buttonNo (string): ボタン番号
            strength (string): 電波強度

        Returns:
            bool: チェック結果
        """
        ret = False
        m = re.fullmatch(r"[1-9A-Za-z]", modelNo[1:2])
        if m is None:
            self.logger.error("recv error (Illegal modelNo). data = {0}, modelNo={1}".format(line.hex(), modelNo))
            return ret

        buttonNum = int(modelNo[1:2]) if modelNo[1:2].isdecimal() else ord(modelNo[1:2].upper()) - ord('A') + 10

        if int(buttonNo) < 1 :
            self.logger.error("recv error (Illegal buttonNo). data = {0}, modelNo={1}, buttonNo={2}".format(line.hex(), modelNo, buttonNo))
        elif buttonNum < int(buttonNo):
            self.logger.error("recv error (buttonNo is too big). data = {0}, modelNo={1}, buttonNo={2}".format(line.hex(), modelNo, buttonNo))
        else:
            ret = True
        return ret

    def verify_checksum(self, data, checksum):
        """チェックサム検証
        
        Args:
            data (bytes): 受信データ
            checksum (bytes): チェックサム（4バイト）
        """
        try:
            checksum = checksum.decode("utf-8")
            return checksum.upper() ==  self.__calc_checksum(data).upper()
        except:
            self.logger.error("verify_checksum failed.", exc_info=True)
            return False

    def __calc_checksum(self, data):
        """チェックサム計算
        
        Args:
            data (bytes): 受信データ
        
        Returns:
            string: チェックサム
        """
        sum = 0
        for d in data[:-6]:
            sum += int(d)

        return "{:04x}".format(sum)

    def check_duplication(self, nowTime, address, modelNo, buttonNo, strength):
        """重複受信チェック
        
        Args:
            nowTime (float): データ受信時のEPOCHTime（秒）
            address (string): アドレス
            modelNo (string): モデル番号
            buttonNo (string): ボタン番号
            strength (string): 電波強度
        """
        ret = False
        commonSetting = CallerSetting().common
        dupCheckIntervalMs = commonSetting["dupCheckIntervalMs"]
        mapkey = "{0}_{1}_{2}".format(address, modelNo, buttonNo)
        if mapkey in self.__dupCheckManage:
            prevTime = self.__dupCheckManage[mapkey]["time"]
            interval = nowTime - prevTime 
            if interval <= float(dupCheckIntervalMs)/1000:
                # 重複
                ret = False
            else:
                ret = True
        else:
            ret = True

        # [S2]へ
        self.__dupCheckManage[mapkey] = {
            "time": nowTime,
        }
        return ret


    def handle_recvdata(self, nowTime, address, modelNo, buttonNo, strength):
        """受信データ処理
        
        Args:
            nowTime (float): データ受信時のEPOCHTime（秒）
            address (string): アドレス
            modelNo (string): モデル番号
            buttonNo (string): ボタン番号
            strength (string): 電波強度
        """
        if modelNo[0] == "A":
            # モデルA
            self.__handle_recvdata_modelA(nowTime, address, modelNo, buttonNo, strength)

        else:
            # その他モデル
            self.__handle_recvdata_modelOther(nowTime, address, modelNo, buttonNo, strength)

    def __handle_recvdata_modelA(self, nowTime, address, modelNo, buttonNo, strength):
        """受信データ処理（モデルA）
        
        Args:
            nowTime (float): データ受信時のEPOCHTime（秒）
            address (string): アドレス
            modelNo (string): モデル番号
            buttonNo (string): ボタン番号
            strength (string): 電波強度
        """


        modelASetting = CallerSetting().modelA
        timeout = modelASetting["timeout"]
        longPressTime = modelASetting["longPressTime"]

        # print(self.__modelAManage)

        mapkey = "{0}_{1}".format(address, modelNo)
        if longPressTime == 0:  #長押し消去無効
            #ボタンの戻りを無視する時間(2秒固定)
            ignoreTime = 2
            
            #押したとき
            if buttonNo == "20":
                self.recvCb(address, modelNo, '01', strength, False)

                self.__modelAManage[mapkey] = {
                    "time": nowTime,
                }
            #離したとき
            else:
                if mapkey in self.__modelAManage:
                    prevTime = self.__modelAManage[mapkey]["time"]
                    interval = nowTime - prevTime 
                    if interval > ignoreTime:
                        #2秒以上の長押しは呼出扱い（2秒以内は二重に呼び出されるのを防止するた目に無視する。）
                        self.recvCb(address, modelNo, '01', strength, False)
                else:
                    # 「押し」を検出できなかった場合も呼び出し扱い
                    self.recvCb(address, modelNo, '01', strength, False)

                self.__modelAManage[mapkey] = {
                    "time":0,
                }

        else:   #長押し消去有効

            # mapkey = "{0}_{1}".format(address, modelNo)
            #[E1] 押したとき 
            if buttonNo == "20":
                if mapkey in self.__modelAManage:
                    prevTime = self.__modelAManage[mapkey]["time"]
                    interval = nowTime - prevTime 
                    if prevTime == 0:
                        # [S1]初期状態
                        pass
                    elif interval <= timeout:
                        # [S2]消去判定タイムアウト待ち
                        self.logger.warn("E1(ModelA Push Event) happened on Status S2.")
                    else:
                        # [S3]10秒タイムアウト待ち
                        self.logger.warn("E1(ModelA Push Event) happened on Status S3.")

                else:
                    # [S1]初期状態
                    pass

                
                # [S2]へ
                self.__modelAManage[mapkey] = {
                    "time": nowTime,
                }


            #離したとき
            else:
                if mapkey in self.__modelAManage:
                    prevTime = self.__modelAManage[mapkey]["time"]
                    interval = nowTime - prevTime 
                    if prevTime == 0:
                        # [S1]初期状態
                        #「押した」が届かず（空振り）した場合   
                        # 「離した」で 呼出とする。
                        self.recvCb(address, modelNo, '01', strength, False)
                        self.logger.warn("E2(ModelA Release Event) happened on Status S1.")
                    elif interval <= timeout:
                        # [S2]消去判定タイムアウト待ち
                        # [S3]10秒タイムアウト待ち

                        #・[S2]「押した」から長押し秒数以内に「離した」が届いた：呼出と判定　
                        #・[S3]「押した」から長押し秒数以上、かつ１０秒以内に「離した」が届いた：消去
                        
                        # 長押しか？
                        longPress = interval > longPressTime
                        # ボタンが押されたことを通知する。
                        self.recvCb(address, modelNo, '01', strength, longPress)
                    else:
                        # timeout
                        # [S1]初期状態として扱う
                        self.recvCb(address, modelNo, '01', strength, False)
                        self.logger.warn("E2(ModelA Release Event) happened on Status S1.")

                    self.__modelAManage[mapkey] = {
                        "time":0,
                    }

                else:
                    # [S1]初期状態
                    #「押した」が届かず（空振り）した場合に
                    # 「離した」で 呼出とする。
                    self.recvCb(address, modelNo, '01', strength, False)
                    self.logger.warn("E2(ModelA Release Event) happened on Status S1.")
                    self.__modelAManage[mapkey] = {
                        "time":0,
                    }

    def __handle_recvdata_modelOther(self, nowTime, address, modelNo, buttonNo, strength):
        """受信データ処理（その他モデル）
        
        Args:
            nowTime (float): データ受信時のEPOCHTime（秒）
            address (string): アドレス
            modelNo (string): モデル番号
            buttonNo (string): ボタン番号
            strength (string): 電波強度
        """
        # ボタンが押されたことを通知する。
        self.recvCb(address, modelNo, buttonNo, strength)

if __name__ == "__main__":
    # テスト
    # dev = DummyRecvDevice()
    import time
    dev = RecvDevice()
    def callback(address, modelNo, buttonNo, strength):
        print('address={0}, modelNo={1}, buttonNo={2}'.format(address, modelNo, buttonNo))

    try:
        dev.start(callback)
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        dev.stop()
