import hashlib
import pathlib
import sqlite3
import uuid
from contextlib import closing

from setup_logger import setup_logger_devel

DBNAME=pathlib.Path('data.db').resolve()

class Database():
    """データベース
    """

    def __init__(self):
        """コンストラクタ
        """
        self.logger = setup_logger_devel(self.__class__.__name__)

    def install(self):
        """インストール（テーブル生成）
        """
        conn = None
        try:
            with closing(sqlite3.connect(DBNAME)) as conn:
                createTables = [
                    Caller.TABLE_DEF,
                    CallerButton.TABLE_DEF,
                    Eraser.TABLE_DEF,
                    Group.TABLE_DEF,
                    Display.TABLE_DEF,
                    CallList.TABLE_DEF,
                    DelList.TABLE_DEF,
                    WebUser.TABLE_DEF,
                    IntervalJob.TABLE_DEF,
                    CallLog.TABLE_DEF
                ]

                createViews = [
                    CallerButtonView.TABLE_DEF,
                    CallListView.TABLE_DEF,
                    DelListView.TABLE_DEF
                ]

                c = conn.cursor()
                for createTable in createTables:
                    c.execute(createTable)

                for createView in createViews:
                    c.execute(createView)

                conn.commit()
        except:
            self.logger.error("install failed.", exc_info=True)

    def upgrade(self):
        upgradeTables = [
            Caller,
            CallList,
            DelList,
            CallListView,
            DelListView,
            CallerButton,
            CallerButtonView,
        ]
        
        for T in upgradeTables:
            t = T()
            if hasattr(t, "upgrade"):
                t.upgrade()


    def setup(self):
        """セットアップ（初期設定）
        """
        setupTables = [
            WebUser,
            IntervalJob
        ]
        for T in setupTables:
            t = T()
            if hasattr(t, "setup"):
                t.setup()


    def uninstall(self):
        """アンインストール（テーブル削除）
        """
        try:
            with closing(sqlite3.connect(DBNAME)) as conn:
                tableNames = [
                    Caller.TABLE_NAME,
                    CallerButton.TABLE_NAME,
                    Eraser.TABLE_NAME,
                    Group.TABLE_NAME,
                    Display.TABLE_NAME,
                    CallList.TABLE_NAME,
                    DelList.TABLE_NAME,
                    WebUser.TABLE_NAME,
                    IntervalJob.TABLE_NAME,
                    CallLog.TABLE_NAME
                ]

                viewNames = [
                    CallerButtonView.TABLE_NAME,
                    CallListView.TABLE_NAME,
                    DelListView.TABLE_NAME
                ]

                c = conn.cursor()
                for tableName in tableNames:
                    c.execute('DROP TABLE IF EXISTS ' + tableName)

                for viewName in viewNames:
                    c.execute('DROP VIEW IF EXISTS ' + viewName)

                conn.commit()
        except:
            self.logger.error("uninstall failed.", exc_info=True)


class BaseDataAccess():
    """データアクセス基底クラス
    """
    TABLE_NAME="xxxxxx"
    TABLE_DEF = ""

    def __init__(self):
        """コンストラクタ
        """
        self.logger = setup_logger_devel(self.__class__.__name__)

    def insert(self, data):
        """登録
        
        Args:
            data (dictionary): 登録データ
                    登録データの形式
                        =  { "address": addressValue,
                            "modelNo": modelNoValue,
                            "buttonNo": buttonNoValue
                            }
        """
        try:
            with closing(sqlite3.connect(DBNAME)) as conn:
                c = conn.cursor()
                insert = self.__create_insert(data)
                c.execute(insert, data)

                conn.commit()
        except:
            self.logger.error("insert failed.", exc_info=True)

    def insertmany(self, datalist):
        """登録
        
        Args:
            datalist (list): 登録データのリスト

            登録データのリストの形式
                =  [{ "address": addressValue,
                    "modelNo": modelNoValue,
                    "buttonNo": buttonNoValue
                    },
                    ...]
        """
        try:
            with closing(sqlite3.connect(DBNAME)) as conn:
                c = conn.cursor()
                for data in datalist:
                    insert = self.__create_insert(data)
                    c.execute(insert, data)

                conn.commit()
        except:
            self.logger.error("insertmany failed.", exc_info=True)

    def delete(self, data):
        """削除
        
        Args:
            data (dictionary): 削除データ

            削除データの形式
                =   { "address": addressValue,
                    "modelNo": modelNoValue,
                    "buttonNo": buttonNoValue
                    }
        """
        try:
            with closing(sqlite3.connect(DBNAME)) as conn:
                c = conn.cursor()
                delete = self.__create_delete(data)
                c.execute(delete, data)
                conn.commit()
        except:
            self.logger.error("delete failed.", exc_info=True)

    def deletemany(self, datalist):
        """削除
        
        Args:
            datalist (list): 削除データのリスト

            削除データのリストの形式
                =   [{ "address": addressValue,
                    "modelNo": modelNoValue,
                    "buttonNo": buttonNoValue
                    },
                    ...]
        """
        try:
            with closing(sqlite3.connect(DBNAME)) as conn:
                c = conn.cursor()
                for data in datalist:
                    delete = self.__create_delete(data)
                    c.execute(delete, data)

                conn.commit()
        except:
            self.logger.error("deletemany failed.", exc_info=True)

    def select(self, condVal, orderBy = {}, greaterThanUpdDt = None):
        """select
        
        Args:
            condVal (dictionary): 条件

            条件の形式
                = { "address": addressValue,
                    "modelNo": modelNoValue,
                    "buttonNo": buttonNoValue
                    ...
                    }
            orderBy (dictionary, optional): ソート順. Defaults to {}.

            ソート順の形式
                = { "address": "DESC",
                    ...
                    }

            greaterThanUpdDt (datetime), optional): ＞＝更新日付. Defaults to None.
        
        Returns:
            list: selectの結果
        """
        result = []
        try:
            with closing(sqlite3.connect(DBNAME)) as conn:
                conn.row_factory = sqlite3.Row
                c = conn.cursor()
                select = self.__create_select(condVal, orderBy, greaterThanUpdDt)
                c.execute(select, condVal)
                for r in c:
                    val = {}
                    for column in r.keys():
                        val[column] = r[column]

                    result.append(val)

        except:
            self.logger.error("select failed.", exc_info=True)

        return result

    def update(self, key, value):
        """更新
        
        Args:
            key (dict): キー

                キーの形式
                    = { "address": addressValue,
                        "modelNo": modelNoValue,
                        "buttonNo": buttonNoValue
                        }
            value (dict): 値

                値の形式
                    = { "function": 1,
                        "name1": name1Value,
                        "name2": name2Value
                        }
        """

        try:
            with closing(sqlite3.connect(DBNAME)) as conn:
                c = conn.cursor()
                update = self.__create_update(key, value)
                c.execute(update[0], update[1])
                conn.commit()
        except:
            self.logger.error("update failed.", exc_info=True)

    def updatemany(self, key_value_list):
        """更新
        
        Args:
            key_value_list (list): キーと値のタプルのリスト

            キーと値のタプルのリストの形式 = [
                (キー, 値),
                (キー, 値),
                ...
            ]

            キーの形式
                = { "address": addressValue,
                    "modelNo": modelNoValue,
                    "buttonNo": buttonNoValue
                    }

            値の形式
                = { "function": 1,
                    "name1": name1Value,
                    "name2": name2Value
                    }
        """
        try:
            with closing(sqlite3.connect(DBNAME)) as conn:
                c = conn.cursor()
                for (key, value) in key_value_list:
                    update = self.__create_update(key, value)
                    c.execute(update[0], update[1])

                conn.commit()
        except:
            self.logger.error("updatemany failed.", exc_info=True)

    def __create_insert(self, insVal):
        """insert句生成
        
        Args:
            insVal (dict): 値
        
        Returns:
            string: insert句
        """
        return 'INSERT INTO '+ self.TABLE_NAME + self.__create_values(insVal)

    def __create_values(self, insVal):
        """value句生成
        
        Args:
            insVal (dict): 値
        
        Returns:
            string: value句
        """

        values = ""
        clm = []
        vls = []
        for k in insVal:
            clm.append(k)
            vls.append(":" + k)
        if len(clm) > 0:
            values = " (" +", ".join(clm) + ") values (" + ", ".join(vls) + ") "
        return values

    def __create_delete(self, delData):
        """delete句生成
        
        Args:
            delData (dict): 値
        
        Returns:
            string: delete句
        """
        return 'DELETE FROM '+ self.TABLE_NAME + self.__create_where(delData)

    def __create_select(self, condVal, orderBy, greaterThanUpdDt):
        """select句生成
        
        Args:
            condVal (dict)): 条件
            orderBy (dict): ソート順
            greaterThanUpdDt (datetime): ＞＝更新日付
        
        Returns:
            string : select句
        """
        obs = ""
        obl = []
        for k in orderBy:
            obl.append(k + " " + orderBy[k])

        if len(obl) > 0:
            obs = " ORDER BY " + ", ".join(obl)
        where = self.__create_where(condVal)

        if greaterThanUpdDt is not None:
            dt = "{0:%Y-%m-%d %H:%M:%S.%f}".format(greaterThanUpdDt)
            if len(where) > 0:
                where = where + " AND upd_dt > '{}' ".format(dt)
            else:
                where = " WHERE upd_dt > '{}' ".format(dt)

        return 'SELECT * FROM '+ self.TABLE_NAME + where + obs

    def __create_where(self, condVal):
        """where句生成
        
        Args:
            condVal (dict): 条件
        
        Returns:
            string: where句
        """
        where = ""
        whl = []
        for k in condVal:
            whl.append(k + "=:" + k)
        if len(whl) > 0:
            where = " WHERE " + " AND ".join(whl) + " "

        return where

    def _create_where(self, condVal):
        """where句生成(public)
        
        Args:
            condVal (dict): 条件
        
        Returns:
            string: where句
        """
        return self.__create_where(condVal)

    def __create_update(self, key, values):
        """update句生成
        
        Args:
            key (dict): キー
            values (dict): 値
        
        Returns:
            string: update句
        """
        stl = []
        for k in values:
            stl.append(k + "=:___upd___" + k)
        setStr = " SET " + ", ".join(stl) + " "
        where = self.__create_where(key)
        params = {}
        for k in values:
            params["___upd___" + k] = values[k]
        for k in key:
            params[k] = key[k]

        return ('UPDATE '+ self.TABLE_NAME + setStr + where, params)

    def getColumnInfo(self):
        """カラム情報取得

        Returns:
            list: カラム情報
            [(cid, name, type, notnull, default value, pk), ...]
            ex: [(0, 'address', 'text', 0, None, 1), (1, 'modelNo', 'text', 0, None, 0), (2, 'line1', 'text', 0, "''", 0), (3, 'line2', 'text', 0, "''", 0)]
        """
        cols = []
        try:
            with closing(sqlite3.connect(DBNAME)) as conn:
                c = conn.cursor()
                c.execute("PRAGMA TABLE_INFO(" + self.TABLE_NAME + ")")
                cols = c.fetchall()
        except:
            self.logger.error("getColumnInfo failed.", exc_info=True)
        return cols

    def hasColumn(self, columnName):
        """カラムの存在確認

        Args:
            columnName (string): カラム名

        Returns:
            bool: 存在する=True
        """
        cols = self.getColumnInfo()
        return columnName in [items[1] for items in cols]

class Caller(BaseDataAccess):
    """呼出機マスタ
    """
    TABLE_NAME="m_caller"
    TABLE_DEF = "CREATE TABLE IF NOT EXISTS " + TABLE_NAME + '''\
            (
                address    text,
                modelNo     text CHECK(length(modelNo) = 2),
                line1       text    DEFAULT '',
                line2       text    DEFAULT '',
                ins_dt      text,
                PRIMARY KEY (address)
            )
            '''
    def upgrade(self):
        if self.hasColumn('ins_dt') == False:
            ADD_COLUMN="ALTER TABLE " + self.TABLE_NAME + " ADD COLUMN ins_dt text"
            try:
                with closing(sqlite3.connect(DBNAME)) as conn:
                    c = conn.cursor()
                    c.execute(ADD_COLUMN)

                    conn.commit()
            except:
                self.logger.error("upgrade failed.", exc_info=True)

class CallerButton(BaseDataAccess):
    """呼出機ボタンマスタ
    """
    TABLE_NAME="m_caller_button"
    TABLE_DEF = "CREATE TABLE IF NOT EXISTS " + TABLE_NAME + '''\
            (
                address     text,
                buttonNo    text,
                call        integer DEFAULT 0,
                color       text    DEFAULT 'R',
                sound       text    DEFAULT '01',
                priority    integer DEFAULT 0,
                line1       text    DEFAULT '',
                line2       text    DEFAULT '',
                PRIMARY KEY (address, buttonNo)
            )
            '''
    def upgrade(self):
        if self.hasColumn('line1') == False:
            ADD_COLUMN_1="ALTER TABLE " + self.TABLE_NAME + " ADD COLUMN  line1       text    DEFAULT '' "
            ADD_COLUMN_2="ALTER TABLE " + self.TABLE_NAME + " ADD COLUMN  line2       text    DEFAULT '' "
            try:
                with closing(sqlite3.connect(DBNAME)) as conn:
                    c = conn.cursor()
                    c.execute(ADD_COLUMN_1)
                    c.execute(ADD_COLUMN_2)
                    c.execute('DROP VIEW IF EXISTS ' + CallerButtonView.TABLE_NAME)

                    updatesql = ''' UPDATE {B}
                    SET line1 = (SELECT {C}.line1 FROM {C} WHERE {B}.address = {C}.address ) 
                        ,line2 = (SELECT {C}.line2 FROM {C} WHERE {B}.address = {C}.address )
                    '''.format(B=self.TABLE_NAME, C=Caller.TABLE_NAME)

                    c.execute(updatesql)
                    conn.commit()
            except:
                self.logger.error("upgrade failed.", exc_info=True)

class CallerButtonView(BaseDataAccess):
    """呼出機ボタンビュー
    """
    TABLE_NAME="v_caller_button"
    TABLE_DEF = "CREATE VIEW IF NOT EXISTS " + TABLE_NAME + " AS " +'''\
    SELECT 
        C.address,
        C.modelNo,
        B.line1,
        B.line2,
        C.ins_dt,
        B.buttonNo,
        B.call,
        B.color,
        B.sound,
        B.priority
    FROM m_caller C, m_caller_button B
    WHERE
        C.address = B.address
    '''
    def upgrade(self):
        try:
            with closing(sqlite3.connect(DBNAME)) as conn:
                c = conn.cursor()
                c.execute(self.TABLE_DEF)

                conn.commit()
        except:
            self.logger.error("upgrade failed.", exc_info=True)


class Eraser(BaseDataAccess):
    """消去機マスタ
    """
    TABLE_NAME="m_eraser"
    TABLE_DEF = "CREATE TABLE IF NOT EXISTS " + TABLE_NAME + '''\
            (
                address    text,
                modelNo     text CHECK(length(modelNo) = 2),
                name       text    DEFAULT '',
                ins_dt      text,
                PRIMARY KEY (address)
            )
            '''


class Group(BaseDataAccess):
    """グループマスタ
    """
    TABLE_NAME="m_group"
    TABLE_DEF = "CREATE TABLE IF NOT EXISTS " + TABLE_NAME + '''\
    (
        groupId integer PRIMARY KEY,
        name    text    DEFAULT ''
    )
    '''

    def nextGroupId(self):
        """次のGroupIdを取得する。（自動採番用）
        
        Returns:
            int: 次のgroupId
        """
        next = 1
        try:
            with closing(sqlite3.connect(DBNAME)) as conn:
                conn.row_factory = sqlite3.Row
                c = conn.cursor()
                c.execute('SELECT max(groupId) as max FROM '+ self.TABLE_NAME)
                val = {}
                for r in c:
                    for column in r.keys():
                        val[column] = r[column]

                if "max" in val:
                    next = val["max"] + 1
        except:
            self.logger.error("nextGroupId failed.", exc_info=True)

        return next


class Display(BaseDataAccess):
    """表示機リスト
    """
    TABLE_NAME="t_display"
    TABLE_DEF = "CREATE TABLE IF NOT EXISTS " + TABLE_NAME + '''\
    (
        ipAddress   text PRIMARY KEY,
        datetime    text,
        name        text,
        groups      text
    )
    '''

class CallList(BaseDataAccess):
    """呼び出しリスト
    """
    TABLE_NAME="t_call_list"
    TABLE_DEF = "CREATE TABLE IF NOT EXISTS " + TABLE_NAME + '''\
    (
        address    text,
        modelNo     text,
        buttonNo    text,
        groupId     integer,
        recall      integer DEFAULT 0,
        strength     integer DEFAULT 0,
        ins_dt      text,
        upd_dt      text,
        displayed   integer DEFAULT 0
    )
    '''

    def upgrade(self):
        if self.hasColumn('displayed') == False:
            ADD_COLUMN="ALTER TABLE " + self.TABLE_NAME + " ADD COLUMN displayed integer DEFAULT 0"
            try:
                with closing(sqlite3.connect(DBNAME)) as conn:
                    c = conn.cursor()
                    c.execute(ADD_COLUMN)

                    conn.commit()
            except:
                self.logger.error("upgrade failed.", exc_info=True)

class DelList(BaseDataAccess):
    """取消リスト
    """
    TABLE_NAME="t_del_list"
    TABLE_DEF = "CREATE TABLE IF NOT EXISTS " + TABLE_NAME + '''\
    (
        address    text,
        modelNo     text,
        buttonNo    text,
        groupId     integer,
        displayName   text,
        recall      integer DEFAULT 0,
        strength     integer DEFAULT 0,
        ins_dt      text,
        upd_dt      text,
        displayed   integer DEFAULT 0
    )
    '''

    def upgrade(self):
        if self.hasColumn('displayed') == False:
            ADD_COLUMN="ALTER TABLE " + self.TABLE_NAME + " ADD COLUMN displayed integer DEFAULT 0"
            try:
                with closing(sqlite3.connect(DBNAME)) as conn:
                    c = conn.cursor()
                    c.execute(ADD_COLUMN)

                    conn.commit()
            except:
                self.logger.error("upgrade failed.", exc_info=True)


class CallListView(BaseDataAccess):
    """呼び出しリストビュー
    """
    TABLE_NAME="v_call_list"
    TABLE_DEF = "CREATE VIEW IF NOT EXISTS " + TABLE_NAME + " AS " +'''\
    SELECT
        CL.*,
        ifnull(C.line1, '') as line1,
        ifnull(C.line2, '') as line2,
        ifnull(C.color, 'R') as color,
        ifnull(C.sound, '01') as sound,
        ifnull(C.priority, '0') as priority,
        ifnull(G.name, '') as groupName
    FROM t_call_list CL
    LEFT OUTER JOIN v_caller_button C
        ON CL.address = C.address 
        AND CL.buttonNo = C.buttonNo
    LEFT OUTER JOIN m_group G
        ON CL.groupId = G.groupId 
    '''

    def upgrade(self):
        try:
            with closing(sqlite3.connect(DBNAME)) as conn:
                c = conn.cursor()
                c.execute('DROP VIEW IF EXISTS ' + self.TABLE_NAME)
                c.execute(self.TABLE_DEF)

                conn.commit()
        except:
            self.logger.error("upgrade failed.", exc_info=True)

class DelListView(BaseDataAccess):
    """取消リストビュー
    """

    TABLE_NAME="v_del_list"
    TABLE_DEF = "CREATE VIEW IF NOT EXISTS " + TABLE_NAME + " AS " +'''\
    SELECT
        DL.*,
        ifnull(C.line1, '') as line1,
        ifnull(C.line2, '') as line2,
        ifnull(C.color, 'R') as color,
        ifnull(C.sound, '01') as sound,
        ifnull(C.priority, '0') as priority,
        ifnull(G.name, '') as groupName
    FROM t_del_list DL
    LEFT OUTER JOIN v_caller_button C
        ON DL.address = C.address 
        AND DL.buttonNo = C.buttonNo
    LEFT OUTER JOIN m_group G
        ON DL.groupId = G.groupId 
    '''

    def upgrade(self):
        try:
            with closing(sqlite3.connect(DBNAME)) as conn:
                c = conn.cursor()
                c.execute('DROP VIEW IF EXISTS ' + self.TABLE_NAME)
                c.execute(self.TABLE_DEF)

                conn.commit()
        except:
            self.logger.error("upgrade failed.", exc_info=True)


class WebUser(BaseDataAccess):
    """WEBユーザーマスタ
    """

    TABLE_NAME="m_web_user"
    TABLE_DEF = "CREATE TABLE IF NOT EXISTS " + TABLE_NAME + '''\
    (
        userId  text PRIMARY KEY,
        hash    text,
        salt    text,
        privilege   integer
    )
    '''
    SECRET = "ecoeco"

    def setup(self):
        """セットアップ
        """

        users = [ ("admin", "admin", 2), ("user", "", 1),("special", uuid.uuid4().hex, 9)]
        insertData = []
        for user in users:
            d = self.select({"userId": user[0]})
            if len(d) > 0:
                continue

            hash = self.create_hash_salt(user[1])
            data = {
                "userId":user[0],
                "hash":hash[0],
                "salt":hash[1],
                "privilege":user[2]
            }

            insertData.append(data)

        if len(insertData) > 0:
            self.insertmany(insertData)

    def create_hash_salt(self, passwd):
        """ハッシュ、ソルト生成
        
        Args:
            passwd (string): パスワード
        
        Returns:
            bytes, bytes: ハッシュ、ソルト
        """
        salt = uuid.uuid4().hex
        
        hash = hashlib.sha256(passwd.encode("utf-8") + self.SECRET.encode("utf-8") +bytes.fromhex(salt)).hexdigest()

        return hash, salt

    def calc_hash(self, passwd, salt):
        """ハッシュ算出
        
        Args:
            passwd (string): パスワード
            salt (bytes): ソルト
        
        Returns:
            bytes: ハッシュ
        """
        hash = hashlib.sha256(passwd.encode("utf-8") + self.SECRET.encode("utf-8") + bytes.fromhex(salt)).hexdigest()
        return hash

class IntervalJob(BaseDataAccess):
    """定期実行ジョブ管理
    """

    TABLE_NAME="t_job"
    TABLE_DEF = "CREATE TABLE IF NOT EXISTS " + TABLE_NAME + '''\
    (
        jobId       text PRIMARY KEY,
        datetime    text
    )
    '''
    # def upgrade(self):
    #     """アップグレード
    #     """
    #     self.update({"jobId":"ntp"}, {"jobId":"NtpJob"})

    def setup(self):
        """セットアップ
        """
        jobList = [ "NtpJob", "LogClearJob" ]
        insertData = []
        for job in jobList:
            d = self.select({"jobId": job})
            if len(d) > 0:
                continue

            data = {
                "jobId":job
            }

            insertData.append(data)

        if len(insertData) > 0:
            self.insertmany(insertData)

class CallLog(BaseDataAccess):
    """呼び出しログ
    """

    TABLE_NAME="t_call_log"
    TABLE_DEF = "CREATE TABLE IF NOT EXISTS " + TABLE_NAME + '''\
    (
        datetime    text,
        category    text,
        message     text,
        sent       integer
    )
    '''
    
    CATEGORY_CALL="call"
    CATEGORY_PAIRING="pairing mode"
    CATEGORY_RESPONSE="response"

    def deleteOld(self, dt, cond):
        """古いデータを削除
        
        Args:
            dt (datetime) :　削除対象日時
            cond (dictionary): 削除条件

            削除条件の形式
                =   { "sent": 1}
        """
        try:
            with closing(sqlite3.connect(DBNAME)) as conn:
                c = conn.cursor()
                delete = self.__create_delete_old(dt, cond)
                c.execute(delete, cond)
                conn.commit()
        except:
            self.logger.error("deleteOld failed.", exc_info=True)

    def __create_delete_old(self, dt, cond):
        """delete句生成
        
        Args:
            dt (datetime) :　削除対象日時
            cond (dictionary): 削除条件
        
        Returns:
            string : delete句
        """
        where = self._create_where(cond)

        if dt is not None:
            dt = "{0:%Y-%m-%d %H:%M:%S.%f}".format(dt)
            if len(where) > 0:
                where = where + " AND datetime < '{}' ".format(dt)
            else:
                where = " WHERE datetime < '{}' ".format(dt)

        return 'DELETE FROM '+ self.TABLE_NAME + where 


if __name__ == "__main__":
    # テスト
    d = Database()
    d.uninstall()
    d.install()
    c = Caller()
    c.insert({'address': '0123456701234567', 'modelNo': "A3"})
    c.insert({'address': '0123456701234599', 'modelNo': "A2"})
    print(c.select({}))

    cb = CallerButton()
    cb.insert({'address': '0123456701234567', 'buttonNo': "01"})
    cb.insert({'address': '0123456701234567', 'buttonNo': "02"})

    cbv = CallerButtonView()
    print(cbv.select({}))

    cl = CallList()
    cl.insert({'address': '0123456701234567', 'modelNo': "A3", 'buttonNo': "01", "groupId": 0, "recall": 0, "strength": 10})
    print(cl.select({}))
    clv = CallListView()
    print(clv.select({}))

    # c.update({'address': '0123456701234567', 'modelNo': "01", 'buttonNo': "01"}, {'function': 1})
    # c.update({'address': '0123456701234567', 'modelNo': "01", 'buttonNo': "02"}, {'function': 2})
    # print(c.select({}))
