import select
import socket
import threading
import time
from datetime import datetime
from enum import IntEnum

from absthread import AbstractThread
from recv_dev.config import TerminalInfo
from recv_dev.storage import (CallList, CallListView, DelList, DelListView,
                              Display, Group, CallLog)
from resources import Resources
from telegram import Telegram


class Status(IntEnum):
    S1 = 1
    S2 = 2
    S3 = 3

class TCPServerThread(AbstractThread):
    """TCPサーバスレッド
    """

    def __init__(self, address=("", 5001), backlog=5):
        """コンストラクタ
        
        Args:
            address (tuple, optional): 待ち受けアドレス、ポート番号. Defaults to ("", 5001).
            backlog (int, optional): 保留中の接続のキューの最大長. Defaults to 5.
        """
        super(TCPServerThread, self).__init__()
        self.address = address
        self.backlog = backlog
        self.sock = None
        self.poll = None
        self.conn_manage = {}
        self.conn_manage_lock = threading.Lock()
        self.display = Display()
        self.res = Resources()
        

    def pre_proc(self):
        """スレッドループ前処理
        """
        try:
            self.display.delete({})
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.sock.bind(self.address)
            self.sock.listen(self.backlog)
            self.logger.info("Server is listening at {}:{}".format(*self.address))
            self.sock.setblocking(False)


            # ポーリングオブジェクトの作成
            self.poll = select.poll()
            # 監視対象にサーバ自身のソケットオブジェクトを追加
            # 監視イベントは読み出し可能なデータの存在
            self.poll.register(self.sock, select.POLLIN) 
        except:
            self.logger.error("pre_proc failed.", exc_info=True)
            if self.sock is not None:
                self.sock.close()
                self.sock = None
            self.poll = None

    def thread_proc(self):
        """スレッドループ処理
        受信処理を行う。

        Returns:
            bool: 処理結果：Falseを返すとスレッド終了
        """
        try:
            if self.poll is None:
                return False

            # ポーリングを行った結果をもとに(fd, event)のタプルを返す
            ready = self.poll.poll()

            if not ready:
                return True

            for fd, event in ready:
                self.logger.debug("fd ={0}, fileno()={1}".format(fd, self.sock.fileno()))
                if not event & select.POLLIN: 
                    continue

                if fd == self.sock.fileno():
                    # サーバ自身のソケットオブジェクトに対し読み込みデータがある

                    # 新規コネクションを確立しようとしているので、コネクションをアクセプト
                    conn_sock, client_address = self.sock.accept()
                    self.logger.info("Accepted from {}".format(client_address))

                    self.set_socket_to_conn_manage(conn_sock, client_address)

                    # 新規ソケットを監視対象に追加
                    self.poll.register(conn_sock, select.POLLIN)

                    # 受信機情報通知
                    self.notify_sysinfo(conn_sock)

                else:
                    # 既存のコネクションソケットオブジェクトに対し読み込みデータがある

                    # クライアントからのパケットを受信
                    conn_sock = self.get_socket_from_conn_manage(fd)
                    try:
                        raw_data = conn_sock.recv(1024)
            
                        # データがなければクライアントとの接続は切れたということなので、監視対象から外してコネクションを閉じる。
                        if not raw_data:
                            self.logger.info("Closed connection to {}".format(conn_sock.getpeername()) )
                            self.disconnect_client(conn_sock)
                            # self.poll.unregister(conn_sock)
                            # self.remove_socket_from_conn_manage(conn_sock)
                            # conn_sock.close()
                            return True

                        self.logger.info("recv data = {0} from {1}".format(raw_data.hex(), conn_sock.getpeername()))
                        self.logger.info("status = {0}".format(self.getStatus(conn_sock)))

                        self.handle_telegram(conn_sock, raw_data)
                    except:
                        self.logger.error("thread_proc failed.", exc_info=True)
                        self.disconnect_client(conn_sock)
                        # self.poll.unregister(conn_sock)
                        # self.remove_socket_from_conn_manage(conn_sock)
                        # conn_sock.close()
                        return True


        except:
            self.logger.error("thread_proc failed.", exc_info=True)
            return False

    def disconnect_client(self, socket):
        """クライアントからの接続を切断
        
        Args:
            socket (object): ソケット
        """
        self.poll.unregister(socket)
        self.remove_socket_from_conn_manage(socket)
        socket.close()

    def post_proc(self):
        """スレッドループ後処理
        """
        if self.sock is not None:
            self.sock.close()
            self.sock = None
        self.poll = None
        self.conn_manage = {}
        self.display.delete({})

    def handle_telegram(self, socket, raw_data):
        """電文処理
        
        Args:
            socket (object): ソケット
            raw_data (bytes): 電文データ
        """
        pos = 0
        while True:
            idx = raw_data.find(b'\r\n', pos)
            if idx >= 0:
                _raw_data = raw_data[pos:idx+2]

                t = Telegram()
                if t.pre_check_telegram(_raw_data) == False:
                    continue

                recv_data = t.parse_recv_data(_raw_data)
                if recv_data["telgNo"] == Telegram.TELEG_HEALTH_CHECK:
                    self.healthCheck(socket, recv_data)
                elif recv_data["telgNo"] == Telegram.TELEG_REGISTER:
                    self.register(socket, recv_data)
                elif recv_data["telgNo"] == Telegram.TELEG_DEL_DATA:
                    self.eraseData(socket, recv_data)
                elif recv_data["telgNo"] == Telegram.TELEG_SYNC_DATA:
                    self.syncData(socket, recv_data)
                pos = idx + 2
            else:
                break

    def notify_sysinfo(self, socket):
        """受信機情報通知
        
        Args:
            socket (object): ソケット
        """
        termInfo = TerminalInfo()
        g = Group()
        groups = g.select({})
        logLevel = termInfo.logLevel
        t = Telegram()
        send_data = t.create_telegram_sysinfo_R(logLevel, groups, self.get_seqNo(socket))
        self.send_async(socket, send_data)

    def healthCheck(self, socket, data):
        """ヘルスチェック
        
        Args:
            socket (object): ソケット
            data (dict): ヘルスチェック電文データ
        """
        if self.getStatus(socket) == Status.S3:
            groups = data["groups"]
            self.update_groups_to_conn_manage(socket, groups)
            self.update_name_to_conn_manage(socket, data["displayName"])

            t = Telegram()
            send_data = t.create_telegram_healthcheck_R(self.get_seqNo(socket))
            self.send_async(socket, send_data)
        else:
            self.logger.error("status error. status = {}".format(self.getStatus(socket)))

    def register(self, socket, data):
        """登録
        
        Args:
            socket (object): ソケット
            data (dict): 登録データ
        """
        if self.getStatus(socket) == Status.S2:
            groups = data["groups"]
            name = data["displayName"]

            self.register_groups_to_conn_manage(socket, groups, name)

            t = Telegram()
            send_data = t.create_telegram_register_R(self.get_seqNo(socket))
            self.send_async(socket, send_data)
        else:
            self.logger.error("status error. status = {}".format(self.getStatus(socket)))

    def eraseData(self, socket, data):
        """取消
        
        Args:
            socket (object): ソケット
            data (dict): 取消データ
        """
        if self.getStatus(socket) == Status.S3:
            key = {
                "address" : data["address"],
                "modelNo" : data["modelNo"],
                "buttonNo" : data["buttonNo"]
            }

            cl = CallList()
            calllist = cl.select(key)
            # ログ用
            clv = CallListView()
            vcalllist = clv.select(key)
            now = datetime.now()
            cl.delete(key)

            dl = DelList()
            for called in calllist:
                called["displayName"] = data["displayName"]
                called["upd_dt"] = "{0:%Y-%m-%d %H:%M:%S.%f}".format(now)
                called["displayed"] = "0"
                dl.insert(called)

            # 呼出ログ出力
            self.__insert_call_log(now, socket, data, vcalllist)

            self.update_name_to_conn_manage(socket, data["displayName"])

            t = Telegram()
            send_data = t.create_telegram_del_res_R(self.get_seqNo(socket))
            self.send_async(socket, send_data)

        else:
            self.logger.error("status error. status = {}".format(self.getStatus(socket)))

    def syncData(self, socket, data):
        """同期
        
        Args:
            socket (object): ソケット
            data (dict): 同期要求データ
        """
        if self.getStatus(socket) == Status.S3:
            groups = data["groups"]

            hasAllGroup = len([g for g in groups if g==0]) > 0

            clv = CallListView()
            callList = []

            if hasAllGroup:
                # 全てのグループ
                key = {}
                orderBy = {
                    "priority": "DESC",
                    "ins_dt" : ""
                    }
                callList.extend(clv.select(key, orderBy))
            else:
                orderBy = {
                    "priority": "DESC",
                    "ins_dt" : ""
                    }
                for g in groups:
                    key = {"groupId": g}
                    callList.extend(clv.select(key, orderBy))

                key = {"groupId": 0}
                callList.extend(clv.select(key, orderBy))

            callList = sorted(callList, key=lambda x: x["ins_dt"])
            callList = sorted(callList, key=lambda x: x["priority"], reverse=True)
            callList = callList[:100]

            dlv = DelListView()
            delList = []
            if hasAllGroup:
                # 全てのグループ
                key = {}
                orderBy = {
                    "upd_dt" : ""
                    }
                delList.extend(dlv.select(key, orderBy))
            else:
                orderBy = {
                    "upd_dt" : ""
                    }
                for g in groups:
                    key = {"groupId": g}
                    delList.extend(dlv.select(key, orderBy))
                key = {"groupId": 0}
                delList.extend(dlv.select(key, orderBy))

            delList = sorted(delList, key=lambda x: x["upd_dt"])
            # delList = delList[:100]
            delList = delList[-100:]


            self.update_groups_to_conn_manage(socket, groups)
            self.update_name_to_conn_manage(socket, data["displayName"])

            t = Telegram()
            send_data = t.create_telegram_sync_data_R(callList, delList, self.get_seqNo(socket))
            self.send_async(socket, send_data)
        else:
            self.logger.error("status error. status = {}".format(self.getStatus(socket)))
        

    def send_async(self, socket, data, callback=None):
        """非同期送信
        
        Args:
            socket (object): ソケット
            data (bytes): 送信データ
            callback (function, optional): コールバック関数. Defaults to None.
        """

        def __proc_async():
            result = True
            retry = True
            cnt = 0
            while retry:
                try:
                    self.logger.info("send data = {0} to {1}".format(data.hex(), socket.getpeername()))
                    socket.sendall(data)
                    retry = False
                except:
                    if cnt < 2:                        
                        self.logger.warn("socket.sendall failed.", exc_info=True)
                        cnt += 1
                    else:
                        self.logger.error("send_async failed.", exc_info=True)
                        self.disconnect_client(socket)
                        result = False
                        retry = False

            if callback is not None:
                callback(result)

        th = threading.Thread(target=__proc_async)
        th.start()

    def set_socket_to_conn_manage(self, socket, address):
        """接続管理設定
        
        Args:
            socket (object): ソケット
            address (tuple): アドレス
        """
        with self.conn_manage_lock:
            self.conn_manage[socket.fileno()] = {
                "socket": socket,
                "address": address,
                "groupId": [], 
                "seqNo": 0,
                "status": Status.S2
            }
        self.display.insert({"ipAddress": address[0], "datetime": "{0:%Y/%m/%d %H:%M:%S}".format(datetime.now())})

    def register_groups_to_conn_manage(self, socket, groups, name):
        """接続管理設定(グループ登録)
        
        Args:
            socket (object): ソケット
            groups (list): グループ
            name (string): 端末名称
        """

        with self.conn_manage_lock:
            if socket.fileno() in self.conn_manage:
                if len(groups) == 0:
                    groups.append(0)
                self.conn_manage[socket.fileno()]["groupId"] = groups
                self.conn_manage[socket.fileno()]["status"] = Status.S3
                ipAddress = self.conn_manage[socket.fileno()]["address"][0]

                dg = Group()
                groupArr = []
                for g in groups:
                    if g == 0:
                        groupArr.append(self.res.getResource("tcpserverthread:allGroup"))
                    else:
                        key = {"groupId": g}
                        gl = dg.select(key)
                        if len(gl) > 0:
                            groupArr.append(gl[0]["name"])
                        else:
                            groupArr.append(str(g))

                key = {"ipAddress": ipAddress}
                value = {"groups": ", ".join(groupArr), "name": name}
                self.display.update(key, value)

    def update_groups_to_conn_manage(self, socket, groups):
        """接続管理設定(グループ更新)
        
        Args:
            socket (object): ソケット
            groups (list): グループ
        """
        with self.conn_manage_lock:
            if socket.fileno() in self.conn_manage:
                if len(groups) == 0:
                    groups.append(0)
                self.conn_manage[socket.fileno()]["groupId"] = groups
                ipAddress = self.conn_manage[socket.fileno()]["address"][0]

                dg = Group()
                groupArr = []
                for g in groups:
                    if g == 0:
                        groupArr.append(self.res.getResource("tcpserverthread:allGroup"))
                    else:
                        key = {"groupId": g}
                        gl = dg.select(key)
                        if len(gl) > 0:
                            groupArr.append(gl[0]["name"])
                        else:
                            groupArr.append(str(g))

                key = {"ipAddress": ipAddress}
                value = {"groups": ", ".join(groupArr)}
                self.display.update(key, value)

    def update_name_to_conn_manage(self, socket, name):
        """接続管理設定(表示機名称更新)
        
        Args:
            socket (object): ソケット
            groups (list): グループ
        """
        with self.conn_manage_lock:
            if socket.fileno() in self.conn_manage:
                ipAddress = self.conn_manage[socket.fileno()]["address"][0]

                key = {"ipAddress": ipAddress}
                value = {"name": name}
                self.display.update(key, value)

    def get_socket_from_conn_manage(self, fd):
        """接続管理情報からソケット取得
        
        Args:
            fd (int): ファイルディスクリプタ
        
        Returns:
            object: ソケット
        """
        try:
            self.conn_manage_lock.acquire()
            return self.conn_manage[fd]["socket"]
        except:
            self.logger.error("get_socket_from_conn_manage failed.", exc_info=True)
            return None
        finally:
            self.conn_manage_lock.release()

    def remove_socket_from_conn_manage(self, socket):
        """接続管理情報削除
        
        Args:
            socket (object): ソケット
        """
        self.remove_socket_from_conn_manage_by_fd(socket.fileno())

    def remove_socket_from_conn_manage_by_fd(self, fd):
        """接続管理情報削除
        
        Args:
            fd (int): ファイルディスクリプタ
        """
        with self.conn_manage_lock:
            if fd in self.conn_manage:
                ipaddr = self.conn_manage[fd]["address"][0]
                otherFds  = [d for d, m in self.conn_manage.items() if d != fd and m["address"][0] == ipaddr]
                if len(otherFds) == 0:
                    # 同一IPアドレスでのソケット接続が残っていない場合、接続中一覧表示から削除する
                    self.display.delete({"ipAddress": ipaddr})
                del self.conn_manage[fd]

    def get_seqNo(self, socket):
        """シーケンス番号取得
        
        Args:
            socket (object): ソケット
        
        Returns:
            int: シーケンス番号
        """

        fd = socket.fileno()
        with self.conn_manage_lock:
            if fd in self.conn_manage:
                seqNo = self.conn_manage[fd]["seqNo"]
                seqNo += 1
                if seqNo > 255:
                    seqNo = 1
                self.conn_manage[fd]["seqNo"] = seqNo
                return seqNo
            else:
                return 1
        
    def getStatus(self, socket):
        """ステータス取得
        
        Args:
            socket (object): ソケット
        
        Returns:
            enum: ステータス
        """

        fd = socket.fileno()
        with self.conn_manage_lock:
            if fd in self.conn_manage:
                return self.conn_manage[fd]["status"]
            else:
                return None

    def setStatus(self, socket, status):
        """ステータス設定
        
        Args:
            socket (object): ソケット
            status (enum): ステータス
        """
        fd = socket.fileno()
        with self.conn_manage_lock:
            if fd in self.conn_manage:
                self.conn_manage[fd]["status"] = status

    def __insert_call_log(self, dt, socket, data, vcalllist):
        """呼出ログ出力

        Args:
            dt (datetime): 呼出日時
            socket (object): socket
            data (dict): 消去要求データ
            vcalllist (list): 消去対象データのリスト
        """
        try:
            calllog = CallLog()
            self.conn_manage_lock.acquire()
            for called in vcalllist:
                groupName = ""
                if called["groupId"] == 0:
                    groupName = "all"
                else:
                    groupName = called["groupName"]
                ipAddress = self.conn_manage[socket.fileno()]["address"][0]

                message = "address={0},model={1},button={2},strength=-{3}dBm,caller name1={4}, caller name2={5},group={6},color={7},device ip={8},device name={9}".format(
                    data["address"], data["modelNo"], data["buttonNo"], called["strength"], called["line1"], called["line2"], groupName, called["color"], ipAddress, data["displayName"] )

                calllog.insert({
                    "datetime": "{0:%Y-%m-%d %H:%M:%S.%f}".format(dt),
                    "category": CallLog.CATEGORY_RESPONSE,
                    "message": message,
                    "sent": 0
                })
        except:
            self.logger.error("__insert_call_log failed.", exc_info=True)
        finally:
            self.conn_manage_lock.release()

    def getConnInfoListByGroupId(self, groupId):
        """接続情報取得
        
        Args:
            groupId (int): グループID
        
        Returns:
            [dictionary]: 接続情報
        """
        # 接続管理情報
        with self.conn_manage_lock:
            conn_info_arr = self.conn_manage.values()
            if groupId == 0:
                # 全てのグループ
                return conn_info_arr
            else:
                return [connInfo for connInfo in conn_info_arr if groupId in connInfo["groupId"] or  0 in connInfo["groupId"]]

if __name__ == "__main__":
    address = ('', 5001)
    # handler = MyTCPHandler
    # serverThread = TCPServerThread(address)
    serverThread = TCPServerThread(address)
    # serverThread = TCPSendThread()
    try:
        serverThread.start()
        serverThread.join()
    except KeyboardInterrupt:
        serverThread.stop()
