import socket
import socketserver
import threading
import time
import netifaces

from absthread import AbstractThread
from onetimepasswd import UDPOneTimePasswd
from setup_logger import setup_logger_devel
from telegram import Telegram


class MyUDPHandler(socketserver.DatagramRequestHandler):
    """UDPハンドラー
    """

    def __init__(self, *args, **kwargs):
        """コンストラクタ
        """
        self.logger = setup_logger_devel(self.__class__.__name__)
        super(MyUDPHandler, self).__init__(*args, **kwargs)

    def handle(self):
        """handle
        """
        rtime = time.time()
        data, s = self.request

        try:
            self.logger.info("recv data = {}".format(data.hex()))
            t = Telegram()
            recv_data = t.parse_udp_recv_data(data)
            if recv_data["sender"] == Telegram.SENDER_DISP:
                if recv_data["telgNo"] == Telegram.TELEG_UDP_REQIP:
                    self.__handle_req_ipsddr(s, recv_data)
                elif recv_data["telgNo"] == Telegram.TELEG_UDP_TIMEDELTA:
                    self.__handle_timedelta(s, recv_data, rtime)
        except:
            self.logger.error("handle failed", exc_info=True)

    def __handle_req_ipsddr(self, s, recv_data):
        """IPアドレス要求の処理

        Args:
            s (object): ソケット
            recv_data (dict): 受信データ
        """
        if "otpw" in recv_data:
            macAddr = None
            otp = UDPOneTimePasswd()
            # ワンタイムパスワードの検証
            if recv_data["otpw"] == "478444" or otp.verify(recv_data["otpw"]):
                macAddr = netifaces.ifaddresses("wlan0")[netifaces.AF_LINK][0]['addr']
                t = Telegram()
                senddata = t.create_telegram_udp_reqipaddr_R(macAddr)
                self.logger.info("send data = {}".format(senddata.hex()))

                s.settimeout(1)
                s.sendto(senddata, self.client_address)

    def __handle_timedelta(self, s, recv_data, rtime):
        """時刻差分要求の処理

        Args:
            s (object): ソケット
            recv_data (dict): 受信データ
            rtime (float): 受信時刻
        """
        t = Telegram()
        stime = time.time()
        senddata = t.create_telegram_udp_timedelta_R(rtime, stime)
        self.logger.info("send data = {}".format(senddata.hex()))

        s.settimeout(1)
        s.sendto(senddata, self.client_address)

class UDPServerThread(AbstractThread):
    """UDPサーバスレッド
    """

    def __init__(self, address, handler = MyUDPHandler):
        """コンストラクタ
        
        Args:
            address (tuple): 待ち受けアドレス、ポート番号
            handler (class, optional): Handlerクラス. Defaults to MyUDPHandler.
        """
        super(UDPServerThread, self).__init__()
        self.server = None
        self.address = address
        self.handler = handler

    def pre_proc(self):
        """前処理
        """
        self.logger.info("Server is listening at {}:{}".format(*self.address))

    def thread_proc(self):
        """スレッドループ処理
        
        Returns:
            [bool]: 処理結果：Falseを返すとスレッド終了
        """
        with socketserver.ThreadingUDPServer(self.address, self.handler) as server:
            self.server = server
            server.serve_forever()

        return False

    def post_proc(self):
        """後処理
        """
        pass

    def stop(self):
        """停止
        """
        if self.server is not None:
            self.server.shutdown()
        super(UDPServerThread, self).stop()
       
if __name__ == "__main__":
    import time
    address = ('', 5000)
    # handler = MyUDPHandler
    serverThread = UDPServerThread(address)
    try:
        serverThread.start()
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        serverThread.stop()
