import logging
import logging.handlers
import os
import pathlib
import threading
import traceback

import common_util as util

if util.is_android_device():
    from jnius import autoclass, cast


def setup_logger_csv(name, logfile='./log/LOGFILENAME.csv'):
    """CSV用loggerのセットアップ
    
    Args:
        name (string): 名称
        logfile (str, optional): ログファイル名. Defaults to './log/LOGFILENAME.csv'.
    
    Returns:
        object: logger
    """
    logger = logging.getLogger(name)
    if len(logger.handlers) == 0:
        logger.setLevel(logging.DEBUG)

        # create file handler which logs even DEBUG messages
        fh = logging.FileHandler(logfile)
        fh.setLevel(logging.DEBUG)
        fh_formatter = logging.Formatter('%(asctime)s.%(msecs)03d,%(message)s', '%Y-%m-%d %H:%M:%S')
        fh.setFormatter(fh_formatter)

        # create console handler with a INFO log level
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)
        ch_formatter = logging.Formatter('%(asctime)s.%(msecs)03d,%(message)s', '%Y-%m-%d %H:%M:%S')
        ch.setFormatter(ch_formatter)

        # add the handlers to the logger
        logger.addHandler(fh)
        logger.addHandler(ch)
    return logger

CALL_LIST_LOGFILE=pathlib.Path("./log/recvdata.csv").resolve()

def setup_logger_for_call(name):
    """呼び出しログ用loggerのセットアップ
    
    Args:
        name (string): 名称
    
    Returns:
        object: logger
    """
    return setup_logger_csv(name=name, logfile=CALL_LIST_LOGFILE)

DEVEL_LOGFILE=pathlib.Path("./log/devel.log").resolve()
MEGA_BYTES_10=10485760
def setup_logger_devel(name):
    """開発用内部ログセットアップ
    
    Args:
        name (string): 名称
        logfile (string, optional): ログファイル名. Defaults to DEVEL_LOGFILE.
    
    Returns:
        object: logger
    """
    logger = logging.getLogger(name)

    if len(logger.handlers) == 0:

        logfile=DEVEL_LOGFILE
        level=logging.DEBUG
        maxBytes = MEGA_BYTES_10
        if util.is_android_device():
            logfile = __get_android_logPath()
            from disp_dev.config_setting import ConfigSetting
            level = ConfigSetting().logLevel
        else:
            from recv_dev.config import TerminalInfo
            level = TerminalInfo().logLevel

        logger.setLevel(level)

        # fh = logging.handlers.RotatingFileHandler(logfile, maxBytes=maxBytes, backupCount=10, delay=True)
        fh = ThreadSafeRotatingFileHandler(logfile, maxBytes=maxBytes, backupCount=10)
        fh.setLevel(level)
        fh_formatter = logging.Formatter('[%(asctime)s] [%(levelname)s] [%(pathname)s:%(lineno)d] [%(name)s] [%(funcName)s] [%(thread)d] [%(message)s]')
        fh.setFormatter(fh_formatter)

        ch = logging.StreamHandler()
        ch.setLevel(level)
        ch_formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(filename)s:%(lineno)d - %(message)s', '%Y-%m-%d %H:%M:%S')
        ch.setFormatter(ch_formatter)

        # add the handlers to the logger
        logger.addHandler(fh)
        logger.addHandler(ch)

    return logger

def __get_android_logPath():
    """Androidのログのパスを取得する
    
    Returns:
        string: ログファイルパス
    """    
    try:
        mActivity = autoclass('org.kivy.android.PythonActivity').mActivity
        if mActivity is not None:
            app_context = mActivity.getApplication().getApplicationContext()
        else:
            service = autoclass('org.kivy.android.PythonService').mService
            if service is not None:
                app_context = service.getApplication().getApplicationContext()

        logdir = os.path.join(app_context.getExternalFilesDir(None).getPath(), "log") 
        os.makedirs(logdir, exist_ok = True)
        return os.path.join(logdir, "devel.log")
    except:
        traceback.print_exc()
        return DEVEL_LOGFILE

class ThreadSafeRotatingFileHandler(logging.handlers.RotatingFileHandler):
    """RotatingFileHandlerのThreadSave版
    """

    _rollover_lock = threading.Lock()

    def emit(self, record):
        """
        Emit a record.
        Output the record to the file, catering for rollover as described
        in doRollover().
        """
    
        try:
            if self.shouldRollover(record):
                try:
                    self._rollover_lock.acquire()
                    if self.shouldRollover(record):
                        self.doRollover()
                except:
                    raise
                finally:
                    self._rollover_lock.release()

            logging.FileHandler.emit(self, record)
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)

    def shouldRollover(self, record):
        if self._should_rollover():
            # if some other process already did the rollover we might
            # checked log.1, so we reopen the stream and check again on
            # the right log file
            if self.stream:
                self.stream.close()
                self.stream = self._open()

            return self._should_rollover()

        return 0

    def _should_rollover(self):
        if self.maxBytes > 0:
            self.stream.seek(0, 2)
            if self.stream.tell() >= self.maxBytes:
                return True

        return False
