import struct
from datetime import datetime
import time

from setup_logger import setup_logger_devel


class Telegram:
    TELEG_SYSINFO = "1"
    TELEG_REGISTER = "2"
    TELEG_HEALTH_CHECK = "3"
    TELEG_CALL_DATA = "4"
    TELEG_DEL_DATA = "5"
    TELEG_SYNC_DATA = "6"
    TELEG_UDP_REQIP = "1"
    TELEG_UDP_TIMEDELTA = "2"
    SENDER_DISP = "D"
    SENDER_RECV = "R"
    COMMAND_S = "S"
    COMMAND_C = "C"
    #発信者
    DEF_SENDER = [SENDER_DISP, SENDER_RECV]
    #電文識別番号
    DEF_TELEG_NUM = ["{}".format(n) for n in range(1, 10) ] # "1"～"9"

    def __init__(self):
        """コンストラクタ
        """
        self.logger = setup_logger_devel(self.__class__.__name__)

    def pre_check_telegram(self, raw_data):
        """電文プレチェック
        
        Args:
            raw_data (bytes)): 電文データ
        
        Returns:
            bool: チェック結果
        """
        if len(raw_data) < 12:
            self.logger.error("length error. data = {}".format(raw_data.hex()))
            return False

        if raw_data[0:1] != b'@':
            self.logger.error("startcode error. data = {}".format(raw_data.hex()))
            return False

        if raw_data[-2:] != b'\r\n':
            self.logger.error("delimiter error. data = {}".format(raw_data.hex()))
            return False

        data_len = int(raw_data[1:6].decode())
        if data_len != len(raw_data):
            self.logger.error("data_len error. data = {}".format(raw_data.hex()))
            return False

        if not raw_data[6:7].decode() in self.DEF_SENDER:
            self.logger.error("sender error. data = {}".format(raw_data.hex()))
            return False

        if not raw_data[7:8].decode() in self.DEF_TELEG_NUM:
            self.logger.error("telgNo error. data = {}".format(raw_data.hex()))
            return False
        
        check_sum1 = raw_data[-4:-2]
        check_sum2 = self.calc_check_sum(raw_data)
        if check_sum1 != check_sum2:
            self.logger.error("checksum error. data = {}".format(raw_data.hex()))
            return False

        return True

    def calc_check_sum(self, raw_data):
        """チェックサム算出
        
        Args:
            raw_data (bytes): 電文データ
        
        Returns:
            bytes: チェックサム
        """
        dd = raw_data[:-4]
        chksum = 0x00
        for d in dd:
            chksum = chksum^ d

        chksum = chksum & 0x00ff
        return self.__int_2hex(chksum).encode()

    def create_telegram(self, sender, telgNo, bytes_data):
        """電文生成
        
        Args:
            sender (string): 送信者ID
            telgNo (string): 電文No
            bytes_data (bytes): データ
        
        Returns:
            bytes: 電文データ
        """
        data_len = 12 + len(bytes_data)
        data = ("@{0:05d}".format(data_len) + sender[0:1] + telgNo[0:1]).encode() + bytes_data + b'FF\r\n'
        chksum = self.calc_check_sum(data)
        return data[:-4] + chksum + b'\r\n'

    def create_telegram_register_D(self, groups, name, seqNo):
        """登録電文生成(D)
        
        Args:
            groups (list): グループ
            name (string): 名称
            seqNo (int): シーケンス番号
        
        Returns:
            bytes: 電文データ
        """
        sender = self.SENDER_DISP

        group_data = self.__create_group_data(groups)
        name_data = name.ljust(40, " ").encode("utf-16-le")[0:40]
        data = group_data + name_data + self.__int_2hex(seqNo).encode()

        return self.create_telegram(sender, self.TELEG_REGISTER, data)

    def __create_group_data(self, groups):
        """グループデータ生成
        
        Args:
            groups (list): グループIDのリスト
        
        Returns:
            bytes: グループデータ
        """
        groupData = ['  ' for i in range(0, 10)]

        for i in range(0, len(groups)):
            if i >= len(groupData):
                break

            groupData[i] = "{0:02d}".format(groups[i])
        
        group_data = ''.join(groupData).encode('utf-8')

        return group_data

    def create_telegram_register_R(self, seqNo):
        """登録応答電文生成(R)
        
        Args:
            seqNo (int): シーケンス番号
        
        Returns:
            bytes: 電文データ
        """
        sender = self.SENDER_RECV

        stime = struct.pack(">d", time.time())

        data = stime + self.__int_2hex(seqNo).encode()

        return self.create_telegram(sender, self.TELEG_REGISTER, data)

    def create_telegram_healthcheck_R(self, seqNo):
        """接続確認応答電文(R)生成
        
        Args:
            seqNo (int): シーケンス番号
        
        Returns:
            bytes: 電文データ
        """
        stime = struct.pack(">d", time.time())
        dataSeqNo = self.__int_2hex(seqNo).encode()
        data = stime + dataSeqNo 

        return self.create_telegram(self.SENDER_RECV, self.TELEG_HEALTH_CHECK, data)

    def create_telegram_healthcheck_D(self, groups, name, seqNo):
        """接続確認電文(D)生成
        
        Args:
            groups (list): グループ
            name (string): 表示機名称
            seqNo (int): シーケンス番号
        
        Returns:
            bytes: 電文データ
        """

        group_data = self.__create_group_data(groups)
        name_data = name.ljust(40, " ").encode("utf-16-le")[0:40]
        data = group_data + name_data + self.__int_2hex(seqNo).encode()

        return self.create_telegram(self.SENDER_DISP, self.TELEG_HEALTH_CHECK, data)

    def create_telegram_call_data_R(self, command, data, seqNo):
        """呼出データ電文生成(R)
        
        Args:
            command (string): コマンド
            data (dictionary): 呼び出しデータ
            seqNo (int): シーケンス番号
        
        Returns:
            bytes: 電文データ
        """
        call_data = self.__conv_callData_to_rawdata(command, data)
        stime = struct.pack(">d", time.time())

        data = command[0:1].encode() + call_data + stime + self.__int_2hex(seqNo).encode()
        return self.create_telegram(self.SENDER_RECV, self.TELEG_CALL_DATA, data)

    def create_telegram_del_data_D(self, command, data, displayName, seqNo):
        """消去データ電文生成(D)
        
        Args:
            command (string): コマンド
            data (dictionary): 消去データ
            displayName (string): 表示機名
            seqNo (int): シーケンス番号
        
        Returns:
            bytes: 電文データ
        """
        data["displayName"] = displayName
        if "upd_dt" in data:
            del data["upd_dt"]

        del_data = self.__conv_callData_to_rawdata(command, data)

        data = command[0:1].encode() + del_data + self.__int_2hex(seqNo).encode()
        return self.create_telegram(self.SENDER_DISP, self.TELEG_DEL_DATA, data)

    def create_telegram_del_res_R(self, seqNo):
        """消去応答電文(R)生成
        
        Args:
            seqNo (int): シーケンス番号
        
        Returns:
            bytes: 電文データ
        """
        stime = struct.pack(">d", time.time())
        dataSeqNo = self.__int_2hex(seqNo).encode()
        data = stime + dataSeqNo

        return self.create_telegram(self.SENDER_RECV, self.TELEG_DEL_DATA, data)

    def create_telegram_sync_req_D(self, groups, name, seqNo):
        """同期要求電文(D)生成
        
        Args:
            groups (list): グループ
            name (string): 表示機名称
            seqNo (int): シーケンス番号
        
        Returns:
            bytes: 電文データ
        """
        group_data = self.__create_group_data(groups)
        name_data = name.ljust(40, " ").encode("utf-16-le")[0:40]
        dataSeqNo = self.__int_2hex(seqNo).encode()
        data = group_data + name_data + dataSeqNo

        return self.create_telegram(self.SENDER_DISP, self.TELEG_SYNC_DATA, data)

    def create_telegram_sync_data_R(self, callList, delList, seqNo):
        """同期データ電文(R)生成
        
        Args:
            callList (list): 呼び出しデータ
            delList (list): 取消データ
            seqNo (int): シーケンス番号
        
        Returns:
            bytes: 電文データ
        """
        dataSeqNo = self.__int_2hex(seqNo).encode()

        call_list_data = self.__create_call_data_list(callList)
        del_list_data = self.__create_del_data_list(delList)

        stime = struct.pack(">d", time.time())

        data = call_list_data + del_list_data + stime + dataSeqNo

        return self.create_telegram(self.SENDER_RECV, self.TELEG_SYNC_DATA, data)

    def create_telegram_sysinfo_R(self, logLevel, groups, seqNo):
        """受信機情報通知電文生成(R)
        
        Args:
            logLevel (int): ログレベル
            groups (list): グループ
            seqNo (int): シーケンス番号
        
        Returns:
            bytes: 電文データ
        """
        sender = self.SENDER_RECV

        level_data = "{:02d}".format(logLevel).encode()
        glen = "{:02d}".format(len(groups)).encode()
        gdata = bytearray(b'')
        for g in groups:
            gdata += "{:02d}".format(g["groupId"]).encode()
            gdata += g["name"].ljust(68, " ").encode("utf-16-le")[0:68]
        stime = struct.pack(">d", time.time())
        data = level_data + glen + gdata + stime +  self.__int_2hex(seqNo).encode()

        return self.create_telegram(sender, self.TELEG_SYSINFO, data)

    def create_telegram_udp_reqipaddr_D(self, otpw):
        """IPアドレス要求(D)
        
        Args:
            otpw (string): ワンタイムパスワード
        
        Returns:
            bytes: 電文データ
        """
        send_data = (self.SENDER_DISP + self.TELEG_UDP_REQIP + otpw.zfill(6)[0:6]).encode("utf-8")
 
        return send_data

    def create_telegram_udp_reqipaddr_R(self, mac):
        """IPアドレス応答(R)
        
        Args:
            mac (string): MACアドレス("xx:xx:xx:xx:xx:xx"形式)
        
        Returns:
            bytes: 電文データ
        """
        senddata = (self.SENDER_RECV + self.TELEG_UDP_REQIP).encode('utf-8') +  bytes.fromhex( "".join(mac.split(":")))
 
        return senddata

    def create_telegram_udp_timedelta_D(self):
        """時刻差分要求(D)
        
        Returns:
            bytes: 電文データ
        """
        send_data = (self.SENDER_DISP + self.TELEG_UDP_TIMEDELTA).encode("utf-8")
 
        return send_data

    def create_telegram_udp_timedelta_R(self, rtime, stime):
        """時刻差分応答(R)
        
        Args:
            rtime (float): 受信時刻（UNIX時間）
            stime (float): 送信時刻（UNIX時間）
        
        Returns:
            bytes: 電文データ
        """
        timedata = struct.pack(">dd", rtime, stime)
        senddata = (self.SENDER_RECV + self.TELEG_UDP_TIMEDELTA).encode('utf-8') +  timedata
 
        return senddata

    def __create_call_data_list(self, callList):
        """呼び出しデータリスト生成
        
        Args:
            callList (list): 呼び出しデータ
        
        Returns:
            bytes: データ
        """
        data = bytearray(b'')
        num = "{:03d}".format(len(callList))
        data += num.encode()
        for call in callList:
            raw_call_data = self.__conv_callData_to_rawdata(self.COMMAND_S, call)
            data += raw_call_data

        return data

    def __create_del_data_list(self, delList):
        """取消データリスト生成
        
        Args:
            delList (list): 取消データ
        
        Returns:
            bytes: データ
        """
        data = bytearray(b'')
        num = "{:03d}".format(len(delList))
        data += num.encode()
        for delData in delList:
            raw_del_data = self.__conv_callData_to_rawdata(self.COMMAND_C, delData)
            data += raw_del_data

        return data

    def parse_recv_data(self, raw_data):
        """受信データの解析
        
        Args:
            raw_data (bytes): 受信データ
        
        Returns:
            dictionary: 解析後受信データ
        """
        rtime = time.time()
        (sender, telgNo) = struct.unpack_from("@" "1s1s", raw_data, 6)

        sender = sender.decode()
        telgNo = telgNo.decode()
        data = {
            "sender": sender,
            "telgNo": telgNo,
        }

        if ((telgNo == self.TELEG_HEALTH_CHECK and sender == self.SENDER_RECV) or (telgNo == self.TELEG_DEL_DATA and sender == self.SENDER_RECV) 
            or (telgNo == self.TELEG_REGISTER and sender == self.SENDER_RECV)):
            # 接続確認応答電文(R)
            # 取消応答電文(R)
            # 登録応答電文(R)
            (stime, seqNo,) = struct.unpack_from(">d2s", raw_data, 8)
            data["seqNo"] = int(seqNo.decode(), 16)
            data["stime"] = stime
            data["timeDiff"] =  True if abs(rtime - stime) > 10 else False

        elif telgNo == self.TELEG_HEALTH_CHECK and sender == self.SENDER_DISP:
            # 接続確認電文(D)
            (groups, name, seqNo, ) = struct.unpack_from("@" "20s40s2s", raw_data, 8)

            groupsStr = groups.decode("utf-8")

            groupList = [ groupsStr[i: i+2] for i in range(0, 20, 2)]
            groupList = list(filter( lambda g: len(g.strip()) > 0,groupList))
            groupList = [ int(g) for g in groupList]

            data["groups"] = groupList
            data["displayName"] = name.decode("utf-16-le").strip()
            data["seqNo"] = int(seqNo.decode(), 16)

        elif telgNo == self.TELEG_CALL_DATA and sender == self.SENDER_RECV :
            # 呼出データ電文(R)

            (command, ) = struct.unpack_from("@" "1s", raw_data, 8)

            (data, n) = self.__raw_data_to_calldata(raw_data[9:])

            (stime, seqNo,) = struct.unpack_from(">d2s", raw_data, 9 + n)

            data["command"] = command.decode()
            data["seqNo"] = int(seqNo.decode(), 16)
            data["sender"] = sender
            data["telgNo"] = telgNo
            data["stime"] = stime
            data["timeDiff"] =  True if abs(rtime - stime) > 10 else False

        elif telgNo == self.TELEG_DEL_DATA and sender == self.SENDER_DISP:
            # 取消データ電文(D)

            (command, ) = struct.unpack_from("@" "1s", raw_data, 8)

            (data, n) = self.__raw_data_to_calldata(raw_data[9:])

            (seqNo,) = struct.unpack_from("@" "2s", raw_data, 9 + n)

            data["command"] = command.decode()
            data["seqNo"] = int(seqNo.decode(), 16)
            data["sender"] = sender
            data["telgNo"] = telgNo

        elif telgNo == self.TELEG_SYNC_DATA and sender == self.SENDER_DISP:
            # 同期要求電文(D)

            (groups, name, seqNo, ) = struct.unpack_from("@" "20s40s2s", raw_data, 8)
            groupsStr = groups.decode("utf-8")

            groupList = [ groupsStr[i: i+2] for i in range(0, 20, 2)]
            groupList = list(filter( lambda g: len(g.strip()) > 0,groupList))
            groupList = [ int(g) for g in groupList]

            data["groups"] = groupList
            data["displayName"] = name.decode("utf-16-le").strip()
            data["seqNo"] = int(seqNo.decode(), 16)

        elif telgNo == self.TELEG_SYNC_DATA and sender == self.SENDER_RECV:
            # 同期データ電文(R)

            callList = []
            delList = []
            (callLen,) = struct.unpack_from("@" "3s", raw_data, 8)
            callLen = int(callLen.decode(), 10)
            raw_data = raw_data[8 + 3:]
            for i in range(callLen):
                (calldata, n) = self.__raw_data_to_calldata(raw_data)
                callList.append(calldata)
                raw_data = raw_data[n:]

            (delLen,) = struct.unpack_from("@" "3s", raw_data)
            delLen = int(delLen.decode(), 10)
            raw_data = raw_data[3:]
            for i in range(delLen):
                (deldata, n) = self.__raw_data_to_calldata(raw_data)
                delList.append(deldata)
                raw_data = raw_data[n:]

            (stime, seqNo,) = struct.unpack_from(">d2s", raw_data)

            data = {
                "sender": sender,
                "telgNo": telgNo,
                "callList": callList,
                "delList":delList,
                "seqNo": int(seqNo.decode(), 16),
                "stime": stime,
                "timeDiff": True if abs(rtime - stime) > 10 else False
            }
        elif telgNo == self.TELEG_REGISTER and sender == self.SENDER_DISP:
            # 登録電文(D)

            (groups, name, seqNo) = struct.unpack_from("@" "20s40s2s", raw_data, 8)
            groupsStr = groups.decode("utf-8")

            groupList = [ groupsStr[i: i+2] for i in range(0, 20, 2)]
            groupList = list(filter( lambda g: len(g.strip()) > 0,groupList))
            groupList = [ int(g) for g in groupList]

            data["groups"] = groupList
            data["displayName"] = name.decode("utf-16-le").strip()
            data["seqNo"] = int(seqNo.decode(), 16)

        elif telgNo == self.TELEG_SYSINFO:
            # 受信機情報通知電文(R)

            (logLevel, group_num, ) = struct.unpack_from("@" "2s2s", raw_data, 8)

            logLevel = int(logLevel.decode())
            data["logLevel"] = logLevel

            group_num = int(group_num.decode())
            groups = []
            for i in range(group_num):
                (groupId, group_name, ) = struct.unpack_from("@" "2s68s", raw_data, 12 + i*70)
                groupId = int(groupId.decode())
                group_name = group_name.decode("utf-16-le").strip()
                groups.append({
                    "groupId": groupId,
                    "name": group_name
                })

            data["groups"] = groups
            # (seqNo,  ) = struct.unpack_from("@" "2s", raw_data, 14 + group_num*70)
            (stime, seqNo,  ) = struct.unpack_from(">d2s", raw_data, 14 + group_num*70)
            data["seqNo"] = int(seqNo.decode(), 16)
            data["stime"] = stime
            data["timeDiff"] =  True if abs(rtime - stime) > 10 else False


        self.logger.debug("parse data = {}".format(data))
        return data

    def parse_udp_recv_data(self, raw_data):
        """受信データの解析(UDP)
        
        Args:
            raw_data (bytes): 受信データ
        
        Returns:
            dictionary: 解析後受信データ
        """
        (sender, telgNo) = struct.unpack_from("@" "1s1s", raw_data, 0)

        sender = sender.decode()
        telgNo = telgNo.decode()
        data = {
            "sender": sender,
            "telgNo": telgNo,
        }
        if telgNo == self.TELEG_UDP_REQIP and sender == self.SENDER_DISP:
            pw, = struct.unpack_from("@6s", raw_data, 2)
            pw = pw.decode('utf-8')
            data["otpw"] = pw
        elif telgNo == self.TELEG_UDP_REQIP and sender == self.SENDER_RECV:
            mac, = struct.unpack_from("@6s", raw_data, 0)
            macHex = mac.hex()
            macAddress = ":".join([macHex[0:2], macHex[2:4], macHex[4:6], macHex[6:8], macHex[8:10], macHex[10:12]]).upper()
            data["macAddress"] = macAddress
        elif telgNo == self.TELEG_UDP_TIMEDELTA and sender == self.SENDER_DISP:
            pass
        elif telgNo == self.TELEG_UDP_TIMEDELTA and sender == self.SENDER_RECV:
            rtime, stime = struct.unpack_from(">dd", raw_data, 2)
            data["rtime"] = rtime
            data["stime"] = stime

        return data

    def __raw_data_to_calldata(self, raw_data):
        """呼び出しデータ解析
        
        Args:
            raw_data (bytes)): 受信データ（呼び出しデータの部分）
        
        Returns:
            (dictionary, int): (解析後データ, 次のデータの位置)
        """
        fmt = "8s2s2s1s2s1s1s14s3s40s24s68s40s14s"
        (address, modelNo, buttonNo, color, sound, priority, recall, callDt, strength, line1, line2, groupName, displayName, delDt) = \
            struct.unpack_from("@"  + fmt, raw_data)

        data = {}
        data["address"] = address.decode()
        data["modelNo"] = modelNo.decode()
        data["buttonNo"] = buttonNo.decode()
        data["color"] = color.decode()
        data["sound"] = sound.decode()
        data["priority"] = int(priority.decode(), 10)
        data["recall"] = int(recall.decode(), 10)
        data["ins_dt"] = "{:%Y-%m-%d %H:%M:%S.%f}".format(datetime.strptime(callDt.decode(), "%Y%m%d%H%M%S")) 
        data["strength"] = int(strength.decode(), 10)
        data["line1"] = line1.decode("utf-16-le").strip()
        data["line2"] = line2.decode("utf-16-le").strip()
        data["groupName"] = groupName.decode("utf-16-le").strip()
        data["displayName"] = displayName.decode("utf-16-le").strip()
        delDt = delDt.decode().strip()
        if len(delDt) > 0:
            data["upd_dt"] = "{:%Y-%m-%d %H:%M:%S.%f}".format(datetime.strptime(delDt, "%Y%m%d%H%M%S")) 
        else:
            data["upd_dt"] = delDt

        c = 0
        for  n in fmt.split("s"):
            if len(n) > 0:
                c += int(n)
        return (data, c)


    def __conv_callData_to_rawdata(self, command, data):
        """呼び出しデータを電文用のバイナリデータに変換
        
        Args:
            data (dictionary): 呼び出しデータ
        
        Returns:
            bytes: 変換後データ
        """
        address = ""
        modelNo = ""
        buttonNo = ""
        color = "0"
        sound = "01"
        priority = "0"
        recall = '0'
        callDt = datetime.min
        strength = 0
        line1 = ""
        line2 = ""
        groupName = ""
        displayName = ""
        delDt = None
        if "address" in data: address = data["address"] 
        if "modelNo" in data: modelNo = data["modelNo"] 
        if "buttonNo" in data: buttonNo = data["buttonNo"] 
        if "color" in data: color = data["color"] 
        if "sound" in data: sound = data["sound"] 
        if "priority" in data: priority = data["priority"] 
        if "recall" in data: recall = data["recall"] 
        if "ins_dt" in data: callDt = datetime.strptime(data["ins_dt"], "%Y-%m-%d %H:%M:%S.%f")
        if "strength" in data: strength = data["strength"] 
        if "line1" in data: line1 = data["line1"] 
        if "line2" in data: line2 = data["line2"] 
        if "groupName" in data: groupName = data["groupName"] 
        if "displayName" in data and command == self.COMMAND_C: displayName = data["displayName"] 
        if "upd_dt" in data and command == self.COMMAND_C: delDt = datetime.strptime(data["upd_dt"], "%Y-%m-%d %H:%M:%S.%f") 

        raw_data = struct.pack("@8s2s2s1s2s1s1s14s3s40s24s68s40s14s", \
                address.zfill(8)[0:8].encode(),
                modelNo.zfill(2)[0:2].encode(),
                buttonNo.zfill(2)[0:2].encode(),
                color[0:1].encode(),
                sound.zfill(2)[0:2].encode(),
                str(priority)[0:1].encode(),
                str(recall)[0:1].encode(),
                "{:%Y%m%d%H%M%S}".format(callDt).encode(),
                str(strength).zfill(3)[0:3].encode(),
                line1.ljust(40, " ").encode("utf-16-le")[0:40],
                line2.ljust(24, " ").encode("utf-16-le")[0:24],
                groupName.ljust(68, " ").encode("utf-16-le")[0:68],
                displayName.ljust(40, " ").encode("utf-16-le")[0:40],
                (" " * 14).encode() if delDt is None else "{:%Y%m%d%H%M%S}".format(delDt).encode()
        )
        return raw_data


    def __int_2hex(self, i):
        """int hex変換（2バイト分）
        
        Args:
            i (int)): int値
        
        Returns:
            string: 16進文字列
        """
        return "{0:02x}".format(i).upper()

    def __int_1hex(self, i):
        """int hex変換（1バイト分）
        
        Args:
            i (int)): int値
        
        Returns:
            string: 16進文字列
        """
        return "{0:01x}".format(i).upper()



if __name__ == "__main__":
    t = Telegram()
    t.pre_check_telegram(b'@00012D106\r\n')
    data = t.create_telegram_sync_req_D(10)
    print(data)
    print(t.calc_check_sum(data))
    print(t.pre_check_telegram(data))
    recv_data = t.parse_recv_data(data)
    print(recv_data)

    # from recv_dev.storage import CallListView, DelListView
    # cl = CallListView()
    # dl = DelListView()
    # callList = cl.select({}) 
    # delList = dl.select({}) 
    # print (callList)
    # print (delList)
    # raw_data = t.create_telegram_call_data_R("S",callList[0], 12)
    # print(raw_data)
    # data = t.parse_recv_data(raw_data)
    # print(data)
    # raw_data = t.create_telegram_del_data_D("C",delList[0], 12)
    # print(raw_data)
    # data = t.parse_recv_data(raw_data)
    # print(data)
    # raw_data = t.create_telegram_sync_data_R( callList, delList, 10)
    # print(raw_data)
    # data = t.parse_recv_data(raw_data)
    # print(data)
